/*
##############################################################################
#
# MODULE:       r.avaflow.main.c
#
# AUTHORS:      Martin Mergili and Shiva P. Pudasaini
# CONTRIBUTORS: Massimiliano Alvioli, Wolfgang Fellin, Jan-Thomas Fischer,
#               Markus Metz, Alexander Ostermann, Matthias Rauter
#
# PURPOSE:      The simulation model for avalanche and debris flows
#               Flow routing script
#
# COPYRIGHT:    (c) 2008 - 2020 by the authors
#               (c) 2013 - 2020 by the BOKU University, Vienna
#               (c) 2015 - 2020 by the University of Vienna
#               (c) 2014 - 2020 by the University of Bonn
#               (c) 2000 - 2020 by the GRASS Development Team
#
# VERSION:      20200402 (2 April 2020)
#
#               This program is free software under the GNU General Public
#               License (>=v2). Read the file COPYING that comes with GRASS
#               for details.
#
##############################################################################
*/


// Use of GRASS GIS


#define WITHGRASS


// Libraries and structures

#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>


#ifdef WITHGRASS


    #include <grass/gmath.h>
    #include <grass/gis.h>
    #include <grass/gprojects.h>
    #include <grass/glocale.h>
    #include <grass/raster.h>
    #include <grass/segment.h>


#endif


struct ico { // input constants and control variables

    char *MAINMAPSET; int MULT; int LIMITER; int MODEL; int M; int N; int IMAX; int ENTR; int ENTR2; int ENTR3; int CENTR; int CDEP;
    int RELM; int RELM2; int RELM3; int RELV; int RELV2; int RELV3; int PHI; int PHI2; int PHI3; int DELTA; int DELTA2; int DELTA3; int NYSS; int NYFS; int NYFF;
    int AMBDRAG; int FLUFRI; int TRANSSSFS; int TRANSSSFF; int TRANSFSFF; int TRELEASE; int TRELSTOP; double PI; double UNDEF; double HFLOWMIN; double CSZ; double BDWEST;
    double BDNORTH; double BDSOUTH; double GRAVITY; double CFL[2]; double IMPTHR[3]; int CORRHEIGHT; int DIFFCTRL; int NOOSC;
    int ENTRAINMENT; int STOPPING; int BETACTRL; int NVECTMIN; int PMAX; int PHASES[3]; double EKINREF; double FRICMIN; double DYNFRIC; int MESH; int NMESH; double SLOMO;
    int HYDADD; int ORIGINAL;


#ifdef WITHGRASS


    double NSEGRC; double NSEGS;


#endif


};

struct flow { // flow parameters

    double PHI[3]; double PHI0[3]; double RHO1; double RHO2; double DELTA[3]; double DELTA0[3];
    double REP; int DRAG_je; double NY0[3]; double TAUY[3]; double AMBDRAG0; double AMBDRAG; double FLUFRI0; double FLUFRI;
    double RHO3; double J; double NY[3]; double TUFRI; 
    double TRANSSSFS0; double TRANSSSFF0; double TRANSFSFF0; double TRANSSSFS; double TRANSSSFF; double TRANSFSFF;
    double CENTR0; double CENTR; double CDEP0; double CDEP; double VM_n0; double VM_l; double VM_n; 
    double VM_fact; double DRAG_k; double DRAG_m; int DRAG_n; double DRAG_rep; double DRAG_vterm;
    double VIS_chi[3]; double VIS_xi[3]; double VIS_a[3]; double VIS_ry; double FRI_exp; double NY_exp; double KPMAX[3]; double NMESH0; double NMESH;
};


// Functions

double fmax ( double gval1, double gval2 ) { // maximum value

    double gmax;
    if ( gval2 > gval1 ) gmax = gval2;
    else gmax = gval1;
    return gmax;
}

double fmin ( double gval1, double gval2 ) { // minimum value

    double gmin;
    if ( gval2 < gval1 ) gmin = gval2;
    else gmin = gval1;
    return gmin;
}

double fabsmin ( double gval1, double gval2 ) { // minimum value

    double gmin;
    if ( fabs(gval2) < fabs(gval1) ) gmin = gval2;
    else gmin = gval1;
    return gmin;
}

double fsign ( double gval ) { // sign

    double gsign;
    if ( gval < 0 ) gsign = -1;
    else if ( gval == 0 ) gsign = 0;
    else gsign = 1;
    return gsign;
}

int fvalid ( double ghflow, double ghflow2, double ghflow3, struct ico sico ) { // minimum flow height

    int gvalid;

    gvalid = 0;
    if ( ghflow > sico.HFLOWMIN ) gvalid = 1;
    if ( sico.MODEL > 3 && ghflow2 > sico.HFLOWMIN ) gvalid = 1;
    if ( sico.MODEL == 7 && ghflow3 > sico.HFLOWMIN ) gvalid = 1;

    return gvalid;
}

int **alloc_imatrix(int mrows, int mcols) { // function for allocating 2D integer arrays
    int **mm;
    int mi;
    mm = (int **) calloc(mrows, sizeof(int *));
    mm[0] = (int *) calloc(mrows * mcols, sizeof(int));
    for (mi = 1; mi < mrows; mi++)
    mm[mi] = mm[mi - 1] + mcols;
    return mm;
}

double **alloc_dmatrix (int mrows, int mcols) { // function for allocating 2D double arrays
    double **mm;
    int mi;
    mm = (double **) calloc(mrows, sizeof(double *));
    mm[0] = (double *) calloc(mrows * mcols, sizeof(double));
    for (mi = 1; mi < mrows; mi++)
    mm[mi] = mm[mi - 1] + mcols;
    return mm;
}

char **alloc_cmatrix (int mrows, int mcols) { // function for allocating 2D char arrays
    char **mm;
    int mi;
    mm = (char **) calloc(mrows, sizeof(char *));
    mm[0] = (char *) calloc(mrows * mcols, sizeof(char));
    for (mi = 1; mi < mrows; mi++)
    mm[mi] = mm[mi - 1] + mcols;
    return mm;
}

double ***alloc_dmatrix3 (int mrows, int mcols, int mdepths) { // function for allocating 3D double arrays
    double ***mm;
    int mi, mj;
    mm = (double***)calloc(mrows, sizeof(double**));

    for (mi=0; mi<mrows; mi++) {
        mm[mi] = (double**)calloc(mcols, sizeof(double*));

        for (mj=0; mj<mcols; mj++) {
            mm[mi][mj] = (double*)calloc(mdepths, sizeof(double));
        }
    }
    return mm;
}
void free_dmatrix3 (double ***mm,int mi,int mj)
{
    int i,j;

    for(i=0;i<mi;i++)
    {
        for(j=0;j<mj;j++)
        {
            free(mm[i][j]);
        }
        free(mm[i]);
    }
    free(mm);
}

int fiparam ( FILE *gfparam ) { // function for reading integer parameters

    int gparam;
    char gsparam[50];
    char **strtodhlp=NULL;
    fgets(gsparam, 50, gfparam);
    gparam=strtod(gsparam, strtodhlp);
    return gparam;
}

double fdparam ( FILE *gfparam ) { // function for reading double parameters

    double gparam;
    char gsparam[50];
    char **strtodhlp=NULL;
    fgets(gsparam, 50, gfparam);
    gparam=strtod(gsparam, strtodhlp);
    return gparam;
}

char *fcparam ( FILE *gfparam ) { // function for reading character parameters

    char *gsparam = (char*) calloc ( 1000, sizeof(char));
    fgets(gsparam, 1000, gfparam);
    gsparam[strnlen(gsparam, 1010) - 1] = '\0';
    return gsparam;
}

char *flparam ( FILE *gfparam ) { // function for reading ascii raster line

    int asclinemax = 100000;
    char *gsparam2 = (char*) calloc (( asclinemax + 20 ), sizeof(char));
    fgets(gsparam2, asclinemax, gfparam);
    gsparam2[strnlen(gsparam2, asclinemax+10) - 1] = '\0';
    return gsparam2;
}

int *fin ( int gi, double *gpelev, struct ico sico ) { // function for identifying pixel environment

    int *gin;

    gin = (int*) calloc( 9, sizeof(int));

    gin[0] = gi;
    if ( gi%(sico.N / sico.MESH) == 0 || (gi+1)%(sico.N / sico.MESH) == 0 || gi < (sico.N / sico.MESH) || gi > sico.IMAX - (sico.N / sico.MESH) || gpelev[gi] == sico.UNDEF ) {
        // edge pixels (to be excluded from computation):
        gin[1] = sico.IMAX;
        gin[2] = sico.IMAX;
        gin[3] = sico.IMAX;
        gin[4] = -1;
        gin[5] = -1;
        gin[6] = -1;
        gin[7] = sico.IMAX;
        gin[8] = -1;
    }
    else { // pixels to be included in computation:
        gin[1] = gi + 1;
        gin[2] = gi + sico.N / sico.MESH;
        gin[3] = gi + sico.N / sico.MESH + 1;
        gin[4] = gi - 1;
        gin[5] = gi - sico.N / sico.MESH;
        gin[6] = gi - sico.N / sico.MESH - 1;
        gin[7] = gi + sico.N / sico.MESH - 1;
        gin[8] = gi - sico.N / sico.MESH + 1;
    }

    return gin;
}

double fdiv ( double gabove, double gbelow, double gcriterion ) { // function to avoid division by zero

    double gdiv;

    if ( gbelow > gcriterion ) gdiv = gabove / gbelow; else gdiv = 0;

    return gdiv;
}

double fdynfric ( double gfric, double gfricmin, double ggekin, struct ico sico ) { // function for dynamically varying the frictions

    double gdynfric;

    gdynfric = gfricmin + ( gfric - gfricmin ) * exp( 1 - ( ggekin / sico.EKINREF ));
    return gdynfric;

}

double fbeta ( double gelevm, double gelevp, double gspan, struct ico sico ) { // function for topographic slopes in x and y direction

    double gbeta;
    gbeta = atan(( gelevm - gelevp ) / ( gspan * sico.CSZ ));
    return gbeta;
}

double fbetaxy ( double gbetax, double gbetay ) { // function for topographic slope

    double gbetaxy;
    if ( gbetax != 0 || gbetay != 0 ) gbetaxy = atan( pow( pow (tan ( gbetax) , 2 ) + pow( tan( gbetay ), 2 ), 0.5 ));
    else gbetaxy = 0;
    return gbetaxy;
}

double falpha ( double gbetax, double gbetay, struct ico sico ) { // function for topographic aspect

    double galpha;

    if ( gbetax == 0 && gbetay == 0 ) galpha = 0;
    else if ( gbetax == 0 && gbetay > 0 ) galpha = sico.PI / 2;
    else if ( gbetax == 0 && gbetay < 0 ) galpha = 3 * sico.PI / 2;
    else if ( gbetax >= 0 ) galpha = atan( tan( gbetay ) / tan( gbetax ));
    else if ( gbetax < 0 ) galpha = atan( tan( gbetay ) / tan( gbetax )) + sico.PI;

    if ( galpha < 0 ) galpha += 2 * sico.PI;

    return galpha;
}

double falphav ( double gmomx, double gmomy, struct ico sico ) { // function for direction of movement

    double galphav;

    if ( gmomx == 0 && gmomy == 0 ) galphav = 0;
    else if ( gmomx == 0 && gmomy > 0 ) galphav = sico.PI / 2;
    else if ( gmomx == 0 && gmomy < 0 ) galphav = 3 * sico.PI / 2;
    else if ( gmomx >= 0 ) galphav = atan( gmomy / gmomx );
    else if ( gmomx < 0 ) galphav = atan( gmomy / gmomx ) + sico.PI;

    if ( galphav < 0 ) galphav += 2 * sico.PI;

    return galphav;
}

double **finhyd ( char *gname, int ghydtmax, int ghydtmaxx ) { // function for input of hydrograph

    FILE *gfhyd;
    char gpath[200], *gtesth, *gtesth2;
    int gi, gj;
    char *tok;
    char gdelim[] = " \t\n";
    char **strtodhlp=NULL;

    double **ghyd = alloc_dmatrix( ghydtmaxx+1, 7 ); // allocating memory to array of input hydrograph data

    sprintf(gpath, "%s", gname );
    gfhyd=fopen(gpath, "r");

    gtesth = fcparam ( gfhyd );
    free( gtesth );
    for ( gi=0; gi<=ghydtmax; gi++ ) {
        gtesth = fcparam ( gfhyd );
        tok=strtok(gtesth, gdelim);
        gj = 0;
        while ( tok != NULL ) {
            gtesth2=tok;
            tok=strtok(NULL, gdelim);
            ghyd[gi][gj] = strtod (gtesth2, strtodhlp);
            gj += 1;
        }
        free(gtesth);
    }

    fclose(gfhyd);
    return ghyd;
}

int *fhydp ( int ghydx, int ghydy, double ghydl, int *gpx, int *gpy, double galpha,
    double ghydlmax, struct ico sico ) { // function for input hydrograph profile

    int gdistx, gdisty, gi, gj, gx0, gy0, gdprex, gdprey, gctrl, *gx, *gy;
    double gd, gdxtest, gdytest;

    int *ghydp = (int*) calloc( (int)(2*ghydlmax/sico.CSZ+2), sizeof(int) );
        // allocating memory to variable for internal coordinates of hydrograph profile

    gdistx = (int)( ghydl * fabs( cos( galpha )) / ( 2 * sico.CSZ ) + 0.5 );
    gdisty = (int)( ghydl * fabs( sin( galpha )) / ( 2 * sico.CSZ ) + 0.5 );

    gdxtest = fabs((float)( gdistx ) / cos( galpha ));
    gdytest = fabs((float)( gdisty ) / sin( galpha ));

    gx = (int*) calloc( 4 * (( gdistx + 1 ) * ( gdisty + 1 )), sizeof(int));
    gy = (int*) calloc( 4 * (( gdistx + 1 ) * ( gdisty + 1 )), sizeof(int));

    gctrl = 1;

    if ( galpha == 0 || galpha == sico.PI / 2 || gdytest == 0 ) {

        gx0 = ghydx - gdistx;
        gy0 = ghydy;

        gx[gctrl] = gx0;
        gy[gctrl] = gy0;

        while ( gx[gctrl] <= ghydx + gdistx ) {

            gctrl += 1;
            gx[gctrl] = gx0 + gctrl;
            gy[gctrl] = gy0;
        }

    }
    else if ( galpha == sico.PI || galpha == 3 * sico.PI / 2 || gdxtest == 0 ) {

        gx0 = ghydx;
        gy0 = ghydy - gdisty;

        gx[gctrl] = gx0;
        gy[gctrl] = gy0;

        while ( gy[gctrl] <= ghydy + gdisty ) {

            gctrl += 1;
            gx[gctrl] = gx0;
            gy[gctrl] = gy0 + gctrl;
        }

    }
    else {

        if ( gdxtest > gdytest ) {

            gx0 = ghydx - gdistx;

            if (( galpha > 0 && galpha < sico.PI / 2 ) || ( galpha > sico.PI && galpha < 3 * sico.PI / 2 ))
                gy0 = ghydy - gdisty;//  (int)( gdxtest * fabs(sin( galpha )) + 0.5 );
            else gy0 = ghydy + gdisty;// (int)( gdxtest * fabs(sin( galpha )) + 0.5 );

            gdprex = 1;
            if (( galpha > 0 && galpha < sico.PI / 2 ) || ( galpha > sico.PI && galpha < 3 * sico.PI / 2 )) gdprey = 1;
            else gdprey = -1;
        }
        else {

            if (( galpha > 0 && galpha < sico.PI / 2 ) || ( galpha > sico.PI && galpha < 3 * sico.PI / 2 ))
                gx0 = ghydx - gdistx;//  (int)( gdytest * fabs(cos( galpha )) + 0.5 );
            else gx0 = ghydx + gdistx;// (int)( gdytest * fabs(cos( galpha )) + 0.5 );
            gy0 = ghydy - gdisty;
            if (( galpha > 0 && galpha < sico.PI / 2 ) || ( galpha > sico.PI && galpha < 3 * sico.PI / 2 )) gdprex = 1;
            else gdprex = -1;
            gdprey = 1;
        }
        if ( fabs( 1 / cos( galpha )) < fabs(1 / sin ( galpha ))) gd = fabs( 1 / cos( galpha )); else gd = fabs(1 / sin ( galpha ));

        gx[gctrl] = gx0;
        gy[gctrl] = gy0;

        while ( gx[gctrl] <= ghydx + gdistx && gy[gctrl] <= ghydy + gdisty
            && gx[gctrl] >= ghydx - gdistx && gy[gctrl] >= ghydy - gdisty ) {

            gctrl += 1;
            gx[gctrl] = gx0 + gdprex * (int)( gctrl * gd * fabs(cos( galpha )) + 0.5 );
            gy[gctrl] = gy0 + gdprey * (int)( gctrl * gd * fabs(sin( galpha )) + 0.5 );
        }
    }
    gctrl -= 1;
    ghydp[0] = gctrl;

    for ( gi=1; gi<sico.IMAX; gi++ ) {
        for ( gj=1; gj<=gctrl; gj++ ) {
            if ( gpx[gi] == gx[gj] && gpy[gi] == gy[gj] ) {
                ghydp[gj] = gi;
            }
        }
    }

    free(gx); free(gy);

    return ghydp;
}


// F1 -- START -- Conversion of heights to depths ------------------------------------------------------------------

double fconvin ( double *ghflow, double *ghflow2, double *ghflow3, double ghflowmax, int gmaterial, int *gpx, int *gpy, double *gpelev, double *gbetax,
    double *gbetay, double *gbetaxy, int gi, struct ico sico ) { // function for converting heights into depths

    int *gin, gj, gl, gctrlv;
    double gdflowi, ghflowi, ghflowj, grflowi, grflowj, gdxtest, gdytest, gdxytest, gdxytestmin, gdz;

    if ( sico.MODEL <= 3 ) { // depth at target pixel:
        ghflowi = ghflow[gi];
        grflowi = 1;
        if ( ghflowi != 0 ) gdflowi = ghflowi * cos ( gbetaxy[gi] );
        else gdflowi = 0;
    }
    else if ( sico.MODEL > 3 ) {

        if ( sico.MODEL > 3 && sico.MODEL < 7 ) ghflowi = ghflow[gi] + ghflow2[gi];
        else ghflowi = ghflow[gi] + ghflow2[gi] + ghflow3[gi];

        if ( ghflowi != 0 ) {
            if ( gmaterial == 1 ) grflowi = ghflow[gi] / ghflowi;
            else if ( gmaterial == 2 ) grflowi = ghflow2[gi] / ghflowi;
            else if ( gmaterial == 3 ) grflowi = ghflow3[gi] / ghflowi;
            gdflowi = ghflowi * grflowi * cos ( gbetaxy[gi] );
        }
        else gdflowi = 0;
    }

    if ( sico.CORRHEIGHT == 0 ) gdflowi = ghflowi * grflowi;
    else if ( sico.CORRHEIGHT == 2 ) {

        gin = fin( gi, gpelev, sico ); // identifying pixel neighbourhood
        gctrlv = 1; for ( gl=1; gl<9; gl++ ) if ( gin[gl] < 0 || gin[gl] >= sico.IMAX ) gctrlv = 0;
        free(gin);

        if ( gctrlv == 0 ) gdflowi = 0;
        else if ( sico.CORRHEIGHT == 2 && sico.MODEL < 4 && ghflowi * tan( gbetaxy[gi] ) > 0.5 * sico.CSZ ) { // if advanced conversion is necessary:

            gdxytestmin = sico.UNDEF*-1; // initial value of comparison of horizontal component of depth and shift

            for ( gj=0; gj<sico.IMAX; gj++ ) { // loop over all pixels:

                if ( gpelev[gj] != sico.UNDEF ) {

                    if (( sico.MODEL <= 3 && ghflow[gj] != 0 ) || (( sico.MODEL > 3 && sico.MODEL < 7 ) && ghflow[gj] + ghflow2[gj] != 0 )) {
                        if ( pow( sico.CSZ * sico.CSZ * (double)( pow( gpx[gj] - gpx[gi], 2 ) + pow( gpy[gj] - gpy[gi], 2 )), 0.5 )
                            <= ghflowmax * sin( gbetaxy[gi] )) {

                            if ( sico.MODEL <= 3 ) { // height and ratio at shift location:
                                ghflowj = ghflow[gj];
                                if ( ghflowj != 0 ) grflowj = 1;
                            }
                            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {
                                ghflowj = ghflow[gj] + ghflow2[gj];
                                if ( ghflowj != 0 && gmaterial == 1 ) grflowj = ghflow[gj] / ghflowj;
                                else if ( ghflowj != 0 && gmaterial == 2 ) grflowj = ghflow2[gj] / ghflowj;
                            }

                            gdz = gpelev[gj] - gpelev[gi]; // elevation difference between shift location and target pixel
                            gdxtest = sico.CSZ * (double)( gpx[gj] - gpx[gi] ) - ( ghflowj + gdz ) * tan( gbetax[gi] );
                            gdytest = sico.CSZ * (double)( gpy[gj] - gpy[gi] ) - ( ghflowj + gdz ) * tan( gbetay[gi] );
                            gdxytest = pow( pow( gdxtest , 2 ) + pow( gdytest , 2 ), 0.5 );
                                // comparison of horizontal component of depth and shift

                            if ( gdxytest < gdxytestmin ) {
                                if ( ghflowj != 0 ) gdflowi = grflowj * ( ghflowj + gdz ) / cos ( gbetaxy[gi] ); // depth
                                gdxytestmin = gdxytest;
                            }
                        }
                    }
                }         
            }
        }
    }
    return gdflowi;
}

// F1 -- STOP --- Conversion of heights to depths ------------------------------------------------------------------

double fk ( double gh, double ghflow, double gdvflow, double gekin, int gp, struct ico sico, struct flow sflow ) { // function for earth pressure coefficients

    double ggk, gka, gkp, galpha, gdelta, gphi;

    if ( sico.MODEL == 0 ) {

        gdelta = sflow.DELTA[gp];
        gphi = sflow.PHI[gp];

    } else {

        if ( ghflow > sico.HFLOWMIN ) galpha = ghflow / gh; else galpha = 0;
        gdelta = sflow.DELTA[gp] * pow( galpha, sflow.FRI_exp );
        gphi = sflow.PHI[gp] * pow( galpha, sflow.FRI_exp );

        if ( sico.DYNFRIC == 1 ) gdelta = fdynfric( gdelta, sico.FRICMIN, gekin, sico );
        if ( sico.DYNFRIC == 1 ) gphi = fdynfric( gphi, sico.FRICMIN, gekin, sico );

        if ( gphi < gdelta ) gphi = gdelta;
    }

    gka = 2 * ( 1 - pow( 1 - pow( cos( gphi ) / cos( gdelta ), 2 ), 0.5 )) / pow( cos( gphi ), 2 ) - 1;
    gkp = 2 * ( 1 + pow( 1 - pow( cos( gphi ) / cos( gdelta ), 2 ), 0.5 )) / pow( cos( gphi ), 2 ) - 1;

    if ( gdvflow >= 0 ) ggk=gka;
    else ggk = fmin(sflow.KPMAX[gp], gkp );

    return ggk;
}

double *fcurv ( double gvflowx, double gvflowx2, double gvflowx3, double gvflowy, double gvflowy2, double gvflowy3, double *gelev, double *ggrav, struct ico sico ) {
    // function for curvature

    int gl;
    double gu[3], gv[3], gkappax, gkappaxy, gkappay, *gkappau;
    gkappau = (double*) calloc( sico.PMAX, sizeof(double));

    gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
    gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

    gkappax = 8 * sico.CSZ * ( gelev[2] + gelev[5] - 2 * gelev[0] )
        / (( 4 * pow ( sico.CSZ, 2 ) + pow( gelev[2] - gelev[5], 2 ))
        * (pow( 4 * pow ( sico.CSZ, 2 ) + pow( gelev[2] - gelev[5] , 2 ) + pow( gelev[1] - gelev[4] , 2 ), 0.5 )));  

    gkappaxy = 2 * sico.CSZ * ( gelev[3] + gelev[6] - gelev[8] - gelev[7] )
	/ ( pow( 4 * pow ( sico.CSZ, 2 ) + pow( gelev[2] - gelev[5], 2 ), 0.5 ) * pow( 4 * pow ( sico.CSZ, 2 ) 
        + pow( gelev[1] - gelev[4] , 2 ), 0.5 ) * pow( 4 * pow( sico.CSZ, 2 ) + pow( gelev[2] - gelev[5] , 2 ) 
        + pow( gelev[1] - gelev[4] , 2 ), 0.5 ));

    gkappay = 8 * sico.CSZ * ( gelev[1] + gelev[4] - 2 * gelev[0] )
	/ (( 4 * pow ( sico.CSZ, 2 ) + pow( gelev[1] - gelev[4], 2 )) 
        * (pow( 4 * pow ( sico.CSZ, 2 ) + pow( gelev[2] - gelev[5] , 2 ) + pow( gelev[1] - gelev[4] , 2 ), 0.5 ))); 

    for ( gl = 0; gl < sico.PMAX; gl++ ) {

        if ( gu[gl] != 0 || gv[gl] != 0 ) gkappau[gl] = ( pow( gu[gl], 2 ) * gkappax + 2 * gu[gl] * gv[gl] * gkappaxy + pow( gv[gl], 2 ) * gkappay );

        else gkappau[gl] = 0;
        if ( gkappau[gl] < -ggrav[4] ) gkappau[gl] = -ggrav[4];
    }

    return gkappau;
}

double *fvm ( double ghflow, double ghflow2, double ghflow3, double gvflowx, double gvflowx2, double gvflowx3, double gvflowy, double gvflowy2, double gvflowy3,
    struct ico sico, struct flow sflow ) { // function for virtual mass

    double *gvm, gh, galpha[3], ggamma[3], gu[3], gv[3];
    gvm = (double*) calloc( 21, sizeof(double));

    if ( sico.MODEL > 3 ) {

        gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
        gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

        gh = ghflow + ghflow2 + ghflow3;
        if ( gh > sico.HFLOWMIN ) {

            if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh; else galpha[0] = 0;
            if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh; else galpha[1] = 0;
            if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh; else galpha[2] = 0;

        } else { galpha[0] = 0; galpha[1] = 0; galpha[2] = 0; }

        if ( galpha[0] > 0  && galpha[2] > 0 ) ggamma[0] = sflow.RHO3 / sflow.RHO1; else ggamma[0] = 0;
        if ( galpha[0] > 0  && galpha[1] > 0 ) ggamma[1] = sflow.RHO2 / sflow.RHO1; else ggamma[1] = 0;
        if ( galpha[1] > 0  && galpha[2] > 0 ) ggamma[2] = sflow.RHO3 / sflow.RHO2; else ggamma[2] = 0;

        if ( galpha[2] > 0 && galpha[0] > 0 )
            gvm[9] = ( sflow.VM_n0 * ( sflow.VM_l + pow( galpha[0], sflow.VM_n )) - 1 ) / ( galpha[0] / galpha[2] + ggamma[0] ); else gvm[9] = 0;
        if ( galpha[1] > 0 && galpha[0] > 0 )
            gvm[10] = ( sflow.VM_n0 * ( sflow.VM_l + pow( galpha[0], sflow.VM_n )) - 1 ) / ( galpha[0] / galpha[1] + ggamma[1] ); else gvm[10] = 0;
        if ( galpha[2] > 0 && galpha[1] > 0 )
            gvm[11] = ( sflow.VM_n0 * ( sflow.VM_l + pow( galpha[1], sflow.VM_n )) - 1 ) / ( galpha[1] / galpha[2] + ggamma[2] ); else gvm[11] = 0;

        if ( galpha[0] > 0 ) {

            gvm[0] = ggamma[0] * gvm[9] * ( gu[2] - gu[0] ) + ggamma[1] * gvm[10] * ( gu[1] - gu[0] );
            gvm[1] = ggamma[0] * gvm[9] * ( pow( gu[2], 2 ) - pow( gu[0], 2 )) + ggamma[1] * gvm[10] * ( pow( gu[1], 2 ) - pow( gu[0], 2 ));
            gvm[2] = ggamma[0] * gvm[9] * ( gu[2] * gv[2] - gu[0] * gv[0] ) + ggamma[1] * gvm[10] * ( gu[1] * gv[1] - gu[0] * gv[0] );

            if ( fabs( gvm[1] ) > sflow.VM_fact * fabs( gu[2] - gu[0] ) && fabs( gvm[1] ) > sflow.VM_fact * fabs( gu[1] - gu[0] ))
                gvm[1] = sflow.VM_fact * fsign( fmax( fabs( gu[2] - gu[0] ), fabs( gu[1] - gu[0] )));
            if ( fabs( gvm[2] ) > sflow.VM_fact * fabs( gv[2] - gv[0] ) && fabs( gvm[2] ) > sflow.VM_fact * fabs( gv[1] - gv[0] ))
                gvm[2] = sflow.VM_fact * fsign( fmax( fabs( gv[2] - gv[0] ), fabs( gv[1] - gv[0] )));

            gvm[12] = ggamma[0] * gvm[9] * ( gv[2] - gv[0] ) + ggamma[1] * gvm[10] * ( gv[1] - gv[0] );
            gvm[13] = ggamma[0] * gvm[9] * ( pow( gv[2], 2 ) - pow( gv[0], 2 )) + ggamma[1] * gvm[10] * ( pow( gv[1], 2 ) - pow( gv[0], 2 ));
            gvm[14] = ggamma[0] * gvm[9] * ( gu[2] * gv[2] - gu[0] * gv[0] ) + ggamma[1] * gvm[10] * ( gu[1] * gv[1] - gu[0] * gv[0] );

            if ( fabs( gvm[13] ) > sflow.VM_fact * fabs( gv[2] - gv[0] ) && fabs( gvm[13] ) > sflow.VM_fact * fabs( gv[1] - gv[0] ))
                gvm[13] = sflow.VM_fact * fsign( fmax( fabs( gv[2] - gv[0] ), fabs( gv[1] - gv[0] )));
            if ( fabs( gvm[14] ) > sflow.VM_fact * fabs( gu[2] - gu[0] ) && fabs( gvm[14] ) > sflow.VM_fact * fabs( gu[1] - gu[0] ))
                gvm[14] = sflow.VM_fact * fsign( fmax( fabs( gu[2] - gu[0] ), fabs( gu[1] - gu[0] )));

        }
        else { gvm[0] = 0; gvm[1] = 0; gvm[2] = 0; gvm[12] = 0; gvm[13] = 0; gvm[14] = 0; }

        if ( galpha[1] > 0 ) {

            gvm[3] = ggamma[2] * gvm[11] * ( gu[2] - gu[1] ) - galpha[0] / galpha[1] * gvm[10] * ( gu[1] - gu[0] );
            gvm[4] = ggamma[2] * gvm[11] * ( pow( gu[2], 2 ) - pow( gu[1], 2 )) - galpha[0] / galpha[1] * gvm[10] * ( pow( gu[1], 2 ) - pow( gu[0], 2 ));
            gvm[5] = ggamma[2] * gvm[11] * ( gu[2] * gv[2] - gu[1] * gv[1] ) - galpha[0] / galpha[1] * gvm[10] * ( gu[1] * gv[1] - gu[0] * gv[0] );

            if ( fabs( gvm[4] ) > sflow.VM_fact * fabs( gu[2] - gu[1] ) && fabs( gvm[4] ) > sflow.VM_fact * fabs( gu[0] - gu[1] ))
                gvm[4] = sflow.VM_fact * fsign( fmax( fabs( gu[2] - gu[1] ), fabs( gu[0] - gu[1] )));
            if ( fabs( gvm[5] ) > sflow.VM_fact * fabs( gv[2] - gv[1] ) && fabs( gvm[5] ) > sflow.VM_fact * fabs( gv[0] - gv[1] ))
                gvm[5] = sflow.VM_fact * fsign( fmax( fabs( gv[2] - gv[1] ), fabs( gv[0] - gv[1] )));

            gvm[15] = ggamma[2] * gvm[11] * ( gv[2] - gv[1] ) - galpha[0] / galpha[1] * gvm[10] * ( gv[1] - gv[0] );
            gvm[16] = ggamma[2] * gvm[11] * ( pow( gv[2], 2 ) - pow( gv[1], 2 )) - galpha[0] / galpha[1] * gvm[10] * ( pow( gv[1], 2 ) - pow( gv[0], 2 ));
            gvm[17] = ggamma[2] * gvm[11] * ( gu[2] * gv[2] - gu[1] * gv[1] ) - galpha[0] / galpha[1] * gvm[10] * ( gu[1] * gv[1] - gu[0] * gv[0] );

            if ( fabs( gvm[16] ) > sflow.VM_fact * fabs( gv[2] - gv[1] ) && fabs( gvm[16] ) > sflow.VM_fact * fabs( gv[0] - gv[1] ))
                gvm[16] = sflow.VM_fact * fsign( fmax( fabs( gv[2] - gv[1] ), fabs( gv[0] - gv[1] )));
            if ( fabs( gvm[17] ) > sflow.VM_fact * fabs( gu[2] - gu[1] ) && fabs( gvm[17] ) > sflow.VM_fact * fabs( gu[0] - gu[1] ))
                gvm[17] = sflow.VM_fact * fsign( fmax( fabs( gu[2] - gu[1] ), fabs( gu[0] - gu[1] )));

        }
        else { gvm[3] = 0; gvm[4] = 0; gvm[5] = 0; gvm[15] = 0; gvm[16] = 0; gvm[17] = 0; }

        if ( galpha[2] > 0 ) {

            gvm[6] = galpha[0] / galpha[2] * gvm[9] * ( gu[2] - gu[0] ) + galpha[1] / galpha[2] * gvm[11] * ( gu[2] - gu[1] );
            gvm[7] = galpha[0] / galpha[2] * gvm[9] * ( pow( gu[2], 2 ) - pow( gu[0], 2 )) + galpha[1] / galpha[2] * gvm[11] * ( pow( gu[2], 2 ) - pow( gu[1], 2 ));
            gvm[8] = galpha[0] / galpha[2] * gvm[9] * ( gu[2] * gv[2] - gu[0] * gv[0] ) + galpha[1] / galpha[2] * gvm[11] * ( gu[2] * gv[2] - gu[1] * gv[1] );

            if ( fabs( gvm[7] ) > sflow.VM_fact * fabs( gu[2] - gu[0] ) && fabs( gvm[7] ) > sflow.VM_fact * fabs( gu[2] - gu[1] ))
                gvm[7] = sflow.VM_fact * fsign( fmax( fabs( gu[2] - gu[0] ), fabs( gu[2] - gu[1] )));
            if ( fabs( gvm[8] ) > sflow.VM_fact * fabs( gv[2] - gv[0] ) && fabs( gvm[8] ) > sflow.VM_fact * fabs( gv[2] - gv[1] ))
                gvm[8] = sflow.VM_fact * fsign( fmax( fabs( gv[2] - gv[0] ), fabs( gv[2] - gv[1] )));

            gvm[18] = galpha[0] / galpha[2] * gvm[9] * ( gv[2] - gv[0] ) + galpha[1] / galpha[2] * gvm[11] * ( gv[2] - gv[1] );
            gvm[19] = galpha[0] / galpha[2] * gvm[9] * ( pow( gv[2], 2 ) - pow( gv[0], 2 )) + galpha[1] / galpha[2] * gvm[11] * ( pow( gv[2], 2 ) - pow( gv[1], 2 ));
            gvm[20] = galpha[0] / galpha[2] * gvm[9] * ( gu[2] * gv[2] - gu[0] * gv[0] ) + galpha[1] / galpha[2] * gvm[11] * ( gu[2] * gv[2] - gu[1] * gv[1] );

            if ( fabs( gvm[19] ) > sflow.VM_fact * fabs( gv[2] - gv[0] ) && fabs( gvm[19] ) > sflow.VM_fact * fabs( gv[2] - gv[1] ))
                gvm[19] = sflow.VM_fact * fsign( fmax( fabs( gv[2] - gv[0] ), fabs( gv[2] - gv[1] )));
            if ( fabs( gvm[20] ) > sflow.VM_fact * fabs( gu[2] - gu[0] ) && fabs( gvm[20] ) > sflow.VM_fact * fabs( gu[2] - gu[1] ))
                gvm[20] = sflow.VM_fact * fsign( fmax( fabs( gu[2] - gu[0] ), fabs( gu[2] - gu[1] )));
        }
        else { gvm[6] = 0; gvm[7] = 0; gvm[8] = 0; gvm[18] = 0; gvm[19] = 0; gvm[20] = 0; }
    }

    return gvm;
}

double *fdrag ( double ghflow, double ghflow2, double ghflow3, struct ico sico, struct flow sflow ) { // function for drag

    double *gcdrag, gh, galpha[3], galphac[3], ggamma[3], gf, gg, gp, gsp;
    gcdrag = (double*) calloc( 3, sizeof(double));

    if ( sico.MODEL > 3 ) {

        gh = ghflow + ghflow2 + ghflow3;
        if ( gh > sico.HFLOWMIN ) {

            if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh; else galpha[0] = 0;
            if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh; else galpha[1] = 0;
            if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh; else galpha[2] = 0;

        } else { galpha[0] = 0; galpha[1] = 0; galpha[2] = 0; }

        if ( galpha[0] > 0 && galpha[2] > 0 ) ggamma[0] = sflow.RHO3 / sflow.RHO1; else ggamma[0] = 0;
        if ( galpha[0] > 0 && galpha[1] > 0 ) ggamma[1] = sflow.RHO2 / sflow.RHO1; else ggamma[1] = 0;
        if ( galpha[1] > 0 && galpha[2] > 0 ) ggamma[2] = sflow.RHO3 / sflow.RHO2; else ggamma[2] = 0;

        galphac[0] = galpha[0] * galpha[2];
        galphac[1] = galpha[0] * galpha[1];
        galphac[2] = galpha[1] * galpha[2];

        if ( galpha[0] > 0 && galpha[2] > 0 ) {

            gf = ggamma[0] / 180 * pow( galpha[2] / galpha[0], 3 ) * sflow.DRAG_rep;
            gg = pow( galpha[2], sflow.DRAG_m - 1 );
            gp = pow( galpha[0], sflow.DRAG_n );
            gsp = ( gp / galpha[0] + ( 1 - gp ) / galpha[2] ) * sflow.DRAG_k;
            gcdrag[0] = pow( 1, sflow.DRAG_je )
                * galphac[0] * ( 1 - ggamma[0] ) * sico.GRAVITY / pow( sflow.DRAG_vterm * ( gp * gf + ( 1 - gp ) * gg ) + gsp, sflow.DRAG_je );

        } else gcdrag[0] = 0;

        if ( galpha[0] > 0 && galpha[1] > 0 ) {

            gf = ggamma[0] / 180 * pow( galpha[1] / galpha[0], 3 ) * sflow.DRAG_rep;
            gg = pow( galpha[1], sflow.DRAG_m - 1 );
            gp = pow( galpha[0], sflow.DRAG_n );
            gsp = ( gp / galpha[0] + ( 1 - gp ) / galpha[1] ) * sflow.DRAG_k;
            gcdrag[1] = pow( 1, sflow.DRAG_je )
                * galphac[1] * ( 1 - ggamma[1] ) * sico.GRAVITY / pow( sflow.DRAG_vterm * ( gp * gf + ( 1 - gp ) * gg ) + gsp, sflow.DRAG_je );

        } else gcdrag[1] = 0;

        if ( galpha[1] > 0 && galpha[2] > 0 ) {

            gf = ggamma[1] / 180 * pow( galpha[2] / galpha[1], 3 ) * sflow.DRAG_rep;
            gg = pow( galpha[2], sflow.DRAG_m - 1 );
            gp = pow( galpha[1], sflow.DRAG_n );
            gsp = ( gp / galpha[1] + ( 1 - gp ) / galpha[2] ) * sflow.DRAG_k;
            gcdrag[2] = pow( 1, sflow.DRAG_je )
                * galphac[2] * ( 1 - ggamma[2] ) * sico.GRAVITY / pow( sflow.DRAG_vterm * ( gp * gf + ( 1 - gp ) * gg ) + gsp, sflow.DRAG_je );

        } else gcdrag[2] = 0;

    } else { gcdrag[0] = 0; gcdrag[1] = 0; gcdrag[2] = 0; }

    return gcdrag;
}

double *fgze ( double ghflow, double ghflow2, double ghflow3, double gvflowx, double gvflowx2, double gvflowx3, double gvflowy,
    double gvflowy2, double gvflowy3, double *gnh, double gghx, double gghy, double *ggux, double *ggvy, double *gguy, double *ggvx, 
    double *gnuss, double *gnvss, double *gnufs, double *gnvfs, double *gnuff, double *gnvff, double *ggalpha10, double *ggalpha20, double *ggalpha30, double gdx, double gdy, 
    double *ggrav, double gbetax, double gbetay, double *gcdrag, double gtsum, struct ico sico, struct flow sflow ) { // function for enhanced gravity

    int ga, gb, gj, gm, go;
    double *ggze, gh, gu[3], gv[3], gnu[3][9], gnv[3][9], ggalpha[3][9], gw[3], gww[3], guvw[3], galpha[3], ggamma[3], ggamma2[3], gdrag[3], gght, ggtlength[3];
    ggze = (double*) calloc( 12, sizeof(double));

    gh = ghflow + ghflow2 + ghflow3;

    if ( gh > sico.HFLOWMIN ) {

        if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh; else galpha[0] = 0;
        if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh; else galpha[1] = 0;
        if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh; else galpha[2] = 0;

    } else { galpha[0] = 0; galpha[1] = 0; galpha[2] = 0; }

    if ( galpha[0] > 0 && galpha[2] > 0 ) ggamma[0] = sflow.RHO3 / sflow.RHO1; else ggamma[0] = 0;
    if ( galpha[0] > 0 && galpha[1] > 0 ) ggamma[1] = sflow.RHO2 / sflow.RHO1; else ggamma[1] = 0;
    if ( galpha[1] > 0 && galpha[2] > 0 ) ggamma[2] = sflow.RHO3 / sflow.RHO2; else ggamma[2] = 1;

    if ( galpha[0] > 0 && galpha[2] > 0 ) ggamma2[0] = sflow.RHO3 / sflow.RHO1; else ggamma2[0] = 1;
    if ( galpha[0] > 0 && galpha[1] > 0 ) ggamma2[1] = sflow.RHO2 / sflow.RHO1; else ggamma2[1] = 1;
    if ( galpha[1] > 0 && galpha[2] > 0 ) ggamma2[2] = sflow.RHO3 / sflow.RHO2; else ggamma2[2] = 1;

    if ( sico.MODEL > 3 && sico.NOOSC == 2 && gtsum > 1 ) {

        for( gj=0; gj<9; gj++ ) { 

            gnu[0][gj] = gnuss[gj]; gnu[1][gj] = gnufs[gj]; gnu[2][gj] = gnuff[gj]; gnv[0][gj] = gnvss[gj]; gnv[1][gj] = gnvfs[gj]; gnv[2][gj] = gnvff[gj];
            ggalpha[0][gj] = ggalpha10[gj]; ggalpha[1][gj] = ggalpha20[gj]; ggalpha[2][gj] = ggalpha30[gj];
        }

        gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
        gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

        gght = -0.5 / gdx * ( gnh[2] * ( ggalpha[0][2] * gnu[0][2] + ggalpha[1][2] * gnu[1][2] + ggalpha[2][2] * gnu[2][2] ) - gnh[5] * ( ggalpha[0][5] * gnu[0][5] 
            + ggalpha[1][5] * gnu[1][5] + ggalpha[2][5] * gnu[2][5] )) - 0.5 / gdy * ( gnh[1] * ( ggalpha[0][1] * gnv[0][1] + ggalpha[1][1] * gnv[1][1] 
            + ggalpha[2][1] * gnv[2][1] ) - gnh[4] * ( ggalpha[0][4] * gnv[0][4] + ggalpha[1][4] * gnv[1][4] + ggalpha[2][4] * gnv[2][4] ));

        for ( go = 0; go < 3; go++ ) {

            ggtlength[go] = -0.5 * ( -2 * gh * ggvx[go] * gguy[go] + gght * ( ggux[go] + ggvy[go] ) + ( gu[go] * gghx + gv[go] * gghy ) * ( ggux[go] + ggvy[go] ));
            gw[go] = gu[go] * tan(gbetax) + gv[go] * tan(gbetay);
            gww[go] = (gu[go] * tan(gbetax) + gv[go] * tan( gbetay )) - ( ggux[go] + ggvy[go] ) * gh / 2;
            guvw[go] = pow( pow( gu[go], 2 ) + pow( gv[go], 2 ) + pow( gw[go], 2 ), 0.5 );
        }

        gdrag[0] = gcdrag[0] * pow( fabs( guvw[2] - guvw[0] ), sflow.DRAG_je - 1 );
        gdrag[1] = gcdrag[1] * pow( fabs( guvw[1] - guvw[0] ), sflow.DRAG_je - 1 );
        gdrag[2] = gcdrag[2] * pow( fabs( guvw[2] - guvw[1] ), sflow.DRAG_je - 1 );

        ga = 0; gb = 0;

        for ( gm=0; gm<3; gm++ ) {

            if ( galpha[0] > 0 ) ggze[gb+0] = ggrav[ga] * ( 1 - ggamma[0] )
                + ggtlength[0] - 1 / galpha[0] * ( gdrag[0] * ( gww[2] - gww[0] ) + gdrag[1] * ( gww[1] - gww[0] ));
            else ggze[gb+0] = ggrav[ga];

            if ( galpha[1] > 0 ) ggze[gb+1] = ggrav[ga] * ggamma[2] 
                + ggtlength[1] - 1 / galpha[1] * ( -1 / ggamma2[1] * gdrag[1] * ( gww[1] - gww[0] ) + gdrag[2] * ( gww[2] - gww[1] ));
            else ggze[gb+1] = ggrav[ga];

            if ( galpha[2] > 0 ) ggze[gb+2] = ggrav[ga] 
                + ggtlength[2] + 1 / galpha[2] * ( 1 / ggamma2[0] * gdrag[0] * ( gww[2] - gww[0] ) + 1 / ggamma2[2] * gdrag[2] * ( gww[2] - gww[1] ));
            else ggze[gb+2] = ggrav[ga];

            ggze[gb+3] = ggze[gb+0] * ggamma2[0] / ( 1 - ggamma[0] );

            if ( gm == 1 ) { ga = 1; gb = 4; }
            else if ( gm == 2 ) { ga = 4; gb = 8; }
        }

    } else { 

        ggze[0] = ggrav[0] * ( 1 - ggamma[0] );
        ggze[1] = ggrav[0] * ggamma[2];
        ggze[2] = ggrav[0];
        ggze[3] = ggrav[0] * ggamma2[0];
        ggze[4] = ggrav[1] * ( 1 - ggamma[0] );
        ggze[5] = ggrav[1] * ggamma[2];
        ggze[6] = ggrav[1];
        ggze[7] = ggrav[1] * ggamma2[0];
        ggze[8] = ggrav[4] * ( 1 - ggamma[0] );
        ggze[9] = ggrav[4] * ggamma[2];
        ggze[10] = ggrav[4];
        ggze[11] = ggrav[4] * ggamma2[0];
    }

    return ggze;
}

double *fdisp ( double ghflow, double ghflow2, double ghflow3, double gvflowx, double gvflowx2, double gvflowx3, double gvflowy, double gvflowy2, double gvflowy3, 
    double *ggux, double *ggvy, double *gguy, double *ggvx, double gbetax, double gbetay, double *gkx, double *gky, double *gvm, double *gcdrag, double gtsum, 
    struct ico sico, struct flow sflow ) { // function for dispersion

    int go;
    double *gdisp, gh, gu[3], gv[3], gw[3], guvw[3], guu[3], galpha[3], galphassfs, galphassff, galphafsff,
        ggamma[3], ggamma2[3], gcvm[3], gdrag[3], ggtlength[3];
    gdisp = (double*) calloc( 6, sizeof(double));

    if ( sico.MODEL > 3 && sico.NOOSC == 2 && gtsum > 1 ) {

        gh = ghflow + ghflow2 + ghflow3;
        gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
        gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

        for ( go = 0; go < 3; go++ ) {

            gw[go] = gu[go] * tan(gbetax) + gv[go] * tan(gbetay);
            guvw[go] = pow( pow( gu[go], 2 ) + pow( gv[go], 2 ) + pow( gw[go], 2 ), 0.5 );
            guu[go] = ggux[go] + ggvy[go];
            ggtlength[go] = 2 * pow( guu[go], 2 ) - 2 * ggux[go] * ggvy[go] + 2 * ggvx[go] * gguy[go];
        }

        if ( gh > sico.HFLOWMIN ) {

            if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh; else galpha[0] = 0;
            if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh; else galpha[1] = 0;
            if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh; else galpha[2] = 0;

        } else { galpha[0] = 0; galpha[1] = 0; galpha[2] = 0; }

        if ( galpha[1] != 0 ) galphassfs = galpha[0] / galpha[1]; else galphassfs = 0;
        if ( galpha[2] != 0 ) galphassff = galpha[0] / galpha[2]; else galphassff = 0;
        if ( galpha[2] != 0 ) galphafsff = galpha[1] / galpha[2]; else galphafsff = 0;

        if ( galpha[0] > 0 && galpha[2] > 0 ) ggamma[0] = sflow.RHO3 / sflow.RHO1; else ggamma[0] = 0;
        if ( galpha[0] > 0 && galpha[1] > 0 ) ggamma[1] = sflow.RHO2 / sflow.RHO1; else ggamma[1] = 0;
        if ( galpha[1] > 0 && galpha[2] > 0 ) ggamma[2] = sflow.RHO3 / sflow.RHO2; else ggamma[2] = 0;

        if ( galpha[0] > 0 && galpha[2] > 0 ) ggamma2[0] = sflow.RHO3 / sflow.RHO1; else ggamma2[0] = 1;
        if ( galpha[0] > 0 && galpha[1] > 0 ) ggamma2[1] = sflow.RHO2 / sflow.RHO1; else ggamma2[1] = 1;
        if ( galpha[1] > 0 && galpha[2] > 0 ) ggamma2[2] = sflow.RHO3 / sflow.RHO2; else ggamma2[2] = 1;

        gcvm[0] = gvm[9]; gcvm[1] = gvm[10]; gcvm[2] = gvm[11];

        gdrag[0] = gcdrag[0] * pow( fabs( guvw[2] - guvw[0] ), sflow.DRAG_je - 1 );
        gdrag[1] = gcdrag[1] * pow( fabs( guvw[1] - guvw[0] ), sflow.DRAG_je - 1 );
        gdrag[2] = gcdrag[2] * pow( fabs( guvw[2] - guvw[1] ), sflow.DRAG_je - 1 );

        if ( galpha[0] > 0 )
            gdisp[0] = gkx[0] * gh * gh / 12 * (( ggtlength[0] - ggamma[0] * gcvm[0] * ( pow( guu[2], 2 ) - pow(guu[0], 2 )) - ggamma[1] * gcvm[1]
                * ( pow( guu[1], 2 ) - pow(guu[0], 2 ))))
                + 1 / galpha[0] * gkx[0] * gh * gh / 6 * ( gdrag[0] * ( guu[2] - guu[0] ) + gdrag[1] * ( guu[1] - guu[0] ));
        else gdisp[0] = 0;

        if ( galpha[1] > 0 )
            gdisp[1] = gh * gh / 12 * (( ggtlength[1] - ggamma[2] * gcvm[2] * ( pow( guu[2], 2 ) - pow(guu[1], 2 )) + galphassfs * gcvm[1]
                * ( pow( guu[1], 2 ) - pow(guu[0], 2 ))))
                + 1 / galpha[1] * gh * gh / 6 * ( - 1 / ggamma2[1] * gdrag[1] * ( guu[1] - guu[0] ) + gdrag[2] * ( guu[2] - guu[1] ));
        else gdisp[1] = 0;

        if ( galpha[2] > 0 )
            gdisp[2] = gh * gh / 12 * (( ggtlength[2] + galphassff * gcvm[0] * ( pow( guu[2], 2 ) - pow(guu[0], 2 )) + galphafsff * gcvm[2]
                * ( pow( guu[2], 2 ) - pow(guu[1], 2 ))))
                - 1 / galpha[2] * gh * gh / 6 * ( 1 / ggamma2[0] * gdrag[0] * ( guu[2] - guu[0] ) + 1 / ggamma2[2] * gdrag[2] * ( guu[2] - guu[1] ));
        else gdisp[2] = 0;

        if ( galpha[0] > 0 )
            gdisp[3] = gky[0] * gh * gh / 12 * (( ggtlength[0] - ggamma[0] * gcvm[0] * ( pow( guu[2], 2 ) - pow(guu[0], 2 )) - ggamma[1] * gcvm[1]
                * ( pow( guu[1], 2 ) - pow(guu[0], 2 ))))
                + 1 / galpha[0] * gky[0] * gh * gh / 6 * ( gdrag[0] * ( guu[2] - guu[0] ) + gdrag[1] * ( guu[1] - guu[0] ));
        else gdisp[3] = 0;

        if ( galpha[1] > 0 )
            gdisp[4] = gh * gh / 12 * (( ggtlength[1] - ggamma[2] * gcvm[2] * ( pow( guu[2], 2 ) - pow(guu[1], 2 )) + galphassfs * gcvm[1]
                * ( pow( guu[1], 2 ) - pow(guu[0], 2 ))))
                + 1 / galpha[1] * gh * gh / 6 * ( - 1 / ggamma2[1] * gdrag[1] * ( guu[1] - guu[0] ) + gdrag[2] * ( guu[2] - guu[1] ));
        else gdisp[4] = 0;

        if ( galpha[2] > 0 )
            gdisp[5] = gh * gh / 12 * (( ggtlength[2] + galphassff * gcvm[0] * ( pow( guu[2], 2 ) - pow(guu[0], 2 )) + galphafsff * gcvm[2]
                * ( pow( guu[2], 2 ) - pow(guu[1], 2 ))))
                - 1 / galpha[2] * gh * gh / 6 * ( 1 / ggamma2[0] * gdrag[0] * ( guu[2] - guu[0] ) + 1 / ggamma2[2] * gdrag[2] * ( guu[2] - guu[1] ));
        else gdisp[5] = 0;

    } else { gdisp[0] = 0; gdisp[1] = 0; gdisp[2] = 0; gdisp[3] = 0; gdisp[4] = 0; gdisp[5] = 0; }

    return gdisp;
}

double *ff ( double ghflow, double ghflow2, double ghflow3, double gvflowx, double gvflowx2, double gvflowx3, double gvflowy,
    double gvflowy2, double gvflowy3, double *gkx, double *ggz, double *gkappau, double *gvm, double *gdisp, 
    struct ico sico ) { // function for fluxes in x direction

    int gp;
    double *ggf, gh, gu[3], gv[3], galpha[3], gbetax[3];
    ggf = (double*) calloc( sico.NVECTMIN, sizeof(double));

    gh = ghflow + ghflow2 + ghflow3;
    gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
    gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

    if ( gh > sico.HFLOWMIN ) {

        if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh; else galpha[0] = 0;
        if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh; else galpha[1] = 0;
        if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh; else galpha[2] = 0;

    } else { galpha[0] = 0; galpha[1] = 0; galpha[2] = 0; }


    // Hydraulic pressure coefficients

    for ( gp = 0; gp < sico.PMAX; gp++ ) {

        if ( sico.PHASES[gp] == 1 ) gbetax[gp] = gkx[gp] * ( ggz[gp] + 0*gkappau[gp] );
        else gbetax[gp] = 0;
    }


    // Flux terms

    if ( sico.MODEL == 0 ) {

        ggf[0] = gh * gu[0];
        ggf[1] = gh * pow( gu[0], 2 ) + gbetax[0] * pow( gh, 2 ) / 2;
        ggf[2] = gh * gu[0] * gv[0];

    } else {

        if ( galpha[0] > 0 ) {

            ggf[0] = galpha[0] * gh * gu[0];
            ggf[1] = galpha[0] * gh * ( pow( gu[0], 2 ) - gvm[1] + gbetax[0] * gh / 2 + gdisp[0] );
            ggf[2] = galpha[0] * gh * ( gu[0] * gv[0] - gvm[2] );

        } else { ggf[0] = 0; ggf[1] = 0; ggf[2] = 0; }
    }

    if ( sico.MODEL > 3 ) {

        if ( galpha[1] > 0 ) {

            ggf[3] = galpha[1] * gh * gu[1];
            ggf[4] = galpha[1] * gh * ( pow( gu[1], 2 ) - gvm[4] + gdisp[1] );
            ggf[5] = galpha[1] * gh * ( gu[1] * gv[1] - gvm[5] );

        } else { ggf[3] = 0; ggf[4] = 0; ggf[5] = 0; }
    }

    if ( sico.MODEL == 7 ) {

        if ( galpha[2] > 0 ) {

            ggf[6] = galpha[2] * gh * gu[2];
            ggf[7] = galpha[2] * gh * ( pow( gu[2], 2 ) + gvm[7] + gdisp[2] );
            ggf[8] = galpha[2] * gh * ( gu[2] * gv[2] + gvm[8] ); 

        } else { ggf[6] = 0; ggf[7] = 0; ggf[8] = 0; }
    }

    return ggf;
}

double *fg ( double ghflow, double ghflow2, double ghflow3, double gvflowx, double gvflowx2, double gvflowx3, double gvflowy, 
    double gvflowy2, double gvflowy3, double *gky, double *ggz, double *gkappau, double *gvm, double *gdisp, 
    struct ico sico ) { // function for fluxes in y direction

    int gp;
    double *ggg, gh, gu[3], gv[3], galpha[3], gbetay[3];
    ggg = (double*) calloc( sico.NVECTMIN, sizeof(double));

    gh = ghflow + ghflow2 + ghflow3;

    gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
    gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

    if ( gh > sico.HFLOWMIN ) {

        if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh; else galpha[0] = 0;
        if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh; else galpha[1] = 0;
        if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh; else galpha[2] = 0;

    } else { galpha[0] = 0; galpha[1] = 0; galpha[2] = 0; }


    // Hydraulic pressure coefficients

    for ( gp = 0; gp < sico.PMAX; gp++ ) {

        if ( sico.PHASES[gp] == 1 ) gbetay[gp] = gky[gp] * ( ggz[gp+4] + 0*gkappau[gp] );
        else gbetay[gp] = 0;
    }
   

    // Flux terms

    if ( sico.MODEL == 0 ) {

        ggg[0] = gh * gv[0];
        ggg[1] = gh * gu[0] * gv[0];
        ggg[2] = gh * pow( gv[0], 2 ) + gbetay[0] * pow( gh, 2 ) / 2;

    } else {

        if ( galpha[0] > 0 ) {

            ggg[0] = galpha[0] * gh * gv[0];
            ggg[1] = galpha[0] * gh * ( gu[0] * gv[0] - gvm[14] );
            ggg[2] = galpha[0] * gh * ( pow( gv[0], 2 ) - gvm[13] + gbetay[0] * gh / 2 + gdisp[3] );

        } else { ggg[0] = 0; ggg[1] = 0; ggg[2] = 0; }
    }

    if ( sico.MODEL > 3 ) {

        if ( galpha[1] > 0 ) {

            ggg[3] = galpha[1] * gh * gv[1];
            ggg[4] = galpha[1] * gh * ( gu[1] * gv[1] - gvm[17] ); 
            ggg[5] = galpha[1] * gh * ( pow( gv[1], 2 ) - gvm[16] + gdisp[4] );

        } else { ggg[3] = 0; ggg[4] = 0; ggg[5] = 0; }
    }

    if ( sico.MODEL == 7 ) {

        if ( galpha[2] > 0 ) {

            ggg[6] = galpha[2] * gh * gv[2];
            ggg[7] = galpha[2] * gh * ( gu[2] * gv[2] + gvm[20] );
            ggg[8] = galpha[2] * gh * ( pow( gv[2], 2 ) + gvm[19] + gdisp[5] );

        } else { ggg[6] = 0; ggg[7] = 0; ggg[8] = 0; }
    }

    return ggg;
}

double *fs ( double *gh, double ghflow, double ghflow2, double ghflow3, double gvflowx, double gvflowx2, double gvflowx3,
    double gvflowy, double gvflowy2, double gvflowy3, double gghx, double gghy, double ggalphassx, double ggalphassy, 
    double ggalphafsx, double ggalphafsy, double ggalphaffx, double ggalphaffy, double *ggrav, double *ggz,
    double gdx, double gdy, double *gkappau, double *gcdrag, struct ico sico, struct flow sflow ) { // function for source terms

    int gl, go;
    double *ggs, gu[3], gv[3], gw[3], galpha[3], ggalphax[3], ggalphay[3], ggamma0, ggamma1, ggamma2,
        gpbx[3], gpby[3], guvw[3], gdragx[3], gdragy[3], gcompx[3], gcompy[3], guratio, gvratio;
    ggs = (double*) calloc( sico.NVECTMIN, sizeof(double));

    gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
    gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

    if ( sico.NOOSC == 2 ) {

        gw[0] = gu[0] * tan(gdx) + gv[0] * tan(gdy);
        gw[1] = gu[1] * tan(gdx) + gv[1] * tan(gdy);
        gw[2] = gu[2] * tan(gdx) + gv[2] * tan(gdy);

    } else { gw[0] = 0; gw[1] = 0; gw[2] = 0; }

    if ( gh[0] > sico.HFLOWMIN ) {

        if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh[0]; else galpha[0] = 0;
        if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh[0]; else galpha[1] = 0;
        if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh[0]; else galpha[2] = 0;

    } else { galpha[0] = 0; galpha[1] = 0; galpha[2] = 0; }

    if ( galpha[0] > 0  && galpha[2] > 0 ) ggamma0 = sflow.RHO3 / sflow.RHO1; else ggamma0 = 1;
    if ( galpha[0] > 0  && galpha[1] > 0 ) ggamma1 = sflow.RHO2 / sflow.RHO1; else ggamma1 = 1;
    if ( galpha[1] > 0  && galpha[2] > 0 ) ggamma2 = sflow.RHO3 / sflow.RHO2; else ggamma2 = 1;

    for ( go = 0; go < 3; go++ ) guvw[go] = pow( pow( gu[go], 2 ) + pow( gv[go], 2 ) + pow( gw[go], 2 ), 0.5 );

    if ( gh[2] > sico.HFLOWMIN && gh[5] > sico.HFLOWMIN ) ggalphax[0] = ggalphassx; else ggalphax[0] = 0;
    if ( gh[2] > sico.HFLOWMIN && gh[5] > sico.HFLOWMIN ) ggalphax[1] = ggalphafsx; else ggalphax[1] = 0;
    if ( gh[2] > sico.HFLOWMIN && gh[5] > sico.HFLOWMIN ) ggalphax[2] = ggalphaffx; else ggalphax[2] = 0;
    if ( gh[1] > sico.HFLOWMIN && gh[4] > sico.HFLOWMIN ) ggalphay[0] = ggalphassy; else ggalphay[0] = 0;
    if ( gh[1] > sico.HFLOWMIN && gh[4] > sico.HFLOWMIN ) ggalphay[1] = ggalphafsy; else ggalphay[1] = 0;
    if ( gh[1] > sico.HFLOWMIN && gh[4] > sico.HFLOWMIN ) ggalphay[2] = ggalphaffy; else ggalphay[2] = 0;


    // Effective basal pressures and phase-dependent components

    if ( sico.MODEL > 0 ) {

        for ( gl = 0; gl < sico.PMAX; gl++ ) {

            if ( sico.PHASES[gl] == 1 ) {

                gcompx[gl] = galpha[gl] * ( ggrav[2] - ( ggz[3] + 0*gkappau[gl] ) * gghx );
                gcompy[gl] = galpha[gl] * ( ggrav[3] - ( ggz[7] + 0*gkappau[gl] ) * gghy );

            } else if ( sico.PHASES[gl] == 2 ) {

                gpbx[gl] = ggz[gl] + 0*gkappau[gl];
                gpby[gl] = ggz[gl+4] + 0*gkappau[gl];

                gcompx[gl] = galpha[gl] * ( ggrav[2] - ggz[gl] * gghx - ( -0.5 * gpbx[gl] * gh[0] / galpha[gl] * ggalphax[gl] ));
                gcompy[gl] = galpha[gl] * ( ggrav[3] - ggz[gl+4] * gghy - ( -0.5 * gpby[gl] * gh[0] / galpha[gl] * ggalphay[gl] ));

            } else if ( sico.PHASES[gl] == 3 ) {

                gpbx[gl] = ggz[gl] + 0*gkappau[gl];
                gpby[gl] = ggz[gl+4] + 0*gkappau[gl];

                gcompx[gl] = galpha[gl] * ( ggrav[2] - ggz[gl] * gghx - ( -0.5 * gpbx[gl] * gh[0] / galpha[gl] * ggalphax[gl] ));
                gcompy[gl] = galpha[gl] * ( ggrav[3] - ggz[gl+4] * gghy - ( -0.5 * gpby[gl] * gh[0] / galpha[gl] * ggalphay[gl] ));
            }
        }
    }


    // Drag terms

    if ( sico.MODEL > 3 ) {

        gdragx[0] = gcdrag[0] * ( gu[2] - gu[0] ) * pow( fabs( guvw[2] - guvw[0] ), sflow.DRAG_je - 1 );
        gdragx[1] = gcdrag[1] * ( gu[1] - gu[0] ) * pow( fabs( guvw[1] - guvw[0] ), sflow.DRAG_je - 1 );
        gdragx[2] = gcdrag[2] * ( gu[2] - gu[1] ) * pow( fabs( guvw[2] - guvw[1] ), sflow.DRAG_je - 1 );

        gdragy[0] = gcdrag[0] * ( gv[2] - gv[0] ) * pow( fabs( guvw[2] - guvw[0] ), sflow.DRAG_je - 1 );
        gdragy[1] = gcdrag[1] * ( gv[1] - gv[0] ) * pow( fabs( guvw[1] - guvw[0] ), sflow.DRAG_je - 1 );
        gdragy[2] = gcdrag[2] * ( gv[2] - gv[1] ) * pow( fabs( guvw[2] - guvw[1] ), sflow.DRAG_je - 1 );

    } else { gdragx[0] = 0; gdragx[1] = 0; gdragx[2] = 0; gdragy[0] = 0; gdragy[1] = 0; gdragy[2] = 0; }


    // Source terms

    if ( sico.MODEL == 0 ) {

        if ( guvw[0] > 0 ) guratio = gu[0] / guvw[0]; else guratio = 0;
        if ( guvw[0] > 0 ) gvratio = gv[0] / guvw[0]; else gvratio = 0;

        ggs[0] = 0; 
        ggs[1] = ( ggrav[2] * gh[0] - guratio * ( tan( sflow.DELTA[0] ) * ( ggrav[0] ) * gh[0] + sico.GRAVITY * pow( guvw[0], 2 ) / sflow.TUFRI ));
        ggs[2] = ( ggrav[3] * gh[0] - gvratio * ( tan( sflow.DELTA[0] ) * ( ggrav[1] ) * gh[0] + sico.GRAVITY * pow( guvw[0], 2 ) / sflow.TUFRI ));

    } else {

        if ( galpha[0] > 0 ) {
 
            ggs[0] = 0;
            ggs[1] = gh[0] * ( gcompx[0] + gdragx[0] + gdragx[1] );
            ggs[2] = gh[0] * ( gcompy[0] + gdragy[0] + gdragy[1] );

        } else { ggs[0] = 0; ggs[1] = 0; ggs[2] = 0; }
    }

    if ( sico.MODEL > 3 ) {

        if ( galpha[1] > 0 ) {

            ggs[3] = 0; 
            ggs[4] = gh[0] * ( gcompx[1] - 1 / ggamma1 * gdragx[1] + gdragx[2] );
            ggs[5] = gh[0] * ( gcompy[1] - 1 / ggamma1 * gdragy[1] + gdragy[2] );

        } else { ggs[3] = 0; ggs[4] = 0; ggs[5] = 0; }
    }

    if ( sico.MODEL == 7 ) {

        if ( galpha[2] > 0 ) {

            ggs[6] = 0; 
            ggs[7] = gh[0] * ( gcompx[2] - 1 / ggamma0 * gdragx[0] - 1 / ggamma2 * gdragx[2] );
            ggs[8] = gh[0] * ( gcompy[2] - 1 / ggamma0 * gdragy[0] - 1 / ggamma2 * gdragy[2] );

        } else { ggs[6] = 0; ggs[7] = 0; ggs[8] = 0; }
    }

    return ggs;
}

double *fd ( double *gh, double gghx, double gghy, double *gnuss, double *gnvss, double *gnufs, double *gnvfs,
    double *gnuff, double *gnvff, double *ggalpha10, double *ggalpha20, double *ggalpha30, double gbetax, double gbetay, double *ggz, double gdx, double gdy,
    double *gd0, double *gkappau, double gekin, struct ico sico, struct flow sflow ) { // function for deceleration terms

    int gg, gj, gl, go, gq;
    double *ggd, gd[3][6], gw[3][6], guvw[3][6], ggux[3][6], gguy[3][6], ggvx[3][6], ggvy[3][6], ggalpha[3][9], ggalphax[3][6], 
        ggalphay[3][6], gfricx[3], gfricy[3], gvisx[3], gvisy[3], gcambdragx, gcambdragy, gp[3], gpx[3], gpy[3], gtauy[3], gnye[3][6], gcuf[3][6], gguzb[3][6], gcvf[3][6], 
        ggvzb[3][6], gtaunnx[3], gtaunny[3], gflufri[3], gdelta, gphi, gny[6], difuxx[6], difvxx[6], difuyx[6], difvyy[6], difuyy[6], difvxy[6], difax, difay, gpbx[3], gpby[3], 
        guratio[3], gvratio[3], gnu[3][9], gnv[3][9], gxterm1, gyterm1, gxterm2, gyterm2, gxterm3, gyterm3;

    ggd = (double*) calloc( sico.NVECTMIN+18, sizeof(double));

    if ( sico.MODEL > 0 ) {

        for( gj=0; gj<9; gj++ ) { 

            gnu[0][gj] = gnuss[gj]; gnu[1][gj] = gnufs[gj]; gnu[2][gj] = gnuff[gj]; gnv[0][gj] = gnvss[gj]; gnv[1][gj] = gnvfs[gj]; gnv[2][gj] = gnvff[gj];
            ggalpha[0][gj] = ggalpha10[gj]; ggalpha[1][gj] = ggalpha20[gj]; ggalpha[2][gj] = ggalpha30[gj];
        }

        for ( gq=0; gq<6; gq++ ) {

            gd[0][gq] = gd0[3*gq]; gd[1][gq] = gd0[3*gq+1]; gd[2][gq] = gd0[3*gq+2];

            for ( go = 0; go < 3; go++ ) {

                if ( sico.NOOSC == 2 ) gw[0][gq] = gnu[go][gq] * tan(gbetax) + gnv[go][gq] * tan(gbetay); //!!!CHECK neighbourhood of gbeta needed
                else gw[go][gq] = 0;

                guvw[go][gq] = pow( pow( gnu[go][gq], 2 ) + pow( gnv[go][gq], 2 ) + pow( gw[go][gq], 2 ), 0.5 );
                if ( guvw[go][0] > 0 ) guratio[go] = gnu[go][0] / guvw[go][0]; else guratio[go] = 0;
                if ( guvw[go][0] > 0 ) gvratio[go] = gnv[go][0] / guvw[go][0]; else gvratio[go] = 0;
            }
        }


        // Gradients of velocities and fractions

        for ( go=0; go<3; go++ ) {

            ggux[go][0] = ( gnu[go][2] - gnu[go][5] ) / ( 2 * gdx );
            gguy[go][0] = ( gnu[go][1] - gnu[go][4] ) / ( 2 * gdy );
            ggvx[go][0] = ( gnv[go][2] - gnv[go][5] ) / ( 2 * gdx );
            ggvy[go][0] = ( gnv[go][1] - gnv[go][4] ) / ( 2 * gdy );

            ggux[go][1] = ( gnu[go][3] - gnu[go][8] ) / ( 2 * gdx );
            gguy[go][1] = ( gnu[go][1] - gnu[go][0] ) / ( 1 * gdy );
            ggvx[go][1] = ( gnv[go][3] - gnv[go][8] ) / ( 2 * gdx );
            ggvy[go][1] = ( gnv[go][1] - gnv[go][0] ) / ( 1 * gdy );

            ggux[go][2] = ( gnu[go][2] - gnu[go][0] ) / ( 1 * gdx );
            gguy[go][2] = ( gnu[go][3] - gnu[go][7] ) / ( 2 * gdy );
            ggvx[go][2] = ( gnv[go][2] - gnv[go][0] ) / ( 1 * gdx );
            ggvy[go][2] = ( gnv[go][3] - gnv[go][7] ) / ( 2 * gdy );

            ggux[go][4] = ( gnu[go][7] - gnu[go][6] ) / ( 2 * gdx );
            gguy[go][4] = ( gnu[go][0] - gnu[go][4] ) / ( 1 * gdy );
            ggvx[go][4] = ( gnv[go][7] - gnv[go][6] ) / ( 2 * gdx );
            ggvy[go][4] = ( gnv[go][0] - gnv[go][4] ) / ( 1 * gdy );

            ggux[go][5] = ( gnu[go][0] - gnu[go][5] ) / ( 1 * gdx );
            gguy[go][5] = ( gnu[go][8] - gnu[go][6] ) / ( 2 * gdy );
            ggvx[go][5] = ( gnv[go][0] - gnv[go][5] ) / ( 1 * gdx );
            ggvy[go][5] = ( gnv[go][8] - gnv[go][6] ) / ( 2 * gdy );

            ggalphax[go][0] = ( ggalpha[go][2] - ggalpha[go][5] ) / ( 2 * gdx );
            ggalphay[go][0] = ( ggalpha[go][1] - ggalpha[go][4] ) / ( 2 * gdy );

            ggalphax[go][1] = ( ggalpha[go][3] - ggalpha[go][8] ) / ( 2 * gdx );
            ggalphay[go][1] = ( ggalpha[go][1] - ggalpha[go][0] ) / ( 1 * gdy );

            ggalphax[go][2] = ( ggalpha[go][2] - ggalpha[go][0] ) / ( 1 * gdx );
            ggalphay[go][2] = ( ggalpha[go][3] - ggalpha[go][7] ) / ( 2 * gdy );

            ggalphax[go][4] = ( ggalpha[go][7] - ggalpha[go][6] ) / ( 2 * gdx );
            ggalphay[go][4] = ( ggalpha[go][0] - ggalpha[go][4] ) / ( 1 * gdy );

            ggalphax[go][5] = ( ggalpha[go][0] - ggalpha[go][5] ) / ( 1 * gdx );
            ggalphay[go][5] = ( ggalpha[go][8] - ggalpha[go][6] ) / ( 2 * gdy );
        }


        // Effective ambient drag coefficients

        if ( gh[0] > sico.HFLOWMIN ) {

            if ( fsign( gnu[0][0] + gnu[1][0] + gnu[2][0] ) == fsign( gghx )) gcambdragx = 0;
            else gcambdragx = gdx * fabs( gghx ) / gh[0] * sflow.AMBDRAG;
            gcambdragx = fmin( sflow.AMBDRAG, gcambdragx );

            if ( fsign( gnv[0][0] + gnv[1][0] + gnv[2][0] ) == fsign( gghy )) gcambdragy = 0;
            else gcambdragy = gdy * fabs( gghy ) / gh[0] * sflow.AMBDRAG;
            gcambdragy = fmin( sflow.AMBDRAG, gcambdragy );

        } else { gcambdragx = 0; gcambdragy = 0; }


        // Phase-dependent components

        for ( gl = 0; gl < sico.PMAX; gl++ ) {

            if ( sico.PHASES[gl] == 1 ) {

                gdelta = sflow.DELTA[gl] * pow( ggalpha[gl][0], sflow.FRI_exp );
                if ( sico.DYNFRIC == 1 ) gdelta = fdynfric( gdelta, sico.FRICMIN, gekin, sico );

                gpbx[gl] = ggz[gl] + gkappau[gl];
                gpby[gl] = ggz[gl+4] + gkappau[gl];

                gfricx[gl] = guratio[gl] * tan( gdelta ) * gpbx[gl];
                gfricy[gl] = gvratio[gl] * tan( gdelta ) * gpby[gl];


                gflufri[gl] = 0;
                gvisx[gl] = 0;
                gvisy[gl] = 0;

            } else if ( sico.PHASES[gl] == 2 ) {

                gfricx[gl] = 0;
                gfricy[gl] = 0;

                gdelta = sflow.DELTA[gl] * pow( ggalpha[gl][0], sflow.FRI_exp );
                if ( sico.DYNFRIC == 1 ) gdelta = fdynfric( gdelta, sico.FRICMIN, gekin, sico );

                gphi = sflow.PHI[gl] * pow( ggalpha[gl][0], sflow.FRI_exp );
                if ( sico.DYNFRIC == 1 ) gphi = fdynfric( gphi, sico.FRICMIN, gekin, sico );

                gflufri[gl] = sflow.FLUFRI;
                if ( sico.DYNFRIC == 1 ) gflufri[gl] = fdynfric( gflufri[gl], 0, gekin, sico );

                gp[gl] = ggz[gl+8] + gkappau[gl];
                gpx[gl] = ggz[gl] + gkappau[gl];
                gpy[gl] = ggz[gl+4] + gkappau[gl];
                if ( sflow.TAUY[gl] == -9999 ) gtauy[gl] = sin( gphi ) * gp[gl]; //!!!CHECK neighbourhood ggz and friction parameters needed
                else gtauy[gl] = sflow.TAUY[gl];

                for ( gq=0; gq<6; gq++ ) {

                    gny[gq] = 1 / sico.SLOMO * sflow.NY[gl] * pow( ggalpha[gl][gq], sflow.NY_exp );

                    if ( gd[gl][gq] != 0 ) gnye[gl][gq] = gny[gq] + gtauy[gl] / gd[gl][gq] * ( 1 - exp( -sflow.VIS_ry * gd[gl][gq] )); else gnye[gl][gq] = gny[gq];

                    if ( gphi > 0 && gdelta > 0 ) {

                        if ( guvw[gl][gq] != 0 ) gcuf[gl][gq] = -gnu[gl][gq] / fabs( guvw[gl][gq] ) * tan( gdelta ); else gcuf[gl][gq] = 0;
                        if ( gnye[gl][gq] != 0 ) gguzb[gl][gq] = gcuf[gl][gq] / gnye[gl][gq] * gp[gl] + 2 * gcuf[gl][gq] * ggux[gl][gq]; else gguzb[gl][gq] = 0;

                        if ( guvw[gl][gq] != 0 ) gcvf[gl][gq] = -gnv[gl][gq] / fabs( guvw[gl][gq] ) * tan( gdelta ); else gcvf[gl][gq] = 0;
                        if ( gnye[gl][gq] != 0 ) ggvzb[gl][gq] = gcvf[gl][gq] / gnye[gl][gq] * gp[gl] + 2 * gcvf[gl][gq] * ggvy[gl][gq]; else ggvzb[gl][gq] = 0;

                    } else { 

                        if ( gh[gq] > sico.HFLOWMIN ) gguzb[gl][gq] = sflow.VIS_chi[gl] * gnu[gl][gq] / gh[gq]; else gguzb[gl][gq] = 0;
                        if ( gh[gq] > sico.HFLOWMIN ) ggvzb[gl][gq] = sflow.VIS_chi[gl] * gnv[gl][gq] / gh[gq]; else ggvzb[gl][gq] = 0;
                    }
                }

                for ( gq=0; gq<6; gq++ ) {

                    if ( gh[gq] > sico.HFLOWMIN )

                        gd[gl][gq] = 1 / sico.SLOMO * 
                            pow( fabs( 4 * ggux[gl][gq] * ggvy[gl][gq] - pow( gguy[gl][gq] + ggvx[gl][gq], 2 ) - pow( gguzb[gl][gq], 2 ) - pow( ggvzb[gl][gq], 2 )), 0.5 );

                    else gd[gl][gq] = 0;
                }

                if ( ggalpha[gl][0] > 0 ) {

                    gtaunnx[gl] = 0; gtaunny[gl] = 0;

                    for ( gg = 0; gg < sico.PMAX; gg++ ) {

                        if ( gg != gl && ( sico.PHASES[gg] == 1 )) {

                            difuxx[5] = ggalphax[gg][5] * ( gnu[gl][5] - gnu[gg][5] );
                            difuxx[2] = ggalphax[gg][2] * ( gnu[gl][2] - gnu[gg][2] );

                            difvxx[4] = ggalphax[gg][4] * ( gnv[gl][4] - gnv[gg][4] );
                            difvxx[1] = ggalphax[gg][1] * ( gnv[gl][1] - gnv[gg][1] );

                            difuyx[4] = ggalphay[gg][4] * ( gnu[gl][4] - gnu[gg][4] );
                            difuyx[1] = ggalphay[gg][1] * ( gnu[gl][1] - gnu[gg][1] );

                            difvyy[4] = ggalphay[gg][4] * ( gnv[gl][4] - gnv[gg][4] );
                            difvyy[1] = ggalphay[gg][1] * ( gnv[gl][1] - gnv[gg][1] );

                            difuyy[5] = ggalphay[gg][5] * ( gnu[gl][5] - gnu[gg][5] );
                            difuyy[2] = ggalphay[gg][2] * ( gnu[gl][2] - gnu[gg][2] );

                            difvxy[5] = ggalphax[gg][5] * ( gnv[gl][5] - gnv[gg][5] );
                            difvxy[2] = ggalphax[gg][2] * ( gnv[gl][2] - gnv[gg][2] );

                            difax = gnu[gl][0] - gnu[gg][0];
                            difay = gnv[gl][0] - gnv[gg][0];

                            gxterm1 = gnye[gl][2] * difuxx[2] - gnye[gl][5] * difuxx[5];
                            gxterm2 = gnye[gl][1] * ( difvxx[1] + difuyx[1] ) - gnye[gl][4] * ( difvxx[4] - difuyx[4] );
                            gyterm1 = gnye[gl][1] * difvyy[1] - gnye[gl][4] * difvyy[4];
                            gyterm2 = gnye[gl][2] * ( difuyy[2] + difvxy[2] ) - gnye[gl][5] * ( difuyy[5] - difvxy[5] );

                            gtaunnx[gl] += pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * ( 2 / gdx * gxterm1 + 1 / gdy * gxterm2 ) 
                                - pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * sflow.VIS_xi[gg] * ggalpha[gg][0] * gnye[gl][0] * difax / pow( gh[0], 2);
                                //!!!CHECK sflow.VIS_a

                            gtaunny[gl] += pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * ( 2 / gdy * gyterm1 + 1 / gdx * gyterm2 ) 
                                - pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * sflow.VIS_xi[gg] * ggalpha[gg][0] * gnye[gl][0] * difay / pow( gh[0], 2);
                        }
                    }

                } else { gtaunnx[gl] = 0; gtaunny[gl] = 0; }

                if ( ggalpha[gl][0] > 0 && guvw[gl][0] != 0 ) {

                    gxterm1 = 2 / gdx * ( gnye[gl][2] * ggux[gl][2] - gnye[gl][5] * ggux[gl][5] );
                    gxterm2 = 1 / gdy * ( gnye[gl][1] * ggvx[gl][1] - gnye[gl][4] * ggvx[gl][4] );
                    gxterm3 = 1 / gdy * ( gnye[gl][1] * gguy[gl][1] - gnye[gl][4] * gguy[gl][4] );

                    gyterm1 = 2 / gdy * ( gnye[gl][1] * ggvy[gl][1] - gnye[gl][4] * ggvy[gl][4] );
                    gyterm2 = 1 / gdx * ( gnye[gl][2] * gguy[gl][2] - gnye[gl][5] * gguy[gl][5] );
                    gyterm3 = 1 / gdx * ( gnye[gl][2] * ggvx[gl][2] - gnye[gl][5] * ggvx[gl][5] );

                    gvisx[gl] = -( gxterm1 + gxterm2 + gxterm3 - ( -tan(gdelta) * gpx[gl] * gnu[gl][0]/fabs(guvw[gl][0])
                        - 2 * tan( gdelta ) * gnye[gl][0] * gnu[gl][0] / fabs( guvw[gl][0] ) * ggux[gl][0] ) / gh[0] ) + gtaunnx[gl];

                    gvisy[gl] = -( gyterm1 + gyterm2 + gyterm3 - ( -tan(gdelta) * gpy[gl] * gnv[gl][0]/fabs(guvw[gl][0])
                        - 2 * tan( gdelta ) * gnye[gl][0] * gnv[gl][0] / fabs( guvw[gl][0] ) * ggvy[gl][0] ) / gh[0] ) + gtaunny[gl];

                } else { gvisx[gl] = 0; gvisy[gl] = 0; }

            } else if ( sico.PHASES[gl] == 3 ) {

                gfricx[gl] = 0;
                gfricy[gl] = 0;

                gflufri[gl] = sflow.FLUFRI;
                if ( sico.DYNFRIC == 1 ) gflufri[gl] = fdynfric( gflufri[gl], 0, gekin, sico );

                if ( sflow.TAUY[gl] == -9999 ) gtauy[gl] = 0;
                else gtauy[gl] = sflow.TAUY[gl];

                for ( gq=0; gq<6; gq++ ) {

                    gny[gq] = 1 / sico.SLOMO * sflow.NY[gl] * pow( ggalpha[gl][gq], sflow.NY_exp );

                    if ( gh[gq] > sico.HFLOWMIN ) gguzb[gl][gq] = sflow.VIS_chi[gl] * gnu[gl][gq] / gh[gq]; else gguzb[gl][gq] = 0;
                    if ( gh[gq] > sico.HFLOWMIN ) ggvzb[gl][gq] = sflow.VIS_chi[gl] * gnv[gl][gq] / gh[gq]; else ggvzb[gl][gq] = 0;

                    if ( gh[gq] > sico.HFLOWMIN )

                        gd[gl][gq] = 1 / sico.SLOMO *  
                            pow( fabs( 4 * ggux[gl][gq] * ggvy[gl][gq] - pow( gguy[gl][gq] + ggvx[gl][gq], 2 ) - pow( gguzb[gl][gq], 2 ) - pow( ggvzb[gl][gq], 2 )), 0.5 );

                    else gd[gl][gq] = 0;

                    if ( gd[gl][gq] != 0 ) gnye[gl][gq] = gny[gq] + gtauy[gl] / gd[gl][gq] * ( 1 - exp( -sflow.VIS_ry * gd[gl][gq] )); else gnye[gl][gq] = gny[gq];
                }

                if ( ggalpha[gl][0] > 0 ) {

                    gtaunnx[gl] = 0; gtaunny[gl] = 0;

                    for ( gg = 0; gg < sico.PMAX; gg++ ) {

                        if ( gg != gl && ( sico.PHASES[gg] == 1 || sico.PHASES[gg] == 2 )) {

                            difuxx[5] = ggalphax[gg][5] * ( gnu[gl][5] - gnu[gg][5] );
                            difuxx[2] = ggalphax[gg][2] * ( gnu[gl][2] - gnu[gg][2] );

                            difvxx[4] = ggalphax[gg][4] * ( gnv[gl][4] - gnv[gg][4] );
                            difvxx[1] = ggalphax[gg][1] * ( gnv[gl][1] - gnv[gg][1] );

                            difuyx[4] = ggalphay[gg][4] * ( gnu[gl][4] - gnu[gg][4] );
                            difuyx[1] = ggalphay[gg][1] * ( gnu[gl][1] - gnu[gg][1] );

                            difvyy[4] = ggalphay[gg][4] * ( gnv[gl][4] - gnv[gg][4] );
                            difvyy[1] = ggalphay[gg][1] * ( gnv[gl][1] - gnv[gg][1] );

                            difuyy[5] = ggalphay[gg][5] * ( gnu[gl][5] - gnu[gg][5] );
                            difuyy[2] = ggalphay[gg][2] * ( gnu[gl][2] - gnu[gg][2] );

                            difvxy[5] = ggalphax[gg][5] * ( gnv[gl][5] - gnv[gg][5] );
                            difvxy[2] = ggalphax[gg][2] * ( gnv[gl][2] - gnv[gg][2] );

                            difax = gnu[gl][0] - gnu[gg][0];
                            difay = gnv[gl][0] - gnv[gg][0];

                            gxterm1 = gnye[gl][2] * difuxx[2] - gnye[gl][5] * difuxx[5];
                            gxterm2 = gnye[gl][1] * ( difvxx[1] + difuyx[1] ) - gnye[gl][4] * ( difvxx[4] - difuyx[4] );
                            gyterm1 = gnye[gl][1] * difvyy[1] - gnye[gl][4] * difvyy[4];
                            gyterm2 = gnye[gl][2] * ( difuyy[2] + difvxy[2] ) - gnye[gl][5] * ( difuyy[5] - difvxy[5] );

                            gtaunnx[gl] += pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * ( 2 / gdx * gxterm1 + 1 / gdy * gxterm2 ) 
                                - pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * sflow.VIS_xi[gg] * ggalpha[gg][0] * gnye[gl][0] * difax / pow( gh[0], 2);
                                //!!!CHECK sflow.VIS_a

                            gtaunny[gl] += pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * ( 2 / gdy * gyterm1 + 1 / gdx * gyterm2 ) 
                                - pow( 1 - ggalpha[gg][0], sflow.VIS_a[gg] ) / ggalpha[gl][0] * sflow.VIS_xi[gg] * ggalpha[gg][0] * gnye[gl][0] * difay / pow( gh[0], 2);
                        }
                    }

                } else { gtaunnx[gl] = 0; gtaunny[gl] = 0; }

                if ( ggalpha[gl][0] > 0 ) {

                    gxterm1 = 2 / gdx * ( gnye[gl][2] * ggux[gl][2] - gnye[gl][5] * ggux[gl][5] );
                    gxterm2 = 1 / gdy * ( gnye[gl][1] * ggvx[gl][1] - gnye[gl][4] * ggvx[gl][4] );
                    gxterm3 = 1 / gdy * ( gnye[gl][1] * gguy[gl][1] - gnye[gl][4] * gguy[gl][4] );

                    gyterm1 = 2 / gdy * ( gnye[gl][1] * ggvy[gl][1] - gnye[gl][4] * ggvy[gl][4] );
                    gyterm2 = 1 / gdx * ( gnye[gl][2] * gguy[gl][2] - gnye[gl][5] * gguy[gl][5] );
                    gyterm3 = 1 / gdx * ( gnye[gl][2] * ggvx[gl][2] - gnye[gl][5] * ggvx[gl][5] );

                    gvisx[gl] = -( gxterm1 + gxterm2 + gxterm3 /*- gnye[gl][0] * gguzb[gl][0] / gh[0]*/ ) + gtaunnx[gl]; //!!!CHECK asymmetric results
                    gvisy[gl] = -( gyterm1 + gyterm2 + gyterm3 /*- gnye[gl][0] * ggvzb[gl][0] / gh[0]*/ ) + gtaunny[gl];

                } else { gvisx[gl] = 0; gvisy[gl] = 0; }
            }
        }


        // Deceleration terms

        ggd[0] = 0;

        if ( gnu[0][0] != 0 )
            ggd[1] = gh[0] * ggalpha[0][0] * (( gcambdragx + gflufri[0] / ( 8 * fmax(sico.HFLOWMIN, gh[0] ))) * gnu[0][0] * guvw[0][0] + gfricx[0] + gvisx[0] );
        else ggd[1] = 0;

        if ( gnv[0][0] != 0 )
            ggd[2] = gh[0] * ggalpha[0][0] * (( gcambdragy + gflufri[0] / ( 8 * fmax(sico.HFLOWMIN, gh[0] ))) * gnv[0][0] * guvw[0][0] + gfricy[0] + gvisy[0] );
        else ggd[2] = 0;
    
        if ( sico.MODEL > 3 ) {

            ggd[3] = 0; 

            if ( gnu[1][0] != 0 )
                ggd[4] = gh[0] * ggalpha[1][0] * (( gcambdragx + gflufri[1] / ( 8 * fmax(sico.HFLOWMIN, gh[0] ))) * gnu[1][0] * guvw[1][0] + gfricx[1] + gvisx[1] );
            else ggd[4] = 0;

            if ( gnv[1][0] != 0 )
                ggd[5] = gh[0] * ggalpha[1][0] * (( gcambdragy + gflufri[1] / ( 8 * fmax(sico.HFLOWMIN, gh[0] ))) * gnv[1][0] * guvw[1][0] + gfricy[1] + gvisy[1] );
            else ggd[5] = 0;
        }
 
        if ( sico.MODEL == 7 ) {

            ggd[6] = 0;

            if ( gnu[2][0] != 0 )
                ggd[7] = gh[0] * ggalpha[2][0] * (( gcambdragx + gflufri[2] / ( 8 * fmax(sico.HFLOWMIN, gh[0] ))) * gnu[2][0] * guvw[2][0] + gfricx[2] + gvisx[2] );
            else ggd[7] = 0;

            if ( gnv[2][0] != 0 )
                ggd[8] = gh[0] * ggalpha[2][0] * (( gcambdragy + gflufri[2] / ( 8 * fmax(sico.HFLOWMIN, gh[0] ))) * gnv[2][0] * guvw[2][0] + gfricy[2] + gvisy[2] );
            else ggd[8] = 0;
        }

        for ( gq=0; gq<6; gq++ ) {

            ggd[sico.NVECTMIN+3*gq] = gd[0][gq];
            ggd[sico.NVECTMIN+3*gq+1] = gd[1][gq];
            ggd[sico.NVECTMIN+3*gq+2] = gd[2][gq];
        }

    } else {

        ggd[0] = 0; 
        ggd[1] = 0;
        ggd[2] = 0;

        for ( go=3; go<sico.NVECTMIN+18; go++ ) ggd[go] = 0;
    }

    return ggd;
}

double fsigma ( double gv, double gvm, double gvp, int gdir, double *gdx, double *gdy, int gi, struct ico sico ) {
    // function for slope of vector

    double gsigma, gd;

    if ( gdir == 1 ) gd = gdx[gi]; // pixel spacing in appropriate direction
    else if ( gdir == 2 ) gd = gdy[gi];

    if ( gvp == gv ) gsigma = 0;
    else {

        gsigma = ( gv - gvm ) / ( gvp - gv ); // input for limiter

        if ( sico.LIMITER == 1 ) gsigma = fmax( 0, fmin( 1, gsigma )); // Minmod limiter

        else if ( sico.LIMITER == 2 ) {

            gsigma = fmax( 0, fmin( 1, 2 * gsigma ));
            gsigma = fmax( gsigma, fmin( gsigma, 2 )); // Superbee limiter
        }

        else if ( sico.LIMITER == 3 ) {

            gsigma = fmin( 2 * gsigma, 0.5 * ( 1 + gsigma ));
            gsigma = fmax( 0, fmin ( 2, gsigma )); // Woodward limiter
        }
        else if ( sico.LIMITER == 4 ) gsigma = ( gsigma + fabs( gsigma )) / ( 1 + fabs( gsigma )); // van Leer limiter

        gsigma *= ( gvp - gv ) / gd; // slope of vector
    }
    return gsigma;
}

double fcvelr2 ( double ghflow, double ghflow2, double ghflow3, double gvflowx, double gvflowy, double gvflowx2, double gvflowy2,
    double gvflowx3, double gvflowy3, struct flow sflow, double *gkx, double *gky, double *gbetaxy, int gi, struct ico sico ) {
    // function for determining time step length

    int gl;
    double gcelx, gcely, gcel, gcelr, gh, galpha[3], ggamma[3], gu[3], gv[3];

    gh = ghflow + ghflow2 + ghflow3;
    gu[0] = gvflowx; gu[1] = gvflowx2; gu[2] = gvflowx3;
    gv[0] = gvflowy; gv[1] = gvflowy2; gv[2] = gvflowy3;

    if ( gh > sico.HFLOWMIN ) {

        if ( ghflow > sico.HFLOWMIN ) galpha[0] = ghflow / gh; else galpha[0] = 0;
        if ( ghflow2 > sico.HFLOWMIN ) galpha[1] = ghflow2 / gh; else galpha[1] = 0;
        if ( ghflow3 > sico.HFLOWMIN ) galpha[2] = ghflow3 / gh; else galpha[2] = 0;

        if ( galpha[0] > 0  && galpha[2] > 0 ) ggamma[0] = sflow.RHO3 / sflow.RHO1; else ggamma[0] = 0;
        if ( galpha[0] > 0  && galpha[1] > 0 ) ggamma[1] = sflow.RHO2 / sflow.RHO1; else ggamma[1] = 0;
        if ( galpha[1] > 0  && galpha[2] > 0 ) ggamma[2] = sflow.RHO3 / sflow.RHO2; else ggamma[2] = 1;

        gcel = 0;

        for ( gl = 0; gl < sico.PMAX; gl++ ) {

            if ( sico.PHASES[gl] == 1 ) {

                gcelx  = pow( sico.GRAVITY * fabs( cos( gbetaxy[gi] )) * ( 1 - ggamma[0] ) * gkx[gl] * ( gh * ( galpha[gl] + 0.5 * ( 1 - galpha[gl] ))), 0.5 );
                gcely  = pow( sico.GRAVITY * fabs( cos( gbetaxy[gi] )) * ( 1 - ggamma[0] ) * gky[gl] * ( gh * ( galpha[gl] + 0.5 * ( 1 - galpha[gl] ))), 0.5 );

            } else if ( sico.PHASES[gl] == 2 ) {

                gcelx = pow( sico.GRAVITY * fabs( cos( gbetaxy[gi] )) * ggamma[2] * ( gh * ( galpha[gl] + 0.5 * ( 1 - galpha[gl] ))), 0.5 );
                gcely = pow( sico.GRAVITY * fabs( cos( gbetaxy[gi] )) * ggamma[2] * ( gh * ( galpha[gl] + 0.5 * ( 1 - galpha[gl] ))), 0.5 );

            } else if ( sico.PHASES[gl] == 3 ) {

                gcelx = pow( sico.GRAVITY * fabs( cos( gbetaxy[gi] )) * ( gh * ( galpha[gl] + 0.5 * ( 1 - galpha[gl] ))), 0.5 );
                gcely = pow( sico.GRAVITY * fabs( cos( gbetaxy[gi] )) * ( gh * ( galpha[gl] + 0.5 * ( 1 - galpha[gl] ))), 0.5 );
            }

            gcelx = fabs( gu[gl] ) + fmin( gcelx, 0.0 + 20 * gcelx * fabs( -0.5 * tanh( 1.0 * gcelx  )));
            gcely = fabs( gv[gl] ) + fmin( gcely, 0.0 + 20 * gcely * fabs( -0.5 * tanh( 1.0 * gcely  )));
            if ( gcelx > gcely ) gcelr = gcelx;  else gcelr =  gcely;
            if ( gcelr > gcel ) gcel = gcelr;
        }

    } else gcel = 0;

    return gcel;
}


// F1 -- START -- Conversion of depths to heights ------------------------------------------------------------------

double fconvout ( int ggi, double **gdflow, double ghflowmax, int gmaterial, int gtype, int *gpx, int *gpy, double *gpelev, double *gbetax,
    double *gbetay, double *gbetaxy, struct ico sico ) {
    // function for converting depths into heights

    int *gin, ggj, gl, gctrlv, gm1, gm2, gm3, gsign;
    double ghflowi, gdflowi, gdflowj, grflowi, grflowj, gdxtest, gdytest, gdxytest, gdxytestmin, gdz, gbetaxyi, gbetaxj, gbetayj, gbetaxyj;

    gsign = 1;

    gbetaxyi = gbetaxy[ggi];

    if ( sico.MODEL <= 3 ) { // height at target pixel:

        if ( gtype == 1 ) gm1 = 0;
        else if ( gtype == 2 ) gm1 = 3;
        else if ( gtype == 3 ) gm1 = 7;
        gdflowi = gsign * gdflow[ggi][gm1];

        if ( gdflowi != 0 ) ghflowi = gdflowi / cos ( gbetaxyi );
        else ghflowi = 0;

        grflowi = 1;
    }
    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

        if ( gtype == 1 ) { gm1 = 0; gm2 = 3; }
        else if ( gtype == 2 ) { gm1 = 6; gm2 = 7; }
        else if ( gtype == 3 ) { gm1 = 18; gm2 = 20; }
        gdflowi = gsign * ( gdflow[ggi][gm1] + gdflow[ggi][gm2] );

        if ( gdflowi != 0 ) {
            if ( gmaterial == 1 ) grflowi = gsign * gdflow[ggi][gm1] / gdflowi;
            else if ( gmaterial == 2 ) grflowi = gsign * gdflow[ggi][gm2] / gdflowi;
            else if ( gmaterial == 4 ) grflowi = 1;
            ghflowi = gdflowi * grflowi / cos ( gbetaxyi );
        }
        else ghflowi = 0;
    }
    else if ( sico.MODEL == 7 ) {

        if ( gtype == 1 ) { gm1 = 0; gm2 = 3; gm3 = 6; }
        else if ( gtype == 2 ) { gm1 = 9; gm2 = 10; gm3 = 11; }
        else if ( gtype == 3 ) { gm1 = 25; gm2 = 27; gm3 = 29; }
        gdflowi = gsign * ( gdflow[ggi][gm1] + gdflow[ggi][gm2] + gdflow[ggi][gm3] );

        if ( gdflowi != 0 ) {
            if ( gmaterial == 1 ) grflowi = gsign * gdflow[ggi][gm1] / gdflowi;
            else if ( gmaterial == 2 ) grflowi = gsign * gdflow[ggi][gm2] / gdflowi;
            else if ( gmaterial == 3 ) grflowi = gsign * gdflow[ggi][gm3] / gdflowi;
            else if ( gmaterial == 4 ) grflowi = 1;
            ghflowi = gdflowi * grflowi / cos ( gbetaxyi );
        }
        else ghflowi = 0;
    }

    if ( sico.CORRHEIGHT == 0 ) ghflowi = gdflowi * grflowi;
    else if ( sico.CORRHEIGHT == 2 ) {

        gin = fin( ggi, gpelev, sico ); // identifying pixel neighbourhood
        gctrlv = 1; for ( gl=1; gl<9; gl++ ) if ( gin[gl] < 0 || gin[gl] >= sico.IMAX ) gctrlv = 0;
        free(gin);

        if ( gctrlv == 0 ) ghflowi = 0;
        else if ( sico.MODEL < 4 && gdflowi * sin( gbetaxyi ) > 0.5 * sico.CSZ ) { // if advanced conversion is necessary:

            gdxytestmin = sico.UNDEF*-1; // initial value of comparison of horizontal component of depth and shift

            for ( ggj=0; ggj<sico.IMAX; ggj++) { // loop over all pixels:

                if ( gpelev[ggj] != sico.UNDEF ) {

                    if (( sico.MODEL <= 3 && gdflow[ggj][gm1] != 0 ) || (( sico.MODEL > 3 && sico.MODEL < 7 ) && gdflow[ggj][gm1] + gdflow[ggj][gm2] != 0 )) {
                        if ( pow( sico.CSZ * sico.CSZ * (double)(pow( gpx[ggj] - gpx[ggi], 2 ) + pow( gpy[ggj] - gpy[ggi], 2 )), 0.5 )
                            <= ghflowmax ) {

                            gin = fin( ggj, gpelev, sico ); // identifying pixel neighbourhood
                            gctrlv = 1; for ( gl=1; gl<9; gl++ ) if ( gin[gl] < 0 || gin[gl] >= sico.IMAX ) gctrlv = 0;
                            free(gin);

                            if ( gctrlv == 1 ) {

                                if ( sico.MODEL <= 3 ) { // depth and ratio at shift location:
                                    gdflowj = gsign * gdflow[ggj][gm1];
                                    if ( gdflowj != 0 ) grflowj = 1;
                                }
                                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                                    gdflowj = gsign * ( gdflow[ggj][gm1] + gdflow[ggj][gm2] );
                                    if ( gdflowj != 0 && gmaterial == 1 ) grflowj = gsign * gdflow[ggj][gm1] / gdflowj;
                                    else if ( gdflowj != 0 && gmaterial == 2 ) grflowj = gsign * gdflow[ggj][gm2] / gdflowj;
                                    else if ( gmaterial == 4 ) grflowj = 1;
                                }

                                gbetaxj = gbetax[ggj];
                                gbetayj = gbetay[ggj];
                                gbetaxyj = gbetaxy[ggj];

                                gdz = gpelev[ggj] - gpelev[ggi]; // elevation difference between shift location and target pixel
                                gdxtest = sico.CSZ * (double)( gpx[ggj] - gpx[ggi] ) + ( gdflowj * sin( gbetaxj ));
                                gdytest = sico.CSZ * (double)( gpy[ggj] - gpy[ggi] ) + ( gdflowj * sin( gbetayj ));
                                gdxytest = pow( pow( gdxtest , 2 ) + pow( gdytest , 2 ), 0.5 );
                                    // comparison of horizontal component of depth and shift

                                if ( gdxytest < gdxytestmin ) {
                                    if ( gdflowj != 0 ) ghflowi = grflowj * ( gdflowj * cos ( gbetaxyj ) + gdz ); // height at target pixel
                                    gdxytestmin = gdxytest;
                                }
                            }
                        }
                    }   
                }
            }
        }
    }
    return ghflowi;
}

// F1 -- STOP --- Conversion of depths to heights ------------------------------------------------------------------

void foutasc ( double **gparam, int *gpx, int *gpy, char *goutmaps, char gname[20], double ghflowmax, double *gpelev, double *gbetax,
    double *gbetay, double *gbetaxy, int gk, struct ico sico ) {
    //!!!CHECK for values out of aoi - function for output of ascii raster maps

    FILE *gfascii;
    char gpath[200];
    int gi, gx, gy;
    double gpout;

    sprintf(gpath, "%s%s.asc", goutmaps, gname ); // writing name of ascii file to string
    gfascii=fopen(gpath, "w"); // opening ascii file

    fprintf( gfascii, "ncols %i\n", sico.N / sico.MESH ); // writing header to ascii file
    fprintf( gfascii, "nrows %i\n", sico.M / sico.MESH );
    fprintf( gfascii, "xllcenter %.2f\n", sico.BDWEST );
    fprintf( gfascii, "yllcenter %.2f\n", sico.BDSOUTH );
    fprintf( gfascii, "cellsize %.2f\n", sico.CSZ * sico.MESH );
    fprintf( gfascii, "NODATA_value %.3f\n", sico.UNDEF );

    gi = 0;
    for ( gx=0; gx<sico.M / sico.MESH; gx+=1 ) { for ( gy=0; gy<sico.N / sico.MESH; gy+=1 ) { // loop over all pixels:

        if ( gpx[gi] == gx && gpy[gi] == gy ) { // converting depths into heights:

            if ( sico.MODEL <= 3 && gk == 0 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 1,
		gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL <= 3 && gk == 3 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL <= 3 && gk == 7 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );

            else if ( sico.MODEL <= 3 && gk == 1 ) gpout = fdiv( gparam[gi][gk], gparam[gi][0], sico.HFLOWMIN );
            else if ( sico.MODEL <= 3 && gk == 2 ) gpout = fdiv( gparam[gi][gk], gparam[gi][0], sico.HFLOWMIN );

            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 0 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 1,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 3 ) gpout = fconvout( gi, gparam, ghflowmax, 2, 1,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 6 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 7 ) gpout = fconvout( gi, gparam, ghflowmax, 2, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 10 ) gpout = fconvout( gi, gparam, ghflowmax, 4, 1,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 17 ) gpout = fconvout( gi, gparam, ghflowmax, 4, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 18 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 20 ) gpout = fconvout( gi, gparam, ghflowmax, 2, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 22 ) gpout = fconvout( gi, gparam, ghflowmax, 4, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );

            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 1 ) gpout = fdiv( gparam[gi][gk], gparam[gi][0], sico.HFLOWMIN );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 2 ) gpout = fdiv( gparam[gi][gk], gparam[gi][0], sico.HFLOWMIN );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 4 ) gpout = fdiv( gparam[gi][gk], gparam[gi][3], sico.HFLOWMIN );
            else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && gk == 5 ) gpout = fdiv( gparam[gi][gk], gparam[gi][3], sico.HFLOWMIN );

            else if ( sico.MODEL == 7 && gk == 0 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 1,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 3 ) gpout = fconvout( gi, gparam, ghflowmax, 2, 1,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 6 ) gpout = fconvout( gi, gparam, ghflowmax, 3, 1,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 9 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 10 ) gpout = fconvout( gi, gparam, ghflowmax, 2, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 11 ) gpout = fconvout( gi, gparam, ghflowmax, 3, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 15 ) gpout = fconvout( gi, gparam, ghflowmax, 4, 1,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 24 ) gpout = fconvout( gi, gparam, ghflowmax, 4, 2,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 25 ) gpout = fconvout( gi, gparam, ghflowmax, 1, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 27 ) gpout = fconvout( gi, gparam, ghflowmax, 2, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 29 ) gpout = fconvout( gi, gparam, ghflowmax, 3, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );
            else if ( sico.MODEL == 7 && gk == 31 ) gpout = fconvout( gi, gparam, ghflowmax, 4, 3,
                gpx, gpy, gpelev, gbetax, gbetay, gbetaxy, sico );

            else if ( sico.MODEL == 7 && gk == 1 ) gpout = fdiv( gparam[gi][gk], gparam[gi][0], sico.HFLOWMIN );
            else if ( sico.MODEL == 7 && gk == 2 ) gpout = fdiv( gparam[gi][gk], gparam[gi][0], sico.HFLOWMIN );
            else if ( sico.MODEL == 7 && gk == 4 ) gpout = fdiv( gparam[gi][gk], gparam[gi][3], sico.HFLOWMIN );
            else if ( sico.MODEL == 7 && gk == 5 ) gpout = fdiv( gparam[gi][gk], gparam[gi][3], sico.HFLOWMIN );
            else if ( sico.MODEL == 7 && gk == 7 ) gpout = fdiv( gparam[gi][gk], gparam[gi][6], sico.HFLOWMIN );
            else if ( sico.MODEL == 7 && gk == 8 ) gpout = fdiv( gparam[gi][gk], gparam[gi][6], sico.HFLOWMIN );

            else gpout = gparam[gi][gk];

            fprintf(gfascii, "%.3f ", gpout ); // writing output to ascii file
            gi += 1;
        }
        else fprintf(gfascii, "0.000 "); // writing zero to file if pixel is not in area of interest
    } fprintf(gfascii, "\n"); }

    fclose(gfascii);

    return;
}

void foutdircoord ( FILE *gf_directions, int gintmode, int *gpx, int *gpy, struct ico sico ) {
    // function for coordinates for output flow direction file

    int gi, git, gy0, gx, gy, intv, intv2;
    double gxmetric, gymetric;

    intv = (int)( sico.N / ( sico.MESH * 75 ) + 1 ); // pixel interval for writing output

    if ( gintmode == 1 ) { intv2 = 1; gy0 = 0; } // defining shift depending on type of parameter
    else if ( gintmode == 2 ) { intv2 = 2; gy0 = 0; }
    else if ( gintmode == 3 ) { intv2 = 2; gy0 = intv; }
    else if ( gintmode == 4 ) { intv2 = 3; gy0 = 0; }
    else if ( gintmode == 5 ) { intv2 = 3; gy0 = intv; }
    else if ( gintmode == 6 ) { intv2 = 3; gy0 = 2*intv; }

    gi = 0;
    for ( gx=0; gx<sico.M / sico.MESH; gx+=intv ) { for ( gy=gy0; gy<sico.N / sico.MESH; gy+=(intv2*intv)) { // writing x coordinate to file:

        git = gi;
        while (( gpx[git] != gx || gpy[git] != gy ) && git < sico.IMAX ) git+= 1;
        if ( git < sico.IMAX ) {

            gi = git;
            gxmetric = ( float ) gpy[gi] * sico.CSZ + sico.BDWEST;
            fprintf( gf_directions, "%.0f\t", gxmetric );
        }
    }

    if ( sico.MODEL < 4 ) { if ( gy0 == intv ) gy0 = 0; else gy0 = intv; }
    else { if ( gy0 == 2*intv ) gy0 = intv; else if ( gy0 == intv ) gy0 = 0; else gy0 = 2*intv; }}

    fprintf( gf_directions, "\n" );

    gi = 0;
    for ( gx=0; gx<sico.M / sico.MESH; gx+=intv ) { for ( gy=gy0; gy<sico.N / sico.MESH; gy+=(intv2*intv)) { // writing y coordinate to file:

        git = gi;
        while (( gpx[git] != gx || gpy[git] != gy ) && git < sico.IMAX ) git+= 1;
        if ( git < sico.IMAX ) {

            gi = git;
            gymetric = sico.BDNORTH - ( float ) gpx[gi] * sico.CSZ;
            fprintf( gf_directions, "%.0f\t", gymetric );
        }
    }

    if ( sico.MODEL < 4 ) { if ( gy0 == intv ) gy0 = 0; else gy0 = intv; }
    else { if ( gy0 == 2*intv ) gy0 = intv; else if ( gy0 == intv ) gy0 = 0; else gy0 = 2*intv; }}

    fprintf( gf_directions, "\n" );

    return;
}

void foutdir ( FILE *gf_directions, double **gparam, int gl, int gintmode, int *gpx, int *gpy, struct ico sico ) {
    // function for output flow direction file

    int gi, git, gy0, gx, gy, intv, intv2;
    double ghflowi, gparami;

    intv = (int)( sico.N / ( 75 * sico.MESH ) + 1 ); // pixel interval for writing output

    if ( gintmode == 1 ) { intv2 = 1; gy0 = 0; } // defining shift depending on type of parameter
    else if ( gintmode == 2 ) { intv2 = 2; gy0 = 0; }
    else if ( gintmode == 3 ) { intv2 = 2; gy0 = intv; }
    else if ( gintmode == 4 ) { intv2 = 3; gy0 = 0; }
    else if ( gintmode == 5 ) { intv2 = 3; gy0 = intv; }
    else if ( gintmode == 6 ) { intv2 = 3; gy0 = 2*intv; }

    gi = 0;
    for ( gx=0; gx<sico.M / sico.MESH; gx+=intv ) { for ( gy=gy0; gy<sico.N / sico.MESH; gy+=(intv2*intv)) { // writing parameter to file:

        git = gi;
        while (( gpx[git] != gx || gpy[git] != gy ) && git < sico.IMAX ) git+= 1;
        if ( git < sico.IMAX ) {

            gi = git;

            if ( gintmode == 1 ) ghflowi = gparam[gi][0];
            else if ( gintmode == 2 ) ghflowi = gparam[gi][0];
            else if ( gintmode == 3 ) ghflowi = gparam[gi][3];
            else if ( gintmode == 4 ) ghflowi = gparam[gi][0];
            else if ( gintmode == 5 ) ghflowi = gparam[gi][3];
            else if ( gintmode == 6 ) ghflowi = gparam[gi][6];

            if ( ghflowi >= sico.IMPTHR[0] ) {

                if ( gintmode == 1 || gintmode == 2 ) gparami = gparam[gi][gl] / gparam[gi][0];
                else if ( gintmode == 3 ) gparami = gparam[gi][gl] / gparam[gi][3];
                else if ( gintmode == 4 ) gparami = gparam[gi][gl] / gparam[gi][0];
                else if ( gintmode == 5 ) gparami = gparam[gi][gl] / gparam[gi][3];
                else if ( gintmode == 6 ) gparami = gparam[gi][gl] / gparam[gi][6];
                fprintf( gf_directions, "%.2f\t", gparami );
            }
            else fprintf( gf_directions, "0\t" );
        }
    }

    if ( sico.MODEL < 4 ) { if ( gy0 == intv ) gy0 = 0; else gy0 = intv; }
    else { if ( gy0 == 2*intv ) gy0 = intv; else if ( gy0 == intv ) gy0 = 0; else gy0 = 2*intv; }}

    fprintf( gf_directions, "\n" );

    return;
}


#ifdef WITHGRASS


    SEGMENT finrasti ( char gm[1000], struct ico sico ) { // input of integer GRASS raster maps

        int gf, gx, gy, gin;
        CELL *gc;
        SEGMENT gseg_in;

        Segment_open ( &gseg_in, G_tempfile(), sico.M, sico.N, sico.NSEGRC, sico.NSEGRC, sizeof(int), sico.NSEGS );

        gc = Rast_allocate_c_buf();
        gf = Rast_open_old ( gm, sico.MAINMAPSET );

        for ( gx = 0; gx < sico.M; gx++ ) {

            Rast_get_c_row ( gf, gc, gx );

            for ( gy = 0; gy < sico.N; gy++ ) {

                if ( !Rast_is_c_null_value ( gc + gy ) ) gin = ( int ) gc[gy];
                else gin = (int)sico.UNDEF;
                Segment_put(&gseg_in, &gin, gx, gy);
	    }
        }
        Segment_flush(&gseg_in);
        G_free ( gc );
        Rast_close ( gf );

        return gseg_in;
    }

    SEGMENT finrastd ( char gm[1000], struct ico sico ) { // input of double GRASS raster maps

        int gf, gx, gy;
        double gin;
        DCELL *gc;
        SEGMENT gseg_in;

        Segment_open ( &gseg_in, G_tempfile(), sico.M, sico.N, sico.NSEGRC, sico.NSEGRC, sizeof(double), sico.NSEGS );

        gc = Rast_allocate_d_buf();
        gf = Rast_open_old ( gm, sico.MAINMAPSET );

        for ( gx = 0; gx < sico.M; gx++ ) {

            Rast_get_d_row ( gf, gc, gx );

            for ( gy = 0; gy < sico.N; gy++ ) {

                if ( !Rast_is_d_null_value ( gc + gy ) ) gin = ( double ) gc[gy];
                else gin = sico.UNDEF;
                Segment_put(&gseg_in, &gin, gx, gy);
	    }
        }

        Segment_flush(&gseg_in);
        G_free ( gc );
        Rast_close ( gf );

        return gseg_in;
    }

    void foutrast ( char gm[1000], double ***gtestv, struct ico sico, int gk, double gtsum ) { // output of GRASS raster maps

        int gf, gx, gy;
        double gv;
        DCELL *gc;

        if ( gm != NULL ) {
            gc = Rast_allocate_d_buf();
	    gf = Rast_open_new( gm, DCELL_TYPE );
        }

        for ( gx = 0; gx < sico.M; gx++ ) {

            if ( gm != NULL ) {

	        for ( gy = 0; gy < sico.N; gy++ ) {

                    //Segment_get(&gseg_v, &gv, gx, gy);
                    gc[gy] = (DCELL) gtestv[gx][gy][gk];
                }
	        Rast_put_d_row( gf, gc );
            }
        }
        Rast_close( gf );
        G_free( gc );

        return;
    }


#else


    double *finaschdr ( char *gname ) { // function for input of ascii raster map header

        FILE *gfascii;
        char gpath[200], *tok;
        char delim[] = " \t\n";
        int gi, gx, gy;
        double gtest3, *gparam, gtestthrsh;
        char *gtest, *gtest2;
        char **strtodhlp=NULL;

        gparam = (double*) calloc( 7, sizeof(double));

        sprintf(gpath, "%s", gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "r"); // opening ascii file

        for ( gi=0; gi<6; gi++ ) { // reading header lines and writing to array:
            gtest = fcparam ( gfascii );
            tok=strtok(gtest, delim);
            while ( tok != NULL ) {
                gtest2=tok;
                tok=strtok(NULL, delim);
            }
            gparam[gi] = strtod (gtest2, strtodhlp);
            free(gtest);
        }

        gtestthrsh = gparam[5]; // threshold for validity

        gi = 0;
        for ( gx=0; gx<(int)gparam[1]; gx++ ) { // counting number of pixels of reduced coordinate system:

            gtest = flparam ( gfascii );

            tok = strtok(gtest, delim);
            for ( gy=0; gy<(int)gparam[0]; gy++ ) {
                gtest2 = tok;
                tok = strtok(NULL, delim);
                gtest3 = strtod (gtest2, strtodhlp);
                if ( gtest3 >= gtestthrsh ) gi += 1;
            }
            free(gtest);
        }
        gparam[6] = (double)gi;
        fclose( gfascii );
        return gparam;
    }

    int *finascxy ( char *gname, int gcontrol, struct ico sico ) { // function for input of ascii raster x and y coordinates

        FILE *gfascii;
        char *gpath, *tok;
        char delim[] = " \t\n";
        int gi, gx, gy, *gpxy;
        double gtest3, gtestthrsh;
        char **strtodhlp=NULL;

        char *gtest; // allocating memory to arrays
        char *gtest2;

        gpath = (char*) calloc ( 1000, sizeof(char)); // allocating memory to arrays:
        gpxy = (int*) calloc ( sico.IMAX, sizeof(int));

        gtestthrsh = sico.UNDEF; // threshold for validity

        sprintf(gpath, "%s", gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "r"); // opening ascii file
        for ( gi=0; gi<6; gi++ ) {
            gtest = fcparam ( gfascii ); // reading header lines
            free(gtest);
        }

        gi = 0;
        for ( gx=0; gx<sico.M; gx+=sico.MESH ) { // writing internal coordinates of all rows and columns to array:

            gtest = flparam ( gfascii );
            tok = strtok(gtest, delim);
            for ( gy=0; gy<sico.N; gy+=sico.MESH ) {

                gtest2 = tok;
                tok = strtok(NULL, delim);
                gtest3 = strtod (gtest2, strtodhlp);

                if ( gtest3 >= gtestthrsh ) {

                    if ( gcontrol == 1 ) gpxy[gi] = gx;
                    else gpxy[gi] = gy;

                    gi += 1;
                }
            }
            free(gtest);
        }
        fclose( gfascii );
        free(gpath);
        return gpxy;
    }

    double *finascval ( char *gname, int *gpx, int *gpy, struct ico sico ) { // function for input of ascii raster map values

        FILE *gfascii;
        char *gpath, *tok;
        char delim[] = " \t\n";
        int gi, gx, gy;
        double gtest3, *gparam;
        char **strtodhlp=NULL;

        char *gtest; // allocating memory to arrays
        char *gtest2;

        gpath = (char*) calloc ( 1000, sizeof(char)); // allocating memory to arrays:
        gparam = (double*) calloc( sico.IMAX, sizeof(double));

        sprintf(gpath, "%s", gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "r"); // opening ascii file
        for ( gi=0; gi<6; gi++ ) {
            gtest = fcparam ( gfascii ); // reading header lines
            free(gtest);
        }

        gi = 0;
        for ( gx=0; gx<sico.M; gx+=sico.MESH ) { // writing raster values of all rows and columns to array:

            gtest = flparam ( gfascii );
            tok = strtok(gtest, delim);
            for ( gy=0; gy<sico.N; gy+=sico.MESH ) {

                gtest2 = tok;
                tok = strtok(NULL, delim);
                gtest3 = strtod(gtest2, strtodhlp);

                if ( gx == gpx[gi] && gy == gpy[gi] ) {
                    gparam[gi] = gtest3;
                    gi += 1;
                }
            }
            free(gtest);
        }
        fclose( gfascii );
        free(gpath);
        return gparam;
    }


#endif


// Calling main function

int main ( int argc, char *argv[] ) {


// Variables

    FILE *fparam;

    char path[1000], wkdir[1000], *prefix, *outmaps, *outfiles, *nmeshname, *elevname,
        *hreleasename, *hreleasename2, *hreleasename3, *hentrmaxname, *hentrmaxname2, *hentrmaxname3, *vinxname, *vinyname, *vinxname2, *vinyname2,
        *vinxname3, *vinyname3, *centrname, *cdepname, *phiname, *phi2name, *phi3name, *deltaname, *delta2name, *delta3name, *nyssname, *nyfsname, *nyffname,
        *ambdragname, *flufriname, *transssfsname, *transssffname, *transfsffname, *hydreleasename0, *treleasename,
        *trelstopname, mv[1000], **mv0;

    char *madd = (char*) calloc(1000, sizeof(char));

    int *px, *py, *iin, *inn, *innn, hydrograph, hydnin, hydnout, *hydi, hydj, hydk, *hydx, *hydy, hydt, *hydtmax,
        hydtmaxx, *hydp0, **hydp, nvect_all, nvect_red, lmax, xint, ccontinue, csuccess, nsum, nout, time_start,
        time_stop, ctrlr, ctrlv, ctrlvv, ctrlvvv, ccfl, fj[4][2], i, j, jj, jjj, k, l, ll, p, prec_hflow, prec_vol, prec_ekin,
        imax, nalpha, ctrl_hydout, ix, i2, hydcols, ctrl_lake;

    double *pelev, *pelev0, gkx[3], gky[3], *kappau, *vm, *cdrag, *disp, *gze, *gf, *gg, *gs, *flowpar, tout, tmax, vflowx, vflowy, vflow, vflowx2, vflowy2, vflowx3, vflowy3, 
        time_elapsed, tlength, tlength0, tlengthx, tlengthpre, tint, tsum, cfl, cflmax, grav[5], dw_dt[9], hflow_maxmax, hflow_max0, hflow_max, hflow_max02, hflow_max2, 
        hflow_max03, hflow_max3, vflow_max, vflow_max2, vflow_max3, **hydhyd0, ***hydhyd, vol_flow0, vol_flow, vol_flow02, vol_flow2, vol_flow03, vol_flow3, vol_entr, 
        vol_entr2, vol_entr3, vol_edge, vol_edge2, whx, why, vcelr0, vcelr, *hydelev, *hydalpha, *hydx_metric, *hydy_metric, *hydl, hhyd0, hhyd, hyde, hyde2, hyde3,
        hydfalpha, hydout_xmin_metric, hydout_xmax_metric, hydout_ymin_metric, hydout_ymax_metric, hydh, hydh2, hydh3, hydv, hydv2, hydv3, hydq, hydq2, hydq3,
        hydlmax, ekin, ekin_flow, fflux, ffluxn, vol_noflux, vol_noflux2, vol_noflux3, carea, hydmalpha, hydnalpha, hydbeta, hydm0, hydmx, hydmy, hydm, hydfcorr,
        wu[9], wv[9], wu2[9], wv2[9], wu3[9], wv3[9], walpha[9], walpha2[9], walpha3[9], welev[9],
        potentr, potentr2, potentr3, potdep, potdep2, potdep3, corrfact, qentr0, qentr, qentr2, qentr3, qdep, qdep2, qdep3, alpha, alphav, alphavg, betavg, 
        momx, momy, mom, mass, vbulkx, vbulky, rpotentr,  rpotentr3, walphax, walphay, walphax2, walphay2, walphax3, walphay3, wbetax, wbetay, wbetaxy,
        wdx, wdy, wgrav[5], vol_hydbef, vol_hydaft, pelevhydtest, vol_hydbef2, vol_hydaft2, vol_hydbef3, vol_hydaft3, vol_hyd, vol_hyd2, vol_hyd3,
        *gdecel, hflow, hflown, trat, mompre, ttrelease, ttrelstop, tfact, trans, wwd[18], wdu[3], wdv[3], xwdu[3], xwdv[3], wh[9], whflow, whflow2, whflow3,
        dw_dttest[9], wintbtest, wintctest, awttest[9], hekin, hekin_sum, hekin_max, hekin_thr, hydbef[9], minlevel, wmx, wmy;

    struct ico sico;
    struct flow sflow;


    #ifdef WITHGRASS


        SEGMENT seg_nmesh, seg_elev, seg_hrelease, seg_hrelease2, seg_hrelease3, seg_vinx, seg_viny, seg_vinx2, seg_viny2, seg_vinx3, seg_viny3,
        seg_hentrmax, seg_hentrmax2, seg_hentrmax3, seg_centr, seg_cdep, seg_phi, seg_phi2, seg_phi3, seg_delta, seg_delta2, seg_delta3, seg_nyss, seg_nyfs, seg_nyff,
        seg_ambdrag, seg_flufri, seg_transssfs, seg_transssff, seg_transfsff, seg_trelease, seg_trelstop;
        char *yint = (char*) calloc(1000, sizeof(char));
        int x, y, aoi;
        double hflowi, hflowi2, hflowi3, hentri, hentri2, hentri3, pout, nmesh, elev, hrelease, hrelease2, hrelease3, vinx, viny, vinx2, viny2, vinx3, viny3,
        hentrmax, hentrmax2, hentrmax3, centr, cdep, phi, phi2, phi3, delta, delta2, delta3, nyss, nyfs, nyff, ambdrag, flufri,
        transssfs, transssff, transfsff, trelease, trelstop;

        struct GModule *module;
        struct Cell_head cellhd;


    #else


        double *aschdr;


    #endif


// Timer

    time_start = clock(); // start time


    #ifdef WITHGRASS


        printf("Starting model execution.\n\n");


    #else


        printf("Starting model execution with r.avaflow.main.c from 2 April 2020\n");


    #endif


    fflush(stdout); // forcing immediate display


// Preparing environment


    #ifdef WITHGRASS


        sprintf(wkdir, "%s/%s/.tmp/rtemp", G_location_path(),G_mapset()); // defining working directory
        G_init_tempfile(); // initializing generation of temporary files
        sico.NSEGRC = 1; // defining segment parameters
        sico.NSEGS = 1;
        xint=10;
        yint=getenv("XINT");
        xint=atoi(yint);


    #else


        if ( argc > 1 ) sprintf( wkdir, "%s", argv[1] ); // defining working directory
        else {
            printf( "ERROR: Please provide a parameter file.\n" );
            fflush(stdout); // forcing immediate display
            exit( EXIT_SUCCESS );
        }
        xint=1;


    #endif


// Defining constants

    sico.PI = 3.1415926536;


// Reading parameters from file and deriving further parameters

    sprintf( path, "%s/param%d.txt", wkdir, xint ); // writing name of parameter file to string
    fparam=fopen( path, "r" ); // opening parameter file

    if( !fparam ) { // testing whether parameter file exists:
        printf( "ERROR: Unable to open parameter file: '%s'\n", path );
        fflush(stdout); // forcing immediate display
        exit( EXIT_SUCCESS );
    }

    sico.MAINMAPSET = fcparam ( fparam ); // name of main mapset
    sico.MULT = fiparam ( fparam ); // identifier for single model run (0) or multiple model runs (1)
    sico.MESH = fiparam ( fparam ); // mesh spacing
    sico.NMESH = fiparam ( fparam ); // identifier for use of mesh spacing map as input (1) or not (0)
    sico.MODEL = fiparam ( fparam ); // model (1 = one-phase, 4 = two-phase, 7 = three-phase)

    sico.PMAX = fiparam ( fparam ); // number of phases
    for ( l=0; l<sico.PMAX; l++ ) sico.PHASES[l] = fiparam ( fparam ); // phases

    sico.LIMITER = fiparam ( fparam ); // numerical limiter (1 = Minmod, 2 = Superbee, 3 = Woodward, 4 = van Leer)
    sico.CORRHEIGHT = fiparam ( fparam ); // correction of flow height for slope (0=none, 1=basic, 2=advanced)
    sico.DIFFCTRL = fiparam ( fparam ); // diffusion control (0=no, 1=yes)
    sico.NOOSC = fiparam ( fparam ); // surface control (0=no, 1=yes)
    sico.ENTRAINMENT = fiparam ( fparam ); // entrainment and deposition (0=no, 1-3=type of approach)
    sico.STOPPING = fiparam ( fparam ); // stopping of flow (0=no, 1=yes)
    sico.DYNFRIC = fiparam ( fparam ); // dynamic adaptation of friction (0=no, 1=yes)

    if ( sico.MODEL <= 3 ) { // size of arrays

        nvect_all = 11;
        nvect_red = 7;
        sico.NVECTMIN = 3;
    }
    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

        nvect_all = 29;
        nvect_red = 18;
        sico.NVECTMIN = 6;
    }
    else if ( sico.MODEL == 7 ) {

        nvect_all = 40;
        nvect_red = 25;
        sico.NVECTMIN = 9;
    }

    sico.RELM = fiparam ( fparam ); // identifier for use of (PHASE 1) release height map as input (1) or not (0)
    if ( sico.MODEL > 3 ) sico.RELM2 = fiparam ( fparam ); // identifier for use of PHASE 3 release height map as input (1) or not (0)
    if ( sico.MODEL == 7 ) sico.RELM3 = fiparam ( fparam ); // identifier for use of PHASE 3 release height map as input (1) or not (0)
    sico.RELV = fiparam ( fparam ); // identifier for use of (PHASE 1) release velocity maps as input (1) or not (0)
    if ( sico.MODEL > 3 ) sico.RELV2 = fiparam ( fparam ); // identifier for use of PHASE 3 release velocity maps as input (1) or not (0)
    if ( sico.MODEL == 7 ) sico.RELV3 = fiparam ( fparam ); // identifier for use of PHASE 3 release velocity maps as input (1) or not (0)
    sico.ENTR = fiparam ( fparam ); // identifier for use of maximum height of (PHASE 1) entrainment map as input (1) or not (0)
    if ( sico.MODEL > 3 ) sico.ENTR2 = fiparam ( fparam );
    if ( sico.MODEL == 7 ) sico.ENTR3 = fiparam ( fparam );
        // identifier for use of max. height of PHASE 3 entrainment map as input (1) or not (0)
    sico.CENTR = fiparam ( fparam ); // identifier for use of entrainment coefficient map as input (1) or not (0)
    sico.CDEP = fiparam ( fparam ); // identifier for use of deposition coefficient map as input (1) or not (0)
    sico.PHI = fiparam ( fparam ); // identifier for use of internal friction angle of PHASE 1 map as input (1) or not (0)
    sico.PHI2 = fiparam ( fparam ); // identifier for use of internal friction angle of PHASE 2 map as input (1) or not (0)
    sico.PHI3 = fiparam ( fparam ); // identifier for use of internal friction angle of PHASE 3 map as input (1) or not (0)
    sico.DELTA = fiparam ( fparam ); // identifier for use of basal friction angle of PHASE 1 map as input (1) or not (0)
    sico.DELTA2 = fiparam ( fparam ); // identifier for use of basal friction angle of PHASE 2 map as input (1) or not (0)
    sico.DELTA3 = fiparam ( fparam ); // identifier for use of basal friction angle of PHASE 3 map as input (1) or not (0)
    sico.NYSS = fiparam ( fparam ); // identifier for use of nyss map as input (1) or not (0)
    sico.NYFS = fiparam ( fparam ); // identifier for use of nyfs map as input (1) or not (0)
    sico.NYFF = fiparam ( fparam ); // identifier for use of nyff map as input (1) or not (0)
    sico.AMBDRAG = fiparam ( fparam ); // identifier for use of ambient drag map as input (1) or not (0)
    sico.FLUFRI = fiparam ( fparam ); // identifier for use of fluid friction coefficient map as input (1) or not (0)
    sico.TRANSSSFS = fiparam ( fparam ); // identifier for use of rate of PHASE 1-PHASE 2 transformation map as input (1) or not (0)
    sico.TRANSSSFF = fiparam ( fparam ); // identifier for use of rate of PHASE 1-PHASE 3 transformation map as input (1) or not (0)
    sico.TRANSFSFF = fiparam ( fparam ); // identifier for use of rate of PHASE 2-PHASE 3 transformation map as input (1) or not (0)
    sico.TRELEASE = fiparam ( fparam ); // identifier for use of release time map as input (1) or not (0)
    sico.TRELSTOP = fiparam ( fparam ); // identifier for use of release hydrograph map as input (1) or not (0)
    hydrograph = fiparam ( fparam ); // identifier for use of hydrograph(s) for input and/or output (1) or not (0)

    sico.HYDADD = 0; // identifier whether to impose hydrograph upon flow (1) or reset flow before putting hydrograph (0) //!!! CHECK still hardcoded
    sico.ORIGINAL = 0; //identifier whether to suppress change of direction (0) or not (1) //!!! CHECK still hardcoded

    if ( sico.NMESH == 1 ) nmeshname = fcparam ( fparam ); // name of mesh spacing map
    elevname = fcparam ( fparam ); // name of elevation map


// Setting spatial extent and pixel size


    #ifdef WITHGRASS


    // If GRASS rasters are used:

        G_gisinit( argv[0] );
        module = G_define_module();
        G_get_set_window( &cellhd );

        sico.CSZ=cellhd.ew_res; // pixel size
        sico.BDWEST=cellhd.west; // western boundary
        sico.BDNORTH=cellhd.north; // northern boundary
        sico.BDSOUTH=cellhd.south; // southern boundary
        sico.N=cellhd.cols; // number of pixels in y direction (W-E)
        sico.M=cellhd.rows; // number of pixels in x direction (S-N)
        sico.UNDEF = -9999; // no data value


    #else


    // If Ascii rasters are used:

        aschdr = finaschdr( elevname ); // reading header of Ascii raster
        sico.N = (int)aschdr[0]; // number of pixels in y direction (W-E)
        sico.M = (int)aschdr[1]; // number of pixels in x direction (S-N)
        sico.CSZ = aschdr[4] * sico.MESH; // pixel size
        sico.BDWEST = aschdr[2]; // western boundary
        sico.BDSOUTH = aschdr[3]; // southern boundary
        sico.BDNORTH = aschdr[3] + sico.M * sico.CSZ / sico.MESH; // northern boundary
        sico.UNDEF = aschdr[5]; // no data value
        sico.IMAX = (int)aschdr[6] / sico.MESH; // number of pixels in reduced coordinate system

        free(aschdr);


    #endif


// Continuing reading parameters from file and deriving further parameters

    if ( sico.RELM == 1 ) hreleasename = fcparam ( fparam ); // name of (PHASE 1) release height map
    if ( sico.MODEL > 3 && sico.RELM2 == 1 ) hreleasename2 = fcparam ( fparam ); // name of PHASE 3 release height map
    if ( sico.MODEL == 7 && sico.RELM3 == 1 ) hreleasename3 = fcparam ( fparam ); // name of PHASE 3 release height map
    if ( sico.RELV == 1 ) vinxname = fcparam ( fparam ); // name of (PHASE 1) release velocity in x direction map
    if ( sico.MODEL > 3 && sico.RELV2 == 1 ) vinxname2 = fcparam ( fparam ); // name of PHASE 3 release velocity in x direction map
    if ( sico.MODEL == 7 && sico.RELV3 == 1 ) vinxname3 = fcparam ( fparam ); // name of PHASE 3 release velocity in x direction map
    if ( sico.RELV == 1 ) vinyname = fcparam ( fparam ); // name of (PHASE 1) release velocity in y direction map
    if ( sico.MODEL > 3 && sico.RELV2 == 1 ) vinyname2 = fcparam ( fparam ); // name of PHASE 3 release velocity in y direction map
    if ( sico.MODEL == 7 && sico.RELV3 == 1 ) vinyname3 = fcparam ( fparam ); // name of PHASE 3 release velocity in y direction map
    if ( sico.ENTR == 1 ) hentrmaxname = fcparam ( fparam ); // name of maximum height of (PHASE 1 ) entrainment map
    if ( sico.MODEL > 3 && sico.ENTR2 == 1 ) hentrmaxname2 = fcparam ( fparam ); // name of maximum height of PHASE 3 entrainment map
    if ( sico.MODEL == 7 && sico.ENTR3 == 1 ) hentrmaxname3 = fcparam ( fparam ); // name of maximum height of PHASE 3 entrainment map
    if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) centrname = fcparam ( fparam ); // name of entrainment coefficient map
    if ( sico.STOPPING == 6 && sico.CDEP == 1 ) cdepname = fcparam ( fparam ); // name of deposition coefficient map
    if ( sico.PHI == 1 ) phiname = fcparam ( fparam ); // name of internal friction angle of PHASE 1 map
    if ( sico.PHI2 == 1 ) phi2name = fcparam ( fparam ); // name of internal friction angle of PHASE 2 map
    if ( sico.PHI3 == 1 ) phi3name = fcparam ( fparam ); // name of internal friction angle of PHASE 3 map
    if ( sico.DELTA == 1 ) deltaname = fcparam ( fparam ); // name of basal friction angle of PHASE 1 map
    if ( sico.DELTA2 == 1 ) delta2name = fcparam ( fparam ); // name of basal friction angle of PHASE 2 map
    if ( sico.DELTA3 == 1 ) delta3name = fcparam ( fparam ); // name of basal friction angle of PHASE 3 map
    if ( sico.NYSS == 1 ) nyssname = fcparam ( fparam ); // name of nyss map
    if ( sico.NYFS == 1 ) nyfsname = fcparam ( fparam ); // name of nyfs map
    if ( sico.NYFF == 1 ) nyffname = fcparam ( fparam ); // name of nyff map
    if ( sico.AMBDRAG == 1 ) ambdragname = fcparam ( fparam ); // name of ambient drag map
    if ( sico.FLUFRI == 1 ) flufriname = fcparam ( fparam ); // name of fluid friction coefficient map
    if ( sico.TRANSSSFS == 1 ) transssfsname = fcparam ( fparam );
    if ( sico.TRANSSSFF == 1 ) transssffname = fcparam ( fparam );
    if ( sico.TRANSFSFF == 1 ) transfsffname = fcparam ( fparam );
    if ( sico.TRELEASE == 1 ) treleasename = fcparam ( fparam ); // name of release time map
    if ( sico.TRELSTOP == 1 ) trelstopname = fcparam ( fparam ); // name of release hydrograph map

    if ( hydrograph == 1 ) {

        hydnin = fiparam ( fparam ); // number of input hydrographs
        hydnout = fiparam ( fparam ); // number of output hydrographs

        hydtmax = (int*) calloc( hydnin, sizeof(int)); // allocating memory to variable for number(s) of hydrograph time steps
        hydi = (int*) calloc(( hydnin + hydnout ), sizeof(int));
            // allocating memory to variable for internal coordinate of hydrograph point
        hydelev = (double*) calloc(( hydnin + hydnout ), sizeof(double));
            // allocating memory to variable for elevation of hydrograph point
        hydalpha = (double*) calloc(( hydnin + hydnout ), sizeof(double));
            // allocating memory to variable for aspect of hydrograph point
        hydx_metric = (double*) calloc(( hydnin + hydnout ), sizeof(double));
            // allocating memory to variable for x coordinate of hydrograph
        hydy_metric = (double*) calloc(( hydnin + hydnout ), sizeof(double));
            // allocating memory to variable for y coordinate of hydrograph
        hydl = (double*) calloc(( hydnin + hydnout ), sizeof(double)); // allocating memory to variable for length of hydrograph profile
        hydx = (int*) calloc(( hydnin + hydnout ), sizeof(int)); // allocating memory to variable for internal x coordinate of hydrograph
        hydy = (int*) calloc(( hydnin + hydnout ), sizeof(int)); // allocating memory to variable for internal y coordinate of hydrograph

        hydtmaxx = 0; // initializing maximum number of hydrograph time steps
    }
    else hydnin = 1;

    char *hydreleasename[hydnin]; // declaring pointers to input hydrograph files

    if ( hydrograph == 1 ) {

        for ( hydj = 0; hydj < hydnin; hydj++ ) { // loop over all input hydrographs:

            hydreleasename[hydj] = (char*) calloc( 1000 * hydnin, sizeof(char)); // allocating memory to pointers to input hydrograph files

            hydreleasename0 = fcparam ( fparam ); // name of hydrograph file
            strcpy( hydreleasename[hydj], hydreleasename0 );
            free( hydreleasename0 );
            hydtmax[hydj] = fiparam ( fparam ); // number of hydrograph time steps
            if ( hydtmax[hydj] > hydtmaxx ) hydtmaxx = hydtmax[hydj]; // updating maximum number of hydrograph time steps
        }

        for ( hydj = 0; hydj < hydnin + hydnout; hydj++ ) { // loop over all hydrographs:

            hydx_metric[hydj] = fdparam ( fparam ); // x coordinate of hydrograph
            hydy_metric[hydj] = fdparam ( fparam ); // y coordinate of hydrograph
            hydl[hydj] = fdparam ( fparam ); // length of hydrograph profile
            hydalpha[hydj] = fdparam ( fparam ); // aspect of hydrograph profile
        }

        hydhyd = alloc_dmatrix3( hydtmaxx+1, 7, hydnin ); // allocating memory to array of input hydrograph data
    }
    else { hydnin = 1; hydnout = 1; }


    // Flow parameters

    lmax = fiparam ( fparam ); // number of flow parameters
    flowpar = (double*) calloc( lmax, sizeof(double)); // allocating memory for storage of flow parameter values
    for ( l=0; l<lmax; l++ ) flowpar[l] = fdparam ( fparam ); // flow parameters

    if ( sico.MODEL == 0 ) {

        sflow.RHO1 = flowpar[0]; // density
        sflow.PHI0[0] = flowpar[1] * sico.PI / 180; // internal friction angle
        sflow.DELTA0[0] = flowpar[2] * sico.PI / 180; // basal friction angle
        sflow.TUFRI = pow(10, flowpar[3]); // turbulent friction
        sflow.CENTR0 = flowpar[4]; // entrainment coefficient
        sflow.CDEP0 = flowpar[5]; // deposition coefficient

        sflow.KPMAX[0] = 9999;

    } else if ( sico.MODEL <= 3 ) {

        sflow.RHO1 = flowpar[0]; // PHASE 1 density

        sflow.PHI0[0] = flowpar[1] * sico.PI / 180; // internal friction angle (PHASE 1)
        sflow.DELTA0[0] = flowpar[2] * sico.PI / 180; // basal friction angle (PHASE 1)

        sflow.NY0[0] = pow(10, flowpar[3]); // viscosity (PHASE 1)
        sflow.TAUY[0] = flowpar[4]; // yield strength (PHASE 1)

        sflow.AMBDRAG0 = flowpar[5]; // coefficient for ambient drag
        sflow.FLUFRI0 = flowpar[6]; // fluid friction coefficient
        sflow.CENTR0 = flowpar[7]; // entrainment coefficient
        sflow.CDEP0 = flowpar[8]; // deposition coefficient

        sflow.VIS_chi[0] = flowpar[9]; //1;
        sflow.VIS_ry = flowpar[10]; //10;

        sflow.FRI_exp = 0; //0;
        sflow.NY_exp = 0; //0;
        sflow.KPMAX[0] = flowpar[11]; //1;

    } else if ( sico.MODEL < 7 ) {

        sflow.RHO1 = flowpar[0]; // PHASE 1 density
        sflow.RHO2 = flowpar[1]; // PHASE 2 density

        sflow.PHI0[0] = flowpar[2] * sico.PI / 180; // internal friction angle (PHASE 1)
        sflow.DELTA0[0] = flowpar[3] * sico.PI / 180; // basal friction angle (PHASE 1)
        sflow.PHI0[1] = flowpar[4] * sico.PI / 180; // internal friction angle (PHASE 2)
        sflow.DELTA0[1] = flowpar[5] * sico.PI / 180; // basal friction angle (PHASE 2)

        sflow.NY0[0] = pow(10, flowpar[6]); // viscosity (PHASE 1)
        sflow.TAUY[0] = flowpar[7]; // yield strength (PHASE 1)
        sflow.NY0[1] = pow(10, flowpar[8]); // viscosity (PHASE 2)
        sflow.TAUY[1] = flowpar[9]; // yield strength (PHASE 2)

        sflow.AMBDRAG0 = flowpar[10]; // coefficient for ambient drag
        sflow.FLUFRI0 = flowpar[11]; // fluid friction coefficient
        sflow.CENTR0 = flowpar[12]; // entrainment coefficient
        sflow.CDEP0 = flowpar[13]; // deposition coefficient

        sflow.TRANSSSFS0 = flowpar[14]; // transformation coefficient PHASE 1 - PHASE 2

        sflow.VM_n0 = flowpar[15]; //10;
        sflow.VM_l = flowpar[16]; //0.12;
        sflow.VM_n = flowpar[17]; //1;
        sflow.VM_fact = flowpar[18]; //1;

        sflow.DRAG_k = flowpar[19]; //1;
        sflow.DRAG_m = flowpar[20]; //3;
        sflow.DRAG_n = flowpar[21]; //1;
        sflow.DRAG_vterm = flowpar[22]; //0.1;
        sflow.DRAG_rep = flowpar[23]; //1;
        sflow.DRAG_je = flowpar[24]; //1;

        sflow.VIS_chi[0] = flowpar[25]; //1;
        sflow.VIS_chi[1] = flowpar[26]; //1;
        sflow.VIS_xi[0] = flowpar[27]; //0;
        sflow.VIS_xi[1] = flowpar[28]; //0;
        sflow.VIS_a[0] = flowpar[29]; //1;
        sflow.VIS_a[1] = sflow.VIS_a[0]; //1;
        sflow.VIS_ry = flowpar[30]; //10;

        sflow.FRI_exp = flowpar[31]; //0;
        sflow.NY_exp = flowpar[32]; //0;
        sflow.KPMAX[0] = flowpar[33]; //1;
        sflow.KPMAX[1] = flowpar[34]; //1;

    } else if ( sico.MODEL == 7 ) {

        sflow.RHO1 = flowpar[0]; // PHASE 1 density
        sflow.RHO2 = flowpar[1]; // PHASE 2 density
        sflow.RHO3 = flowpar[2]; // PHASE 3 density

        sflow.PHI0[0] = flowpar[3] * sico.PI / 180; // internal friction angle (PHASE 1)
        sflow.DELTA0[0] = flowpar[4] * sico.PI / 180; // basal friction angle (PHASE 1)
        sflow.PHI0[1] = flowpar[5] * sico.PI / 180; // internal friction angle (PHASE 2)
        sflow.DELTA0[1] = flowpar[6] * sico.PI / 180; // basal friction angle (PHASE 2)
        sflow.PHI0[2] = flowpar[7] * sico.PI / 180; // internal friction angle (PHASE 3)
        sflow.DELTA0[2] = flowpar[8] * sico.PI / 180; // basal friction angle (PHASE 3)

        sflow.NY0[0] = pow(10, flowpar[9]); // viscosity (PHASE 1)
        sflow.TAUY[0] = flowpar[10]; // yield strength (PHASE 1)
        sflow.NY0[1] = pow(10, flowpar[11]); // viscosity (PHASE 2)
        sflow.TAUY[1] = flowpar[12]; // yield strength (PHASE 2)
        sflow.NY0[2] = pow(10, flowpar[13]); // viscosity (PHASE 3)
        sflow.TAUY[2] = flowpar[14]; // yield strength (PHASE 3)

        sflow.AMBDRAG0 = flowpar[15]; // coefficient for ambient drag
        sflow.FLUFRI0 = flowpar[16]; // fluid friction coefficient
        sflow.CENTR0 = flowpar[17]; // entrainment coefficient
        sflow.CDEP0 = flowpar[18]; // deposition coefficient

        sflow.TRANSSSFF0 = flowpar[19]; // transformation coefficient PHASE 1 - PHASE 3
        sflow.TRANSSSFS0 = flowpar[20]; // transformation coefficient PHASE 1 - PHASE 2
        sflow.TRANSFSFF0 = flowpar[21]; // transformation coefficient PHASE 2 - PHASE 3

        sflow.VM_n0 = flowpar[22]; //10;
        sflow.VM_l = flowpar[23]; //0.12;
        sflow.VM_n = flowpar[24]; //1;
        sflow.VM_fact = flowpar[25]; //1;

        sflow.DRAG_k = flowpar[26]; //1;
        sflow.DRAG_m = flowpar[27]; //3;
        sflow.DRAG_n = flowpar[28]; //1;
        sflow.DRAG_vterm = flowpar[29]; //0.1;
        sflow.DRAG_rep = flowpar[30]; //1;
        sflow.DRAG_je = flowpar[31]; //1;

        sflow.VIS_chi[0] = flowpar[32]; //1;
        sflow.VIS_chi[1] = flowpar[33]; //1;
        sflow.VIS_chi[2] = flowpar[34]; //1;
        sflow.VIS_xi[0] = flowpar[35]; //0;
        sflow.VIS_xi[1] = flowpar[36]; //0;
        sflow.VIS_xi[2] = flowpar[37]; //0;
        sflow.VIS_a[0] = flowpar[38]; //1;
        sflow.VIS_a[1] = flowpar[39]; //1;
        sflow.VIS_a[2] = flowpar[40]; //1;
        sflow.VIS_ry = flowpar[41]; //10;

        sflow.FRI_exp = flowpar[42]; //0;
        sflow.NY_exp = flowpar[43]; //0;
        sflow.KPMAX[0] = flowpar[44]; //1;
        sflow.KPMAX[1] = flowpar[45]; //1;
        sflow.KPMAX[2] = flowpar[46]; //1;
    }

    if ( sflow.CENTR0 == sico.UNDEF && sflow.CENTR0 == 0 ) sflow.CENTR0 = 0.0; // default value (no entrainment)
    else sflow.CENTR0 = pow(10, sflow.CENTR0 ); // entrainment coefficient

    if ( sflow.CDEP0 == sico.UNDEF || sflow.CDEP0 == 0 ) sflow.CDEP0 = 0.0; // default value (no deposition)
    else if ( sico.STOPPING == 6 ) sflow.CDEP0 = pow(10, sflow.CDEP0 ); // deposition coefficient

    if ( sico.DYNFRIC == 1 ) {

        sico.FRICMIN = fdparam ( fparam ) * sico.PI / 180; // minimum friction for dynamic adaptation of friction
        sico.EKINREF = fdparam ( fparam ); // reference kinetic energy for dynamic adaptation of friction
    }


    // Thresholds

    sico.HFLOWMIN = fdparam ( fparam ); // minimum flow depth for which momentum is computed
    sico.IMPTHR[0] = fdparam ( fparam ); // threshold for display of flow height
    sico.IMPTHR[1] = fdparam ( fparam ); // threshold for display of flow kinetic energy
    sico.IMPTHR[2] = fdparam ( fparam ); // threshold for display of flow pressure

    tout = fdparam ( fparam ); // time interval for writing output
    tmax = fdparam ( fparam ); // time at which to stop simulation
    sico.SLOMO = fdparam ( fparam ); // factor for slow motion

    sico.GRAVITY = 9.81 / pow( sico.SLOMO, 2 );

    trat = round( tmax / tout );
    tmax = trat * tout; // ensuring that time to stop is multiple of output time step length

    sico.CFL[0] = fdparam ( fparam ); // CFL criterion
    sico.CFL[1] = fdparam ( fparam ); // maximum time step length

    prefix = fcparam ( fparam ); // prefix for output
    outmaps = fcparam ( fparam ); // path and prefix for storing output maps
    outfiles = fcparam ( fparam ); // path and prefix for storing output files

    fclose(fparam); // closing parameter file

    sico.BETACTRL = 1; // control for adapting slope to shifted coordinate system (0=no, 1=yes)


// Input hydrograph data and internal coordinates of hydrographs

    if ( hydrograph == 1 ) { // if at least one input hydrograph is provided:

        if ( sico.MODEL <= 3 ) hydcols = 3;
        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) hydcols = 5;
        else if ( sico.MODEL == 7 ) hydcols = 7;

        hydlmax = 0;
        for ( hydj = 0; hydj < hydnin + hydnout; hydj++ ) if ( hydl[hydj] > hydlmax ) hydlmax = hydl[hydj];

        for ( hydj = 0; hydj < hydnin; hydj++ ) {
            hydhyd0 = finhyd ( hydreleasename[hydj], hydtmax[hydj], hydtmaxx ); // reading data from all input hydrographs
            for ( i=0; i<hydtmaxx+1; i++ ) { for ( j=0; j<hydcols; j++ ) hydhyd[i][j][hydj] = hydhyd0[i][j]; }
            free( hydhyd0[0] ); free( hydhyd0 );
        }

        for ( hydj = 0; hydj < hydnin + hydnout; hydj++ ) { // internal x and y coordinates of all hydrographs:

            hydy[hydj] = (int)(( hydx_metric[hydj] - sico.BDWEST ) / sico.CSZ + 0.5 );
            hydx[hydj] = (int)(( sico.BDNORTH - hydy_metric[hydj] ) / sico.CSZ + 0.5 );
        }
    }


// Initializing release heights and associated volumes

    hflow_max0 = 0; // initializing maximum (PHASE 1) release height
    if ( sico.MODEL > 3 ) hflow_max02 = 0; // initializing maximum PHASE 3 release height
    if ( sico.MODEL == 7 ) hflow_max03 = 0; // initializing maximum PHASE 3 release height
    vol_flow0 = 0; // initializing (PHASE 1) release volume
    if ( sico.MODEL > 3 ) vol_flow02 = 0; // initializing PHASE 3 release volume
    if ( sico.MODEL == 7 ) vol_flow03 = 0; // initializing PHASE 3 release volume


    #ifdef WITHGRASS


    // Reading input raster maps

        if ( sico.NMESH == 1 ) seg_nmesh = finrastd( nmeshname, sico ); // mesh spacing
        seg_elev = finrastd( elevname, sico ); // elevation
        if ( sico.RELM == 1 ) seg_hrelease = finrastd( hreleasename, sico ); // (PHASE 1) release height
        if ( sico.MODEL > 3 && sico.RELM2 == 1 ) seg_hrelease2 = finrastd( hreleasename2, sico ); // PHASE 3 release height
        if ( sico.MODEL == 7 && sico.RELM3 == 1 ) seg_hrelease3 = finrastd( hreleasename3, sico ); // PHASE 3 release height
        if ( sico.RELV == 1 ) seg_vinx = finrastd( vinxname, sico ); // (PHASE 1) release velocity in x direction
        if ( sico.MODEL > 3 && sico.RELV2 == 1 ) seg_vinx2 = finrastd( vinxname2, sico ); // PHASE 3 release velocity in x direction
        if ( sico.MODEL == 7 && sico.RELV3 == 1 ) seg_vinx3 = finrastd( vinxname3, sico ); // PHASE 3 release velocity in x direction
        if ( sico.RELV == 1 ) seg_viny = finrastd( vinyname, sico ); // (PHASE 1) release velocity in y direction
        if ( sico.MODEL > 3 && sico.RELV2 == 1 ) seg_viny2 = finrastd( vinyname2, sico ); // PHASE 3 release velocity in y direction
        if ( sico.MODEL == 7 && sico.RELV3 == 1 ) seg_viny3 = finrastd( vinyname3, sico ); // PHASE 3 release velocity in y direction
        if ( sico.ENTR == 1 ) seg_hentrmax = finrastd( hentrmaxname, sico ); // maximum height of (PHASE 1) entrainment
        if ( sico.MODEL > 3 && sico.ENTR2 == 1 ) seg_hentrmax2 = finrastd( hentrmaxname2, sico ); // maximum height of PHASE 3 entrainment
        if ( sico.MODEL == 7 && sico.ENTR3 == 1 ) seg_hentrmax3 = finrastd( hentrmaxname3, sico ); // maximum height of PHASE 3 entrainment
        if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) seg_centr = finrastd( centrname, sico ); // entrainment coefficient
        if ( sico.STOPPING == 6 && sico.CDEP == 1 ) seg_cdep = finrastd( cdepname, sico ); // deposition coefficient
        if ( sico.PHI == 1 ) seg_phi = finrastd( phiname, sico ); // internal friction angle of PHASE 1
        if ( sico.PHI2 == 1 ) seg_phi2 = finrastd( phi2name, sico ); // internal friction angle of PHASE 2
        if ( sico.PHI3 == 1 ) seg_phi3 = finrastd( phi3name, sico ); // internal friction angle of PHASE 3
        if ( sico.DELTA == 1 ) seg_delta = finrastd( deltaname, sico ); // basal friction angle of PHASE 1
        if ( sico.DELTA2 == 1 ) seg_delta2 = finrastd( delta2name, sico ); // basal friction angle of PHASE 2
        if ( sico.DELTA3 == 1 ) seg_delta3 = finrastd( delta3name, sico ); // basal friction angle of PHASE 3
        if ( sico.NYSS == 1 ) seg_nyss = finrastd( nyssname, sico ); // Viscosity of PHASE 1
        if ( sico.NYFS == 1 ) seg_nyfs = finrastd( nyfsname, sico ); // Viscosity of PHASE 2
        if ( sico.NYFF == 1 ) seg_nyff = finrastd( nyffname, sico ); // Viscosity of PHASE 3
        if ( sico.AMBDRAG == 1 ) seg_ambdrag = finrastd( ambdragname, sico ); // ambient drag coefficient
        if ( sico.FLUFRI == 1 ) seg_flufri = finrastd( flufriname, sico ); // fluid friction coefficient
        if ( sico.TRANSSSFS == 1 ) seg_transssfs = finrastd( transssfsname, sico );
        if ( sico.TRANSSSFF == 1 ) seg_transssff = finrastd( transssffname, sico );
        if ( sico.TRANSFSFF == 1 ) seg_transfsff = finrastd( transfsffname, sico );
        if ( sico.TRELEASE == 1 ) seg_trelease = finrastd( treleasename, sico ); // release time
        if ( sico.TRELSTOP == 1 ) seg_trelstop = finrastd( trelstopname, sico ); // release hydrograph stop time


    // Computing number of pixels in aoi

        sico.IMAX = 0;
        for ( x = 0; x < sico.M / sico.MESH; x++ ) {
            for ( y = 0; y < sico.N / sico.MESH; y++ ) {

                sico.IMAX += 1; // increasing number of pixels in area of interest by 1 for each pixel
	    }
        }


    #endif


// Vectors

    int **cedge, **cedge2, **cedge3, *cedge0, *cedge02, *cedge03, *cneighbours, *cneighbours2, *cneighbours3;
    double *pnmesh, *phrelease, *phrelease2, *phrelease3, *pvinx, *pvinx2, *pvinx3, *pviny, *pviny2, *pviny3,
    *phentrmax, *phentrmax2, *phentrmax3, *pcentr, *pdep, *pphi, *pphi2, *pphi3, *pdelta, *pdelta2, *pdelta3, *pnyss, *pnyfs, *pnyff, *pambdrag, *pflufri,
    *ptransssfs, *ptransssff, *ptransfsff, *ptrelease,
    *ptrelstop, *dentr, *dentr2, *dentr3, **cready, **cready2, **cready3, *fom, *cinner, *cinner2, *cinner3;

    int *cdomain = (int*) calloc( sico.IMAX, sizeof(int));
    int *cdomain2 = (int*) calloc( sico.IMAX, sizeof(int));
    int *cflux = (int*) calloc( sico.IMAX, sizeof(int));
    int *cstopped = (int*) calloc( sico.IMAX, sizeof(int));

    cinner = (double*) calloc( sico.IMAX, sizeof(double));
    if ( sico.MODEL > 3 ) cinner2 = (double*) calloc( sico.IMAX, sizeof(double));
    if ( sico.MODEL == 7 ) cinner3 = (double*) calloc( sico.IMAX, sizeof(double));

    double *betax = (double*) calloc( sico.IMAX, sizeof(double));
    double *betay = (double*) calloc( sico.IMAX, sizeof(double));
    double *betaxy = (double*) calloc( sico.IMAX, sizeof(double));
    double *betaxy0 = (double*) calloc( sico.IMAX, sizeof(double));

    pelev0 = (double*) calloc( sico.IMAX, sizeof(double));


    #ifdef WITHGRASS


        double *v;
        double ***testv = alloc_dmatrix3( sico.M, sico.N, nvect_all );

        v = (double*) calloc( nvect_all, sizeof(double));
        px = (int*) calloc( sico.IMAX, sizeof(int));
        py = (int*) calloc( sico.IMAX, sizeof(int));

        if ( sico.NMESH == 1 ) pnmesh = (double*) calloc( sico.IMAX, sizeof(double));

        pelev = (double*) calloc( sico.IMAX, sizeof(double));
        phrelease = (double*) calloc( sico.IMAX, sizeof(double));
        pvinx = (double*) calloc( sico.IMAX, sizeof(double));
        pviny = (double*) calloc( sico.IMAX, sizeof(double));
        phentrmax = (double*) calloc( sico.IMAX, sizeof(double));

        if ( sico.MODEL > 3 ) {

            phrelease2 = (double*) calloc( sico.IMAX, sizeof(double));
            pvinx2 = (double*) calloc( sico.IMAX, sizeof(double));
            pviny2 = (double*) calloc( sico.IMAX, sizeof(double));
            phentrmax2 = (double*) calloc( sico.IMAX, sizeof(double));
        }

        if ( sico.MODEL == 7 ) {

            phrelease3 = (double*) calloc( sico.IMAX, sizeof(double));
            pvinx3 = (double*) calloc( sico.IMAX, sizeof(double));
            pviny3 = (double*) calloc( sico.IMAX, sizeof(double));
            phentrmax3 = (double*) calloc( sico.IMAX, sizeof(double));
        }

        if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) pcentr = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.STOPPING == 6 && sico.CDEP == 1 ) pdep = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.PHI == 1 ) pphi = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.PHI2 == 1 ) pphi2 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.PHI3 == 1 ) pphi3 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.DELTA == 1 ) pdelta = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.DELTA2 == 1 ) pdelta2 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.DELTA3 == 1 ) pdelta3 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.NYSS == 1 ) pnyss = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.NYFS == 1 ) pnyfs = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.NYFF == 1 ) pnyff = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.AMBDRAG == 1 ) pambdrag = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.FLUFRI == 1 ) pflufri = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.TRANSSSFS == 1 ) ptransssfs = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.TRANSSSFF == 1 ) ptransssff = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.TRANSFSFF == 1 ) ptransfsff = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.TRELEASE == 1 ) ptrelease = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.TRELSTOP == 1 ) ptrelstop = (double*) calloc( sico.IMAX, sizeof(double));


    #endif


    dentr = (double*) calloc( sico.IMAX, sizeof(double));
    if ( sico.MODEL > 3 ) dentr2 = (double*) calloc( sico.IMAX, sizeof(double));
    if ( sico.MODEL == 7 ) dentr3 = (double*) calloc( sico.IMAX, sizeof(double));

    if ( sico.STOPPING > 0 && sico.STOPPING < 4 ) fom = (double*) calloc( sico.IMAX, sizeof(double));

    int **in = alloc_imatrix( sico.IMAX, 9 );
    int **ibasket = alloc_imatrix( 2, sico.IMAX );
    int **icheck = alloc_imatrix( sico.IMAX, 2 );
    int *ib = (int*) calloc( 2, sizeof(int));

    double *asigma_xelev = calloc( sico.IMAX, sizeof(double));
    double *asigma_yelev = calloc( sico.IMAX, sizeof(double));

    double **aw = alloc_dmatrix( sico.IMAX, nvect_all );
    double **awt = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **af = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **ag = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **as = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **ad = alloc_dmatrix( sico.IMAX, sico.NVECTMIN+18 );

    double **asigma_x = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **asigma_y = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **asigma_f = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **asigma_g = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );

    double **wintelev = alloc_dmatrix( sico.IMAX, 4 );
    double ***winta = alloc_dmatrix3( sico.IMAX, 4, sico.NVECTMIN );
    double ***wintb = alloc_dmatrix3( sico.IMAX, 6, sico.NVECTMIN );
    double ***wintc = alloc_dmatrix3( sico.IMAX, 4, sico.NVECTMIN );
    double **wintd = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );
    double **wintdtest = alloc_dmatrix( sico.IMAX, sico.NVECTMIN );

    double **f = alloc_dmatrix( 4, sico.NVECTMIN );
    double **g = alloc_dmatrix( 4, sico.NVECTMIN );
    double **s = alloc_dmatrix( 4, sico.NVECTMIN );
    double ***d = alloc_dmatrix3( sico.IMAX, 6, sico.NVECTMIN+18 );

    double *dx = (double*) calloc( sico.IMAX, sizeof(double));
    double *dy = (double*) calloc( sico.IMAX, sizeof(double));

    if ( sico.DIFFCTRL != 0 ) { // if diffusion control is activated:

        cedge = alloc_imatrix( sico.IMAX, 9 );
        cready = alloc_dmatrix( sico.IMAX, 9 );
        cedge2 = alloc_imatrix( sico.IMAX, 9 );
        cready2 = alloc_dmatrix( sico.IMAX, 9 );
        cedge3 = alloc_imatrix( sico.IMAX, 9 );
        cready3 = alloc_dmatrix( sico.IMAX, 9 );
        cedge0 = (int*) calloc( sico.IMAX, sizeof(int));
        cedge02 = (int*) calloc( sico.IMAX, sizeof(int));
        cedge03 = (int*) calloc( sico.IMAX, sizeof(int));
        cneighbours = (int*) calloc( sico.IMAX, sizeof(int));
        cneighbours2 = (int*) calloc( sico.IMAX, sizeof(int));
        cneighbours3 = (int*) calloc( sico.IMAX, sizeof(int));
    }


// Variables for names of output raster maps

    mv0 = (char**) calloc( 1000, sizeof(char*));

    if ( sico.MODEL <= 3 ) {

        mv0[0] = "hflow"; mv0[1] = "vflowx"; mv0[2] = "vflowy"; mv0[3] = "basechange"; mv0[4] = "vflow";
        mv0[5] = "tflow"; mv0[6] = "pflow"; mv0[7] = "hflow_max"; mv0[8] = "vflow_max";
        mv0[9] = "tflow_max"; mv0[10] = "pflow_max";
    }
    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

        mv0[0] = "hflow1"; mv0[1] = "vflowx1"; mv0[2] = "vflowy1"; mv0[3] = "hflow2"; mv0[4] = "vflowx2"; mv0[5] = "vflowy2";
        mv0[6] = "basechange1"; mv0[7] = "basechange2"; mv0[8] = "vflow1"; mv0[9] = "vflow2"; mv0[10] = "hflow"; mv0[11] = "tflow1";
        mv0[12] = "tflow2"; mv0[13] = "tflow"; mv0[14] = "pflow1"; mv0[15] = "pflow2"; mv0[16] = "pflow"; mv0[17] = "basechange";
        mv0[18] = "hflow1_max"; mv0[19] = "vflow1_max"; mv0[20] = "hflow2_max"; mv0[21] = "vflow2_max"; mv0[22] = "hflow_max";
        mv0[23] = "tflow1_max"; mv0[24] = "pflow1_max"; mv0[25] = "tflow2_max"; mv0[26] = "pflow2_max"; mv0[27] = "tflow_max";
        mv0[28] = "pflow_max";
    }
    else if ( sico.MODEL == 7 ) {

        mv0[0]  = "hflow1";     mv0[1]  = "vflowx1";    mv0[2]  = "vflowy1";    mv0[3]  = "hflow2";      mv0[4]  = "vflowx2";     mv0[5] = "vflowy2";
        mv0[6]  = "hflow3";     mv0[7]  = "vflowx3";    mv0[8]  = "vflowy3";    mv0[9]  = "basechange1"; mv0[10] = "basechange2"; mv0[11] = "basechange3";
        mv0[12] = "vflow1";     mv0[13] = "vflow2";     mv0[14] = "vflow3";     mv0[15] = "hflow";       mv0[16] = "tflow1";      mv0[17] = "tflow2";
        mv0[18] = "tflow3";     mv0[19] = "tflow";      mv0[20] = "pflow1";     mv0[21] = "pflow2";      mv0[22] = "pflow3";      mv0[23] = "pflow";
        mv0[24] = "basechange"; mv0[25] = "hflow1_max"; mv0[26] = "vflow1_max"; mv0[27] = "hflow2_max";  mv0[28] = "vflow2_max";  mv0[29] = "hflow3_max";
        mv0[30] = "vflow3_max"; mv0[31] = "hflow_max";  mv0[32] = "tflow1_max"; mv0[33] = "pflow1_max";  mv0[34] = "tflow2_max";  mv0[35] = "pflow2_max";
        mv0[36] = "tflow3_max"; mv0[37] = "pflow3_max"; mv0[38] = "tflow_max";  mv0[39] = "pflow_max";
    }


// Defining, opening and preparing output files

    FILE *f_summary, *f_directions, *f_directions2, *f_directions3, *f_nout, *f_hydout, *f_hydinfo[hydnin+hydnout];

    if ( sico.MULT == 0 ) sprintf(path, "%s%ssummary.txt", outfiles, prefix); // summary file
    else sprintf(path, "%s%ssummary%d.txt", outfiles, prefix, xint);
    f_summary=fopen(path, "w");

    if ( hydrograph == 1 ) {

        sprintf(path, "%s%shydprofiles.txt", outfiles, prefix); // hydrograph profiles file
        f_hydout=fopen(path, "w");

        fprintf(f_hydout, "ID\tx1\txC\tx2\ty1\tyC\ty2\n"); // printing header of hydrograph profiles file

        if ( sico.MULT == 0 ) {

            for ( i = 0; i < hydnin + hydnout; i++ ) { // loop over all hydrographs:

                sprintf(path, "%s%shydinfo%i.txt", outfiles, prefix, i+1); // hydrograph info file
                f_hydinfo[i]=fopen(path, "w");

                if ( i >= hydnin ) fprintf(f_hydinfo[i],
                    "T\tH1\tV1\tE1\tQ1\tH2\tV2\tE2\tQ2\tH3\tV3\tE3\tQ3\n0.0\t0.00\t0.00\t0.00\t0.00\t0.00\t0.00\t0.00\t0.00\t0.00\t0.00\t0.00\t0.00\n");
                else fprintf(f_hydinfo[i], "T\tH1\tV1\tE1\tQ1\tH2\tV2\tE2\tQ2\tH3\tV3\tE3\tQ3\n"); // printing header of hydrograph info file
            }

        }
    }

    if ( sico.MULT == 0 ) {
        sprintf(path, "%s%sdirections1.txt", outfiles, prefix); // (PHASE 1) flow direction file (for display as arrows)
        f_directions=fopen(path, "w");

        if ( sico.MODEL > 3 ) {
            sprintf(path, "%s%sdirections2.txt", outfiles, prefix); // PHASE 3 flow direction file (for display as arrows)
            f_directions2=fopen(path, "w");
        }

        if ( sico.MODEL == 7 ) {
            sprintf(path, "%s%sdirections3.txt", outfiles, prefix); // PHASE 3 flow direction file (for display as arrows)
            f_directions3=fopen(path, "w");
        }
    }


    #ifdef WITHGRASS


    // Preparing segment files for processing and output (if GRASS is used)

        //SEGMENT *seg_v = (SEGMENT*) calloc( nvect_all, sizeof(SEGMENT));

        //for ( k=0; k<nvect_all; k++ ) // segment files for all vector components
            //Segment_open (&seg_v[k], G_tempfile(), sico.M, sico.N, sico.NSEGRC, sico.NSEGRC, sizeof(double), sico.NSEGS);


    // Initializing and preparing working and output variables (if GRASS is used)

        i = 0;

        for ( x = 0; x < sico.M; x++ ) { for ( y = 0; y < sico.N; y++ ) { // loop over all pixels:

            aoi = 1;
            if ( sico.NMESH == 1 ) Segment_get(&seg_nmesh, &nmesh, x, y); // mesh spacing
            Segment_get(&seg_elev, &elev, x, y); // elevation
            if ( sico.RELM == 1 ) Segment_get(&seg_hrelease, &hrelease, x, y); // (PHASE 1) release height
            if ( sico.MODEL > 3 && sico.RELM2 == 1 ) Segment_get(&seg_hrelease2, &hrelease2, x, y); //PHASE 3 release height
            if ( sico.MODEL == 7 && sico.RELM3 == 1 ) Segment_get(&seg_hrelease3, &hrelease3, x, y); //PHASE 3 release height
            if ( sico.RELV == 1 ) Segment_get(&seg_vinx, &vinx, x, y); // (PHASE 1) release velocity in x direction
            if ( sico.MODEL > 3 && sico.RELV2 == 1 ) Segment_get(&seg_vinx2, &vinx2, x, y); //PHASE 3 release velocity in x direction
            if ( sico.MODEL == 7 && sico.RELV3 == 1 ) Segment_get(&seg_vinx3, &vinx3, x, y); //PHASE 3 release velocity in x direction
            if ( sico.RELV == 1 ) Segment_get(&seg_viny, &viny, x, y); // (PHASE 1) release velocity in y direction
            if ( sico.MODEL > 3 && sico.RELV2 == 1 ) Segment_get(&seg_viny2, &viny2, x, y); //PHASE 3 release velocity in y direction
            if ( sico.MODEL == 7 && sico.RELV3 == 1 ) Segment_get(&seg_viny3, &viny3, x, y); //PHASE 3 release velocity in y direction
            if ( sico.ENTR == 1 ) Segment_get(&seg_hentrmax, &hentrmax, x, y); // maximum height of (PHASE 1) entrainment
            if ( sico.MODEL > 3 && sico.ENTR2 == 1 ) Segment_get(&seg_hentrmax2, &hentrmax2, x, y); // maximum height of PHASE 3 entrainment
            if ( sico.MODEL == 7 && sico.ENTR3 == 1 ) Segment_get(&seg_hentrmax3, &hentrmax3, x, y); // maximum height of PHASE 3 entrainment
            if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) Segment_get(&seg_centr, &centr, x, y); // entrainment coefficient
            if ( sico.STOPPING == 6 && sico.CDEP == 1 ) Segment_get(&seg_cdep, &cdep, x, y); // deposition coefficient
            if ( sico.PHI == 1 ) Segment_get(&seg_phi, &phi, x, y); // internal friction angle of PHASE 1
            if ( sico.PHI2 == 1 ) Segment_get(&seg_phi2, &phi2, x, y); // internal friction angle of PHASE 2
            if ( sico.PHI3 == 1 ) Segment_get(&seg_phi3, &phi3, x, y); // internal friction angle of PHASE 3
            if ( sico.DELTA == 1 ) Segment_get(&seg_delta, &delta, x, y); // basal friction angle of PHASE 1
            if ( sico.DELTA2 == 1 ) Segment_get(&seg_delta2, &delta2, x, y); // basal friction angle of PHASE 2
            if ( sico.DELTA3 == 1 ) Segment_get(&seg_delta3, &delta3, x, y); // basal friction angle of PHASE 3
            if ( sico.NYSS == 1 ) Segment_get(&seg_nyss, &nyss, x, y); // Viscosity of PHASE 1
            if ( sico.NYFS == 1 ) Segment_get(&seg_nyfs, &nyfs, x, y); // Viscosity of PHASE 2
            if ( sico.NYFF == 1 ) Segment_get(&seg_nyff, &nyff, x, y); // Viscosity of PHASE 3
            if ( sico.AMBDRAG == 1 ) Segment_get(&seg_ambdrag, &ambdrag, x, y); // ambient drag coefficient
            if ( sico.FLUFRI == 1 ) Segment_get(&seg_flufri, &flufri, x, y); // fluid friction coefficient
            if ( sico.TRANSSSFS == 1 ) Segment_get(&seg_transssfs, &transssfs, x, y);
            if ( sico.TRANSSSFF == 1 ) Segment_get(&seg_transssff, &transssff, x, y);
            if ( sico.TRANSFSFF == 1 ) Segment_get(&seg_transfsff, &transfsff, x, y);
            if ( sico.TRELEASE == 1 ) Segment_get(&seg_trelease, &trelease, x, y); // release time
            if ( sico.TRELSTOP == 1 ) Segment_get(&seg_trelstop, &trelstop, x, y); // release hydrograph stop time

            if ( aoi >= 1 ) { // if pixel is in area of interest:

                px[i] = x; // internal x coordinate of pixel
                py[i] = y; // internal y coordinate of pixel

                if ( sico.NMESH == 1 && nmesh != sico.UNDEF ) pnmesh[i] = nmesh;
                else if ( sico.NMESH == 1 ) pnmesh[i] = sico.UNDEF;

                pelev[i] = elev; pelev0[i] = elev;

                if ( sico.RELM == 1 && hrelease != sico.UNDEF ) phrelease[i] = hrelease; // writing input to arrays
                else phrelease[i] = 0;
                if ( sico.MODEL > 3 && sico.RELM2 == 1  && hrelease2 != sico.UNDEF ) phrelease2[i] = hrelease2;
                else if ( sico.MODEL > 3 ) phrelease2[i] = 0;
                if ( sico.MODEL == 7 && sico.RELM3 == 1  && hrelease3 != sico.UNDEF ) phrelease3[i] = hrelease3;
                else if ( sico.MODEL == 7 ) phrelease3[i] = 0;
                if ( sico.RELV == 1 && vinx != sico.UNDEF ) pvinx[i] = vinx;
                else pvinx[i] = 0;
                if ( sico.MODEL > 3 && sico.RELV2 == 1  && vinx2 != sico.UNDEF ) pvinx2[i] = vinx2;
                else if ( sico.MODEL > 3 ) pvinx2[i] = 0;
                if ( sico.MODEL == 7 && sico.RELV3 == 1  && vinx3 != sico.UNDEF ) pvinx3[i] = vinx3;
                else if ( sico.MODEL == 7 ) pvinx3[i] = 0;
                if ( sico.RELV == 1 && viny != sico.UNDEF ) pviny[i] = viny;
                else pviny[i] = 0;
                if ( sico.MODEL > 3 && sico.RELV2 == 1  && viny2 != sico.UNDEF ) pviny2[i] = viny2;
                else if ( sico.MODEL > 3 ) pviny2[i] = 0;
                if ( sico.MODEL == 7 && sico.RELV3 == 1  && viny3 != sico.UNDEF ) pviny3[i] = viny3;
                else if ( sico.MODEL == 7 ) pviny3[i] = 0;
                if ( sico.ENTR == 1 ) phentrmax[i] = hentrmax;
                else phentrmax[i] = 9999;
                if ( sico.MODEL > 3 && sico.ENTR2 == 1 && hentrmax2 != sico.UNDEF ) phentrmax2[i] = hentrmax2;
                else if ( sico.MODEL > 3 ) phentrmax2[i] = 9999;
                if ( sico.MODEL == 7 && sico.ENTR3 == 1 && hentrmax3 != sico.UNDEF ) phentrmax3[i] = hentrmax3;
                else if ( sico.MODEL == 7 ) phentrmax3[i] = 9999;
                if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 && centr != sico.UNDEF ) pcentr[i] = pow( 10, centr );
                else if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 )  pcentr[i] = sico.UNDEF;
                if ( sico.STOPPING == 6 && sico.CDEP == 1 && cdep != sico.UNDEF ) pdep[i] = pow( 10, cdep );
                else if ( sico.STOPPING == 6 && sico.CDEP == 1 ) pdep[i] = sico.UNDEF;

                if ( sico.PHI == 1 && phi != sico.UNDEF ) pphi[i] = phi * sico.PI / 180;
                else if ( sico.PHI == 1 ) pphi[i] = sico.UNDEF;
                if ( sico.PHI2 == 1 && phi2 != sico.UNDEF ) pphi2[i] = phi2 * sico.PI / 180;
                else if ( sico.PHI2 == 1 ) pphi2[i] = sico.UNDEF;
                if ( sico.PHI3 == 1 && phi3 != sico.UNDEF ) pphi3[i] = phi3 * sico.PI / 180;
                else if ( sico.PHI3 == 1 ) pphi3[i] = sico.UNDEF;
                if ( sico.DELTA == 1 && delta != sico.UNDEF ) pdelta[i] = delta * sico.PI / 180;
                else if ( sico.DELTA == 1 ) pdelta[i] = sico.UNDEF;
                if ( sico.DELTA2 == 1 && delta2 != sico.UNDEF ) pdelta2[i] = delta2 * sico.PI / 180;
                else if ( sico.DELTA2 == 1 ) pdelta2[i] = sico.UNDEF;
                if ( sico.DELTA3 == 1 && delta3 != sico.UNDEF ) pdelta3[i] = delta3 * sico.PI / 180;
                else if ( sico.DELTA3 == 1 ) pdelta3[i] = sico.UNDEF;
                if ( sico.NYSS == 1 && nyss != sico.UNDEF ) pnyss[i] = pow( 10, nyss );
                else if ( sico.NYSS == 1 ) pnyss[i] = sico.UNDEF;
                if ( sico.NYFS == 1 && nyfs != sico.UNDEF ) pnyfs[i] = pow( 10, nyfs );
                else if ( sico.NYFS == 1 ) pnyfs[i] = sico.UNDEF;
                if ( sico.NYFF == 1 && nyff != sico.UNDEF ) pnyff[i] = pow( 10, nyff );
                else if ( sico.NYFF == 1 ) pnyff[i] = sico.UNDEF;
                if ( sico.AMBDRAG == 1 && ambdrag != sico.UNDEF ) pambdrag[i] = ambdrag;
                else if ( sico.AMBDRAG == 1 ) pambdrag[i] = sico.UNDEF;
                if ( sico.FLUFRI == 1 && flufri != sico.UNDEF ) pflufri[i] = flufri;
                else if ( sico.FLUFRI == 1 ) pflufri[i] = sico.UNDEF;
                if ( sico.TRANSSSFS == 1 && transssfs != sico.UNDEF ) ptransssfs[i] = transssfs;
                else if ( sico.TRANSSSFS == 1 ) ptransssfs[i] = sico.UNDEF;
                if ( sico.TRANSSSFF == 1 && transssff != sico.UNDEF ) ptransssff[i] = transssff;
                else if ( sico.TRANSSSFF == 1 ) ptransssff[i] = sico.UNDEF;
                if ( sico.TRANSFSFF == 1 && transfsff != sico.UNDEF ) ptransfsff[i] = transfsff;
                else if ( sico.TRANSFSFF == 1 ) ptransfsff[i] = sico.UNDEF;
                if ( sico.TRELEASE == 1 && trelease != sico.UNDEF ) ptrelease[i] = trelease;
                else if ( sico.TRELEASE == 1 ) ptrelease[i] = sico.UNDEF;
                if ( sico.TRELSTOP == 1 && trelstop != sico.UNDEF ) ptrelstop[i] = trelstop;
                else if ( sico.TRELSTOP == 1 ) ptrelstop[i] = sico.UNDEF;

                i += 1;
            }

            for ( k=0; k<nvect_all; k++ ) { // preparing segment files for output rasters:

                //v[k] = 0;
                //Segment_put(&seg_v[k], &v[k], x, y);
            }
        }}
        //for ( k=0; k<nvect_all; k++ ) Segment_flush(&seg_v[k]);


    #else


// Initializing and preparing working and output variables (if GRASS is not used)

        px = finascxy( elevname, 1, sico ); // internal x coordinate of pixel
        py = finascxy( elevname, 2, sico ); // internal y coordinate of pixel

        if ( sico.NMESH == 1 ) pnmesh = finascval( nmeshname, px, py, sico ); // mesh spacing

        pelev = finascval( elevname, px, py, sico ); // elevation
        if ( sico.RELM == 1 ) phrelease = finascval( hreleasename, px, py, sico ); // (PHASE 1) release height
        else phrelease = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL > 3 && sico.RELM2 == 1 ) phrelease2 = finascval( hreleasename2, px, py, sico ); // PHASE 3 release height
        else if ( sico.MODEL > 3 ) phrelease2 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL == 7 && sico.RELM3 == 1 ) phrelease3 = finascval( hreleasename3, px, py, sico ); // PHASE 3 release height
        else if ( sico.MODEL == 7 ) phrelease3 = (double*) calloc( sico.IMAX, sizeof(double));

        if ( sico.RELV == 1 ) pvinx = finascval( vinxname, px, py, sico ); // (PHASE 1) release velocity in x direction
        else pvinx = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL > 3 && sico.RELV2 == 1 ) pvinx2 = finascval( vinxname2, px, py, sico ); // PHASE 3 release velocity in x direction
        else if ( sico.MODEL > 3 ) pvinx2 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL == 7 && sico.RELV3 == 1 ) pvinx3 = finascval( vinxname3, px, py, sico ); // PHASE 3 release velocity in x direction
        else if ( sico.MODEL == 7 ) pvinx3 = (double*) calloc( sico.IMAX, sizeof(double));

        if ( sico.RELV == 1 ) pviny = finascval( vinyname, px, py, sico ); // (PHASE 1) release velocity in y direction
        else pviny = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL > 3 && sico.RELV2 == 1 ) pviny2 = finascval( vinyname2, px, py, sico ); // PHASE 3 release velocity in y direction
        else if ( sico.MODEL > 3 ) pviny2 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL == 7 && sico.RELV3 == 1 ) pviny3 = finascval( vinyname3, px, py, sico ); // PHASE 3 release velocity in y direction
        else if ( sico.MODEL == 7 ) pviny3 = (double*) calloc( sico.IMAX, sizeof(double));

        if ( sico.ENTR == 1 ) phentrmax = finascval( hentrmaxname, px, py, sico ); // maximum height of (PHASE 1) entrainment
        else phentrmax = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL > 3 && sico.ENTR2 == 1 ) phentrmax2 = finascval( hentrmaxname2, px, py, sico ); // max. height of PHASE 3 entrainment
        else if ( sico.MODEL > 3 ) phentrmax2 = (double*) calloc( sico.IMAX, sizeof(double));
        if ( sico.MODEL == 7 && sico.ENTR3 == 1 ) phentrmax3 = finascval( hentrmaxname3, px, py, sico ); // max. height of PHASE 3 entrainment
        else if ( sico.MODEL == 7 ) phentrmax3 = (double*) calloc( sico.IMAX, sizeof(double));

        if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) pcentr = finascval( centrname, px, py, sico ); // entrainment coefficient
        if ( sico.STOPPING == 6 && sico.CDEP == 1 ) pdep = finascval( cdepname, px, py, sico ); // deposition coefficient
        if ( sico.PHI == 1 ) pphi = finascval( phiname, px, py, sico ); // internal friction angle of PHASE 1
        if ( sico.PHI2 == 1 ) pphi2 = finascval( phi2name, px, py, sico ); // internal friction angle of PHASE 2
        if ( sico.PHI3 == 1 ) pphi3 = finascval( phi3name, px, py, sico ); // internal friction angle of PHASE 3
        if ( sico.DELTA == 1 ) pdelta = finascval( deltaname, px, py, sico ); // basal friction angle of PHASE 1
        if ( sico.DELTA2 == 1 ) pdelta2 = finascval( delta2name, px, py, sico ); // basal friction angle of PHASE 2
        if ( sico.DELTA3 == 1 ) pdelta3 = finascval( delta3name, px, py, sico ); // basal friction angle of PHASE 3
        if ( sico.NYSS == 1 ) pnyss = finascval( nyssname, px, py, sico ); // Viscosity of PHASE 1
        if ( sico.NYFS == 1 ) pnyfs = finascval( nyfsname, px, py, sico ); // Viscosity of PHASE 2
        if ( sico.NYFF == 1 ) pnyff = finascval( nyffname, px, py, sico ); // Viscosity of PHASE 3
        if ( sico.AMBDRAG == 1 ) pambdrag = finascval( ambdragname, px, py, sico ); // ambient drag coefficient
        if ( sico.FLUFRI == 1 ) pflufri = finascval( flufriname, px, py, sico ); // fluid friction coefficient
        if ( sico.TRANSSSFS == 1 ) ptransssfs = finascval( transssfsname, px, py, sico );
        if ( sico.TRANSSSFF == 1 ) ptransssff = finascval( transssffname, px, py, sico );
        if ( sico.TRANSFSFF == 1 ) ptransfsff = finascval( transfsffname, px, py, sico );
        if ( sico.TRELEASE == 1 ) ptrelease = finascval( treleasename, px, py, sico ); // release time
        if ( sico.TRELSTOP == 1 ) ptrelstop = finascval( trelstopname, px, py, sico ); // release hydrograph stop time

        for ( i=0; i<sico.IMAX; i++ ) { // finalizing input arrays:

            if ( sico.NMESH == 1 && pnmesh[i] != sico.UNDEF ) pnmesh[i] = pnmesh[i];
            else if ( sico.NMESH == 1 ) pcentr[i] = sico.UNDEF;

            pelev0[i] = pelev[i];
            if ( sico.RELM != 1 || phrelease[i] == sico.UNDEF ) phrelease[i] = 0;
            if ( sico.MODEL > 3 && ( sico.RELM2 != 1 || phrelease2[i] == sico.UNDEF )) phrelease2[i] = 0;
            if ( sico.MODEL == 7 && ( sico.RELM3 != 1 || phrelease3[i] == sico.UNDEF )) phrelease3[i] = 0;

            if ( sico.RELV != 1 || pvinx[i] == sico.UNDEF ) pvinx[i] = 0;
            if ( sico.MODEL > 3 && ( sico.RELV2 != 1 || pvinx2[i] == sico.UNDEF )) pvinx2[i] = 0;
            if ( sico.MODEL == 7 && ( sico.RELV3 != 1 || pvinx3[i] == sico.UNDEF )) pvinx3[i] = 0;

            if ( sico.RELV != 1 || pviny[i] == sico.UNDEF ) pviny[i] = 0;
            if ( sico.MODEL > 3 && ( sico.RELV2 != 1 || pviny2[i] == sico.UNDEF )) pviny2[i] = 0;
            if ( sico.MODEL == 7 && ( sico.RELV3 != 1 || pviny3[i] == sico.UNDEF )) pviny3[i] = 0;

            if ( sico.ENTR != 1 || phentrmax[i] == sico.UNDEF ) phentrmax[i] = 9999;
            if ( sico.MODEL > 3 && ( sico.ENTR2 != 1 || phentrmax2[i] == sico.UNDEF )) phentrmax2[i] = 9999;
            if ( sico.MODEL == 7 && ( sico.ENTR3 != 1 || phentrmax3[i] == sico.UNDEF )) phentrmax3[i] = 9999;

            if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 && pcentr[i] != sico.UNDEF ) pcentr[i] = pow( 10, pcentr[i] );
            else if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) pcentr[i] = sico.UNDEF;
            if ( sico.STOPPING == 6 && sico.CDEP == 1 && pdep[i] != sico.UNDEF ) pdep[i] = pow( 10, pdep[i] );
            else if ( sico.STOPPING == 6 && sico.CDEP == 1 ) pdep[i] = sico.UNDEF;
            if ( sico.PHI == 1 && pphi[i] != sico.UNDEF ) pphi[i] = pphi[i] * sico.PI / 180;
            else if ( sico.PHI == 1 ) pphi[i] = sico.UNDEF;
            if ( sico.PHI2 == 1 && pphi2[i] != sico.UNDEF ) pphi2[i] = pphi2[i] * sico.PI / 180;
            else if ( sico.PHI2 == 1 ) pphi2[i] = sico.UNDEF;
            if ( sico.PHI3 == 1 && pphi3[i] != sico.UNDEF ) pphi3[i] = pphi3[i] * sico.PI / 180;
            else if ( sico.PHI3 == 1 ) pphi3[i] = sico.UNDEF;
            if ( sico.DELTA == 1 && pdelta[i] != sico.UNDEF ) pdelta[i] = pdelta[i] * sico.PI / 180;
            else if ( sico.DELTA == 1 ) pdelta[i] = sico.UNDEF;
            if ( sico.DELTA2 == 1 && pdelta2[i] != sico.UNDEF ) pdelta2[i] = pdelta2[i] * sico.PI / 180;
            else if ( sico.DELTA2 == 1 ) pdelta2[i] = sico.UNDEF;
            if ( sico.DELTA3 == 1 && pdelta3[i] != sico.UNDEF ) pdelta3[i] = pdelta3[i] * sico.PI / 180;
            else if ( sico.DELTA3 == 1 ) pdelta3[i] = sico.UNDEF;
            if ( sico.NYSS == 1 && pnyss[i] != sico.UNDEF ) pnyss[i] = pow( 10, pnyss[i] );
            else if ( sico.NYSS == 1 ) pnyss[i] = sico.UNDEF;
            if ( sico.NYFS == 1 && pnyfs[i] != sico.UNDEF ) pnyfs[i] = pow( 10, pnyfs[i] );
            else if ( sico.NYFS == 1 ) pnyfs[i] = sico.UNDEF;
            if ( sico.NYFF == 1 && pnyff[i] != sico.UNDEF ) pnyff[i] = pow( 10, pnyff[i] );
            else if ( sico.NYFF == 1 ) pnyff[i] = sico.UNDEF;
            if ( sico.AMBDRAG == 1 && pambdrag[i] != sico.UNDEF ) pambdrag[i] = pambdrag[i];
            else if ( sico.AMBDRAG == 1 ) pambdrag[i] = sico.UNDEF;
            if ( sico.FLUFRI == 1 && pflufri[i] != sico.UNDEF ) pflufri[i] = pflufri[i];
            else if ( sico.FLUFRI == 1 ) pflufri[i] = sico.UNDEF;
            if ( sico.TRANSSSFS == 1 && ptransssfs[i] != sico.UNDEF ) ptransssfs[i] = ptransssfs[i];
            else if ( sico.TRANSSSFS == 1 ) ptransssfs[i] = sico.UNDEF;
            if ( sico.TRANSSSFF == 1 && ptransssff[i] != sico.UNDEF ) ptransssff[i] = ptransssff[i];
            else if ( sico.TRANSSSFF == 1 ) ptransssff[i] = sico.UNDEF;
            if ( sico.TRANSFSFF == 1 && ptransfsff[i] != sico.UNDEF ) ptransfsff[i] = ptransfsff[i];
            else if ( sico.TRANSFSFF == 1 ) ptransfsff[i] = sico.UNDEF;
            if ( sico.TRELEASE == 1 && ptrelease[i] != sico.UNDEF ) ptrelease[i] = ptrelease[i];
            else if ( sico.TRELEASE == 1 ) ptrelease[i] = sico.UNDEF;
            if ( sico.TRELSTOP == 1 && ptrelstop[i] != sico.UNDEF ) ptrelstop[i] = ptrelstop[i];
            else if ( sico.TRELSTOP == 1 ) ptrelstop[i] = sico.UNDEF;
        }


    #endif


    for ( i=0; i<sico.IMAX; i++ ) { // updating flow height maxima and volumes:

        if ( sico.RELM == 1 ) {
            if ( phrelease[i] > hflow_max0 ) hflow_max0 = phrelease[i]; // updating maximum (PHASE 1) release height
            vol_flow0 += phrelease[i] * pow ( sico.CSZ, 2 ); // updating (PHASE 1) release volume
        }
        if ( sico.MODEL > 3 && sico.RELM2 == 1 ) {
            if ( phrelease2[i] > hflow_max02 ) hflow_max02 = phrelease2[i]; // updating maximum PHASE 3 release height
            vol_flow02 += phrelease2[i] * pow ( sico.CSZ, 2 ); // updating PHASE 3 release volume
        }
        if ( sico.MODEL == 7 && sico.RELM3 == 1 ) {
            if ( phrelease3[i] > hflow_max03 ) hflow_max03 = phrelease3[i]; // updating maximum PHASE 3 release height
            vol_flow03 += phrelease3[i] * pow ( sico.CSZ, 2 ); // updating PHASE 3 release volume
        }
    }


// Defining computational domains

    // *** cdomain[i] = 0: 1st row edge pixels
    // *** cdomain[i] = 1: all other pixels

    // *** cdomain2[i] = 1: 1st, 2nd or 3rd row edge pixels
    // *** cdomain2[i] = 0: all other pixels

    ib[0] = 0;

    for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels:

        iin = fin( i, pelev, sico ); // pixel neighbourhood
        for ( j=0; j<9; j++ ) in[i][j] = iin[j];
        ctrlv = 1; ctrlvv = 1; ctrlvvv = 1; // resetting controls
        for ( j=1; j<9; j++ ) { // loop over all neighbour pixels:

            if ( in[i][j] < 0 || in[i][j] >= sico.IMAX ) ctrlv = 0; // 1st row edge pixels

            inn = fin( in[i][j], pelev, sico ); // neighbourhood of neighbour pixel
            for ( jj=1; jj<9; jj++ ) { // loop over neighbourhood
                if ( inn[jj] < 0 || inn[jj] >= sico.IMAX ) ctrlvv = 0; // 2nd row edge pixels

                innn = fin( inn[jj], pelev, sico ); // neighbourhood of neighbour pixel
                for ( jjj=1; jjj<9; jjj++ ) // loop over neighbourhood
                    if ( innn[jjj] < 0 || innn[jjj] >= sico.IMAX ) ctrlvvv = 0; // 3rd row edge pixels

                free( innn ); // releasing memory
            }
            free( inn ); // releasing memory
        }
        free( iin ); // releasing memory

        icheck[i][0] = 0;
        icheck[i][1] = 0;

        cflux[i] = 1;

        if ( ctrlv == 0 ) { cdomain[i] = 0; cdomain2[i] = 1; }
        else if ( ctrlvv == 0 || ctrlvvv == 0 ) { cdomain[i] = 1; cdomain2[i] = 1; }
        else { cdomain[i] = 1; cdomain2[i] = 0; }

        if ( cdomain2[i] == 0 ) {

            ibasket[0][ib[0]] = i;
            ib[0] += 1;
        }
    }


// Computing slopes and topography-following pixel sizes

    for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels:

        if ( cdomain[i] != 0 ) { // if pixel is no 1st row edge pixel:

            betax[i] = fbeta( pelev[in[i][5]], pelev[in[i][2]], 2.0, sico ); // slopes
            betay[i] = fbeta( pelev[in[i][4]], pelev[in[i][1]], 2.0, sico );
            betaxy[i] = fbetaxy( betax[i], betay[i] );

        } else { betax[i] = 0; betay[i] = 0; betaxy[i] = 0; }

        betaxy0[i] = betaxy[i];

        if ( sico.CORRHEIGHT != 0 ) {

            dx[i] = sico.CSZ / cos( betax[i] ); // topography-following pixel spacing
            dy[i] = sico.CSZ / cos( betay[i] );

        } else {

            dx[i] = sico.CSZ;
            dy[i] = sico.CSZ;
        }
    }


// Converting heights into depths and deriving release momenta

    for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels of reduced coordinate system:

        if ( sico.TRELEASE == 1 ) ttrelease = ptrelease[i]; // release time
        else ttrelease = 0;

        if ( sico.MODEL <= 3 ) {

            if ( ttrelease == 0 ) {

                aw[i][0] = fconvin( phrelease, phrelease, phrelease, hflow_max0, 1, px, py, pelev, betax, betay, betaxy, i, sico ); // release depth
                aw[i][1] = aw[i][0] * pviny[i] * -1; // release momentum in x direction
                aw[i][2] = aw[i][0] * pvinx[i]; // release momentum in y direction
            }
            else { aw[i][0] = 0; aw[i][1] = 0; aw[i][2] = 0; }
        }

        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

            if ( ttrelease == 0 ) {

                aw[i][0] = fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 1,
                    px, py, pelev, betax, betay, betaxy, i, sico ); // PHASE 1 release depth
                aw[i][3] = fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 2,
                    px, py, pelev, betax, betay, betaxy, i, sico ); // PHASE 3 release depth

                aw[i][1] = aw[i][0] * pviny[i] * -1; // PHASE 1 release momentum in x direction
                aw[i][2] = aw[i][0] * pvinx[i]; // PHASE 1 release momentum in y direction
                aw[i][4] = aw[i][3] * pviny2[i] * -1; // PHASE 3 release momentum in x direction
                aw[i][5] = aw[i][3] * pvinx2[i]; // PHASE 3 release momentum in y direction
            }
            else { aw[i][0] = 0; aw[i][1] = 0; aw[i][2] = 0; aw[i][3] = 0; aw[i][4] = 0; aw[i][5] = 0; }
        }

        else if ( sico.MODEL == 7 ) {

            if ( ttrelease == 0 ) {

                aw[i][0] = fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 1,
                    px, py, pelev, betax, betay, betaxy, i, sico ); // PHASE 1 release depth
                aw[i][3] = fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 2,
                    px, py, pelev, betax, betay, betaxy, i, sico ); // PHASE 3 release depth
                aw[i][6] = fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 3,
                    px, py, pelev, betax, betay, betaxy, i, sico ); // PHASE 3 release depth

                aw[i][1] = aw[i][0] * pviny[i] * -1; // PHASE 1 release momentum in x direction
                aw[i][2] = aw[i][0] * pvinx[i]; // PHASE 1 release momentum in y direction
                aw[i][4] = aw[i][3] * pviny2[i] * -1; // PHASE 3 release momentum in x direction
                aw[i][5] = aw[i][3] * pvinx2[i]; // PHASE 3 release momentum in y direction
                aw[i][7] = aw[i][6] * pviny3[i] * -1; // PHASE 3 release momentum in x direction
                aw[i][8] = aw[i][6] * pvinx3[i]; // PHASE 3 release momentum in y direction
            }
            else { aw[i][0] = 0; aw[i][1] = 0; aw[i][2] = 0; aw[i][3] = 0; aw[i][4] = 0; aw[i][5] = 0;
                aw[i][6] = 0; aw[i][7] = 0; aw[i][8] = 0; }
        }
    }


// Writing input parameters to vectors and initializing volumes

    hflow_max = 0; // initializing maximum (PHASE 1) release depth
    if ( sico.MODEL > 3 ) hflow_max2 = 0; // initializing maximum PHASE 3 release depth
    if ( sico.MODEL == 7 ) hflow_max3 = 0; // initializing maximum PHASE 3 release depth
    hflow_maxmax = 0; // initializing absolute maximum flow depth
    vol_flow = 0; // initializing (PHASE 1) release volume
    vol_entr = 0; // initializing (PHASE 1) entrained or deposited volume

    if ( sico.MODEL > 3 ) {
        vol_flow2 = 0; // initializing PHASE 3 release volume
        vol_entr2 = 0; // initializing PHASE 3 entrained or deposited volume
    }

    if ( sico.MODEL == 7 ) {
        vol_flow3 = 0; // initializing PHASE 3 release volume
        vol_entr3 = 0; // initializing PHASE 3 entrained or deposited volume
    }

    for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels of reduced coordinate system:

        dentr[i] = 0; // depth of (PHASE 1) entrainment

        cstopped[i] = 0; // control for stopping

        if ( sico.CORRHEIGHT == 0 ) carea = pow ( sico.CSZ, 2 );
        else carea = pow ( sico.CSZ , 2 ) * pow( 1 - pow( sin( betax[i] ) , 2 ) * pow ( sin( betay[i] ) , 2 ) , 0.5 )
            / ( cos( betax[i] ) * cos( betay[i] ) ); // topography-following area of pixel

        if ( sico.MODEL <= 3 ) { // for one-phase model:

            for ( k=3; k<7; k++ ) aw[i][k] = 0;
            if ( sico.RELM == 1 ) aw[i][7] = aw[i][0];
            else aw[i][7] = 0;
            for ( k=8; k<11; k++ ) aw[i][k] = 0;
        }
        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:

            dentr2[i] = 0; // depth of PHASE 3 entrainment

            for ( k=6; k<10; k++ ) aw[i][k] = 0;
            aw[i][10] = aw[i][0] + aw[i][3];
            for ( k=11; k<18; k++ ) aw[i][k] = 0;
            if ( sico.RELM == 1 ) aw[i][18] = aw[i][0];
            else aw[i][18] = 0;
            aw[i][19] = 0;
            if ( sico.RELM2 == 1 ) aw[i][20] = aw[i][3];
            else aw[i][20] = 0;
            aw[i][21] = 0;
            aw[i][22] = aw[i][0] + aw[i][3];
            for ( k=23; k<29; k++ ) aw[i][k] = 0;

            if ( aw[i][3] > hflow_max2 ) hflow_max2 = aw[i][3]; // updating maximum PHASE 3 release depth
            if ( hflow_max + hflow_max2 > hflow_maxmax ) hflow_maxmax = hflow_max + hflow_max2; // updating absolute maximum flow depth
            vol_flow2 += aw[i][3] * carea; // updating PHASE 3 release volume
        }
        else if ( sico.MODEL == 7 ) { // for three-phase model:

            dentr2[i] = 0; // depth of PHASE 3 entrainment
            dentr3[i] = 0; // depth of PHASE 3 entrainment

            for ( k=9; k<14; k++ ) aw[i][k] = 0;
            aw[i][15] = aw[i][0] + aw[i][3] + aw[i][6];
            for ( k=16; k<24; k++ ) aw[i][k] = 0;
            if ( sico.RELM == 1 ) aw[i][25] = aw[i][0];
            else aw[i][25] = 0;
            aw[i][26] = 0;
            if ( sico.RELM2 == 1 ) aw[i][27] = aw[i][3];
            else aw[i][27] = 0;
            aw[i][28] = 0;
            if ( sico.RELM3 == 1 ) aw[i][29] = aw[i][6];
            else aw[i][29] = 0;
            aw[i][30] = 0;
            aw[i][31] = aw[i][0] + aw[i][3] + aw [i][6];
            for ( k=32; k<40; k++ ) aw[i][k] = 0;

            if ( aw[i][3] > hflow_max2 ) hflow_max2 = aw[i][3]; // updating maximum PHASE 3 release depth
            if ( hflow_max + hflow_max2 > hflow_maxmax ) hflow_maxmax = hflow_max + hflow_max2; // updating absolute maximum flow depth
            vol_flow2 += aw[i][3] * carea; // updating PHASE 3 release volume


            if ( aw[i][6] > hflow_max3 ) hflow_max3 = aw[i][6]; // updating maximum PHASE 3 release depth
            if ( hflow_max + hflow_max2 + hflow_max3 > hflow_maxmax ) hflow_maxmax = hflow_max + hflow_max2 + hflow_max3; // updating absolute maximum flow depth
            vol_flow3 += aw[i][6] * carea; // updating PHASE 3 release volume
        }

        if ( aw[i][0] > hflow_max ) hflow_max = aw[i][0]; // updating maximum (PHASE 1) release depth
        vol_flow += aw[i][0] * carea; // updating (PHASE 1) release volume
        if ( sico.MODEL <= 3 && hflow_max > hflow_maxmax ) hflow_maxmax = hflow_max; // updating absolute maximum flow depth

        if ( hydrograph == 1 ) { // coordinates and elevation of centres of hydrographs:

            for ( hydj = 0; hydj < ( hydnin + hydnout ); hydj++ ) {
                if ( px[i] == hydx[hydj] && py[i] == hydy[hydj] ) {

                    hydi[hydj] = i;
                    hydelev[hydj] = pelev[i];
                }
            }
        }
    }

    vol_hyd = 0; vol_hyd2 = 0;
    hekin_max = 0;
    mompre = 0;


// Defining hydrograph profiles

    if ( hydrograph == 1 ) { // if at least one input hydrograph is provided:

        hydp = alloc_imatrix( (int)(2*hydlmax/sico.CSZ+2), hydnin+hydnout ); // allocating memory to hydrograph profiles

        for ( hydj = 0; hydj < ( hydnin + hydnout ); hydj++ ) { // loop over all hydrographs:

            if ( hydalpha[hydj] < 0 ) hydalpha[hydj] = falpha ( betax[hydi[hydj]], betay[hydi[hydj]], sico ) + sico.PI / 2;
                // cross-slope direction
            if ( hydalpha[hydj] > 2 * sico.PI ) hydalpha[hydj] -= 2 * sico.PI;
            hydp0 = fhydp ( hydx[hydj], hydy[hydj], hydl[hydj], px, py, hydalpha[hydj], hydlmax, sico );
            for ( i=0; i<(int)(2*hydlmax/sico.CSZ+2); i++ ) hydp[i][hydj] = hydp0[i]; // hydrograph profile
            free( hydp0 );


        // Adjusting central point according to lowest elevation along hydrograph profile

            pelevhydtest = sico.UNDEF * -1;

            for ( hydk=1; hydk<= hydp[0][hydj]; hydk++ ) { // loop over all pixels of hydrograph profile:

               if ( pelev[hydp[hydk][hydj]] < pelevhydtest ) {

                    pelevhydtest = pelev[hydp[hydk][hydj]];
                    hydi[hydj] = hydp[hydk][hydj];
                    hydx[hydj] = px[hydp[hydk][hydj]];
                    hydy[hydj] = py[hydp[hydk][hydj]];
                }
            }

            hydelev[hydj] = pelevhydtest;

            hydx_metric[hydj] = ( float ) hydy[hydj] * sico.CSZ + sico.BDWEST;
            hydy_metric[hydj] = sico.BDNORTH - ( float ) hydx[hydj] * sico.CSZ;

            hydout_xmin_metric = ( float ) py[hydp[1][hydj]] * sico.CSZ + sico.BDWEST; // external coordinates of profile terminal points
            hydout_xmax_metric = ( float ) py[hydp[hydp[0][hydj]][hydj]] * sico.CSZ + sico.BDWEST;
            hydout_ymin_metric = sico.BDNORTH - ( float ) px[hydp[1][hydj]] * sico.CSZ;
            hydout_ymax_metric = sico.BDNORTH - ( float ) px[hydp[hydp[0][hydj]][hydj]] * sico.CSZ;

            if ( hydj < hydnin ) fprintf(f_hydout, "I%i\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\n", hydj+1,
                hydout_xmin_metric, hydx_metric[hydj], hydout_xmax_metric, hydout_ymin_metric, hydy_metric[hydj], hydout_ymax_metric);
            else fprintf(f_hydout, "O%i\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\n", hydj-hydnin+1,
                hydout_xmin_metric, hydx_metric[hydj], hydout_xmax_metric, hydout_ymin_metric, hydy_metric[hydj], hydout_ymax_metric);
                // writing hydrograph profile terminal points and centre to text file

        }
    }


// Initializing control variables

    vcelr = 0; // initial value of flow velocity plus wave speed
    tlength0 = sico.CFL[1]; // initial length of time step (s)
    tlength = tlength0;
    tlengthpre = tlength0;
    tsum = 0; // initial time (s)
    tint = 0; // initial time since last output (s)
    cflmax = 0; // initial maximum CFL value
    nsum = 1; // initial total number of time steps
    nout = 1; // initial number of output time steps
    ccontinue = 1; // control whether simulation should be continued (1 or 0)
    csuccess = 1; // control for success of simulation (1 or 0)
    //coverflow = 0; // control for overflow i.e. part of the flow leaving the study area (1 or 0)
    ctrl_hydout = 1; // control for hydrograph output
    if ( sico.DIFFCTRL != 0 ) { for ( i=0; i<sico.IMAX; i++ ) { for ( j=1; j<9; j++ ) { cready[i][j] = 0; }}}
        // degree of fill of pixel with regard to flux in a given direction (ratio)


// Summary of release and initial state of the flow

    if ( sico.MODEL <= 3 ) { // for one-phase model:

        if ( hflow_max0 >= 1000 ) prec_hflow = 0; // precision for summary of maximum flow heights and depths
        else if ( hflow_max0 >= 100 ) prec_hflow = 1;
        else if ( hflow_max0 >= 10 ) prec_hflow = 2;
        else if ( hflow_max0 >= 1 ) prec_hflow = 3;
        else if ( hflow_max0 >= 0.1 ) prec_hflow = 4;
        else if ( hflow_max0 == 0 ) prec_hflow = 2;
        else prec_hflow = 5;

        if ( vol_flow0 >= 100000000 ) prec_vol = 0; // precision for summary of flow volumes
        else if ( vol_flow0 >= 1000000 ) prec_vol = 1;
        else if ( vol_flow0 >= 100000 ) prec_vol = 2;
        else if ( vol_flow0 >= 10000 ) prec_vol = 3;
        else if ( vol_flow0 >= 1000 ) prec_vol = 4;
        else if ( vol_flow0 == 0 ) prec_vol = 3;
        else prec_vol = 5;

        prec_ekin = prec_vol-3; if ( prec_ekin < 0 ) prec_ekin = 0; // precision for summary of flow kinetic energies

        printf("   nout\tnsum\tcfl\ttlength\ttsum\tdmax\tvmax\tvolume\tekin\n");
        printf("   R\t-----\t-----\t-----\t-----\t*%.*f\t-----\t*%.*f\t-----\n",
            prec_hflow, hflow_max0, prec_vol, vol_flow0/1000 ); // displaying release parameters


        #ifdef WITHGRASS


            printf("   0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t-----\n",
                prec_hflow, hflow_max, prec_vol, vol_flow/1000 ); // displaying initial parameters


        #else


            printf("   0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t-----",
                prec_hflow, hflow_max, prec_vol, vol_flow/1000 ); // displaying initial parameters


        #endif


        fflush(stdout); // forcing immediate display

        fprintf(f_summary, "nout\tnsum\tcfl\ttlength\ttsum\tdmax\tvmax\tvolume\tekin\n");
        fprintf(f_summary, "R\t-----\t-----\t-----\t-----\t*%.*f\t-----\t*%.*f\t-----\n",
            prec_hflow, hflow_max0, prec_vol, vol_flow0 ); // writing release parameters to summary file
        fprintf(f_summary, "0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t\n", prec_hflow, hflow_max, prec_vol, vol_flow );
            // writing initial parameters to summary file
    }
    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:

        if ( hflow_max0 >= 1000 || hflow_max02 >= 1000 ) prec_hflow = 0;  // precision for summary of maximum flow heights and depths
        else if ( hflow_max0 >= 100 || hflow_max02 >= 100 ) prec_hflow = 1;
        else if ( hflow_max0 >= 10 || hflow_max02 >= 10 ) prec_hflow = 2;
        else if ( hflow_max0 >= 1 || hflow_max02 >= 1 ) prec_hflow = 3;
        else if ( hflow_max0 >= 0.1 || hflow_max02 >= 0.1 ) prec_hflow = 4;
        else if ( hflow_max0 == 0 && hflow_max02 == 0 ) prec_hflow = 2;
        else prec_hflow = 5;

        if ( vol_flow0 >= 100000000 || vol_flow02 >= 10000000 ) prec_vol = 0; // precision for summary of flow volumes
        else if ( vol_flow0 >= 1000000 || vol_flow02 >= 1000000 ) prec_vol = 1;
        else if ( vol_flow0 >= 100000 || vol_flow02 >= 100000 ) prec_vol = 2;
        else if ( vol_flow0 >= 10000 || vol_flow02 >= 10000 ) prec_vol = 3;
        else if ( vol_flow0 >= 1000 || vol_flow02 >= 1000 ) prec_vol = 4;
        else if ( vol_flow0 == 0 && vol_flow02 == 0 ) prec_vol = 3;
        else prec_vol = 5;

        prec_ekin = prec_vol-3; if ( prec_ekin < 0 ) prec_ekin = 0; // precision for summary of flow kinetic energies

        printf("   nout\tnsum\tcfl\ttlength\ttsum\tdmax1\tvmax1\tdmax2\tvmax2\tvolume1\tvolume2\tekin\n");
        printf("   R\t-----\t-----\t-----\t-----\t*%.*f\t-----\t*%.*f\t-----\t*%.*f\t*%.*f\t-----\n",
            prec_hflow, hflow_max0, prec_hflow, hflow_max02, prec_vol, vol_flow0/1000, prec_vol, vol_flow02/1000);
            // displaying release parameters


        #ifdef WITHGRASS


            printf("   0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t%.*f\t-----\n", prec_hflow, hflow_max, prec_hflow, hflow_max2,
                prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000); // displaying initial parameters


        #else


            printf("   0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t%.*f\t-----", prec_hflow, hflow_max, prec_hflow, hflow_max2,
                prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000); // displaying initial parameters


        #endif


        fflush(stdout); // forcing immediate display

        fprintf(f_summary, "nout\tnsum\tcfl\ttlength\ttsum\tdmax1\tvmax1\tdmax2\tvmax2\tvolume1\tvolume2\tekin\n");
        fprintf(f_summary, "R\t-----\t-----\t-----\t-----\t*%.*f\t-----\t*%.*f\t-----\t*%.*f\t*%.*f\t-----\n",
            prec_hflow, hflow_max0, prec_hflow, hflow_max02, prec_vol, vol_flow0, prec_vol, vol_flow02 );
            // writing release parameters to summary file
        fprintf(f_summary, "0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t%.*f\t-----\n",
            prec_hflow, hflow_max, prec_hflow, hflow_max2, prec_vol, vol_flow, prec_vol, vol_flow2 );
            // writing initial parameters to summary file
    }
    else if ( sico.MODEL == 7 ) { // for three-phase or multi-phase model:

        if ( hflow_max0 >= 1000 || hflow_max02 >= 1000 || hflow_max03 >= 1000 ) prec_hflow = 0;  // precision for summary of maximum flow heights and depths
        else if ( hflow_max0 >= 100 || hflow_max02 >= 100 || hflow_max03 >= 100 ) prec_hflow = 1;
        else if ( hflow_max0 >= 10 || hflow_max02 >= 10 || hflow_max03 >= 10 ) prec_hflow = 2;
        else if ( hflow_max0 >= 1 || hflow_max02 >= 1 || hflow_max03 >= 1 ) prec_hflow = 3;
        else if ( hflow_max0 >= 0.1 || hflow_max02 >= 0.1 || hflow_max03 >= 0.1 ) prec_hflow = 4;
        else if ( hflow_max0 == 0 && hflow_max02 == 0 && hflow_max03 == 0 ) prec_hflow = 2;
        else prec_hflow = 5;

        if ( vol_flow0 >= 100000000 || vol_flow02 >= 10000000 || vol_flow03 >= 10000000 ) prec_vol = 0; // precision for summary of flow volumes
        else if ( vol_flow0 >= 1000000 || vol_flow02 >= 1000000 || vol_flow03 >= 1000000 ) prec_vol = 1;
        else if ( vol_flow0 >= 100000 || vol_flow02 >= 100000 || vol_flow03 >= 100000 ) prec_vol = 2;
        else if ( vol_flow0 >= 10000 || vol_flow02 >= 10000 || vol_flow03 >= 10000 ) prec_vol = 3;
        else if ( vol_flow0 >= 1000 || vol_flow02 >= 1000  || vol_flow03 >= 1000 ) prec_vol = 4;
        else if ( vol_flow0 == 0 && vol_flow02 == 0 && vol_flow03 == 0 ) prec_vol = 3;
        else prec_vol = 5;

        prec_ekin = prec_vol-3; if ( prec_ekin < 0 ) prec_ekin = 0; // precision for summary of flow kinetic energies

        printf("   nout\tnsum\tcfl\ttlength\ttsum\tdmax1\tvmax1\tdmax2\tvmax2\tdmax3\tvmax3\tvolume1\tvolume2\tvolume3\tekin\n");
        printf("   R\t-----\t-----\t-----\t-----\t*%.*f\t-----\t*%.*f\t-----\t*%.*f\t-----\t*%.*f\t*%.*f\t*%.*f\t-----\n",
            prec_hflow, hflow_max0, prec_hflow, hflow_max02, prec_hflow, hflow_max03, prec_vol, vol_flow0/1000, prec_vol, vol_flow02/1000, prec_vol, vol_flow03/1000);
            // displaying release parameters


        #ifdef WITHGRASS


            printf("   0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t%.*f\t%.*f\t-----\n", prec_hflow, hflow_max, prec_hflow, hflow_max2,
                prec_hflow, hflow_max3, prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000, prec_vol, vol_flow3/1000); // displaying initial parameters


        #else


            printf("   0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t%.*f\t%.*f\t-----", prec_hflow, hflow_max, prec_hflow, hflow_max2,
                prec_hflow, hflow_max3, prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000, prec_vol, vol_flow3/1000); // displaying initial parameters


        #endif


        fflush(stdout); // forcing immediate display

        fprintf(f_summary, "nout\tnsum\tcfl\ttlength\ttsum\tdmax1\tvmax1\tdmax2\tvmax2\tdmax3\tvmax3\tvolume1\tvolume2\tvolume3\tekin\n");
        fprintf(f_summary, "R\t-----\t-----\t-----\t-----\t*%.*f\t-----\t*%.*f\t-----\t*%.*f\t-----\t*%.*f\t*%.*f\t*%.*f\t-----\n",
            prec_hflow, hflow_max0, prec_hflow, hflow_max02, prec_hflow, hflow_max03, prec_vol, vol_flow0, prec_vol, vol_flow02, prec_vol, vol_flow03 );
            // writing release parameters to summary file
        fprintf(f_summary, "0\t0\t-----\t-----\t0.0\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t0.00\t%.*f\t%.*f\t%.*f\t-----\n",
            prec_hflow, hflow_max, prec_hflow, hflow_max2, prec_hflow, hflow_max3, prec_vol, vol_flow, prec_vol, vol_flow2, prec_vol, vol_flow3 );
            // writing initial parameters to summary file
    }


                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cdomain[i] != 0 ) {

                        asigma_xelev[i] = fsigma( pelev[i], pelev[in[i][5]], pelev[in[i][2]], 1, dx, dy, i, sico );
                        asigma_yelev[i] = fsigma( pelev[i], pelev[in[i][4]], pelev[in[i][1]], 2, dx, dy, i, sico );
                    }
                }


                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cdomain[i] != 0 ) {

                        fj[0][0]=1; fj[0][1]=1; fj[1][0]=1; fj[1][1]=-1; fj[2][0]=-1; fj[2][1]=1; fj[3][0]=-1; fj[3][1]=-1; // factors for positive or negative gradients
                        for ( j=0; j<4; j++ ) wintelev[i][j]=pelev[in[i][j]]+fj[j][0]*0.25*dx[i]*asigma_xelev[in[i][j]]+fj[j][1]*0.25*dy[i]*asigma_yelev[in[i][j]];
                    }
                }







// *** Start of loop over time steps until break criterion is met: -------------------------------------------------

    while ( tsum <= tmax && ccontinue == 1 ) {
        if ( tint >= tout ) tint -= tout; // resetting time interval for output, if required


// *** Start of loop over two steps, each moving the system half of a pixel (NOC scheme) ---------------------------

        for ( p=0; p<2; p++ ) {

            if ( p == 0 ) tlength = tlength0; // time step length
            else if ( p == 1 ) tlength = tlength0;

            ib[1] = 0;


        // Updating slopes and topography-following pixel sizes

            for ( ix=0; ix<ib[0]; ix++ ) { // loop over all pixels:

                i = ibasket[0][ix];

                betax[i] = fbeta( pelev[in[i][5]], pelev[in[i][2]], 2.0, sico );
                betay[i] = fbeta( pelev[in[i][4]], pelev[in[i][1]], 2.0, sico );
                betaxy[i] = fbetaxy( betax[i], betay[i] );

                if ( sico.CORRHEIGHT != 0 ) {

                    dx[i] = sico.CSZ / cos( betax[i] ); // topography-following pixel spacing
                    dy[i] = sico.CSZ / cos( betay[i] );
                }
            }


        // Getting data from release hydrograph maps, if applicable

            if ( sico.TRELSTOP == 1 ) tfact = tlengthpre; else tfact = 1;

            for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels of reduced coordinate system:

                if ( sico.TRELEASE == 1 ) ttrelease = ptrelease[i]; // release time
                else ttrelease = 0;

                if ( sico.TRELSTOP == 1 ) ttrelstop = ptrelstop[i]; // release hydrograph stop time
                else ttrelstop = ttrelease;

                if ( sico.TRELEASE == 1 && tsum >= ttrelease && ( ptrelease[i] != 1000 * sico.UNDEF || tsum < ttrelstop )) {

                    if ( sico.MODEL <= 3 ) {

                        aw[i][0] += fconvin( phrelease, phrelease, phrelease, hflow_max0, 1, px, py, pelev, betax, betay, betaxy, i, sico ) * tfact; // release depth

                        aw[i][1] += fconvin( phrelease, phrelease, phrelease, hflow_max0, 1, px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pviny[i] * -1;
                            // release momentum in x direction
                        aw[i][2] += fconvin( phrelease, phrelease, phrelease, hflow_max0, 1, px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pvinx[i];
                            // release momentum in y direction
                    }

                    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                        aw[i][0] += fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 1,
                            px, py, pelev, betax, betay, betaxy, i, sico ) * tfact; // PHASE 1 release depth
                        aw[i][3] += fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 2,
                            px, py, pelev, betax, betay, betaxy, i, sico ) * tfact; // PHASE 2 release depth

                        aw[i][1] += fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 1,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pviny[i] * -1; // PHASE 1 release momentum in x direction
                        aw[i][2] += fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 1,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pvinx[i]; // PHASE 1 release momentum in y direction
                        aw[i][4] += fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 2,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pviny2[i] * -1; // PHASE 2 release momentum in x direction
                        aw[i][5] += fconvin( phrelease, phrelease2, phrelease2, hflow_max0+hflow_max02, 2,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pvinx2[i]; // PHASE 2 release momentum in y direction
                    }

                    else if ( sico.MODEL == 7 ) {

                        aw[i][0] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 1,
                            px, py, pelev, betax, betay, betaxy, i, sico ) * tfact; // PHASE 1 release depth
                        aw[i][3] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 2,
                            px, py, pelev, betax, betay, betaxy, i, sico ) * tfact; // PHASE 2 release depth
                        aw[i][6] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 3,
                            px, py, pelev, betax, betay, betaxy, i, sico ) * tfact; // PHASE 3 release depth

                        aw[i][1] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 1,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pviny[i] * -1; // PHASE 1 release momentum in x direction
                        aw[i][2] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 1,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pvinx[i]; // PHASE 1 release momentum in y direction

                        aw[i][4] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 2,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pviny2[i] * -1; // PHASE 2 release momentum in x direction
                        aw[i][5] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 2,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pvinx2[i]; // PHASE 2 release momentum in y direction

                        aw[i][7] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 3,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pviny3[i] * -1; // PHASE 3 release momentum in x direction
                        aw[i][8] += fconvin( phrelease, phrelease2, phrelease3, hflow_max0+hflow_max02+hflow_max03, 3,
                                px, py, pelev, betax, betay, betaxy, i, sico ) * tfact * pvinx3[i]; // PHASE 3 release momentum in y direction
                    }

                    ptrelease[i] = 1000 * sico.UNDEF;
                }
            }


// ----- START -- Updating flow depths and velocities according to input hydrographs ------------------------------- 


            vol_hydbef = 0; vol_hydaft = 0;
            if ( sico.MODEL > 3 ) { vol_hydbef2 = 0; vol_hydaft2 = 0; }
            if ( sico.MODEL == 7 ) { vol_hydbef3 = 0; vol_hydaft3 = 0; }

            if ( hydrograph == 1 ) {

                if ( tsum == 0 || sico.HYDADD == 0 ) tlengthx = 1;
                else tlengthx = tlengthpre;

                for ( hydj = 0; hydj < hydnin; hydj++ ) { // loop over all input hydrographs:

                    for ( hydk=0; hydk<=hydtmax[hydj]; hydk++ ) { // identifying relevant line of input hydrograph:
                        if ( hydhyd[hydk][0][hydj] <= tsum ) hydt=hydk;
                        else break;
                    }

                    if ( sico.MODEL <= 3 ) hhyd0 = hydhyd[hydt][1][hydj]; // total flow height at centre of hydrograph
                    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) hhyd0 = hydhyd[hydt][1][hydj] + hydhyd[hydt][3][hydj];
                    else if ( sico.MODEL == 7 ) hhyd0 = hydhyd[hydt][1][hydj] + hydhyd[hydt][3][hydj] + hydhyd[hydt][5][hydj];

                    if ( hhyd0 > sico.HFLOWMIN ) {

                        for ( hydk=1; hydk<= hydp[0][hydj]; hydk++ ) { // loop over all pixels of hydrograph profile:

                            if ( sico.CORRHEIGHT == 0 ) carea = pow ( sico.CSZ, 2 );
                            else carea = pow ( sico.CSZ, 2 ) * pow( 1 - pow( sin( betax[hydp[hydk][hydj]] ) , 2 )
                                * pow ( sin( betay[hydp[hydk][hydj]] ) , 2 ) , 0.5 )
                                / ( cos( betax[hydp[hydk][hydj]] ) * cos( betay[hydp[hydk][hydj]] ) ); // topography-following area of pixel

                            hhyd = hhyd0 - ( pelev[hydp[hydk][hydj]] - hydelev[hydj] ); // total flow height at pixel

                            for ( k=0; k<sico.NVECTMIN; k++ ) {
                                if ( sico.HYDADD == 1 ) hydbef[k] = aw[hydp[hydk][hydj]][k]; else hydbef[k] = 0;
                            }

                            vol_hydbef += aw[hydp[hydk][hydj]][0] * carea;
                            if ( sico.MODEL > 3 ) vol_hydbef2 += aw[hydp[hydk][hydj]][3] * carea;
                            if ( sico.MODEL == 7 ) vol_hydbef3 += aw[hydp[hydk][hydj]][6] * carea;

                            if ( hhyd > sico.HFLOWMIN ) {

                                aw[hydp[hydk][hydj]][0] = hydbef[0] + tlengthx * hydhyd[hydt][1][hydj] * hhyd / hhyd0 * cos( betaxy[hydp[hydk][hydj]] );
                                    // (PHASE 1) flow depth
                                aw[hydp[hydk][hydj]][1] = hydbef[1] + tlengthx * hydhyd[hydt][1][hydj] * hhyd / hhyd0 * hydhyd[hydt][2][hydj]
                                    * cos( hydalpha[hydj] - sico.PI / 2 )  * cos( betaxy[hydp[hydk][hydj]] ); // (PHASE 1) x momentum
                                aw[hydp[hydk][hydj]][2] = hydbef[2] + tlengthx * hydhyd[hydt][1][hydj] * hhyd / hhyd0 * hydhyd[hydt][2][hydj]
                                    * sin( hydalpha[hydj] - sico.PI / 2 )  * cos( betaxy[hydp[hydk][hydj]] ); // (PHASE 1) y momentum
                            }

                            if ( hhyd > sico.HFLOWMIN && sico.MODEL > 3 ) {

                                aw[hydp[hydk][hydj]][3] = hydbef[3] + tlengthx * hydhyd[hydt][3][hydj] * hhyd / hhyd0  * cos( betaxy[hydp[hydk][hydj]] );
                                    // PHASE 3 flow depth
                                aw[hydp[hydk][hydj]][4] = hydbef[4] + tlengthx * hydhyd[hydt][3][hydj] * hhyd / hhyd0 * hydhyd[hydt][4][hydj]
                                    * cos( hydalpha[hydj] - sico.PI / 2 ) * cos( betaxy[hydp[hydk][hydj]] ); // PHASE 3 x momentum
                                aw[hydp[hydk][hydj]][5] = hydbef[5] + tlengthx * hydhyd[hydt][3][hydj] * hhyd / hhyd0 * hydhyd[hydt][4][hydj]
                                    * sin( hydalpha[hydj] - sico.PI / 2 ) * cos( betaxy[hydp[hydk][hydj]] ); // PHASE 3 y momentum
                            }

                            if ( hhyd > sico.HFLOWMIN && sico.MODEL == 7 ) {

                                aw[hydp[hydk][hydj]][6] = hydbef[6] + tlengthx * hydhyd[hydt][5][hydj] * hhyd / hhyd0  * cos( betaxy[hydp[hydk][hydj]] );
                                    // PHASE 3 flow depth
                                aw[hydp[hydk][hydj]][7] = hydbef[7] + tlengthx * hydhyd[hydt][5][hydj] * hhyd / hhyd0 * hydhyd[hydt][6][hydj]
                                    * cos( hydalpha[hydj] - sico.PI / 2 ) * cos( betaxy[hydp[hydk][hydj]] ); // PHASE 3 x momentum
                                aw[hydp[hydk][hydj]][8] = hydbef[8] + tlengthx * hydhyd[hydt][5][hydj] * hhyd / hhyd0 * hydhyd[hydt][6][hydj]
                                    * sin( hydalpha[hydj] - sico.PI / 2 ) * cos( betaxy[hydp[hydk][hydj]] ); // PHASE 3 y momentum
                            }

                            vol_hydaft += aw[hydp[hydk][hydj]][0] * carea;
                            if ( sico.MODEL > 3 ) vol_hydaft2 += aw[hydp[hydk][hydj]][3] * carea;
                            if ( sico.MODEL == 7 ) vol_hydaft3 += aw[hydp[hydk][hydj]][6] * carea;

                            if ( fvalid( aw[hydp[hydk][hydj]][0], aw[hydp[hydk][hydj]][3], aw[hydp[hydk][hydj]][6], sico ) == 1
                                && cstopped[hydp[hydk][hydj]] == 0 && cdomain[hydp[hydk][hydj]] > 0 ) { // flow pixels:

                                for ( j=0; j<9; j++ ) { // 1st row flow boundary pixels:

                                    if ( icheck[in[hydp[hydk][hydj]][j]][1] == 0 && cstopped[in[hydp[hydk][hydj]][j]] == 0
                                        && cdomain[in[hydp[hydk][hydj]][j]] > 0 ) {

                                        ibasket[1][ib[1]] = in[hydp[hydk][hydj]][j];
                                        ib[1] += 1;
                                        icheck[in[hydp[hydk][hydj]][j]][1] = 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }


// ----- STOP --- Updating flow depths and velocities according to input hydrographs -------------------------------


// ----- START -- Writing hydrograph infos to files ----------------------------------------------------------------


            if ( p == 1 && hydrograph == 1 ) {

                for ( hydj = 0; hydj < hydnin; hydj++ ) { // loop over all input hydrographs:

                    if ( sico.MULT == 0 && ctrl_hydout == 1 ) { // for first time step after output:

                        hydh = aw[hydi[hydj]][0] / cos( betaxy[hydi[hydj]] ); // (PHASE 1) flow height
                        if ( hydh > sico.HFLOWMIN )
                            hydv = pow( pow( aw[hydi[hydj]][1], 2 ) + pow( aw[hydi[hydj]][2], 2 ), 0.5 ) / aw[hydi[hydj]][0];
                        else hydv= 0; // (PHASE 1) flow velocity

                        if ( sico.MODEL <= 3 ) {

                            hyde = aw[hydi[hydj]][3]; // entrained height
                            hydh2 = 0; hydv2 = 0; hyde2 = 0; hydh3 = 0; hydv3 = 0; hyde3 = 0;
                         }
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                            hydh2 = aw[hydi[hydj]][3] / cos( betaxy[hydi[hydj]] ); // PHASE 2 flow height
                            if ( hydh2 > sico.HFLOWMIN )
                                hydv2 = pow( pow( aw[hydi[hydj]][4], 2 ) + pow( aw[hydi[hydj]][5], 2 ), 0.5 ) / aw[hydi[hydj]][3];
                            else hydv2= 0; // PHASE 2 flow velocity
                            hyde = aw[hydi[hydj]][6]; // entrained PHASE 1 height
                            hyde2 = aw[hydi[hydj]][7]; // entrained PHASE 2 height
                            hydh3 = 0; hydv3 = 0; hyde3 = 0;
                        }
                        else if ( sico.MODEL == 7 ) {

                            hydh2 = aw[hydi[hydj]][3] / cos( betaxy[hydi[hydj]] ); // PHASE 2 flow height
                            if ( hydh2 > sico.HFLOWMIN )
                                hydv2 = pow( pow( aw[hydi[hydj]][4], 2 ) + pow( aw[hydi[hydj]][5], 2 ), 0.5 ) / aw[hydi[hydj]][3];
                            else hydv2= 0; // PHASE 2 flow velocity
                            hydh3 = aw[hydi[hydj]][6] / cos( betaxy[hydi[hydj]] ); // PHASE 3 flow height
                            if ( hydh3 > sico.HFLOWMIN )
                                hydv3 = pow( pow( aw[hydi[hydj]][7], 2 ) + pow( aw[hydi[hydj]][8], 2 ), 0.5 ) / aw[hydi[hydj]][6];
                            else hydv3= 0; // PHASE 3 flow velocity
                            hyde = aw[hydi[hydj]][9]; // entrained PHASE 1 height
                            hyde2 = aw[hydi[hydj]][10]; // entrained PHASE 2 height
                            hyde3 = aw[hydi[hydj]][11]; // entrained PHASE 3 height
                        }

                        hydq = 0; // resetting (PHASE 1) discharge
                        if ( sico.MODEL > 3 ) hydq2 = 0; // resetting PHASE 2 discharge
                        if ( sico.MODEL == 7 ) hydq3 = 0; // resetting PHASE 3 discharge

                        if ( hydalpha[hydj] == 0 || hydalpha[hydj] == sico.PI / 2 || hydalpha[hydj] == sico.PI
                            || hydalpha[hydj] == 3 * sico.PI / 2 ) hydfalpha = 1;
                        else if ( fabs( 1 / sin( hydalpha[hydj] )) < fabs( 1 / cos( hydalpha[hydj] )))
                            hydfalpha = fabs( 1 / sin( hydalpha[hydj] ));
                        else hydfalpha = fabs( 1 / cos( hydalpha[hydj] )); // correction factor for profile direction

                        for ( hydk = 1; hydk <= hydp[0][hydj]; hydk++ ) { // loop over all pixels of profile:

                            alpha = falpha( betax[hydp[hydk][hydj]], betay[hydp[hydk][hydj]], sico ); // aspect
                            hydbeta = atan ( tan( betaxy[hydp[hydk][hydj]] ) * cos ( alpha - hydalpha[hydj] + sico.PI / 2 ));
                                // corrected slope

                            hydfcorr = sico.CSZ * hydfalpha / cos( hydbeta ); // reference length for discharge
                            hydm0 = pow( pow(aw[hydp[hydk][hydj]][1], 2 ) + pow( aw[hydp[hydk][hydj]][2], 2), 0.5 );

                            if ( aw[hydp[hydk][hydj]][0] > sico.HFLOWMIN && hydm0 > 0 ) {

                                hydmx = aw[hydp[hydk][hydj]][1];
                                hydmy = aw[hydp[hydk][hydj]][2];

                                if ( hydmy > 0 ) hydmalpha = acos( hydmx / hydm0 );
                                else hydmalpha = 2 * sico.PI - acos( hydmx / hydm0 );
                                hydnalpha = hydalpha[hydj] - sico.PI / 2; if ( hydnalpha < 0 ) hydnalpha += 2 * sico.PI;
                                hydm = hydm0 * cos ( hydnalpha - hydmalpha ); // momentum in direction perpendicular to profile
                                hydq += hydm * hydfcorr; // updating (PHASE 1) discharge
                            }

                            if ( sico.MODEL > 3 ) {

                                hydm0 = pow( pow(aw[hydp[hydk][hydj]][4], 2 ) + pow( aw[hydp[hydk][hydj]][5], 2), 0.5 );

                                if ( aw[hydp[hydk][hydj]][3] > sico.HFLOWMIN && hydm0 > 0 ) {

                                    hydmx = aw[hydp[hydk][hydj]][4];
                                    hydmy = aw[hydp[hydk][hydj]][5];

                                    if ( hydmy > 0 ) hydmalpha = acos( hydmx / hydm0 );
                                    else hydmalpha = 2 * sico.PI - acos( hydmx / hydm0 );
                                    hydnalpha = hydalpha[hydj] - sico.PI / 2; if ( hydnalpha < 0 ) hydnalpha += 2 * sico.PI;
                                    hydm = hydm0 * cos ( hydnalpha - hydmalpha ); // momentum in direction perpendicular to profile
                                    hydq2 += hydm * hydfcorr; // updating PHASE 2 discharge
                                }
                            }
                            else hydq2 = 0;

                            if ( sico.MODEL == 7 ) {

                                hydm0 = pow( pow(aw[hydp[hydk][hydj]][7], 2 ) + pow( aw[hydp[hydk][hydj]][8], 2), 0.5 );

                                if ( aw[hydp[hydk][hydj]][6] > sico.HFLOWMIN && hydm0 > 0 ) {

                                    hydmx = aw[hydp[hydk][hydj]][7];
                                    hydmy = aw[hydp[hydk][hydj]][8];

                                    if ( hydmy > 0 ) hydmalpha = acos( hydmx / hydm0 );
                                    else hydmalpha = 2 * sico.PI - acos( hydmx / hydm0 );
                                    hydnalpha = hydalpha[hydj] - sico.PI / 2; if ( hydnalpha < 0 ) hydnalpha += 2 * sico.PI;
                                    hydm = hydm0 * cos ( hydnalpha - hydmalpha ); // momentum in direction perpendicular to profile
                                    hydq3 += hydm * hydfcorr; // updating PHASE 3 discharge
                                }
                            }
                            else hydq3 = 0;
                        }

                        fprintf(f_hydinfo[hydj], "%.1f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\n",
                            tsum + tlength0 + tlength0, hydh, hydv, hyde, hydq, hydh2, hydv2, hyde2, hydq2, hydh3, hydv3, hyde3, hydq3);
                            // writing hydrograph info to file !!!CHECK for tlength if time step has to be repeated
                    }
                }
                ctrl_hydout = 0; // resetting control for hydrograph output
            }

            if ( p == 0 ) vol_hyd += ( vol_hydaft - vol_hydbef ); // computing volume added through hydrograph
            if ( p == 0 && sico.MODEL > 3 ) vol_hyd2 += ( vol_hydaft2 - vol_hydbef2 );
            if ( p == 0 && sico.MODEL == 7 ) vol_hyd3 += ( vol_hydaft3 - vol_hydbef3 );


// ----- STOP --- Writing hydrograph infos to files ----------------------------------------------------------------


// F4 -- START -- Surface control ----------------------------------------------------------------------------------


            // *** cflux[i] = 0: fluxes are not computed in order to avoid oscillations of horizontal flow surfaces
            // *** cflux[i] = 1: fluxes are computed

            for ( ix=0; ix<ib[0]; ix++ ) {

                i = ibasket[0][ix];

                if ( sico.NOOSC == 3 ) { // if function is enabled:

                    cflux[i] = 1; // setting initial value of flux control to negative
                    if ( cdomain[i] > 0 ) { // if pixel is no 1st row edge pixel, testing whether fluxes should be computed or not:

                        if ( sico.CORRHEIGHT == 0 ) fflux = 1; // conversion factors between flow depths and heights
                        else fflux = 1 / cos( betaxy0[i] );

                        if ( sico.MODEL <= 3 ) hflow = fflux * aw[i][0];
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) hflow = fflux * ( aw[i][0] + aw[i][3] );
                        else if ( sico.MODEL == 7 ) hflow = fflux * ( aw[i][0] + aw[i][3] + aw[i][6] );

                        ctrlr = 0;
                        for ( j=1; j<9; j++ ) { // loop over neighbourhood:

                            if ( sico.CORRHEIGHT == 0 ) ffluxn = 1; // conversion factors between flow depths and heights
                            else ffluxn = 1 / cos( betaxy0[in[i][j]] );

                            if ( sico.MODEL <= 3 ) hflown = ffluxn * aw[in[i][j]][0];
                            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) hflown = ffluxn * ( aw[in[i][j]][0] + aw[in[i][j]][3] );
                            else if ( sico.MODEL == 7 ) hflown = ffluxn * ( aw[in[i][j]][0] + aw[in[i][j]][3] + aw[in[i][j]][6] );

                            if (( fabs(( pelev[in[i][j]] + hflown ) - ( pelev[i] + hflow )) <= sico.HFLOWMIN
                                || ( hflown <= sico.HFLOWMIN && pelev[in[i][j]] - pelev[i] - hflow > sico.HFLOWMIN ))) ctrlr += 1;
                                // setting flux control to positive if inflow or outflow is possible
                        }
                        if ( ctrlr == 8 ) cflux[i] = 0;
                    }
                }

                else cflux[i] = 1; // if function is disabled, setting flux control to positive
            }


// F4 -- STOP --- Surface control ----------------------------------------------------------------------------------


// ----- START -- Updating computational domains -------------------------------------------------------------------


            // *** cdomain[i] = 0: 1st row edge pixel
            // *** cdomain[i] = 1: no-flux pixel (suppression of oscillations)
            // *** cdomain[i] = 3: included pixels, flow depth < minimum at & around pixel
            // *** cdomain[i] = 4: included pixels, flow depth >= minimum around pixel (flow boundary)
            // *** cdomain[i] = 5: included pixels, flow depth >= minimum (flow)

            for ( ix=0; ix<ib[0]; ix++ ) {

                i = ibasket[0][ix];

                if (( fvalid( aw[i][0], aw[i][3], aw[i][6], sico ) == 1 || ( sico.TRELEASE == 1 && ptrelease[i] > 0 && ptrelease[i] < 999999 ))
                    && cstopped[i] == 0 && cdomain[i] > 0 ) { // flow pixels: !!!CHECK

                    for ( j=0; j<9; j++ ) { // 1st row flow boundary pixels:

                        if ( icheck[in[i][j]][1] == 0 && cstopped[in[i][j]] == 0 && cdomain[in[i][j]] > 0 ) {

                            ibasket[1][ib[1]] = in[i][j];
                            ib[1] += 1;
                            icheck[in[i][j]][1] = 1;
                        }
                    }
                }
            }

            ib[0] = 0;

            for ( ix=0; ix<ib[1]; ix++ ) { // 2nd row flow boundary pixels:

                i = ibasket[1][ix];

                for ( j=0; j<9; j++ ) {

                    if ( icheck[in[i][j]][0] == 0 && cstopped[in[i][j]] == 0 && cdomain[in[i][j]] > 0 ) {

                        ibasket[0][ib[0]] = in[i][j];
                        ib[0] += 1;
                        icheck[in[i][j]][0] = 1;
                    }
                }
            }

            for ( ix=0; ix<ib[0]; ix++ ) { // updating domains and resetting checks:

                i = ibasket[0][ix];

                if ( sico.MODEL <= 3 ) hflow = aw[i][0];
                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) hflow = aw[i][0] + aw[i][3];
                else if ( sico.MODEL == 7 ) hflow = aw[i][0] + aw[i][3] + aw[i][6];

                if ( cdomain[i] != 0 ) { // for all non-edge pixels:

                    ctrlv = 0; ctrlr = 1; // resetting controls
                    for ( j=1; j<9; j++ ) { // loop over neighbourhood:

                        if ( sico.MODEL <= 3 ) hflown = aw[in[i][j]][0];
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) hflown = aw[in[i][j]][0] + aw[in[i][j]][3];
                        else if ( sico.MODEL == 7 ) hflown = aw[in[i][j]][0] + aw[in[i][j]][3] + aw[in[i][j]][6];

                        if ( cflux[i] == 1 && ( cflux[in[i][j]] == 1 && ( hflown > sico.HFLOWMIN || hflow > sico.HFLOWMIN ))) ctrlv = 1;
                            // setting control for pixel to positive if not at or surrounded by no-flux area

                        if ( cstopped[i] == 1 || cstopped[in[i][j]] == 1 ) ctrlr = 0;
                            // setting control for pixel to negative if at or adjacent to stopped area
                    }
                    if ( ctrlv == 1 && ctrlr == 1 ) cdomain[i] = 3; // updating computational domain
                }

                if ( cdomain[i] == 3 ) { // if pixel is part of computational domain:

                    ctrlr = 0; // resetting control

                    for ( j=1; j<9; j++ ) { // loop over neighbourhood:

                        if ( sico.MODEL <= 3 ) hflown = aw[in[i][j]][0];
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) hflown = aw[in[i][j]][0] + aw[in[i][j]][3];
                        else if ( sico.MODEL == 7 ) hflown = aw[in[i][j]][0] + aw[in[i][j]][3] + aw[in[i][j]][6];
                        if ( hflown > sico.HFLOWMIN ) ctrlr = 1; // if neighbour pixel is flow pixel, setting control to positive
                    }
                    if ( ctrlr == 1 ) cdomain[i] = 4; // flow depth >= minimum around pixel
                    if ( hflow > sico.HFLOWMIN ) cdomain[i] = 5; // flow depth >= minimum
                }

                icheck[i][0] = 0;
                icheck[i][1] = 0;
            }


// ----- STOP --- Updating computational domains -------------------------------------------------------------------


// *** Start of loop over various time step lengths until CFL criterion is met -------------------------------------


            ccfl = 0; // resetting control for fulfilment of CFL criterion
            while ( ccfl == 0 ) {


// ----- START -- Flow propagation with NOC scheme -----------------------------------------------------------------


                // *** High-resolution scheme, system is moved half a pixel size each time step
                // *** Technically, the system is shifted one pixel each second time step
                // *** At the end, the new state variables are written to a temporary vector (awt[i][k])
                // *** The values from the temporary vector are written to the permanent vector (aw[i][k])
                // *** as soon as it has been verified that the time step length is sufficiently short
                // *** to fulfil the CFL criterion
                // *** Otherwise, the procedure is repeated with shorter time step length


                // Fluxes, source terms, and gradients (original coordinate system)

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cdomain[i] <= 3 ) {

                        for ( k=0; k<sico.NVECTMIN; k++ ) { af[i][k]=0; ag[i][k]=0; as[i][k]=0; asigma_x[i][k]=0; asigma_y[i][k]=0; }

                    } else {

                        if ( sico.PHI == 1 ) {
                            if ( pphi[i] != sico.UNDEF ) sflow.PHI[0] = pphi[i]; // internal friction angle of PHASE 1
                            else sflow.PHI[0] = sflow.PHI0[0];
                        } else sflow.PHI[0] = sflow.PHI0[0];

                        if ( sico.PHI2 == 1 ) {
                            if ( pphi2[i] != sico.UNDEF ) sflow.PHI[1] = pphi2[i]; // internal friction angle of PHASE 2
                            else sflow.PHI[1] = sflow.PHI0[1];
                        } else sflow.PHI[1] = sflow.PHI0[1];

                        if ( sico.PHI3 == 1 ) {
                            if ( pphi3[i] != sico.UNDEF ) sflow.PHI[2] = pphi3[i]; // internal friction angle of PHASE 3
                            else sflow.PHI[2] = sflow.PHI0[2];
                        } else sflow.PHI[2] = sflow.PHI0[2];

                        if ( sico.DELTA == 1 ) {
                            if ( pdelta[i] != sico.UNDEF ) sflow.DELTA[0] = pdelta[i]; // basal friction angle of PHASE 1
                            else sflow.DELTA[0] = sflow.DELTA0[0];
                        } else sflow.DELTA[0] = sflow.DELTA0[0];

                        if ( sico.DELTA2 == 1 ) {
                            if ( pdelta2[i] != sico.UNDEF ) sflow.DELTA[1] = pdelta2[i]; // basal friction angle of PHASE 2
                            else sflow.DELTA[1] = sflow.DELTA0[1];
                        } else sflow.DELTA[1] = sflow.DELTA0[1];

                        if ( sico.DELTA3 == 1 ) {
                            if ( pdelta3[i] != sico.UNDEF ) sflow.DELTA[2] = pdelta3[i]; // basal friction angle of PHASE 3
                            else sflow.DELTA[2] = sflow.DELTA0[2];
                        } else sflow.DELTA[2] = sflow.DELTA0[2];

                        if ( sico.NYSS == 1 ) {
                            if ( pnyss[i] != sico.UNDEF ) sflow.NY[0] = pnyss[i]; // viscosity of PHASE 1
                            else sflow.NY[0] = sflow.NY0[0];
                        } else sflow.NY[0] = sflow.NY0[0];

                        if ( sico.NYFS == 1 ) {
                            if ( pnyfs[i] != sico.UNDEF ) sflow.NY[1] = pnyfs[i]; // viscosity of PHASE 2
                            else sflow.NY[1] = sflow.NY0[1];
                        } else sflow.NY[1] = sflow.NY0[1];

                        if ( sico.NYFF == 1 ) {
                            if ( pnyff[i] != sico.UNDEF ) sflow.NY[2] = pnyff[i]; // viscosity of PHASE 3
                            else sflow.NY[2] = sflow.NY0[2];
                        } else sflow.NY[2] = sflow.NY0[2];

                        if ( sico.AMBDRAG == 1 ) {
                            if ( pambdrag[i] != sico.UNDEF ) sflow.AMBDRAG = pambdrag[i]; // ambient drag coefficient
                            else sflow.AMBDRAG = sflow.AMBDRAG0;
                        } else sflow.AMBDRAG = sflow.AMBDRAG0;

                        if ( sico.FLUFRI == 1 ) {
                            if ( pflufri[i] != sico.UNDEF ) sflow.FLUFRI = pflufri[i]; // fluid friction coefficient
                            else sflow.FLUFRI = sflow.FLUFRI0;
                        } else sflow.FLUFRI = sflow.FLUFRI0;


                        // Components of gravity

                        grav[0] = sico.GRAVITY * cos( betax[i] );
                        grav[1] = sico.GRAVITY * cos( betay[i] );
                        grav[2] = sico.GRAVITY * sin( betax[i] );
                        grav[3] = sico.GRAVITY * sin( betay[i] );
                        grav[4] = sico.GRAVITY * cos( betaxy[i] );


                        // Flow velocities in x and y directions

                        vflowx = fdiv( aw[i][1], aw[i][0], sico.HFLOWMIN );
                        vflowy = fdiv( aw[i][2], aw[i][0], sico.HFLOWMIN );

                        if ( sico.MODEL > 3 ) vflowx2 = fdiv( aw[i][4], aw[i][3], sico.HFLOWMIN ); else vflowx2 = 0;
                        if ( sico.MODEL > 3 ) vflowy2 = fdiv( aw[i][5], aw[i][3], sico.HFLOWMIN ); else vflowy2 = 0;

                        if ( sico.MODEL == 7 ) vflowx3 = fdiv( aw[i][7], aw[i][6], sico.HFLOWMIN ); else vflowx3 = 0;
                        if ( sico.MODEL == 7 ) vflowy3 = fdiv( aw[i][8], aw[i][6], sico.HFLOWMIN ); else vflowy3 = 0;


                        // Total flow depths, velocities, and fractions of neighbouring cells

                        welev[0] = pelev[i];

                        for ( l=0; l<9; l++ ) {

                            welev[l] = pelev[in[i][l]];

                            if ( sico.MODEL <= 3 ) wh[l] = aw[in[i][l]][0];
                            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) wh[l] = aw[in[i][l]][0] + aw[in[i][l]][3];
                            else if ( sico.MODEL == 7 ) wh[l] = aw[in[i][l]][0] + aw[in[i][l]][3] + aw[in[i][l]][6];

                            wu[l] = fdiv( aw[in[i][l]][1], aw[in[i][l]][0], sico.HFLOWMIN );
                            wv[l] = fdiv( aw[in[i][l]][2], aw[in[i][l]][0], sico.HFLOWMIN );

                            if ( sico.MODEL > 3 ) wu2[l] = fdiv( aw[in[i][l]][4], aw[in[i][l]][3], sico.HFLOWMIN ); else wu2[l] = 0;
                            if ( sico.MODEL > 3 ) wv2[l] = fdiv( aw[in[i][l]][5], aw[in[i][l]][3], sico.HFLOWMIN ); else wv2[l] = 0;

                            if ( sico.MODEL == 7 ) wu3[l] = fdiv( aw[in[i][l]][7], aw[in[i][l]][6], sico.HFLOWMIN ); else wu3[l] = 0;
                            if ( sico.MODEL == 7 ) wv3[l] = fdiv( aw[in[i][l]][8], aw[in[i][l]][6], sico.HFLOWMIN ); else wv3[l] = 0;

                            walpha[l] = fdiv( aw[in[i][l]][0], wh[l], sico.HFLOWMIN );
                            if ( sico.MODEL > 3 ) walpha2[l] = fdiv( aw[in[i][l]][3], wh[l], sico.HFLOWMIN ); else walpha2[l] = 0;
                            if ( sico.MODEL == 7 ) walpha3[l] = fdiv( aw[in[i][l]][6], wh[l], sico.HFLOWMIN ); else walpha3[l] = 0;
                        }


                        // Gradients of total flow depth, flow velocities, and fractions

                        whx = ( wh[2] - wh[5] ) / ( 2 * dx[i] );
                        if ( sico.NOOSC > 0 && ( wh[2] == 0 || wh[5] == 0 )) whx = tan( betax[i] ); 

                        why = ( wh[1] - wh[4] ) / ( 2 * dy[i] );
                        if ( sico.NOOSC > 0 && ( wh[1] == 0 || wh[4] == 0 )) why = tan( betay[i] ); 

                         wdu[0] = ( wu[2] - wu[5] ) / ( 2 * dx[i] );
                        wdv[0] = ( wv[1] - wv[4] ) / ( 2 * dy[i] );

                        if ( sico.MODEL > 3 ) wdu[1] = ( wu2[2] - wu2[5] ) / ( 2 * dx[i] ); else wdu[1] = 0;
                        if ( sico.MODEL > 3 ) wdv[1] = ( wv2[1] - wv2[4] ) / ( 2 * dy[i] ); else wdv[1] = 0;

                        if ( sico.MODEL == 7 ) wdu[2] = ( wu3[2] - wu3[5] ) / ( 2 * dx[i] ); else wdu[2] = 0;
                        if ( sico.MODEL == 7 ) wdv[2] = ( wv3[1] - wv3[4] ) / ( 2 * dy[i] ); else wdv[2] = 0;

                        if ( walpha[2] ) xwdv[0] = ( wv[2] - wv[5] ) / ( 2 * dx[i] );
                        if ( walpha[1] ) xwdu[0] = ( wu[1] - wu[4] ) / ( 2 * dy[i] );

                        if ( sico.MODEL > 3 ) xwdv[1] = ( wv2[2] - wv2[5] ) / ( 2 * dx[i] ); else xwdv[1] = 0;
                        if ( sico.MODEL > 3 ) xwdu[1] = ( wu2[1] - wu2[4] ) / ( 2 * dy[i] ); else xwdu[1] = 0;

                        if ( sico.MODEL == 7 ) xwdv[2] = ( wv3[2] - wv3[5] ) / ( 2 * dx[i] ); else xwdv[2] = 0;
                        if ( sico.MODEL == 7 ) xwdu[2] = ( wu3[1] - wu3[4] ) / ( 2 * dy[i] ); else xwdu[2] = 0;

                        walphax = ( walpha[2] - walpha[5] ) / ( 2 * dx[i] );
                        walphay = ( walpha[1] - walpha[4] ) / ( 2 * dy[i] );

                        if ( sico.MODEL > 3 ) walphax2 = ( walpha2[2] - walpha2[5] ) / ( 2 * dx[i] ); else walphax2 = 0;
                        if ( sico.MODEL > 3 ) walphay2 = ( walpha2[1] - walpha2[4] ) / ( 2 * dy[i] ); else walphay2 = 0;

                        if ( sico.MODEL == 7 ) walphax3 = ( walpha3[2] - walpha3[5] ) / ( 2 * dx[i] ); else walphax3 = 0;
                        if ( sico.MODEL == 7 ) walphay3 = ( walpha3[1] - walpha3[4] ) / ( 2 * dy[i] ); else walphay3 = 0;


                        // Earth pressure coefficients

                        if ( sico.MODEL <= 3 ) { whflow = aw[i][0], whflow2 = 0, whflow3 = 0; }
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { whflow = aw[i][0], whflow2 = aw[i][3], whflow3 = 0; }
                        else if ( sico.MODEL == 7 ) { whflow = aw[i][0], whflow2 = aw[i][3], whflow3 = aw[i][6]; }

                        hekin = 0.5 * whflow * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 ));
                        if ( sico.MODEL > 3 ) hekin += 0.5 * whflow2 * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ));
                        if ( sico.MODEL == 7 ) hekin += 0.5 * whflow3 * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ));

                        for ( l=0; l<sico.PMAX; l++ ) {

                            if ( sico.PHASES[l] < 2 ) {

                                gkx[l] = fk( whflow+whflow2+whflow3, aw[i][3*l], wdu[l], hekin, l, sico, sflow );
                                gky[l] = fk( whflow+whflow2+whflow3, aw[i][3*l], wdv[l], hekin, l, sico, sflow );
                            }
                        }

                        for ( k=0; k<18; k++ ) wwd[k] = ad[i][k+sico.NVECTMIN];


                        // Curvature, flux, source, and deceleration terms

                        kappau = fcurv( vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, welev,
                            grav, sico );

                        vm = fvm( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, sico, sflow );
                        cdrag = fdrag( whflow, whflow2, whflow3, sico, sflow );
                        gze = fgze( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, wh, whx, why,
                            wdu, wdv, xwdu, xwdv, wu, wv, wu2, wv2, wu3, wv3, walpha, walpha2, walpha3, dx[i], dy[i],
                            grav, betax[i], betay[i], cdrag, tsum, sico, sflow );
                        disp = fdisp( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, wdu, wdv, xwdu, xwdv, 
                            betax[i], betay[i], gkx, gky, vm, cdrag, tsum, sico, sflow );

                        gf = ff( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3,
                            gkx, gze, kappau, vm, disp, sico );
                        gg = fg( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3,
                            gky, gze, kappau, vm, disp, sico );
                        gs = fs( wh, whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3,
                            whx, why, walphax, walphay, walphax2, walphay2, walphax3, walphay3, grav, gze, betax[i], betay[i], kappau, cdrag, sico, sflow );
                        gdecel = fd( wh, whx, why, wu, wv, wu2, wv2, wu3, wv3, walpha, walpha2, walpha3, betax[i], betay[i], gze, dx[i], dy[i],
                            wwd, kappau, hekin, sico, sflow );

                        for ( k=0; k<sico.NVECTMIN; k++ ) { 
                            af[i][k] = gf[k]; ag[i][k] = gg[k]; as[i][k] = gs[k];
                        }
                        for ( k=0; k<sico.NVECTMIN+18; k++ ) ad[i][k] = gdecel[k];

                        free( kappau ); free( vm ); free( cdrag ); free( gze ); free( disp ); free( gf ); free( gg ); free( gs ); free( gdecel );


                        // Slopes of the vector components                        

                        asigma_xelev[i] = fsigma( pelev[i], pelev[in[i][5]], pelev[in[i][2]], 1, dx, dy, i, sico );
                        asigma_yelev[i] = fsigma( pelev[i], pelev[in[i][4]], pelev[in[i][1]], 2, dx, dy, i, sico );

                        for (k=0; k<sico.NVECTMIN; k++) {

                            asigma_x[i][k] = fsigma( aw[i][k], aw[in[i][5]][k], aw[in[i][2]][k], 1, dx, dy, i, sico );
                            asigma_y[i][k] = fsigma( aw[i][k], aw[in[i][4]][k], aw[in[i][1]][k], 2, dx, dy, i, sico );
                        }
                    }
                }


                // Slopes of the fluxes

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cdomain[i] <= 3 ) {

                        for ( k=0; k<sico.NVECTMIN; k++) { asigma_f[i][k]=0; asigma_g[i][k]=0; }
		
                    } else {

                        for ( k=0; k<sico.NVECTMIN; k++ ) {

                            asigma_f[i][k] = fsigma ( af[i][k], af[in[i][5]][k], af[in[i][2]][k], 1, dx, dy, i, sico );
                            asigma_g[i][k] = fsigma ( ag[i][k], ag[in[i][4]][k], ag[in[i][1]][k], 2, dx, dy, i, sico );
                        }
                    }
                }


                // Values of vector at quarter of pixel after half time step


                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cdomain[i] > 3 ) {

                        fj[0][0]=1; fj[0][1]=1; fj[1][0]=1; fj[1][1]=-1; fj[2][0]=-1; fj[2][1]=1; fj[3][0]=-1; fj[3][1]=-1; // factors for positive or negative gradients
                        for ( j=0; j<4; j++ ) wintelev[i][j]=pelev[in[i][j]]+fj[j][0]*0.25*dx[i]*asigma_xelev[in[i][j]]+fj[j][1]*0.25*dy[i]*asigma_yelev[in[i][j]];
                    }
                }

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cdomain[i] > 3 ) {

                        fj[0][0]=1; fj[0][1]=1; fj[1][0]=1; fj[1][1]=-1; fj[2][0]=-1; fj[2][1]=1; fj[3][0]=-1; fj[3][1]=-1;
                            // factors for positive or negative gradients

                        for ( j=0; j<6; j++ ) { // loop over adjacent pixels:

                            for ( k=0; k<sico.NVECTMIN; k++ ) { // loop over all components of the vector:

                                dw_dt[k]=-asigma_f[in[i][j]][k]-asigma_g[in[i][j]][k]+as[in[i][j]][k]; // gradient of vector
                                dw_dttest[k]=-asigma_f[in[i][j]][k]-asigma_g[in[i][j]][k]+(as[in[i][j]][k]-ad[in[i][j]][k]); // gradient of vector
                            }

                            for ( k=0; k<sico.NVECTMIN; k++ ) { // loop over all components of the vector:

                                if ( j < 4 ) winta[i][j][k]=aw[in[i][j]][k]+fj[j][0]*0.25*dx[i]*asigma_x[in[i][j]][k]+fj[j][1]
                                    *0.25*dy[i]*asigma_y[in[i][j]][k]; // value at quarter of pixel

                                wintb[i][j][k]=aw[in[i][j]][k]+0.5*tlength*dw_dt[k];
                                wintbtest=aw[in[i][j]][k]+0.5*tlength*dw_dttest[k];

                                if (( fsign(wintb[i][j][k]) == fsign(wintbtest) && fabs(wintb[i][j][k]) > fabs(wintbtest)) || sico.ORIGINAL == 1 ) wintb[i][j][k] = wintbtest;
                                else if ( fsign(wintb[i][j][k]) == fsign(wintbtest)) wintb[i][j][k] = wintb[i][j][k];
                                else if ( k==0 || k==3 || k==6 ) wintb[i][j][k] = aw[in[i][j]][k];
                                else wintb[i][j][k] = 0;

                                if ( j < 4 ) {

                                    wintc[i][j][k]=aw[in[i][j]][k]+0.5*tlength*dw_dt[k]+fj[j][0]*0.25*dx[i]*asigma_x[in[i][j]][k]
                                    +fj[j][1]*0.25*dy[i]*asigma_y[in[i][j]][k];

                                    wintctest=aw[in[i][j]][k]+0.5*tlength*dw_dttest[k]+fj[j][0]*0.25*dx[i]*asigma_x[in[i][j]][k]
                                    +fj[j][1]*0.25*dy[i]*asigma_y[in[i][j]][k];

                                    if (( fsign(wintc[i][j][k]) == fsign(wintctest) && fabs(wintc[i][j][k]) > fabs(wintctest)) || sico.ORIGINAL==1 ) wintc[i][j][k] = wintctest;
                                    else if ( fsign(wintc[i][j][k]) == fsign(wintctest)) wintc[i][j][k] = wintc[i][j][k];
                                    else if ( k==0 || k==3 || k==6 ) wintc[i][j][k] = aw[in[i][j]][k];
                                    else wintc[i][j][k] = 0;

                                }
                            }
                        }

                        if ( sico.NOOSC > 0 ) {

                            ctrl_lake = 0;
                            minlevel = 99999;

                            for ( j=0; j<4; j++ ) {

                                if ( sico.MODEL <= 3 ) {

                                    wh[j] = aw[in[i][j]][0];
                                    wmx = aw[in[i][j]][1];
                                    wmy = aw[in[i][j]][2];

                                } else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                                    wh[j] = aw[in[i][j]][0] + aw[in[i][j]][3];
                                    wmx = aw[in[i][j]][1] + aw[in[i][j]][4];
                                    wmy = aw[in[i][j]][2] + aw[in[i][j]][5];

                                } else if ( sico.MODEL == 7 ) {

                                    wh[j] = aw[in[i][j]][0] + aw[in[i][j]][3] + aw[in[i][j]][6];
                                    wmx = aw[in[i][j]][1] + aw[in[i][j]][4] + aw[in[i][j]][7];
                                    wmy = aw[in[i][j]][2] + aw[in[i][j]][5] + aw[in[i][j]][8];
                                }

                                if ( wh[j] > 0 && wmx == 0 && wmy == 0 ) {

                                    ctrl_lake = 1;
                                    if ( wh[j] + pelev[in[i][j]] < minlevel ) minlevel = wh[j] + pelev[in[i][j]];
                                }
                            }

                            if ( ctrl_lake == 1 ) {

                                for ( j=0; j<4; j++ ) {

                                    if ( sico.MODEL <= 3 ) {

                                        wh[j] = aw[i][0] - ( wintelev[i][j] - pelev[i] );
                                        walpha[j] = 1.0;

                                    } else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                                        wh[j] = aw[i][0] + aw[i][3] - ( wintelev[i][j] - pelev[i] );
                                        walpha[j] = fdiv( aw[i][0], aw[i][0] + aw[i][3], 0 );
                                        walpha2[j] = fdiv( aw[i][3], aw[i][0] + aw[i][3], 0 );

                                    } else if ( sico.MODEL == 7 ) {

                                        wh[j] = aw[i][0] + aw[i][3] + aw[i][6] - ( wintelev[i][j] - pelev[i] );
                                        walpha[j] = fdiv( aw[i][0], aw[i][0] + aw[i][3] + aw[i][6], 0 );
                                        walpha2[j] = fdiv( aw[i][3], aw[i][0] + aw[i][3] + aw[i][6], 0 );
                                        walpha3[j] = fdiv( aw[i][6], aw[i][0] + aw[i][3] + aw[i][6], 0 );
                                    }

                                    wintc[i][j][0] = wh[j] * walpha[j];
                                    if ( wintelev[i][j] + wh[j] > minlevel ) wintc[i][j][0] = ( minlevel - wintelev[i][j] ) * walpha[j];
                                    if ( wintc[i][j][0] < 0 ) wintc[i][j][0] = 0;

                                    if ( sico.MODEL > 3 ) {

                                        wintc[i][j][3] = wh[j] * walpha2[j];
                                        if ( wintelev[i][j] + wh[j] > minlevel ) wintc[i][j][3] = ( minlevel - wintelev[i][j] ) * walpha2[j];
                                        if ( wintc[i][j][3] < 0 ) wintc[i][j][3] = 0;
                                    }
                                    if ( sico.MODEL == 7 ) {

                                        wintc[i][j][6] = wh[j] * walpha3[j];
                                        if ( wintelev[i][j] + wh[j] > minlevel ) wintc[i][j][6] = ( minlevel - wintelev[i][j] ) * walpha3[j];
                                        if ( wintc[i][j][6] < 0 ) wintc[i][j][6] = 0;
                                     }
                                }
                            }
                        }

                    } else {

                        for ( j=0; j<4; j++ ) {
                            for ( k=0; k<sico.NVECTMIN; k++ ) {

                                if ( j < 4 ) winta[i][j][k]=0; wintb[i][j][k]=0; if ( j < 4 ) wintc[i][j][k]=0;
                            }
                        }
                    }
                }


                // Fluxes and source terms (shifted coordinate system)

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cdomain[i] > 3 ) {

                        if ( sico.PHI == 1 ) {
                            if ( pphi[i] != sico.UNDEF ) sflow.PHI[0] = pphi[i]; // internal friction angle of PHASE 1
                            else sflow.PHI[0] = sflow.PHI0[0];
                        } else sflow.PHI[0] = sflow.PHI0[0];

                        if ( sico.PHI2 == 1 ) {
                            if ( pphi2[i] != sico.UNDEF ) sflow.PHI[1] = pphi2[i]; // internal friction angle of PHASE 2
                            else sflow.PHI[1] = sflow.PHI0[1];
                        } else sflow.PHI[1] = sflow.PHI0[1];

                        if ( sico.PHI3 == 1 ) {
                            if ( pphi3[i] != sico.UNDEF ) sflow.PHI[2] = pphi3[i]; // internal friction angle of PHASE 3
                            else sflow.PHI[2] = sflow.PHI0[2];
                        } else sflow.PHI[2] = sflow.PHI0[2];

                        if ( sico.DELTA == 1 ) {
                            if ( pdelta[i] != sico.UNDEF ) sflow.DELTA[0] = pdelta[i]; // basal friction angle of PHASE 1
                            else sflow.DELTA[0] = sflow.DELTA0[0];
                        } else sflow.DELTA[0] = sflow.DELTA0[0];

                        if ( sico.DELTA2 == 1 ) {
                            if ( pdelta2[i] != sico.UNDEF ) sflow.DELTA[1] = pdelta2[i]; // basal friction angle of PHASE 2
                            else sflow.DELTA[1] = sflow.DELTA0[1];
                        } else sflow.DELTA[1] = sflow.DELTA0[1];

                        if ( sico.DELTA3 == 1 ) {
                            if ( pdelta3[i] != sico.UNDEF ) sflow.DELTA[2] = pdelta3[i]; // basal friction angle of PHASE 3
                            else sflow.DELTA[2] = sflow.DELTA0[2];
                        } else sflow.DELTA[2] = sflow.DELTA0[2];

                        if ( sico.NYSS == 1 ) {
                            if ( pnyss[i] != sico.UNDEF ) sflow.NY[0] = pnyss[i]; // viscosity of PHASE 1
                            else sflow.NY[0] = sflow.NY0[0];
                        } else sflow.NY[0] = sflow.NY0[0];

                        if ( sico.NYFS == 1 ) {
                            if ( pnyfs[i] != sico.UNDEF ) sflow.NY[1] = pnyfs[i]; // viscosity of PHASE 2
                            else sflow.NY[1] = sflow.NY0[1];
                        } else sflow.NY[1] = sflow.NY0[1];

                        if ( sico.NYFF == 1 ) {
                            if ( pnyff[i] != sico.UNDEF ) sflow.NY[2] = pnyff[i]; // viscosity of PHASE 3
                            else sflow.NY[2] = sflow.NY0[2];
                        } else sflow.NY[2] = sflow.NY0[2];

                        if ( sico.AMBDRAG == 1 ) {
                            if ( pambdrag[i] != sico.UNDEF ) sflow.AMBDRAG = pambdrag[i]; // ambient drag coefficient
                            else sflow.AMBDRAG = sflow.AMBDRAG0;
                        } else sflow.AMBDRAG = sflow.AMBDRAG0;

                        if ( sico.FLUFRI == 1 ) {
                            if ( pflufri[i] != sico.UNDEF ) sflow.FLUFRI = pflufri[i]; // fluid friction coefficient
                            else sflow.FLUFRI = sflow.FLUFRI0;
                        } else sflow.FLUFRI = sflow.FLUFRI0;

                        for ( j=0; j<4; j++ ) {


                            // Slopes, topography-following cell sizes, and gravity components at quarter of pixel

                            if ( cdomain2[i] == 1 ) {

                                wbetax = betax[i];
                                wbetay = betay[i];
                                wbetaxy = betaxy[i];

                            } else {

                                wbetax = fbeta( wintelev[in[i][5]][j], wintelev[in[i][2]][j], 2.0, sico );
                                wbetay = fbeta( wintelev[in[i][4]][j], wintelev[in[i][1]][j], 2.0, sico );
                                wbetaxy = fbetaxy( wbetax, wbetay );

                            }

                            if ( sico.CORRHEIGHT != 0 ) {

                                wdx = sico.CSZ / cos( wbetax );
                                wdy = sico.CSZ / cos( wbetay );

                            } else {

                                wdx = sico.CSZ;
                                wdy = sico.CSZ;
                            }

                            wgrav[0] = sico.GRAVITY * cos( wbetax );
                            wgrav[1] = sico.GRAVITY * cos( wbetay );
                            wgrav[2] = sico.GRAVITY * sin( wbetax );
                            wgrav[3] = sico.GRAVITY * sin( wbetay );
                            wgrav[4] = sico.GRAVITY * cos( wbetaxy );


                            // Total flow depths

                            for ( l=1; l<9; l++ ) {

                                if ( sico.MODEL <= 3 ) wh[l] = wintb[in[i][l]][j][0];
                                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) wh[l] = wintb[in[i][l]][j][0] + wintb[in[i][l]][j][3];
                                else if ( sico.MODEL == 7 ) wh[l] = wintb[in[i][l]][j][0] + wintb[in[i][l]][j][3] + wintb[in[i][l]][j][6];
                            }


                            // Gradients of total flow depth

                            whx = ( wh[2] - wh[5] ) / ( 2 * dx[i] );
                            if ( sico.NOOSC > 0 && ( wh[2] == 0 || wh[5] == 0 )) whx = tan( betax[i] ); 

                            why = ( wh[1] - wh[4] ) / ( 2 * dy[i] );
                            if ( sico.NOOSC > 0 && ( wh[1] == 0 || wh[4] == 0 )) why = tan( betay[i] ); 


                            // Flow velocities in x and y directions

                            vflowx = fdiv( wintb[i][j][1], wintb[i][j][0], sico.HFLOWMIN );
                            vflowy = fdiv( wintb[i][j][2], wintb[i][j][0], sico.HFLOWMIN );

                            if ( sico.MODEL > 3 ) vflowx2 = fdiv( wintb[i][j][4], wintb[i][j][3], sico.HFLOWMIN ); else vflowx2 = 0;
                            if ( sico.MODEL > 3 ) vflowy2 = fdiv( wintb[i][j][5], wintb[i][j][3], sico.HFLOWMIN ); else vflowy2 = 0;

                            if ( sico.MODEL == 7 ) vflowx3 = fdiv( wintb[i][j][7], wintb[i][j][6], sico.HFLOWMIN ); else vflowx3 = 0;
                            if ( sico.MODEL == 7 ) vflowy3 = fdiv( wintb[i][j][8], wintb[i][j][6], sico.HFLOWMIN ); else vflowy3 = 0;


                            // Gradients of flow velocities

                            for ( l=1; l<9; l++ ) {

                                wu[l] = fdiv( wintb[in[i][l]][j][1], wintb[in[i][l]][j][0], sico.HFLOWMIN );
                                wv[l] = fdiv( wintb[in[i][l]][j][2], wintb[in[i][l]][j][0], sico.HFLOWMIN );

                                if ( sico.MODEL > 3 ) wu2[l] = fdiv( wintb[in[i][l]][j][4], wintb[in[i][l]][j][3], sico.HFLOWMIN ); else wu2[l] = 0;
                                if ( sico.MODEL > 3 ) wv2[l] = fdiv( wintb[in[i][l]][j][5], wintb[in[i][l]][j][3], sico.HFLOWMIN ); else wv2[l] = 0;

                                if ( sico.MODEL == 7 ) wu3[l] = fdiv( wintb[in[i][l]][j][7], wintb[in[i][l]][j][6], sico.HFLOWMIN ); else wu3[l] = 0;
                                if ( sico.MODEL == 7 ) wv3[l] = fdiv( wintb[in[i][l]][j][8], wintb[in[i][l]][j][6], sico.HFLOWMIN ); else wv3[l] = 0;

                                walpha[l] = fdiv( wintb[in[i][l]][j][0], wh[l], sico.HFLOWMIN );
                                if ( sico.MODEL > 3 ) walpha2[l] = fdiv( wintb[in[i][l]][j][3], wh[l], sico.HFLOWMIN ); else walpha2[l] = 0;
                                if ( sico.MODEL == 7 ) walpha3[l] = fdiv( wintb[in[i][l]][j][6], wh[l], sico.HFLOWMIN ); else walpha3[l] = 0;
                            }

                            wdu[0] = ( wu[2] - wu[5] ) / ( 2 * wdx );
                            wdv[0] = ( wv[1] - wv[4] ) / ( 2 * wdy );

                            if ( sico.MODEL > 3  ) wdu[1] = ( wu2[2] - wu2[5] ) / ( 2 * wdx ); else wdu[1] = 0;
                            if ( sico.MODEL > 3  ) wdv[1] = ( wv2[1] - wv2[4] ) / ( 2 * wdy ); else wdv[1] = 0;

                            if ( sico.MODEL == 7 ) wdu[2] = ( wu3[2] - wu3[5] ) / ( 2 * wdx ); else wdu[2] = 0;
                            if ( sico.MODEL == 7 ) wdv[2] = ( wv3[1] - wv3[4] ) / ( 2 * wdy ); else wdv[2] = 0;

                            xwdv[0] = ( wv[2] - wv[5] ) / ( 2 * wdx );
                            xwdu[0] = ( wu[1] - wu[4] ) / ( 2 * wdy );

                            if ( sico.MODEL > 3 ) xwdv[1] = ( wv2[2] - wv2[5] ) / ( 2 * dx[i] ); else xwdv[1] = 0;
                            if ( sico.MODEL > 3 ) xwdu[1] = ( wu2[1] - wu2[4] ) / ( 2 * dy[i] ); else xwdu[1] = 0;

                            if ( sico.MODEL == 7 ) xwdv[2] = ( wv3[2] - wv3[5] ) / ( 2 * dx[i] ); else xwdv[2] = 0;
                            if ( sico.MODEL == 7 ) xwdu[2] = ( wu3[1] - wu3[4] ) / ( 2 * dy[i] ); else xwdu[2] = 0;

 
                            // Earth pressure coefficients

                            if ( sico.MODEL <= 3 ) { whflow = wintb[i][j][0], whflow2 = 0, whflow3 = 0; }
                            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { whflow = wintb[i][j][0], whflow2 = wintb[i][j][3], whflow3 = 0; }
                            else if ( sico.MODEL == 7 ) { whflow = wintb[i][j][0], whflow2 = wintb[i][j][3], whflow3 = wintb[i][j][6]; }

                            hekin = 0.5 * whflow * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 ));
                            if ( sico.MODEL > 3 ) hekin += 0.5 * whflow2 * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ));
                            if ( sico.MODEL == 7 ) hekin += 0.5 * whflow3 * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ));

                            for ( ll=0; ll<sico.PMAX; ll++ ) {

                                if ( sico.PHASES[ll] < 2 ) {

                                    gkx[ll] = fk( whflow+whflow2+whflow3, wintb[i][j][3*ll], wdu[ll], hekin, ll, sico, sflow );
                                    gky[ll] = fk( whflow+whflow2+whflow3, wintb[i][j][3*ll], wdv[ll], hekin, ll, sico, sflow );
                                }
                            }


                            // Flux terms

                            kappau = fcurv( vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, welev, wgrav, sico );

                            vm = fvm( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, sico, sflow );
                            cdrag = fdrag( whflow, whflow2, whflow3, sico, sflow );
                            gze = fgze( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, wh, whx, why, wdu, wdv, xwdu, xwdv, 
                                wu, wv, wu2, wv2, wu3, wv3, walpha, walpha2, walpha3, wdx, wdy, wgrav, wbetax, wbetay, cdrag, tsum, sico, sflow );
                            disp = fdisp( whflow,whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, wdu, wdv, xwdu, xwdv, 
                                wbetax, wbetay, gkx, gky, vm, cdrag, tsum, sico, sflow );

                            gf = ff( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3,
                                gkx, gze, kappau, vm, disp, sico );
                            gg = fg( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3,
                                gky, gze, kappau, vm, disp, sico );

                            for ( k=0; k<sico.NVECTMIN; k++ ) { f[j][k] = gf[k]; g[j][k] = gg[k]; }
                            free( kappau ); free( vm ); free( cdrag ); free( gze ); free( disp ); free( gf ); free( gg );


                            // Flow velocities in x and y directions

                            vflowx = fdiv( wintc[i][j][1], wintc[i][j][0], sico.HFLOWMIN );
                            vflowy = fdiv( wintc[i][j][2], wintc[i][j][0], sico.HFLOWMIN );

                            if ( sico.MODEL > 3 ) vflowx2 = fdiv( wintc[i][j][4], wintc[i][j][3], sico.HFLOWMIN ); else vflowx2 = 0;
                            if ( sico.MODEL > 3 ) vflowy2 = fdiv( wintc[i][j][5], wintc[i][j][3], sico.HFLOWMIN ); else vflowy2 = 0;

                            if ( sico.MODEL == 7 ) vflowx3 = fdiv( wintc[i][j][7], wintc[i][j][6], sico.HFLOWMIN ); else vflowx3 = 0;
                            if ( sico.MODEL == 7 ) vflowy3 = fdiv( wintc[i][j][8], wintc[i][j][6], sico.HFLOWMIN ); else vflowy3 = 0;


                            // Total flow depths, velocities, and fractions of neighbouring cells

                            welev[0] = wintelev[i][0];

                            for ( l=0; l<9; l++ ) {

                                welev[l] = wintelev[in[i][l]][j];

                                if ( sico.MODEL <= 3 ) wh[l] = wintc[in[i][l]][j][0];
                                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) wh[l] = wintc[in[i][l]][j][0] + wintc[in[i][l]][j][3];
                                else if ( sico.MODEL == 7 ) wh[l] = wintc[in[i][l]][j][0] + wintc[in[i][l]][j][3] + wintc[in[i][l]][j][6];

                                wu[l] = fdiv( wintc[in[i][l]][j][1], wintc[in[i][l]][j][0], sico.HFLOWMIN );
                                wv[l] = fdiv( wintc[in[i][l]][j][2], wintc[in[i][l]][j][0], sico.HFLOWMIN );

                                if ( sico.MODEL > 3 ) wu2[l] = fdiv( wintc[in[i][l]][j][4], wintc[in[i][l]][j][3], sico.HFLOWMIN ); else wu2[l] = 0;
                                if ( sico.MODEL > 3 ) wv2[l] = fdiv( wintc[in[i][l]][j][5], wintc[in[i][l]][j][3], sico.HFLOWMIN ); else wv2[l] = 0;

                                if ( sico.MODEL == 7 ) wu3[l] = fdiv( wintc[in[i][l]][j][7], wintc[in[i][l]][j][6], sico.HFLOWMIN ); else wu3[l] = 0;
                                if ( sico.MODEL == 7 ) wv3[l] = fdiv( wintc[in[i][l]][j][8], wintc[in[i][l]][j][6], sico.HFLOWMIN ); else wv3[l] = 0;

                                walpha[l] = fdiv( wintc[in[i][l]][j][0], wh[l], sico.HFLOWMIN );
                                if ( sico.MODEL > 3 ) walpha2[l] = fdiv( wintc[in[i][l]][j][3], wh[l], sico.HFLOWMIN ); else walpha2[l] = 0;
                                if ( sico.MODEL == 7 ) walpha3[l] = fdiv( wintc[in[i][l]][j][6], wh[l], sico.HFLOWMIN ); else walpha3[l] = 0;
                            }


                            // Gradients of total flow depth, flow velocities, and fractions

                            whx = ( wh[2] - wh[5] ) / ( 2 * wdx );
                            if ( sico.NOOSC > 0 && ( wh[2] == 0 || wh[5] == 0 )) whx = tan( wbetax ); 

                            why = ( wh[1] - wh[4] ) / ( 2 * wdy );
                            if ( sico.NOOSC > 0 && ( wh[1] == 0 || wh[4] == 0 )) why = tan( wbetay );

                            wdu[0] = ( wu[2] - wu[5] ) / ( 2 * wdx );
                            wdv[0] = ( wv[1] - wv[4] ) / ( 2 * wdy );

                            if ( sico.MODEL > 3 ) wdu[1] = ( wu2[2] - wu2[5] ) / ( 2 * wdx ); else wdu[1] = 0;
                            if ( sico.MODEL > 3 ) wdv[1] = ( wv2[1] - wv2[4] ) / ( 2 * wdy ); else wdv[1] = 0;

                            if ( sico.MODEL == 7 ) wdu[2] = ( wu3[2] - wu3[5] ) / ( 2 * wdx ); else wdu[2] = 0;
                            if ( sico.MODEL == 7 ) wdv[2] = ( wv3[1] - wv3[4] ) / ( 2 * wdy ); else wdv[2] = 0;

                            xwdv[0] = ( wv[2] - wv[5] ) / ( 2 * wdx );
                            xwdu[0] = ( wu[1] - wu[4] ) / ( 2 * wdy );

                            if ( sico.MODEL > 3 ) xwdv[1] = ( wv2[2] - wv2[5] ) / ( 2 * dx[i] ); else xwdv[1] = 0;
                            if ( sico.MODEL > 3 ) xwdu[1] = ( wu2[1] - wu2[4] ) / ( 2 * dy[i] ); else xwdu[1] = 0;

                            if ( sico.MODEL == 7 ) xwdv[2] = ( wv3[2] - wv3[5] ) / ( 2 * dx[i] ); else xwdv[2] = 0;
                            if ( sico.MODEL == 7 ) xwdu[2] = ( wu3[1] - wu3[4] ) / ( 2 * dy[i] ); else xwdu[2] = 0;

                            walphax = ( walpha[2] - walpha[5] ) / ( 2 * wdx );
                            walphay = ( walpha[1] - walpha[4] ) / ( 2 * wdy );

                            if ( sico.MODEL > 3 ) walphax2 = ( walpha2[2] - walpha2[5] ) / ( 2 * wdx ); else walphax2 = 0;
                            if ( sico.MODEL > 3 ) walphay2 = ( walpha2[1] - walpha2[4] ) / ( 2 * wdy ); else walphay2 = 0;

                            if ( sico.MODEL == 7 ) walphax3 = ( walpha3[2] - walpha3[5] ) / ( 2 * wdx ); else walphax3 = 0;
                            if ( sico.MODEL == 7 ) walphay3 = ( walpha3[1] - walpha3[4] ) / ( 2 * wdy ); else walphay3 = 0;


                            // Source terms

                            if ( sico.MODEL <= 3 ) { whflow = wintc[i][j][0], whflow2 = 0, whflow3 = 0; }
                            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { whflow = wintc[i][j][0], whflow2 = wintc[i][j][3], whflow3 = 0; }
                            else if ( sico.MODEL == 7 ) { whflow = wintc[i][j][0], whflow2 = wintc[i][j][3], whflow3 = wintc[i][j][6]; }

                            hekin = 0.5 * whflow * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 ));
                            if ( sico.MODEL > 3 ) hekin += 0.5 * whflow2 * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ));
                            if ( sico.MODEL == 7 ) hekin += 0.5 * whflow3 * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ));

                            kappau = fcurv( vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, welev, wgrav, sico );
                            vm = fvm( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, sico, sflow );
                            cdrag = fdrag( whflow, whflow2, whflow3, sico, sflow );

                            gze = fgze( whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3, wh, whx, why, wdu, wdv, xwdu, xwdv, 
                                wu, wv, wu2, wv2, wu3, wv3, walpha, walpha2, walpha3, wdx, wdy, wgrav, wbetax, wbetay, cdrag, tsum, sico, sflow );

                            gs = fs( wh, whflow, whflow2, whflow3, vflowx, vflowx2, vflowx3, vflowy, vflowy2, vflowy3,
                                whx, why, walphax, walphay, walphax2, walphay2, walphax3, walphay3, wgrav, gze, wbetax, wbetay, kappau, cdrag, sico, sflow );

                            for ( k=0; k<sico.NVECTMIN; k++ ) s[j][k] = gs[k];
                            free( gs );


                            // Deceleration terms

                            for ( k=0; k<18; k++ ) wwd[k] = d[i][j][k+sico.NVECTMIN];

                            gdecel = fd( wh, whx, why, wu, wv, wu2, wv2, wu3, wv3, walpha, walpha2, walpha3, wbetax, wbetay, gze, wdx, wdy, 
                                wwd, kappau, hekin, sico, sflow );

                            for ( k=0; k<sico.NVECTMIN+18; k++ ) d[i][j][k] = gdecel[k];
                            free( kappau ); free( cdrag ); free( vm ); free( gze ); free( gdecel );
                        }


                        // Value of vector at top right corner of the pixel

                        for ( k=0; k<sico.NVECTMIN; k++ ) {

                            wintd[i][k]=0.25*(winta[i][0][k]+winta[i][1][k]+winta[i][2][k]+winta[i][3][k])
                                -tlength/dx[i]*(0.5*(f[2][k]+f[3][k])-0.5*(f[0][k]+f[1][k]))
                                -tlength/dy[i]*(0.5*(g[1][k]+g[3][k])-0.5*(g[0][k]+g[2][k]))
                                +0.25*tlength*(s[0][k]+s[1][k]+s[2][k]+s[3][k]);

                            if ( sico.NOOSC > 0 && fabs(-tlength/dx[i]*(0.5*(f[2][k]+f[3][k])-0.5*(f[0][k]+f[1][k]))
                                -tlength/dy[i]*(0.5*(g[1][k]+g[3][k])-0.5*(g[0][k]+g[2][k]))
                                 +0.25*tlength*(s[0][k]+s[1][k]+s[2][k]+s[3][k])) < pow( 10, -7 )) {

                                 if ( p == 0 ) wintd[i][k] = aw[in[i][0]][k];
                                 else wintd[i][k] = aw[in[i][3]][k];
                            }

                            wintdtest[i][k]=0.25*(winta[i][0][k]+winta[i][1][k]+winta[i][2][k]+winta[i][3][k])
                                -tlength/dx[i]*(0.5*(f[2][k]+f[3][k])-0.5*(f[0][k]+f[1][k]))
                                -tlength/dy[i]*(0.5*(g[1][k]+g[3][k])-0.5*(g[0][k]+g[2][k]))
                                +0.25*tlength*(s[0][k]+s[1][k]+s[2][k]+s[3][k]-d[i][0][k]-d[i][1][k]-d[i][2][k]-d[i][3][k]);

                            if ( sico.NOOSC > 0 && fabs(-tlength/dx[i]*(0.5*(f[2][k]+f[3][k])-0.5*(f[0][k]+f[1][k]))
                                -tlength/dy[i]*(0.5*(g[1][k]+g[3][k])-0.5*(g[0][k]+g[2][k]))
                                +0.25*tlength*(s[0][k]+s[1][k]+s[2][k]+s[3][k]-d[i][0][k]-d[i][1][k]-d[i][2][k]-d[i][3][k])) < pow( 10, -7 )) {

                                if ( p == 0 ) wintdtest[i][k] = aw[in[i][0]][k];
                                else wintdtest[i][k] = aw[in[i][3]][k];
                            }
                        }

                    } else for ( k=0; k<sico.NVECTMIN; k++ ) { wintd[i][k] = aw[i][k]; wintdtest[i][k] = aw[i][k]; }
                }


                // Moving vector if second sub-timestep and writing values to temporary vector

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    ctrlv = 1; if ( in[i][6] == 0 ) ctrlv = 0;
                    if ( ctrlv == 1 ) {
                        ctrlr = 0;
                        if ( sico.MODEL <= 3 ) {

                            if ( wintd[i][0] > 0  ) ctrlr = 1;
                            if ( wintd[in[i][6]][0] > 0 ) ctrlr = 1;

                        } else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                            if ( wintd[i][0] > 0 || wintd[i][3] > 0  ) ctrlr = 1;
                            if ( wintd[in[i][6]][0] > 0 || wintd[in[i][6]][3] > 0 ) ctrlr = 1;

                        } else if ( sico.MODEL == 7 ) {

                            if ( wintd[i][0] > 0 || wintd[i][3] > 0 || wintd[i][6] > 0 ) ctrlr = 1;
                            if ( wintd[in[i][6]][0] > 0 || wintd[in[i][6]][3] > 0 || wintd[in[i][6]][6] > 0 ) ctrlr = 1;
                        }
                    }

                    if ( ctrlv == 1 && ctrlr == 1 && cdomain[i] > 3 ) {

                        for ( k=0; k<sico.NVECTMIN; k++ ) {
                            if ( p == 1 && ( cdomain[in[i][6]] > 3 || cflux[in[i][6]] == 0 )) {

                                awt[i][k]=wintd[in[i][6]][k];
                                awttest[k]=wintdtest[in[i][6]][k];
                            }

                            else if ( p != 1 && ( cdomain[i] > 3 || cflux[i] == 0 )) {

                                awt[i][k]=wintd[i][k];
                                awttest[k]=wintdtest[i][k];
                            }
                            else if ( k<sico.NVECTMIN ) { awt[i][k] = 0; awttest[k] = 0; }
                            else { awt[i][k] = aw[i][k]; awttest[k] = aw[i][k]; }
                            if (( fsign(awt[i][k]) == fsign(awttest[k]) && fabs(awt[i][k]) > fabs(awttest[k])) || sico.ORIGINAL == 1 ) awt[i][k] = awttest[k];
                            else if ( fsign(awt[i][k]) == fsign(awttest[k])) awt[i][k] = awt[i][k];
                            else if ( k==0 || k==3 || k==6 ) awt[i][k] = aw[i][k];
                            else awt[i][k] = 0;
                        }
                    }
                    else if ( cflux[i] != 0 && cstopped[i] != 1 ) { for ( k=0; k<sico.NVECTMIN; k++ ) awt[i][k] = 0; }
                    else { for ( k=0; k<sico.NVECTMIN; k++ ) awt[i][k] = aw[i][k]; }

                    if ( awt[i][0] < 0 ) { awt[i][0] = 0; awt[i][1] = 0; awt[i][2] = 0; }
                    if ( sico.MODEL > 3 && awt[i][3] < 0 ) { awt[i][3] = 0; awt[i][4] = 0; awt[i][5] = 0; }
                    if ( sico.MODEL == 7 && awt[i][6] < 0 ) { awt[i][6] = 0; awt[i][7] = 0; awt[i][8] = 0; }
                    for ( k=0; k<sico.NVECTMIN; k++ ) { if ( isnan( awt[i][k] ) != 0 ) awt[i][k] = 0; }

                    if ( cdomain2[i] == 1 ) { for ( k=0; k<3; k++ ) awt[i][k] = aw[i][k]; }
                    if ( sico.MODEL > 3 && cdomain2[i] == 1 ) { for ( k=3; k<6; k++ ) awt[i][k] = aw[i][k]; }
                    if ( sico.MODEL == 7 && cdomain2[i] == 1 ) { for ( k=6; k<9; k++ ) awt[i][k] = aw[i][k]; }
                }


// ----- STOP --- Flow propagation with NOC scheme -----------------------------------------------------------------


// ----- START -- Diffusion control --------------------------------------------------------------------------------


                // *** Still in experimental stage
                // *** cedge [i][j] (cedge2 [i][j]): edge pixel with regard to PHASE 1 (PHASE 3) in direction j - 1 = yes, 2 = no
                // *** cready[i][j] (cready2[i][j]): degree of PHASE 1 (PHASE 3) fill of neighbour pixels in direction j - ratio
                // *** cedge0[i] (cedge02[i]): number of pixels which may provide PHASE 1 (PHASE 3) inflow
                // *** cneighbours[i] (cneighbours2[i]): number of neighbour pixels with non-zero flow depth


                if ( sico.DIFFCTRL != 0 ) {

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];

                        cedge0[i] = 0; cedge02[i] = 0; cedge03[i] = 0;
                    }


                // Updating degree of fill of neighbour pixels

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];
                        if ( p == 1 ) i2 = i;//-sico.N-1; !!!CHECK correct consideration
                        else i2 = i;

                        if ( fvalid( aw[i2][0], aw[i2][3], aw[i2][6], sico ) == 1 ) { // if pixel was flow pixel at the beginning of the time step:

                            if ( aw[i2][0] > 0 ) {
                                // identifying directions for which pixel is edge pixel (with regard to PHASE 1):

                                for ( j=1; j<9; j++ ) {
                                    if ( aw[in[i2][j]][0] == 0  ) cedge[i][j] = 1;
                                    else cedge[i][j] = 0;
                                }
                            }
                            if ( sico.MODEL > 3 && aw[i2][3] > 0 ) {
                                // identifying directions for which pixel is edge pixel with regard to PHASE 3:

                                for ( j=1; j<9; j++ ) {
                                    if ( aw[in[i2][j]][3] == 0 ) cedge2[i][j] = 1;
                                    else cedge2[i][j] = 0;
                                }
                            }
                            if ( sico.MODEL == 7 && aw[i2][6] > 0 ) {
                                // identifying directions for which pixel is edge pixel with regard to PHASE 3:

                                for ( j=1; j<9; j++ ) {
                                    if ( aw[in[i2][j]][6] == 0 ) cedge3[i][j] = 1;
                                    else cedge3[i][j] = 0;
                                }
                            }

                            if ( nsum == 1 ) { // for first time step:

                                if ( cedge[i][2] == 1 && aw[i2][0] > 0 ) cready[i][2] = 2; // degree of fill (for PHASE 1)
                                if ( cedge[i][5] == 1 && aw[i2][0] > 0 ) cready[i][5] = 2;
                                if ( cedge[i][1] == 1 && aw[i2][0] > 0 ) cready[i][1] = 2;
                                if ( cedge[i][4] == 1 && aw[i2][0] > 0 ) cready[i][4] = 2;

                                if ( sico.MODEL > 3 && cedge2[i][2] == 1 && aw[i2][3] > 0 ) // degree of fill for PHASE 2
                                    cready2[i][2] = 2;
                                if ( sico.MODEL > 3 && cedge2[i][5] == 1 && aw[i2][3] > 0 )
                                    cready2[i][5] = 2;
                                if ( sico.MODEL > 3 && cedge2[i][1] == 1 && aw[i2][3] > 0 )
                                    cready2[i][1] = 2;
                                if ( sico.MODEL > 3 && cedge2[i][4] == 1 && aw[i2][3] > 0 )
                                    cready2[i][4] = 2;

                                if ( sico.MODEL == 7 && cedge3[i][2] == 1 && aw[i2][6] > 0 ) // degree of fill for PHASE 3
                                    cready3[i][2] = 2;
                                if ( sico.MODEL == 7 && cedge3[i][5] == 1 && aw[i2][6] > 0 )
                                    cready3[i][5] = 2;
                                if ( sico.MODEL == 7 && cedge3[i][1] == 1 && aw[i2][6] > 0 )
                                    cready3[i][1] = 2;
                                if ( sico.MODEL == 7 && cedge3[i][4] == 1 && aw[i2][6] > 0 )
                                    cready3[i][4] = 2;

                            } else { // for all other time steps:

                                if ( cedge[i][2] == 1 && aw[i2][0] > 0 )
                                    cready[i][2] = cready[i2][2] + aw[i2][1] / aw[i2][0] * 2*tlength / sico.CSZ; // degree of fill (for PHASE 1)
                                if ( cedge[i][5] == 1 && aw[i2][0] > 0 )
                                    cready[i][5] = cready[i2][5] - aw[i2][1] / aw[i2][0] * 2*tlength / sico.CSZ;
                                if ( cedge[i][1] == 1 && aw[i2][0] > 0 )
                                    cready[i][1] = cready[i2][1] + aw[i2][2] / aw[i2][0] * 2*tlength / sico.CSZ;
                                if ( cedge[i][4] == 1 && aw[i2][0] > 0 )
                                    cready[i][4] = cready[i2][4] - aw[i2][2] / aw[i2][0] * 2*tlength / sico.CSZ;

                                if ( sico.MODEL > 3 && cedge2[i][2] == 1 && aw[i2][3] > 0 ) // degree of fill for PHASE 2
                                    cready2[i][2] = cready2[i2][2] + aw[i2][4] / aw[i2][3] * 2*tlength / sico.CSZ;
                                if ( sico.MODEL > 3 && cedge2[i][5] == 1 && aw[i2][3] > 0 )
                                    cready2[i][5] = cready2[i2][5] - aw[i2][4] / aw[i2][3] * 2*tlength / sico.CSZ;
                                if ( sico.MODEL > 3 && cedge2[i][1] == 1 && aw[i2][3] > 0 )
                                    cready2[i][1] = cready2[i2][1] + aw[i2][5] / aw[i2][3] * 2*tlength / sico.CSZ;
                                if ( sico.MODEL > 3 && cedge2[i][4] == 1 && aw[i2][3] > 0 )
                                    cready2[i][4] = cready2[i2][4] - aw[i2][5] / aw[i2][3] * 2*tlength / sico.CSZ;

                                if ( sico.MODEL == 7 && cedge3[i][2] == 1 && aw[i2][6] > 0 ) // degree of fill for PHASE 3
                                    cready3[i][2] = cready3[i2][2] + aw[i2][7] / aw[i2][6] * 2*tlength / sico.CSZ;
                                if ( sico.MODEL == 7 && cedge3[i][5] == 1 && aw[i2][6] > 0 )
                                    cready3[i][5] = cready3[i2][5] - aw[i2][7] / aw[i2][6] * 2*tlength / sico.CSZ;
                                if ( sico.MODEL == 7 && cedge3[i][1] == 1 && aw[i2][6] > 0 )
                                    cready3[i][1] = cready3[i2][1] + aw[i2][8] / aw[i2][6] * 2*tlength / sico.CSZ;
                                if ( sico.MODEL == 7 && cedge3[i][4] == 1 && aw[i2][6] > 0 )
                                    cready3[i][4] = cready3[i2][4] - aw[i2][8] / aw[i2][6] * 2*tlength / sico.CSZ;
                            }
                        }
                    }


                    // Number of pixels which may provide inflow to neighbour pixels (cedge0, cedge02)

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];
                        if ( p == 1 ) i2 = i;//-sico.N-1;
                        else i2 = i;

                        cneighbours[i] = 0; cneighbours2[i] = 0; cneighbours3[i] = 0;

                        if ( fvalid( aw[i2][0], aw[i2][3], aw[i2][6], sico ) == 1 ) {

                            for ( j=1; j<3; j++ ) {

                                if (( cedge [i][j] == 0 || cready [i][j] > 1 )
                                    && aw[i2][0] > 0 ) cedge0 [in[i][j]] += 1;
                                if ( sico.MODEL > 3 && ( cedge2[i][j] == 0 || cready2[i][j] > 1 )
                                    && aw[i2][3] > 0 ) cedge02[in[i][j]] += 1;
                                if ( sico.MODEL == 7 && ( cedge3[i][j] == 0 || cready3[i][j] > 1 )
                                    && aw[i2][6] > 0 ) cedge03[in[i][j]] += 1;
                            }
                            for ( j=4; j<6; j++ ) {

                                if (( cedge [i][j] == 0 || cready [i][j] > 1 )
                                    && aw[i2][0] > 0 ) cedge0 [in[i][j]] += 1;
                                if ( sico.MODEL > 3 && ( cedge2[i][j] == 0 || cready2[i][j] > 1 )
                                    && aw[i2][3] > 0 ) cedge02[in[i][j]] += 1;
                                if ( sico.MODEL == 7 && ( cedge3[i][j] == 0 || cready3[i][j] > 1 )
                                    && aw[i2][6] > 0 ) cedge03[in[i][j]] += 1;
                            }
                            if (( cedge[i][3] == 0 || cready[i][1] > 1 || cready[i][2] > 1 )
                                && aw[i2][0] > 0 ) cedge0 [in[i][3]] += 1;
                            if (( cedge[i][6] == 0 || cready[i][4] > 1 || cready[i][5] > 1 )
                                && aw[i2][0] > 0 ) cedge0 [in[i][6]] += 1;
                            if (( cedge[i][7] == 0 || cready[i][2] > 1 || cready[i][4] > 1 )
                                && aw[i2][0] > 0 ) cedge0 [in[i][7]] += 1;
                            if (( cedge[i][8] == 0 || cready[i][1] > 1 || cready[i][5] > 1 )
                                && aw[i2][0] > 0 ) cedge0 [in[i][8]] += 1;

                            if ( sico.MODEL > 3 && ( cedge2[i][3] == 0 || cready2[i][1] > 1
                                || cready2[i][2] > 1 ) && aw[i2][3] > 0 ) cedge02[in[i][3]] += 1;
                            if ( sico.MODEL > 3 && ( cedge2[i][6] == 0 || cready2[i][4] > 1
                                || cready2[i][5] > 1 ) && aw[i2][3] > 0 ) cedge02[in[i][6]] += 1;
                            if ( sico.MODEL > 3 && ( cedge2[i][7] == 0 || cready2[i][2] > 1
                                || cready2[i][4] > 1 ) && aw[i2][3] > 0 ) cedge02[in[i][7]] += 1;
                            if ( sico.MODEL > 3 && ( cedge2[i][8] == 0 || cready2[i][1] > 1
                                || cready2[i][5] > 1 ) && aw[i2][3] > 0 ) cedge02[in[i][8]] += 1;

                            if ( sico.MODEL == 7 && ( cedge3[i][3] == 0 || cready3[i][1] > 1
                                || cready3[i][2] > 1 ) && aw[i2][6] > 0 ) cedge03[in[i][3]] += 1;
                            if ( sico.MODEL == 7 && ( cedge3[i][6] == 0 || cready3[i][4] > 1
                                || cready3[i][5] > 1 ) && aw[i2][6] > 0 ) cedge03[in[i][6]] += 1;
                            if ( sico.MODEL == 7 && ( cedge3[i][7] == 0 || cready3[i][2] > 1
                                || cready3[i][4] > 1 ) && aw[i2][6] > 0 ) cedge03[in[i][7]] += 1;
                            if ( sico.MODEL == 7 && ( cedge3[i][8] == 0 || cready3[i][1] > 1
                                || cready3[i][5] > 1 ) && aw[i2][6] > 0 ) cedge03[in[i][8]] += 1;
                        }
                    }


                    // Number of neighbour pixels with non-zero flow depth (cneighbours, cneighbours2)

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];

                        if ( cdomain[i] >= 3 ) {

                            for ( j=1; j<9; j++ ) { 

                                if ( cedge [i][j] == 1 && cedge0[in[i][j]] == 0 ) {

                                    if ( awt[in[i][j]][0] > 0 ) cneighbours[in[i][j]] += 1;
                                }
                            }
                            for ( j=1; j<9; j++ ) {

                                if ( cedge2 [i][j] == 1 && cedge02[in[i][j]] == 0 ) { 

                                    if ( sico.MODEL > 3 && awt[in[i][j]][3] > 0 ) cneighbours2[in[i][j]] += 1;
                                }
                            }
                            for ( j=1; j<9; j++ ) {

                                if ( cedge3 [i][j] == 1 && cedge03[in[i][j]] == 0 ) { 

                                    if ( sico.MODEL == 7 && awt[in[i][j]][6] > 0 ) cneighbours3[in[i][j]] += 1;
                                }
                            }
                        }
                    }


                    // Reallocating flow depth and momentum from outside pixels to edge pixels

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];

                        if ( cdomain[i] >= 3 ) {

                            for ( j=1; j<9; j++ ) {

                                if ( cedge [i][j] == 1 && cedge0[in[i][j]] == 0 ) {

                                    for ( k=0; k<3; k++ ) {
                                        if ( cneighbours[in[i][j]] > 0 )
                                            awt[i][k] += awt[in[i][j]][k] / cneighbours[in[i][j]];
                                    }
                                }
                                if ( sico.MODEL > 3 && cedge2[i][j] == 1 && cedge02[in[i][j]] == 0 ) {

                                    for ( k=3; k<6; k++ ) {
                                        if ( cneighbours2[in[i][j]] > 0 )
                                            awt[i][k] += awt[in[i][j]][k] / cneighbours2[in[i][j]];
                                    }
                                }
                                if ( sico.MODEL == 7 && cedge3[i][j] == 1 && cedge03[in[i][j]] == 0 ) {

                                    for ( k=6; k<9; k++ ) {
                                        if ( cneighbours3[in[i][j]] > 0 )
                                            awt[i][k] += awt[in[i][j]][k] / cneighbours3[in[i][j]];
                                    }
                                }
                            }
                        }
                    }


                    // Setting flow depth and momentum at outside pixels to zero:

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];

                        if ( cdomain[i] >= 3 ) {

                            for ( j=1; j<9; j++ ) {

                                if ( cedge [i][j] == 1 && cedge0 [in[i][j]] == 0 ) {
                                    for ( k=0; k<3; k++ ) awt[in[i][j]][k] = 0;
                                }
                                if ( sico.MODEL > 3 && cedge2[i][j] == 1 && cedge02[in[i][j]] == 0 ) {
                                    for ( k=3; k<6; k++ ) awt[in[i][j]][k] = 0;
                                }
                                if ( sico.MODEL == 7 && cedge3[i][j] == 1 && cedge03[in[i][j]] == 0 ) {
                                    for ( k=6; k<9; k++ ) awt[in[i][j]][k] = 0;
                                }
                            }
                        }
                    }
                }

 
// ----- STOP --- Diffusion control --------------------------------------------------------------------------------


// ----- START -- Evaluating time step length and validity of time step --------------------------------------------


                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( fvalid( awt[i][0], awt[i][3], awt[i][6], sico ) == 1 ) {


                        // Gradients of flow velocities

                        for ( l=1; l<9; l++ ) {

                            wu[l] = fdiv( awt[in[i][l]][1], awt[in[i][l]][0], sico.HFLOWMIN );
                            wv[l] = fdiv( awt[in[i][l]][2], awt[in[i][l]][0], sico.HFLOWMIN );

                            if ( sico.MODEL > 3 ) wu2[l] = fdiv( awt[in[i][l]][4], awt[in[i][l]][3], sico.HFLOWMIN ); else wu2[l] = 0;
                            if ( sico.MODEL > 3 ) wv2[l] = fdiv( awt[in[i][l]][5], awt[in[i][l]][3], sico.HFLOWMIN ); else wv2[l] = 0;

                            if ( sico.MODEL == 7 ) wu3[l] = fdiv( awt[in[i][l]][7], awt[in[i][l]][6], sico.HFLOWMIN ); else wu3[l] = 0;
                            if ( sico.MODEL == 7 ) wv3[l] = fdiv( awt[in[i][l]][8], awt[in[i][l]][6], sico.HFLOWMIN ); else wv3[l] = 0;
                        }

                        wdu[0] = wu[2] - wu[5];
                        wdv[0] = wv[1] - wv[4];

                        if ( sico.MODEL > 3 ) wdu[1] = wu2[2] - wu2[5];
                        if ( sico.MODEL > 3 ) wdv[1] = wv2[1] - wv2[4];

                        if ( sico.MODEL == 7 ) wdu[2] = wu3[2] - wu3[5];
                        if ( sico.MODEL == 7 ) wdv[2] = wv3[1] - wv3[4];

                        vflowx = fdiv( awt[i][1], awt[i][0], sico.HFLOWMIN );
                        vflowy = fdiv( awt[i][2], awt[i][0], sico.HFLOWMIN );

                        if ( sico.MODEL > 3 ) vflowx2 = fdiv( awt[i][4], awt[i][3], sico.HFLOWMIN ); else vflowx2 = 0;
                        if ( sico.MODEL > 3 ) vflowy2 = fdiv( awt[i][5], awt[i][3], sico.HFLOWMIN ); else vflowy2 = 0;

                        if ( sico.MODEL == 7 ) vflowx3 = fdiv( awt[i][7], awt[i][6], sico.HFLOWMIN ); else vflowx3 = 0;
                        if ( sico.MODEL == 7 ) vflowy3 = fdiv( awt[i][8], awt[i][6], sico.HFLOWMIN ); else vflowy3 = 0;

                        if ( sico.MODEL <= 3 ) { whflow = aw[i][0], whflow2 = 0, whflow3 = 0; }
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { whflow = aw[i][0], whflow2 = aw[i][3], whflow3 = 0; }
                        else if ( sico.MODEL == 7 ) { whflow = aw[i][0], whflow2 = aw[i][3], whflow3 = aw[i][6]; }

                        hekin = 0.5 * whflow * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 ));
                        if ( sico.MODEL > 3 ) hekin+= 0.5 * whflow2 * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ));
                        if ( sico.MODEL == 7 ) hekin += 0.5 * whflow3 * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ));

                        for ( l=0; l<sico.PMAX; l++ ) {

                            if ( sico.PHASES[l] < 2 ) {

                                gkx[l] = fk( whflow+whflow2+whflow3, aw[i][3*l], wdu[l], hekin, l, sico, sflow );
                                gky[l] = fk( whflow+whflow2+whflow3, aw[i][3*l], wdv[l], hekin, l, sico, sflow );
                            }
                        }

                        vcelr0 = fcvelr2( whflow, whflow2, whflow3, vflowx, vflowy, vflowx2, vflowy2, vflowx3, vflowy3, sflow,
                            gkx, gky, betaxy, i, sico );

                        if ( vcelr0 > vcelr ) vcelr = vcelr0;
                    }
                }

                cfl = vcelr * tlength / sico.CSZ; // updating CFL value

                if ( cfl <= sico.CFL[0] || vcelr == 0 ) { // if time step length meets CFL criterion:

                    ccfl = 1; // setting control for fulfilment of CFL criterion to positive
                    if ( cfl > cflmax ) cflmax = cfl; // updating maximum CFL value, if necessary

                    tlengthpre = tlength;

                    tsum += tlength; // updating time since release
                    tint += tlength; // updating time since last output

                    if ( tlength <= 0.0000001 ) { // if time step length indicates numerical failure:
                        ccontinue = 0; // setting control for continuation to negative
                        csuccess = 0;  // setting control for simulation success to negative
                    }

                    if ( vcelr > 0 ) {
                        if ( p == 0 ) tlength0 = sico.CFL[0] * 0.90 * sico.CSZ / vcelr;
                        else if ( p == 1 ) tlength0 = sico.CFL[0] * 0.90 * sico.CSZ / vcelr;
                    }
                    else {
                        if ( p == 0 ) tlength0 = sico.CFL[1];
                        else if ( p == 1 ) tlength0 = sico.CFL[1];
                    } // updating length of time step according to CFL criterion

                    vcelr = 0; // resetting flow velocity plus wave speed
                }

                else { // if time step is too long too meet CFL criterion, reducing time step length:

                    tlength = 0.75 * tlength * sico.CFL[0] / cfl; // defining new time step length
                    vcelr = 0; // resetting flow velocity plus wave speed
                }


// ----- STOP --- Evaluating time step length and validity of time step --------------------------------------------


                #ifdef WITHGRASS


                    printf("   %i\t%i\t%.3f\t%.1f  \t%.1f\t...\r", nout, nsum, cflmax, 100*tlength, tsum ); // displaying progress
                    fflush(stdout); // forcing immediate display


                #endif


// *** End of loop over time step lengths --------------------------------------------------------------------------


            }


            // Applying values of temporary vectors

            for ( ix=0; ix<ib[0]; ix++ ) {

                i = ibasket[0][ix];

                for ( k=0; k<sico.NVECTMIN; k++ ) aw[i][k] = awt[i][k];


                // Phase transformations

                if ( sico.MODEL == 4 || sico.MODEL == 7 ) {
                    if ( sico.TRANSSSFS == 1 ) {
                        if ( ptransssfs[i] != sico.UNDEF ) sflow.TRANSSSFS = ptransssfs[i];
                        else sflow.TRANSSSFS = sflow.TRANSSSFS0;
                    } else sflow.TRANSSSFS = sflow.TRANSSSFS0;
                }

                if ( sico.MODEL == 5 || sico.MODEL == 7 ) {
                    if ( sico.TRANSSSFF == 1 ) {
                        if ( ptransssff[i] != sico.UNDEF ) sflow.TRANSSSFF = ptransssff[i];
                        else sflow.TRANSSSFF = sflow.TRANSSSFF0;
                    } else sflow.TRANSSSFF = sflow.TRANSSSFF0;
                }

                if ( sico.MODEL == 6 || sico.MODEL == 7 ) {
                    if ( sico.TRANSFSFF == 1 ) {
                        if ( ptransfsff[i] != sico.UNDEF ) sflow.TRANSFSFF = ptransfsff[i];
                        else sflow.TRANSFSFF = sflow.TRANSFSFF0;
                    } else sflow.TRANSFSFF = sflow.TRANSFSFF0;
                }

                if ( aw[i][0] > 0 ) { vflowx = aw[i][1] / aw[i][0]; vflowy = aw[i][2] / aw[i][0]; } else { vflowx = 0; vflowy = 0; }
                if ( sico.MODEL > 3 && aw[i][3] > 0 ) { vflowx2 = aw[i][4] / aw[i][3]; vflowy2 = aw[i][5] / aw[i][3]; } else { vflowx2 = 0; vflowy2 = 0; }
                if ( sico.MODEL == 7 && aw[i][6] > 0 ) { vflowx3 = aw[i][7] / aw[i][6]; vflowy3 = aw[i][8] / aw[i][6]; } else { vflowx3 = 0; vflowy3 = 0; }

                if ( sico.MODEL > 3 && sico.MODEL < 7 ) ekin = ( aw[i][0] * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 ))
                    + aw[i][3] * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ))) / 2;

                else if ( sico.MODEL == 7 ) ekin = ( aw[i][0] * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 ))
                    + aw[i][3] * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ))
                    + aw[i][6] * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ))) / 2;

                if (( sico.MODEL == 5 || sico.MODEL == 7 ) && sflow.TRANSSSFF > 0 ) {

                    trans = fmin( aw[i][0], tlength * pow( 10, -fabs( sflow.TRANSSSFF )) * ekin );

                    aw[i][0] -= trans;
                    aw[i][1] -= trans * vflowx;
                    aw[i][2] -= trans * vflowy;

                    aw[i][6] += trans * sflow.RHO1 / sflow.RHO3;
                    aw[i][7] += trans * sflow.RHO1 / sflow.RHO3 * vflowx;
                    aw[i][8] += trans * sflow.RHO1 / sflow.RHO3 * vflowy;
                }

                if (( sico.MODEL == 5 || sico.MODEL == 7 ) && sflow.TRANSSSFF < 0 ) {

                    trans = fmax( -aw[i][6], -tlength * pow( 10, -fabs( sflow.TRANSSSFF )) * ekin );

                    aw[i][0] -= trans * sflow.RHO3 / sflow.RHO1;
                    aw[i][1] -= trans * vflowx3 * sflow.RHO3 / sflow.RHO1;
                    aw[i][2] -= trans * vflowy3 * sflow.RHO3 / sflow.RHO1;

                    aw[i][6] += trans;
                    aw[i][7] += trans * vflowx3;
                    aw[i][8] += trans * vflowy3;
                }

                if (( sico.MODEL == 4 || sico.MODEL == 7 ) && sflow.TRANSSSFS > 0 ) {

                    trans = fmin( aw[i][0], tlength * pow( 10, -fabs( sflow.TRANSSSFS )) * ekin );

                    aw[i][0] -= trans;
                    aw[i][1] -= trans * vflowx;
                    aw[i][2] -= trans * vflowy;

                    aw[i][3] += trans * sflow.RHO1 / sflow.RHO2;
                    aw[i][4] += trans * sflow.RHO1 / sflow.RHO2 * vflowx;
                    aw[i][5] += trans * sflow.RHO1 / sflow.RHO2 * vflowy;
                }

                if (( sico.MODEL == 4 || sico.MODEL == 7 ) && sflow.TRANSSSFS < 0 ) {

                    trans = fmax( -aw[i][3], -tlength * pow( 10, -fabs( sflow.TRANSSSFS )) * ekin );

                    aw[i][0] -= trans * sflow.RHO2 / sflow.RHO1;
                    aw[i][1] -= trans * vflowx2 * sflow.RHO2 / sflow.RHO1;
                    aw[i][2] -= trans * vflowy2 * sflow.RHO2 / sflow.RHO1;

                    aw[i][3] += trans;
                    aw[i][4] += trans * vflowx2;
                    aw[i][5] += trans * vflowy2;
                }         

                if (( sico.MODEL == 6 || sico.MODEL == 7 ) && sflow.TRANSFSFF > 0 ) {

                    trans = fmin( aw[i][3], tlength * pow( 10, -fabs( sflow.TRANSFSFF )) * ekin );

                    aw[i][3] -= trans;
                    aw[i][4] -= trans * vflowx2;
                    aw[i][5] -= trans * vflowy2;

                    aw[i][6] += trans * sflow.RHO2 / sflow.RHO3;
                    aw[i][7] += trans * sflow.RHO2 / sflow.RHO3 * vflowx2;
                    aw[i][8] += trans * sflow.RHO2 / sflow.RHO3 * vflowy2;
                }

                if (( sico.MODEL == 6 || sico.MODEL == 7 ) && sflow.TRANSFSFF < 0 ) {

                    trans = fmax( -aw[i][6], -tlength * pow( 10, -fabs( sflow.TRANSFSFF )) * ekin );

                    aw[i][3] -= trans * sflow.RHO3 / sflow.RHO2;
                    aw[i][4] -= trans * vflowx3 * sflow.RHO3 / sflow.RHO2;
                    aw[i][5] -= trans * vflowy3 * sflow.RHO3 / sflow.RHO2;

                    aw[i][6] += trans;
                    aw[i][7] += trans * vflowx3;
                    aw[i][8] += trans * vflowy3;
                }
            }


// F4 -- START -- Entrainment --------------------------------------------------------------------------------------


            // *** 1 = Entrainment coefficient multiplied with flow kinetic energy
            // *** 1 = Entrainment coefficient multiplied with flow momentum


            for ( ix=0; ix<ib[0]; ix++ ) {

                i = ibasket[0][ix];

                if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) {
                    if ( pcentr[i] != sico.UNDEF ) sflow.CENTR = pcentr[i]; // entrainment coefficient
                    else sflow.CENTR = sflow.CENTR0;
                }
                else sflow.CENTR = sflow.CENTR0;

                if ( sico.ENTRAINMENT != 0 ) {

                    if ( aw[i][0] > 0 ) { vflowx = aw[i][1]/aw[i][0]; vflowy = aw[i][2]/aw[i][0]; } else { vflowx = 0; vflowy = 0; } // computing flow velocities
                    if ( sico.MODEL > 3 && aw[i][3] > 0 ) { vflowx2 = aw[i][4]/aw[i][3]; vflowy2 = aw[i][5]/aw[i][3]; } else { vflowx2 = 0; vflowy2 = 0; }
                    if ( sico.MODEL == 7 && aw[i][6] > 0 ) { vflowx3 = aw[i][7]/aw[i][6]; vflowy3 = aw[i][8]/aw[i][6]; } else { vflowx3 = 0; vflowy3 = 0; }

                    if ( sico.ENTRAINMENT == 1 ) {

                        hekin = 0.5 * aw[i][0] * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 )); // computing flow kinetic energies
                        if ( sico.MODEL > 3 ) hekin+= 0.5 * aw[i][3] * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ));
                        if ( sico.MODEL == 7 ) hekin += 0.5 * aw[i][6] * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ));

                    } else {

                        hekin = aw[i][0] * sflow.RHO1 * pow( pow( vflowx, 2 ) + pow( vflowy, 2 ), 0.5 ); // computing flow momenta
                        if ( sico.MODEL > 3 ) hekin+= aw[i][3] * sflow.RHO2 * pow( pow( vflowx2, 2 ) + pow( vflowy2, 2 ), 0.5 );
                        if ( sico.MODEL == 7 ) hekin += aw[i][6] * sflow.RHO3 * pow( pow( vflowx3, 2 ) + pow( vflowy3, 2 ), 0.5 );
                    }
                }

                if ( sico.MODEL <= 3 && sico.ENTRAINMENT != 0 && aw[i][0] > sico.HFLOWMIN ) {
                    // if entrainment is enabled with one-phase model:

                    if ( sico.CORRHEIGHT == 0 ) corrfact = 1; // correction factor for flow height to depth conversion
                    else corrfact = cos( betaxy[i] );

                    potentr = ( phentrmax[i] - ( pelev0[i] - pelev[i] )) * corrfact;
                    if ( potentr < 0 ) potentr = 0; // potential entrainment rate

                    qentr = tlength * sflow.CENTR * hekin; // entrainment rate

                    if ( qentr * 1.00 > potentr ) qentr = potentr;
                        // constraining PHASE 1 entrainment rate by remaining entrainable PHASE 1 depth

                    aw[i][0] = aw[i][0] + qentr * 1.00; // PHASE 1 flow depth
                    aw[i][1] = aw[i][1] + qentr * 1.00 * aw[i][1] / aw[i][0]; // PHASE 1 x momentum
                    aw[i][2] = aw[i][2] + qentr * 1.00 * aw[i][1] / aw[i][0]; // PHASE 1 y momentum
                    aw[i][3] = aw[i][3] - qentr * 1.00; // change of basal topography
                }

                else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && sico.ENTRAINMENT != 0
                    && ( aw[i][0] > sico.HFLOWMIN || aw[i][3] > sico.HFLOWMIN )) {
                    // if entrainment is enabled with two-phase model:

                    if ( sico.CORRHEIGHT == 0 ) corrfact = 1; // correction factor for flow height to depth conversion
                    else corrfact = cos( betaxy[i] );

                    if ( phentrmax[i] + phentrmax2[i] > 0 ) rpotentr = phentrmax[i] / ( phentrmax[i] + phentrmax2[i] );
                    else rpotentr = 0; // PHASE 1 ratio

                    potentr = ( phentrmax[i] - rpotentr * ( pelev0[i] - pelev[i] )) * corrfact;
                        // potential PHASE 1 entrainment rate
                    potentr2 = ( phentrmax2[i] - ( 1 - rpotentr ) * ( pelev0[i] - pelev[i] )) * corrfact;
                        // potential PHASE 3 entrainment rate
                    if ( potentr < 0 ) potentr = 0; if ( potentr2 < 0 ) potentr2 = 0;

                    qentr = tlength * sflow.CENTR * rpotentr * hekin;
                    qentr2 = tlength * sflow.CENTR * ( 1 - rpotentr ) * hekin;

                    vbulkx = ( aw[i][1] * sflow.RHO1 + aw[i][4] * sflow.RHO2 ) / ( aw[i][0] * sflow.RHO1 + aw[i][3] * sflow.RHO2 );
                    vbulky = ( aw[i][2] * sflow.RHO1 + aw[i][5] * sflow.RHO2 ) / ( aw[i][0] * sflow.RHO1 + aw[i][3] * sflow.RHO2 );

                    if ( qentr * 1.00 > potentr ) qentr = potentr;
                        // constraining PHASE 1 entrainment rate by remaining entrainable PHASE 1 depth
                    if ( qentr2 * 1.00 > potentr2 ) qentr2 = potentr2;
                        // constraining PHASE 3 entrainment rate by remaining entrainable PHASE 3 depth

                    aw[i][0] = aw[i][0] + qentr * 1.00; // PHASE 1 flow depth
                    aw[i][1] = aw[i][1] + qentr * 1.00 * vbulkx; // PHASE 1 x momentum
                    aw[i][2] = aw[i][2] + qentr * 1.00 * vbulky; // PHASE 1 y momentum
                    aw[i][6] = aw[i][6] - qentr * 1.00; // change of basal topography (PHASE 1)

                    aw[i][3] = aw[i][3] + qentr2 * 1.00; // PHASE 3 flow depth
                    aw[i][4] = aw[i][4] + qentr2 * 1.00 * vbulkx; // PHASE 3 x momentum
                    aw[i][5] = aw[i][5] + qentr2 * 1.00 * vbulky; // PHASE 3 y momentum
                    aw[i][7] = aw[i][7] - qentr2 * 1.00; // change of basal topography (PHASE 3)
                }

                else if ( sico.MODEL == 7 && sico.ENTRAINMENT != 0
                    && ( aw[i][0] > sico.HFLOWMIN || aw[i][3] > sico.HFLOWMIN || aw[i][6] > sico.HFLOWMIN )) {
                    // if entrainment is enabled with three-phase model:

                    if ( sico.CORRHEIGHT == 0 ) corrfact = 1; // correction factor for flow height to depth conversion
                    else corrfact = cos( betaxy[i] );

                    if ( phentrmax[i] + phentrmax2[i] + phentrmax3[i] > 0 ) rpotentr = phentrmax[i] / ( phentrmax[i] + phentrmax2[i] + phentrmax3[i] );
                    else rpotentr = 0; // PHASE 1 ratio

                    if ( phentrmax[i] + phentrmax2[i] + phentrmax3[i] > 0 ) rpotentr3 = phentrmax3[i] / ( phentrmax[i] + phentrmax2[i] + phentrmax3[i] );
                    else rpotentr3 = 0; // PHASE 3 ratio

                    potentr = ( phentrmax[i] - rpotentr * ( pelev0[i] - pelev[i] )) * corrfact;
                        // potential PHASE 1 entrainment rate
                    potentr2 = ( phentrmax2[i] - ( 1 - rpotentr - rpotentr3 ) * ( pelev0[i] - pelev[i] )) * corrfact;
                        // potential PHASE 3 entrainment rate
                    potentr3 = ( phentrmax3[i] - rpotentr3 * ( pelev0[i] - pelev[i] )) * corrfact;
                        // potential PHASE 3 entrainment rate
                    if ( potentr < 0 ) potentr = 0; if ( potentr2 < 0 ) potentr2 = 0; if ( potentr3 < 0 ) potentr3 = 0;

                    qentr0 = tlength * sflow.CENTR * hekin;
                    qentr = rpotentr * qentr0;
                    qentr2 = ( 1 - rpotentr - rpotentr3 ) * qentr0;
                    qentr3 = rpotentr3 * qentr0;

                    vbulkx = (aw[i][1] * sflow.RHO1 + aw[i][4] * sflow.RHO2 + aw[i][7] * sflow.RHO3 )
                        / ( aw[i][0] * sflow.RHO1 + aw[i][3] * sflow.RHO2 + aw[i][6] * sflow.RHO3 );
                    vbulky = (aw[i][2] * sflow.RHO1 + aw[i][5] * sflow.RHO2 + aw[i][8] * sflow.RHO3 )
                        / ( aw[i][0] * sflow.RHO1 + aw[i][3] * sflow.RHO2 + aw[i][6] * sflow.RHO3 );

                    if ( qentr * 1.00 > potentr ) qentr = potentr;
                        // constraining PHASE 1 entrainment rate by remaining entrainable PHASE 1 depth
                    if ( qentr2 * 1.00 > potentr2 ) qentr2 = potentr2;
                        // constraining PHASE 3 entrainment rate by remaining entrainable PHASE 3 depth
                    if ( qentr3 * 1.00 > potentr3 ) qentr3 = potentr3;
                        // constraining PHASE 3 entrainment rate by remaining entrainable PHASE 3 depth

                    aw[i][0] = aw[i][0] + qentr * 1.00; // PHASE 1 flow depth
                    aw[i][1] = aw[i][1] + qentr * 1.00 * vbulkx; // PHASE 1 x momentum
                    aw[i][2] = aw[i][2] + qentr * 1.00 * vbulky; // PHASE 1 y momentum
                    aw[i][9] = aw[i][9] - qentr * 1.00; // change of basal topography (PHASE 1)

                    aw[i][3] = aw[i][3] + qentr2 * 1.00; // PHASE 3 flow depth
                    aw[i][4] = aw[i][4] + qentr2 * 1.00 * vbulkx; // PHASE 3 x momentum
                    aw[i][5] = aw[i][5] + qentr2 * 1.00 * vbulky; // PHASE 3 y momentum
                    aw[i][10] = aw[i][10] - qentr2 * 1.00; // change of basal topography (PHASE 3)

                    aw[i][6] = aw[i][6] + qentr3 * 1.00; // PHASE 3 flow depth
                    aw[i][7] = aw[i][7] + qentr3 * 1.00 * vbulkx; // PHASE 3 x momentum
                    aw[i][8] = aw[i][8] + qentr3 * 1.00 * vbulky; // PHASE 3 y momentum
                    aw[i][11] = aw[i][11] - qentr3 * 1.00; // change of basal topography (PHASE 3)
                }
            }


// F4 -- STOP --- Entrainment --------------------------------------------------------------------------------------


// F5 -- START -- Stopping of flow ---------------------------------------------------------------------------------


            // *** 1-3 and 6 still experimental, basal friction of PHASE 2 not considered, approaches 2 and 3 unstable
            // *** 1-3 Stopping criterion by energy balance
            // *** 1-3 Factor of mobility fom: time required for stopping / time step length
            // *** 1 = Global stopping: fom is computed for bulk mass, material is deposited and simulation is terminated
            // ***     when stopping occurs
            // *** 2 = Local stopping with immediate deposition: fom is computed for 3 x 3 neighbourhood,
            // ***     stopped material is deposited immediately and simulation continues
            // *** 3 = Local stopping with delayed deposition: fom is computed for 3 x 3 neighbourhood, stopped material
            // ***     is frozen (momenta = zero) and deposited only at the end of the simulation
            // *** 4 = Global stopping based on fraction of maximum kinetic energy, material is deposited and simulation is terminated
            // ***     when stopping occurs
            // *** 5 = Global stopping based on fraction of maximum momentum, material is deposited and simulation is terminated
            // ***     when stopping occurs
            // *** 6 = Deposition of material, simulation is terminated when all the material has been deposited


            if ( p == 1 && tsum > 0 && sico.STOPPING > 0 && sico.STOPPING < 4 ) { // for stopping mechanisms 1-3:

                if ( sico.STOPPING == 1 ) { // for global stopping:

                    mom = 0; momx = 0; momy = 0; mass = 0; betavg = 0; nalpha = 0;
                        // resetting variables for approach 1 (global stopping)

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];

                        if ( cflux[i] == 1 && fvalid( aw[i][0], aw[i][3], aw[i][6], sico ) == 1 ) {
                            // if pixel is flow pixel:

                            if ( sico.MODEL <= 3 ) { // for one-phase model:
                                momx += aw[i][1] * sflow.RHO1; // updating global x momentum
                                momy += aw[i][2] * sflow.RHO1; // updating global y momentum
                            }
                            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:
                                momx += aw[i][1] * sflow.RHO1 + aw[i][4] * sflow.RHO2; // updating global x momentum
                                momy += aw[i][2] * sflow.RHO1 + aw[i][5] * sflow.RHO2; // updating global y momentum
                            }
                            else if ( sico.MODEL == 7 ) { // for three-phase model:
                                momx += aw[i][1] * sflow.RHO1 + aw[i][4] * sflow.RHO2 + aw[i][7] * sflow.RHO3; // updating global x momentum
                                momy += aw[i][2] * sflow.RHO1 + aw[i][5] * sflow.RHO2 + aw[i][8] * sflow.RHO3; // updating global y momentum
                            }
                            nalpha += 1; // number of pixels
                        }
                    }
                    if ( nalpha > 0 ) { momx = momx / nalpha; momy = momy / nalpha; } // average x and y momenta
                    alphavg = falphav( momx, momy, sico ); // global direction of movement

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];

                        if ( cflux[i] == 1 && fvalid( aw[i][0], aw[i][3], aw[i][6], sico ) == 1 ) {

                            alpha = falpha( betax[i], betay[i], sico ); // aspect
                            betavg += tan( betaxy[i] * cos( alpha - alphavg )); // updating movement-following slope

                            if ( sico.MODEL <= 3 ) { // for one-phase flow:

                                alphav = falphav( aw[i][1] * sflow.RHO1, aw[i][2] * sflow.RHO1, sico ); // direction of movement
                                mom += pow( pow (aw[i][1], 2 ) + pow( aw[i][2], 2 ), 0.5 ) * cos( alphav - alphavg ) * sflow.RHO1;
                                    // updating topography-following momentum
                                mass += aw[i][0] * sflow.RHO1; // updating mass
                            }

                            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase flow:

                                alphav = falphav( aw[i][1] * sflow.RHO1 + aw[i][4] * sflow.RHO2,
                                    aw[i][2] * sflow.RHO1 + aw[i][5] * sflow.RHO2, sico ); // direction of movement

                                mom += pow( pow (aw[i][1], 2 ) + pow( aw[i][2], 2 ), 0.5 ) * cos( alphav - alphavg ) * sflow.RHO1
                                    + pow( pow (aw[i][4], 2 ) + pow( aw[i][5], 2 ), 0.5 ) * cos( alphav - alphavg ) * sflow.RHO2;
                                    // updating topography-following momentum
                                mass += aw[i][0] * sflow.RHO1 + aw[i][3] * sflow.RHO2; // updating mass
                            }

                            else if ( sico.MODEL == 7 ) { // for three-phase flow:

                                alphav = falphav( aw[i][1] * sflow.RHO1 + aw[i][4] * sflow.RHO2 + aw[i][7] * sflow.RHO3,
                                    aw[i][2] * sflow.RHO1 + aw[i][5] * sflow.RHO2 + aw[i][8] * sflow.RHO3, sico ); // direction of movement

                                mom += pow( pow (aw[i][1], 2 ) + pow( aw[i][2], 2 ), 0.5 ) * cos( alphav - alphavg ) * sflow.RHO1
                                    + pow( pow (aw[i][4], 2 ) + pow( aw[i][5], 2 ), 0.5 ) * cos( alphav - alphavg ) * sflow.RHO2
                                    + pow( pow (aw[i][7], 2 ) + pow( aw[i][8], 2 ), 0.5 ) * cos( alphav - alphavg ) * sflow.RHO3;
                                    // updating topography-following momentum
                                mass += aw[i][0] * sflow.RHO1 + aw[i][3] * sflow.RHO2; // updating mass
                            }

                            nalpha += 1; // number of pixels
                        }
                    }
                    if ( nalpha > 0 ) { mom = mom / nalpha; mass = mass / nalpha; betavg = atan( betavg / nalpha ); }
                        // averages
                    vflow = mom / mass; // topography-following bulk flow velocity
 
                }

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cflux[i] == 1 && fvalid( aw[i][0], aw[i][3], aw[i][6], sico ) == 1 ) {

                        if ( sico.DELTA == 1 ) {
                            if ( pdelta[i] != sico.UNDEF ) sflow.DELTA[0] = pdelta[i]; // basal friction angle
                            else sflow.DELTA[0] = sflow.DELTA0[0];
                        }
                        else sflow.DELTA[0] = sflow.DELTA0[0];

                        if ( sico.STOPPING == 1 ) imax = 1; else imax = 9;
                            // preparing loop over neighbourhood only for approaches 2 or 3 (local stopping)

                        if ( sico.STOPPING != 1 ) { // for local stopping:

                            mom = 0; momx = 0; momy = 0; mass = 0; betavg = 0; nalpha = 0;
                                // resetting variables for approach 1 (global stopping)

                            for ( j=0; j<imax; j++ ) { // loop over neighbourhood:

                                if ( cdomain[in[i][j]] > 1 && ( aw[in[i][j]][0] > sico.HFLOWMIN
                                    || ( sico.MODEL > 3 && aw[in[i][j]][3] > sico.HFLOWMIN )
                                    || ( sico.MODEL == 7 && aw[in[i][j]][6] > sico.HFLOWMIN ))) {
                                    // if pixel is flow pixel:

                                    if ( sico.MODEL <= 3 ) { // for one-phase flow:

                                        momx += aw[in[i][j]][1] * sflow.RHO1; // updating global x momentum
                                        momy += aw[in[i][j]][2] * sflow.RHO1; // updating global y momentum
                                    }
                                    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase flow:

                                        momx += aw[in[i][j]][1] * sflow.RHO1 + aw[in[i][j]][4] * sflow.RHO2; // updating global x momentum
                                        momy += aw[in[i][j]][2] * sflow.RHO1 + aw[in[i][j]][5] * sflow.RHO2; // updating global y momentum
                                    }
                                    else if ( sico.MODEL == 7 ) { // for three-phase flow:

                                        momx += aw[in[i][j]][1] * sflow.RHO1 + aw[in[i][j]][4] * sflow.RHO2 + aw[in[i][j]][7] * sflow.RHO3; // updating global x momentum
                                        momy += aw[in[i][j]][2] * sflow.RHO1 + aw[in[i][j]][5] * sflow.RHO2 + aw[in[i][j]][8] * sflow.RHO3; // updating global y momentum
                                    }
                                    nalpha += 1; // number of pixels
                                }
                            }
                            if ( nalpha > 0 ) { momx = momx / nalpha; momy = momy / nalpha; } // average x and y momenta
                            alphavg = falphav( momx, momy, sico ); // global direction of movement

                            for ( j=0; j<imax; j++ ) { // loop over neighbourhood:

                                if ( cdomain[in[i][j]] > 1 && ( aw[in[i][j]][0] > sico.HFLOWMIN
                                    || ( sico.MODEL > 3 && aw[in[i][j]][3] > sico.HFLOWMIN )
                                    || ( sico.MODEL == 7 && aw[in[i][j]][6] > sico.HFLOWMIN ))) {
                                    // if pixel is flow pixel:

                                    alpha = falpha( betax[in[i][j]], betay[in[i][j]], sico ); // aspect
                                    betavg += tan( betaxy[in[i][j]] * cos( alpha - alphavg )); // updating movement-following slope

                                    if ( sico.MODEL <= 3 ) { // for one-phase flow:

                                        alphav = falphav( aw[in[i][j]][1] * sflow.RHO1, aw[in[i][j]][2] * sflow.RHO1, sico );
                                            // direction of movement

                                        mom += pow( pow (aw[in[i][j]][1], 2 )
                                            + pow( aw[in[i][j]][2], 2 ), 0.5 ) * cos( alphav - alphavg ) * sflow.RHO1;
                                            // updating topography-following momentum

                                        mass += aw[in[i][j]][0] * sflow.RHO1; // updating mass
                                    }

                                    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase flow:

                                        alphav = falphav( aw[in[i][j]][1] * sflow.RHO1 + aw[in[i][j]][4] * sflow.RHO2,
                                            aw[in[i][j]][2] * sflow.RHO1 + aw[in[i][j]][5] * sflow.RHO2, sico ); // direction of movement

                                        mom += pow( pow (aw[in[i][j]][1], 2 ) + pow( aw[in[i][j]][2], 2 ), 0.5 )
                                            * cos( alphav - alphavg ) * sflow.RHO1
                                            + pow( pow (aw[in[i][j]][4], 2 ) + pow( aw[in[i][j]][5], 2 ), 0.5 )
                                            * cos( alphav - alphavg ) * sflow.RHO2;

                                            // updating topography-following momentum
                                        mass += aw[in[i][j]][0] * sflow.RHO1 + aw[in[i][j]][3] * sflow.RHO2; // updating mass
                                    }

                                    else if ( sico.MODEL == 7 ) { // for three-phase flow:

                                        alphav = falphav( aw[in[i][j]][1] * sflow.RHO1 + aw[in[i][j]][4] * sflow.RHO2 + aw[in[i][j]][7] * sflow.RHO3,
                                            aw[in[i][j]][2] * sflow.RHO1 + aw[in[i][j]][5] * sflow.RHO2 + aw[in[i][j]][8] * sflow.RHO3, sico ); // direction of movement

                                        mom += pow( pow (aw[in[i][j]][1], 2 ) + pow( aw[in[i][j]][2], 2 ), 0.5 )
                                            * cos( alphav - alphavg ) * sflow.RHO1
                                            + pow( pow (aw[in[i][j]][4], 2 ) + pow( aw[in[i][j]][5], 2 ), 0.5 )
                                            * cos( alphav - alphavg ) * sflow.RHO2
                                            + pow( pow (aw[in[i][j]][7], 2 ) + pow( aw[in[i][j]][8], 2 ), 0.5 )
                                            * cos( alphav - alphavg ) * sflow.RHO3;

                                            // updating topography-following momentum
                                        mass += aw[in[i][j]][0] * sflow.RHO1 + aw[in[i][j]][3] * sflow.RHO2 + aw[in[i][j]][6] * sflow.RHO3; // updating mass
                                    }

                                    nalpha += 1; // number of pixels
                                }
                            }
                            if ( nalpha > 0 ) { mom = mom / nalpha; mass = mass / nalpha; betavg = atan( betavg / nalpha ); }
                                // averages
                            vflow = mom / mass; // topography-following bulk flow velocity
                        }

                        if ( sico.STOPPING != 1 ) { // for approaches 2 or 3 (local stopping):

                            fom[i] = ( pow( 2, 0.5 ) - 1 ) * vflow
                                / ( sico.GRAVITY * cos( betavg ) * ( tan( sflow.DELTA[0] ) - tan( betavg )) * tlength ); // factor of mobility
                        }
                    }
                }

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( cflux[i] == 1 && fvalid ( aw[i][0], aw[i][3], aw[i][6], sico ) == 1 ) {

                        sflow.DELTA[0] = sflow.DELTA0[0];

                        if ( sico.STOPPING == 1 ) { // for global stopping:

                            fom[i] = ( pow( 2, 0.5 ) - 1 ) * vflow
                                / ( sico.GRAVITY * cos( betavg ) * ( tan( sflow.DELTA[0] ) - tan( betavg )) * tlength ); // factor of mobility

                            if ( mom > mompre ) fom[i] = 9999; // avoiding stopping of accelerating flows

                        }
                        if ( sico.STOPPING != 2 && fabs(fom[i]) <= 1 ) { // if factor of mobility is 1 or lower for approach 1 or 3:

                            for ( j=0; j<imax; j++ ) {

                                cstopped[in[i][j]] = 1; // updating control for stopping
                                aw[in[i][j]][1] = 0; aw[in[i][j]][2] = 0; // setting (PHASE 1) momenta to zero
                                if ( sico.MODEL > 3 ) aw[in[i][j]][4] = 0; aw[in[i][j]][5] = 0; // setting PHASE 2 momenta to zero
                                if ( sico.MODEL == 7 ) aw[in[i][j]][7] = 0; aw[in[i][j]][8] = 0; // setting PHASE 3 momenta to zero
                            }
                        }
                        else if ( sico.STOPPING == 2 && fabs(fom[i]) <= 1 ) {
                            // if factor of mobility is 1 or lower for approach 2:

                            for ( j=0; j<imax; j++ ) {

                                if ( sico.MODEL <= 3 ) { // for one-phase model:
                                    aw[in[i][j]][3] += aw[in[i][j]][0]; // updating deposited depth
                                    for ( k=0; k<3; k++ ) aw[i][k] = 0; // setting flow depths and momenta to zero
                                }
                                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:
                                    aw[in[i][j]][6] += aw[in[i][j]][0]; // updating deposited PHASE 1 depth
                                    aw[in[i][j]][7] += aw[in[i][j]][3]; // updating deposited PHASE 3 depth
                                    for ( k=0; k<6; k++ ) aw[i][k] = 0; // setting flow depths and momenta to zero
                                }
                                else if ( sico.MODEL == 7 ) { // for three-phase model:
                                    aw[in[i][j]][9] += aw[in[i][j]][0]; // updating deposited PHASE 1 depth
                                    aw[in[i][j]][10] += aw[in[i][j]][3]; // updating deposited PHASE 2 depth
                                    aw[in[i][j]][11] += aw[in[i][j]][6]; // updating deposited PHASE 3 depth
                                    for ( k=0; k<9; k++ ) aw[i][k] = 0; // setting flow depths and momenta to zero
                                }
                            }
                        }
                    }
                }

                mompre = mom;
            }

            if ( p == 1 && tsum > 0 && ( sico.STOPPING == 4 || sico.STOPPING == 5 )) { // application of stopping through fraction of maximum kinetic energy or momentum:

                hekin_sum = 0; // resetting sum of momentum or kinetic energy or momentum

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( aw[i][0] > 0 ) { vflowx = aw[i][1]/aw[i][0]; vflowy = aw[i][2]/aw[i][0]; } else { vflowx = 0; vflowy = 0; } // computing flow velocities
                    if ( sico.MODEL > 3 && aw[i][3] > 0 ) { vflowx2 = aw[i][4]/aw[i][3]; vflowy2 = aw[i][5]/aw[i][3]; } else { vflowx2 = 0; vflowy2 = 0; }
                    if ( sico.MODEL == 7 && aw[i][6] > 0 ) { vflowx3 = aw[i][7]/aw[i][6]; vflowy3 = aw[i][8]/aw[i][6]; } else { vflowx3 = 0; vflowy3 = 0; }

                    if ( sico.STOPPING == 4 ) {

                        hekin = 0.5 * aw[i][0] * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 )); // computing flow kinetic energies
                        if ( sico.MODEL > 3 ) hekin+= 0.5 * aw[i][3] * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ));
                        if ( sico.MODEL == 7 ) hekin += 0.5 * aw[i][6] * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ));

                    } else {

                        hekin = aw[i][0] * sflow.RHO1 * pow( pow( vflowx, 2 ) + pow( vflowy, 2 ), 0.5 ); // computing flow momenta
                        if ( sico.MODEL > 3 ) hekin+= aw[i][3] * sflow.RHO2 * pow( pow( vflowx2, 2 ) + pow( vflowy2, 2 ), 0.5 );
                        if ( sico.MODEL == 7 ) hekin += aw[i][6] * sflow.RHO3 * pow( pow( vflowx3, 2 ) + pow( vflowy3, 2 ), 0.5 );
                    }

                    hekin_sum += hekin; // updating sum of momentum or kinetic energy
                }

                if ( hekin_sum > hekin_max ) hekin_max = hekin_sum; // updating maximum momentum or kinetic energy
                else if ( hekin_sum < hekin_max * sflow.CDEP0 ) {

                    for ( ix=0; ix<ib[0]; ix++ ) {

                        i = ibasket[0][ix];

                        cstopped[i] = 1; // updating control for stopping
                        aw[i][1] = 0; aw[i][2] = 0; // setting (PHASE 1) momenta to zero
                        if ( sico.MODEL > 3 ) aw[i][4] = 0; aw[i][5] = 0; // setting PHASE 2 momenta to zero
                        if ( sico.MODEL == 7 ) aw[i][7] = 0; aw[i][8] = 0; // setting PHASE 3 momenta to zero
                    }
                }
            }

            if ( sico.STOPPING == 6 ) {

                hekin_thr = 200000;

                for ( ix=0; ix<ib[0]; ix++ ) {

                    i = ibasket[0][ix];

                    if ( aw[i][0] > 0 ) { vflowx = aw[i][1]/aw[i][0]; vflowy = aw[i][2]/aw[i][0]; } else { vflowx = 0; vflowy = 0; }
                    if ( sico.MODEL > 3 && aw[i][3] > 0 ) { vflowx2 = aw[i][4]/aw[i][3]; vflowy2 = aw[i][5]/aw[i][3]; } else { vflowx2 = 0; vflowy2 = 0; }
                    if ( sico.MODEL == 7 && aw[i][6] > 0 ) { vflowx3 = aw[i][7]/aw[i][6]; vflowy3 = aw[i][8]/aw[i][6]; } else { vflowx3 = 0; vflowy3 = 0; }

                    hekin = 0.5 * aw[i][0] * sflow.RHO1 * ( pow( vflowx, 2 ) + pow( vflowy, 2 ));
                    if ( sico.MODEL > 3 ) hekin+= 0.5 * aw[i][3] * sflow.RHO2 * ( pow( vflowx2, 2 ) + pow( vflowy2, 2 ));
                    if ( sico.MODEL == 7 ) hekin += 0.5 * aw[i][6] * sflow.RHO3 * ( pow( vflowx3, 2 ) + pow( vflowy3, 2 ));

                    if ( sico.STOPPING == 6 && sico.CDEP == 1 ) {
                        if ( pdep[i] != sico.UNDEF ) sflow.CDEP = pdep[i]; // deposition coefficient
                        else sflow.CDEP = sflow.CDEP0;
                    }
                    else sflow.CDEP = sflow.CDEP0;

                    if ( sico.MODEL <= 3 && aw[i][0] > sico.HFLOWMIN ) {
                        // if deposition is enabled with one-phase model:

                        potdep = aw[i][0];
                        qdep = tlength * sflow.CDEP * pow( fmax( 0, hekin_thr - hekin ) , 4 );
                        if ( qdep > potdep ) qdep = potdep;

                        aw[i][0] = aw[i][0] - qdep; // PHASE 1 flow depth
                        aw[i][1] = aw[i][1] - qdep * aw[i][1] / aw[i][0]; // PHASE 1 x momentum
                        aw[i][2] = aw[i][2] - qdep * aw[i][2] / aw[i][0]; // PHASE 1 y momentum
                        aw[i][3] = aw[i][3] + qdep; // change of basal topography
                    }

                    else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && ( aw[i][0] > sico.HFLOWMIN || aw[i][3] > sico.HFLOWMIN )) {
                        // if deposition is enabled with two-phase model:

                        potdep = aw[i][0]; // potential PHASE 1 deposition rate
                        potdep2 = aw[i][3]; // potential PHASE 2 deposition rate

                        qdep = tlength * sflow.CDEP * fmax( 0, hekin_thr - hekin );
                        qdep2 = tlength * sflow.CDEP * fmax( 0, hekin_thr - hekin );

                        if ( qdep > potdep ) qdep = potdep;
                        if ( qdep2 > potdep2 ) qdep2 = potdep2;

                        aw[i][0] = aw[i][0] - qdep; // PHASE 1 flow depth
                        aw[i][1] = aw[i][1] - qdep * vflowx; // PHASE 1 x momentum
                        aw[i][2] = aw[i][2] - qdep * vflowy; // PHASE 1 y momentum
                        aw[i][6] = aw[i][6] + qdep; // change of basal topography (PHASE 1)

                        aw[i][3] = aw[i][3] - qdep2; // PHASE 2 flow depth
                        aw[i][4] = aw[i][4] - qdep2 * vflowx2; // PHASE 3 x momentum
                        aw[i][5] = aw[i][5] - qdep2 * vflowy2; // PHASE 3 y momentum
                        aw[i][7] = aw[i][7] + qdep2; // chan ge of basal topography (PHASE 3)
                    }

                    else if ( sico.MODEL == 7 && ( aw[i][0] > sico.HFLOWMIN || aw[i][3] > sico.HFLOWMIN || aw[i][6] > sico.HFLOWMIN )) {
                        // if deposition is enabled with three-phase model:

                        potdep = aw[i][0];
                        potdep2 = aw[i][3];
                        potdep3 = aw[i][6];

                        qdep = tlength * sflow.CDEP * fmax( 0, hekin_thr - hekin );
                        qdep2 = tlength * sflow.CDEP * fmax( 0, hekin_thr - hekin );
                        qdep3 = tlength * sflow.CDEP * fmax( 0, hekin_thr - hekin );

                        if ( qdep > potdep ) qdep = potdep;
                        if ( qdep2 > potdep2 ) qdep2 = potdep2;
                        if ( qdep3 > potdep3 ) qdep3 = potdep3;

                        aw[i][0] = aw[i][0] - qdep; // PHASE 1 flow depth
                        aw[i][1] = aw[i][1] - qdep * vflowx; // PHASE 1 x momentum
                        aw[i][2] = aw[i][2] - qdep * vflowy; // PHASE 1 y momentum
                        aw[i][9] = aw[i][9] + qdep; // change of basal topography (PHASE 1)

                        aw[i][3] = aw[i][3] - qdep2; // PHASE 2 flow depth
                        aw[i][4] = aw[i][4] - qdep2 * vflowx2; // PHASE 3 x momentum
                        aw[i][5] = aw[i][5] - qdep2 * vflowy2; // PHASE 3 y momentum
                        aw[i][10] = aw[i][10] + qdep2; // change of basal topography (PHASE 2)

                        aw[i][6] = aw[i][6] - qdep3; // PHASE 3 flow depth
                        aw[i][7] = aw[i][7] - qdep3 * vflowx3; // PHASE 3 x momentum
                        aw[i][8] = aw[i][8] - qdep3 * vflowy3; // PHASE 3 y momentum
                        aw[i][11] = aw[i][11] + qdep3; // change of basal topography (PHASE 3)
                    }
                }
            }


// F5 -- STOP --- Stopping of flow ---------------------------------------------------------------------------------


        // Updating vectors of composite, maximum and cumulative values, and basal topography

            for ( ix=0; ix<ib[0]; ix++ ) {

                i = ibasket[0][ix];

                if ( sico.CORRHEIGHT == 0 ) corrfact = 1; // correction factor for depth to height conversion
                else corrfact = 1 / cos( betaxy[i] );

                if ( sico.MODEL <= 3 ) { // for one-phase model:

                    pelev[i] = pelev[i] + aw[i][3] * corrfact; // correcting elevation for entrainment and deposition

                    if ( aw[i][0] > sico.HFLOWMIN ) aw[i][4] = pow( pow( aw[i][1]/aw[i][0], 2 ) + pow( aw[i][2]/aw[i][0], 2 ) , 0.5 );
                        // updating flow velocity
                    else aw[i][4] = 0;
                    aw[i][5] = 0.5 * sflow.RHO1 * aw[i][0] * pow( aw[i][4], 2 ); // updating flow kinetic energy
                    aw[i][6] = ( 0.5 * sflow.RHO1 * pow( aw[i][4], 2 ) + sico.GRAVITY * sflow.RHO1 * aw[i][0] );
                        // updating flow pressure

                    if ( aw[i][0] > aw[i][7] ) aw[i][7] = aw[i][0]; //updating maximum flow depth at pixel
                    if ( aw[i][4] > aw[i][8] && aw[i][0] >= sico.IMPTHR[0] ) aw[i][8] = aw[i][4];
                        // updating maximum flow velocity at pixel
                    if ( aw[i][5] > aw[i][9] ) aw[i][9] = aw[i][5]; //updating maximum flow kinetic energy at pixel
                    if ( aw[i][6] > aw[i][10] ) aw[i][10] = aw[i][6]; // updating maximum flow pressure at pixel
                }
                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:

                    pelev[i] = pelev0[i] + ( aw[i][6] + aw[i][7] ) * corrfact; // correcting elevation for entrainment and deposition

                    if ( aw[i][0] > sico.HFLOWMIN ) aw[i][8] = pow( pow( aw[i][1]/aw[i][0], 2 ) + pow( aw[i][2]/aw[i][0], 2 ) , 0.5 );
                    else aw[i][8] = 0;
                    if ( aw[i][3] > sico.HFLOWMIN ) aw[i][9] = pow( pow( aw[i][4]/aw[i][3], 2 ) + pow( aw[i][5]/aw[i][3], 2 ) , 0.5 );
                    else aw[i][9] = 0; // updating flow velocities
                    aw[i][10] = aw[i][0] + aw[i][3]; // updating total flow depth
                    aw[i][11] = 0.5 * sflow.RHO1 * aw[i][0] * pow( aw[i][8], 2 );
                        // updating PHASE 1 flow kinetic energy
                    aw[i][12] = 0.5 * sflow.RHO2 * aw[i][3] * pow( aw[i][9], 2 );
                        // updating PHASE 2 flow kinetic energy
                    aw[i][13] = aw[i][11] + aw[i][12];
                        // updating total flow kinetic energy
                    if ( aw[i][0] > sico.HFLOWMIN ) aw[i][14] = aw[i][11] / aw[i][0] + sico.GRAVITY * sflow.RHO1 * aw[i][0];
                        // updating PHASE 1 flow pressure
                    else aw[i][14] = 0;
                    if ( aw[i][3] > sico.HFLOWMIN ) aw[i][15] = aw[i][12] / aw[i][3] + sico.GRAVITY * sflow.RHO2 * aw[i][3];
                        // updating PHASE 2 flow pressure
                    else aw[i][15] = 0;
                    if ( aw[i][10] > sico.HFLOWMIN ) aw[i][16] = aw[i][13] / aw[i][10]
                        + ( sico.GRAVITY * sflow.RHO1 * aw[i][0] + sico.GRAVITY * sflow.RHO2 * aw[i][3] ); // updating total flow pressure
                    else aw[i][16] = 0;
                    aw[i][17] = aw[i][6] + aw[i][7]; // updating total entrained or deposited depth

                    if ( aw[i][0] > aw[i][18] ) aw[i][18] = aw[i][0]; //updating maximum PHASE 1 flow depth at pixel
                    if ( aw[i][8] > aw[i][19] && aw[i][0] >= sico.IMPTHR[0] ) aw[i][19] = aw[i][8];
                        // updating maximum PHASE 1 flow velocity at pixel
                    if ( aw[i][3] > aw[i][20] ) aw[i][20] = aw[i][3]; //updating maximum PHASE 3 flow depth at pixel
                    if ( aw[i][9] > aw[i][21] && aw[i][3] >= sico.IMPTHR[0] ) aw[i][21] = aw[i][9];
                        // updating maximum PHASE 2 flow velocity at pixel
                    if ( aw[i][10] > aw[i][22] ) aw[i][22] = aw[i][10]; // updating maximum total flow depth at pixel
                    if ( aw[i][11] > aw[i][23] ) aw[i][23] = aw[i][11]; // updating maximum PHASE 1 flow kinetic energy at pixel
                    if ( aw[i][14] > aw[i][24] ) aw[i][24] = aw[i][14]; // updating maximum PHASE 1 flow pressure at pixel
                    if ( aw[i][12] > aw[i][25] ) aw[i][25] = aw[i][12]; // updating maximum PHASE 3 flow kinetic energy at pixel
                    if ( aw[i][15] > aw[i][26] ) aw[i][26] = aw[i][15]; // updating maximum PHASE 3 flow pressure at pixel
                    if ( aw[i][13] > aw[i][27] ) aw[i][27] = aw[i][13]; // updating maximum total flow kinetic energy at pixel
                    if ( aw[i][16] > aw[i][28] ) aw[i][28] = aw[i][16]; // updating maximum total flow pressure at pixel
                }
                else if ( sico.MODEL == 7 ) { // for three-phase model:

                    pelev[i] = pelev0[i] + ( aw[i][9] + aw[i][10] + aw[i][11] ) * corrfact; // correcting elevation for entrainment and deposition

                    if ( aw[i][0] > sico.HFLOWMIN ) aw[i][12] = pow( pow( aw[i][1]/aw[i][0], 2 ) + pow( aw[i][2]/aw[i][0], 2 ) , 0.5 );
                    else aw[i][12] = 0;
                    if ( aw[i][3] > sico.HFLOWMIN ) aw[i][13] = pow( pow( aw[i][4]/aw[i][3], 2 ) + pow( aw[i][5]/aw[i][3], 2 ) , 0.5 );
                    else aw[i][13] = 0;
                    if ( aw[i][6] > sico.HFLOWMIN ) aw[i][14] = pow( pow( aw[i][7]/aw[i][6], 2 ) + pow( aw[i][8]/aw[i][6], 2 ) , 0.5 );
                    else aw[i][14] = 0; // updating flow velocities
                    aw[i][15] = aw[i][0] + aw[i][3] + aw[i][6]; // updating total flow depth
                    aw[i][16] = 0.5 * sflow.RHO1 * aw[i][0] * pow( aw[i][12], 2 ); // updating PHASE 1 flow kinetic energy
                    aw[i][17] = 0.5 * sflow.RHO2 * aw[i][3] * pow( aw[i][13], 2 ); // updating PHASE 2 flow kinetic energy
                    aw[i][18] = 0.5 * sflow.RHO3 * aw[i][6] * pow( aw[i][14], 2 ); // updating PHASE 3 flow kinetic energy
                    aw[i][19] = aw[i][16] + aw[i][17] + aw[i][18]; // updating total flow kinetic energy
                    if ( aw[i][0] > sico.HFLOWMIN ) aw[i][20] = aw[i][16] / aw[i][0] + sico.GRAVITY * sflow.RHO1 * aw[i][0];
                    else aw[i][20] = 0; // updating PHASE 1 flow pressure
                    if ( aw[i][3] > sico.HFLOWMIN ) aw[i][21] = aw[i][17] / aw[i][3] + sico.GRAVITY * sflow.RHO2 * aw[i][3];
                    else aw[i][21] = 0; // updating PHASE 2 flow pressure
                    if ( aw[i][6] > sico.HFLOWMIN ) aw[i][22] = aw[i][18] / aw[i][6] + sico.GRAVITY * sflow.RHO3 * aw[i][6];
                    else aw[i][22] = 0; // updating PHASE 3 flow pressure
                    if ( aw[i][15] > sico.HFLOWMIN ) aw[i][23] = aw[i][19] / aw[i][15]
                        + ( sico.GRAVITY * sflow.RHO1 * aw[i][0] + sico.GRAVITY * sflow.RHO2 * aw[i][3] + + sico.GRAVITY * sflow.RHO3 * aw[i][6] );
                    else aw[i][23] = 0; // updating total flow pressure
                    aw[i][24] = aw[i][9] + aw[i][10] + aw[i][11]; // updating total entrained or deposited depth

                    if ( aw[i][0] > aw[i][25] ) aw[i][25] = aw[i][0]; //updating maximum PHASE 1 flow depth at pixel
                    if ( aw[i][12] > aw[i][26] && aw[i][0] >= sico.IMPTHR[0] ) aw[i][26] = aw[i][12];
                        // updating maximum PHASE 1 flow velocity at pixel
                    if ( aw[i][3] > aw[i][27] ) aw[i][27] = aw[i][3]; //updating maximum PHASE 3 flow depth at pixel
                    if ( aw[i][13] > aw[i][28] && aw[i][3] >= sico.IMPTHR[0] ) aw[i][28] = aw[i][13];
                        // updating maximum PHASE 2 flow velocity at pixel
                    if ( aw[i][6] > aw[i][29] ) aw[i][29] = aw[i][6]; //updating maximum PHASE 3 flow depth at pixel
                    if ( aw[i][14] > aw[i][30] && aw[i][6] >= sico.IMPTHR[0] ) aw[i][30] = aw[i][14];
                        // updating maximum PHASE 3 flow velocity at pixel
                    if ( aw[i][15] > aw[i][31] ) aw[i][31] = aw[i][15]; // updating maximum total flow depth at pixel
                    if ( aw[i][16] > aw[i][32] ) aw[i][32] = aw[i][16]; // updating maximum PHASE 1 flow kinetic energy at pixel
                    if ( aw[i][20] > aw[i][33] ) aw[i][33] = aw[i][20]; // updating maximum PHASE 1 flow pressure at pixel
                    if ( aw[i][17] > aw[i][34] ) aw[i][34] = aw[i][17]; // updating maximum PHASE 2 flow kinetic energy at pixel
                    if ( aw[i][21] > aw[i][35] ) aw[i][35] = aw[i][21]; // updating maximum PHASE 2 flow pressure at pixel
                    if ( aw[i][18] > aw[i][36] ) aw[i][36] = aw[i][18]; // updating maximum PHASE 3 flow kinetic energy at pixel
                    if ( aw[i][22] > aw[i][37] ) aw[i][37] = aw[i][22]; // updating maximum PHASE 3 flow pressure at pixel
                    if ( aw[i][19] > aw[i][38] ) aw[i][38] = aw[i][19]; // updating maximum total flow kinetic energy at pixel
                    if ( aw[i][23] > aw[i][39] ) aw[i][39] = aw[i][23]; // updating maximum total flow pressure at pixel
                }
            }


// *** End of loop over two steps, each moving the system half of a pixel (NOC scheme) -----------------------------


        }

        for ( ix=0; ix<ib[0]; ix++ ) {

            i = ibasket[0][ix];

            betax[i] = fbeta( pelev[in[i][5]], pelev[in[i][2]], 2.0, sico ); // slopes
            betay[i] = fbeta( pelev[in[i][4]], pelev[in[i][1]], 2.0, sico );
            betaxy[i] = fbetaxy( betax[i], betay[i] );

            if ( sico.CORRHEIGHT != 0 ) {

                dx[i] = sico.CSZ / cos( betax[i] ); // topography-following pixel spacing
                dy[i] = sico.CSZ / cos( betay[i] );
            }
        }


// ----- START -- Updating maximum values of flow parameters and volumes -------------------------------------------


        hflow_max = 0; // resetting maximum (PHASE 1) flow depth
        vflow_max = 0; // resetting maximum (PHASE 1) flow velocity
        vol_flow = 0; // resetting (PHASE 1) flow volume
        vol_edge = 0; // resetting (PHASE 1) lost flow volume
        vol_entr = 0; // resetting (PHASE 1) entrained volume
        vol_noflux = 0; // resetting (PHASE 1) volume without fluxes
        ekin_flow = 0; // resetting total kinetic energy of flow


        if ( sico.MODEL > 3 ) { // for two-phase model:

            hflow_max2 = 0; // resetting maximum PHASE 2 flow depth
            vflow_max2 = 0; // resetting maximum PHASE 2 velocity
            vol_flow2 = 0; // resetting PHASE 2 flow volume
            vol_edge2 = 0; // resetting PHASE 2 lost flow volume
            vol_entr2 = 0; // resetting PHASE 2 entrained volume
            vol_noflux2 = 0; // resetting PHASE 2 volume without fluxes
        }
        if ( sico.MODEL == 7 ) { // for three-phase model:

            hflow_max3 = 0; // resetting maximum PHASE 3 flow depth
            vflow_max3 = 0; // resetting maximum PHASE 3 velocity
            vol_flow3 = 0; // resetting PHASE 3 flow volume
            //vol_edge3 = 0; // resetting PHASE 3 lost flow volume
            vol_entr3 = 0; // resetting PHASE 3 entrained volume
            vol_noflux3 = 0; // resetting PHASE 3 volume without fluxes
        }

        for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels:

            if ( cdomain2[i] == 0 ) {

                if ( sico.CORRHEIGHT == 0 ) carea = pow ( sico.CSZ, 2 );
                else carea = pow ( sico.CSZ, 2 ) * pow( 1 - pow( sin( betax[i] ) , 2 ) * pow ( sin( betay[i] ) , 2 ) , 0.5 )
                    / ( cos( betax[i] ) * cos( betay[i] ) ); // topography-following area of pixel

                if ( aw[i][0] > hflow_max ) hflow_max = aw[i][0]; // maximum (PHASE 1) flow depth

                if ( cstopped[i] == 0 && cflux[i] == 1 && aw[i][0] > sico.HFLOWMIN ) vol_flow += aw[i][0] * carea; // (PHASE 1) flow volume

                if (( cflux[i] == 0 || cstopped[i] == 1 ) && aw[i][0] > sico.HFLOWMIN )
                    vol_noflux += aw[i][0] * carea; // (PHASE 1) volume without fluxes

                if ( sico.MODEL <= 3 ) { // for one-phase model:

                    if ( hflow_max > hflow_maxmax ) hflow_maxmax = hflow_max; // absolute maximum flow depth
                    if ( aw[i][0] >= sico.IMPTHR[0] && aw[i][4] > vflow_max ) vflow_max = aw[i][4]; // maximum flow velocity

                    if ( cdomain[i] != 0 && ( sico.ENTRAINMENT != 0 || sico.STOPPING > 0 )) vol_entr += aw[i][4] * carea;
                        // entrained or deposited volume

                    ekin_flow += aw[i][5] * carea; // kinetic energy of flow
                }

                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:

                    if ( aw[i][3] > hflow_max2 ) hflow_max2 = aw[i][3];
                    if ( hflow_max + hflow_max2 > hflow_maxmax ) hflow_maxmax = hflow_max + hflow_max2; // absolute maximum flow depth

                    if ( cdomain[i] != 0 && aw[i][0] >= sico.IMPTHR[0] && aw[i][8] > vflow_max ) vflow_max = aw[i][8];
                    if ( cdomain[i] != 0 && aw[i][3] >= sico.IMPTHR[0] && aw[i][9] > vflow_max2 ) vflow_max2 = aw[i][9];
                        // maximum PHASE 1 and PHASE 2 velocities

                    if ( cstopped[i] == 0 && cflux[i] == 1 && aw[i][3] > sico.HFLOWMIN ) vol_flow2 += aw[i][3] * carea;
                        // PHASE 2 flow volume

                    if (( cflux[i] == 0 || cstopped[i] == 1 ) && aw[i][3] > sico.HFLOWMIN )
                        vol_noflux2 += aw[i][3] * carea; // PHASE 2 volume without fluxes

                    if ( cdomain[i] != 0 && ( sico.ENTRAINMENT != 0 || sico.STOPPING > 0 )) vol_entr += aw[i][6] * carea;
                    if ( cdomain[i] != 0 && ( sico.ENTRAINMENT != 0 || sico.STOPPING > 0 )) vol_entr2 += aw[i][7] * carea;
                        // entrained or deposited volumes

                    ekin_flow += aw[i][13] * carea; // total flow kinetic energy
                }

                else if ( sico.MODEL == 7 ) { // for three-phase model:

                    if ( aw[i][3] > hflow_max2 ) hflow_max2 = aw[i][3];
                    if ( aw[i][6] > hflow_max3 ) hflow_max3 = aw[i][6];
                    if ( hflow_max + hflow_max2 + hflow_max3 > hflow_maxmax ) hflow_maxmax = hflow_max + hflow_max2 + hflow_max3; // absolute maximum flow depth

                    if ( cdomain[i] != 0 && aw[i][0] >= sico.IMPTHR[0] && aw[i][12] > vflow_max ) vflow_max = aw[i][12];
                    if ( cdomain[i] != 0 && aw[i][3] >= sico.IMPTHR[0] && aw[i][13] > vflow_max2 ) vflow_max2 = aw[i][13];
                    if ( cdomain[i] != 0 && aw[i][6] >= sico.IMPTHR[0] && aw[i][14] > vflow_max3 ) vflow_max3 = aw[i][14];
                        // maximum PHASE 1, PHASE 2, and PHASE 3 velocities

                    if ( cstopped[i] == 0 && cflux[i] == 1 && aw[i][3] > sico.HFLOWMIN ) vol_flow2 += aw[i][3] * carea;
                        // PHASE 2 flow volume
                    if ( cstopped[i] == 0 && cflux[i] == 1 && aw[i][6] > sico.HFLOWMIN ) vol_flow3 += aw[i][6] * carea;
                        // PHASE 3 flow volume

                    if (( cflux[i] == 0 || cstopped[i] == 1 ) && aw[i][3] > sico.HFLOWMIN )
                        vol_noflux2 += aw[i][3] * carea; // PHASE 2 volume without fluxes

                    if (( cflux[i] == 0 || cstopped[i] == 1 ) && aw[i][6] > sico.HFLOWMIN )
                        vol_noflux3 += aw[i][6] * carea; // PHASE 3 volume without fluxes

                    if ( cdomain[i] != 0 && ( sico.ENTRAINMENT != 0 || sico.STOPPING > 0 )) vol_entr += aw[i][9] * carea;
                    if ( cdomain[i] != 0 && ( sico.ENTRAINMENT != 0 || sico.STOPPING > 0 )) vol_entr2 += aw[i][10] * carea;
                    if ( cdomain[i] != 0 && ( sico.ENTRAINMENT != 0 || sico.STOPPING > 0 )) vol_entr3 += aw[i][11] * carea;
                        // entrained or deposited volumes

                    ekin_flow += aw[i][19] * carea; // total flow kinetic energy
                }
            }

            if ( cdomain2[i] == 1 && aw[i][0] > sico.HFLOWMIN ) vol_edge += aw[i][0] * carea;
                // (PHASE 1) flow volume leaving area of interest

            if ( sico.MODEL > 3 && cdomain2[i] == 1 && aw[i][3] > sico.HFLOWMIN ) vol_edge2 += aw[i][3] * carea;
                // PHASE 2  flow volume leaving area of interest

            if ( sico.MODEL == 7 && cdomain2[i] == 1 && aw[i][6] > sico.HFLOWMIN ) vol_edge2 += aw[i][6] * carea;
                // PHASE 3  flow volume leaving area of interest
        }


// ----- STOP --- Updating maximum values of flow parameters and volumes -------------------------------------------


    // Indicating stopping of the flow (the entire flow has stopped moving or left the area of interest, termination of the simulation)

        if (( sico.MODEL <= 3 && vol_flow == 0 ) || (( sico.MODEL > 3 && sico.MODEL < 7 ) && vol_flow == 0 && vol_flow2 == 0 )
            || ( sico.MODEL == 7 && vol_flow == 0 && vol_flow2 == 0 && vol_flow3 == 0 )) {

            printf("\n                       \n                       ------------------ FLOW STOPPED ------------------\n\n");
            fflush(stdout); // forcing immediate display
            fprintf(f_summary, "\n                       \n                       ------------------ FLOW STOPPED ------------------\n\n");
            ccontinue = 0; // setting control for continuation to negative
        }


    // Indicating numerical failure (termination of the simulation)

        if ( ccontinue == 0 && csuccess == 0 ) { // stopping simulation if time step length indicates numerical failure:

            printf("\n                       \n                       --------------- NUMERICAL FAILURE ----------------\n\n");
            fflush(stdout); // forcing immediate display
            fprintf(f_summary, "\n                       \n                       --------------- NUMERICAL FAILURE ----------------\n\n");
        }


    //!!!CHECK deactivated: indicating overflow (part of the flow leaving the area of interest, simulation continues)

        /*if (( vol_edge > 0 || ( sico.MODEL > 3 && vol_edge2 > 0 ) || ( sico.MODEL == 7 && vol_edge3 > 0 )) && coverflow == 0 ) {
            printf("\n                       \n                       -------------------- OVERFLOW --------------------\n\n");
            fflush(stdout); // forcing immediate display
            fprintf(f_summary, "\n                       \n                       -------------------- OVERFLOW --------------------\n\n");
            coverflow = 1; // setting control for overflow to positive
        }*/


    // Correcting deposited depth and basal topography for stopping, if necessary

        if (( sico.STOPPING == 1 || sico.STOPPING == 3 || sico.STOPPING == 4 || sico.STOPPING == 5 ) && ccontinue == 0 && csuccess == 1 ) {

            for ( ix=0; ix<ib[0]; ix++ ) {

                i = ibasket[0][ix];

                if ( sico.CORRHEIGHT == 0 ) corrfact = 1; // correction factor for depth to height conversion
                else corrfact = 1 / cos( betaxy[i] );

                if ( sico.MODEL <= 3 ) { // for one-phase model:

                    aw[i][3] += cstopped[i] * aw[i][0]; // updating entrained or deposited depth
                    pelev[i] += cstopped[i] * aw[i][0] * corrfact; // correcting elevation for stopping
                    aw[i][0] -= cstopped[i] * aw[i][0]; // updating flow depth
                }
                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:

                    aw[i][6] += cstopped[i] * aw[i][0]; // updating entrained or deposited PHASE 1 depth
                    aw[i][7] += cstopped[i] * aw[i][3]; // updating entrained or deposited PHASE 2 depth
                    aw[i][17] = aw[i][6] + aw[i][7]; // updating total entrained or deposited depth
                    pelev[i] += ( cstopped[i] * aw[i][0] + cstopped[i] * aw[i][3] ) * corrfact;
                        // correcting elevation for stopping
                    aw[i][0] -= cstopped[i] * aw[i][0]; // updating PHASE 1 flow depth
                    aw[i][3] -= cstopped[i] * aw[i][3]; // updating PHASE 2 flow depth
                }
                else if ( sico.MODEL == 7 ) { // for three-phase model:

                    aw[i][9] += cstopped[i] * aw[i][0]; // updating entrained or deposited PHASE 1 depth
                    aw[i][10] += cstopped[i] * aw[i][3]; // updating entrained or deposited PHASE 2 depth
                    aw[i][11] += cstopped[i] * aw[i][6]; // updating entrained or deposited PHASE 3 depth
                    aw[i][17] = aw[i][9] + aw[i][10] + aw[i][11]; // updating total entrained or deposited depth
                    pelev[i] += ( cstopped[i] * aw[i][0] + cstopped[i] * aw[i][3] + cstopped[i] * aw[i][6] ) * corrfact;
                        // correcting elevation for stopping
                    aw[i][0] -= cstopped[i] * aw[i][0]; // updating PHASE 1 flow depth
                    aw[i][3] -= cstopped[i] * aw[i][3]; // updating PHASE 2 flow depth
                    aw[i][6] -= cstopped[i] * aw[i][6]; // updating PHASE 3 flow depth
                }
            }
        }

        if ( tint >= tout || ccontinue == 0 ) { // if defined interval or last time step is reached:


// ----- START -- Writing hydrograph infos to files ----------------------------------------------------------------


            if ( sico.MULT == 0 && hydrograph == 1 ) { //!!!CHECK

                ctrl_hydout = 1; // control for hydrograph output

                for ( hydj = hydnin; hydj < hydnin + hydnout; hydj++ ) {  // loop over all output hydrographs:

                    hydh = aw[hydi[hydj]][0]  / cos( betaxy[hydi[hydj]] ); // (PHASE 1) flow height
                    if ( hydh > sico.HFLOWMIN )
                        hydv = pow( pow( aw[hydi[hydj]][1], 2 ) + pow( aw[hydi[hydj]][2], 2 ), 0.5 ) / aw[hydi[hydj]][0];
                    else hydv= 0; // (PHASE 1) flow velocity

                    if ( sico.MODEL <= 3 ) {

                        hyde = aw[hydi[hydj]][3]; // entrained height
                        hydh2 = 0; hydv2 = 0; hyde2 = 0; hydh3 = 0; hydv3 = 0; hyde3 = 0;
                    }
                    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                        hydh2 = aw[hydi[hydj]][3] / cos( betaxy[hydi[hydj]] ); // PHASE 3 flow height
                        if ( hydh2 > sico.HFLOWMIN )
                            hydv2 = pow( pow( aw[hydi[hydj]][4], 2 ) + pow( aw[hydi[hydj]][5], 2 ), 0.5 ) / aw[hydi[hydj]][3];

                        else hydv2= 0; // PHASE 3 flow velocity
                        hyde = aw[hydi[hydj]][6]; // entrained PHASE 1 height
                        hyde2 = aw[hydi[hydj]][7]; // entrained PHASE 3 height
                        hydh3 = 0; hydv3 = 0; hyde3 = 0;
                    }
                    else if ( sico.MODEL == 7 ) {

                        hydh2 = aw[hydi[hydj]][3] / cos( betaxy[hydi[hydj]] ); // PHASE 3 flow height
                        hydh3 = aw[hydi[hydj]][6] / cos( betaxy[hydi[hydj]] ); // PHASE 3 flow height
                        if ( hydh2 > sico.HFLOWMIN )
                            hydv2 = pow( pow( aw[hydi[hydj]][4], 2 ) + pow( aw[hydi[hydj]][5], 2 ), 0.5 ) / aw[hydi[hydj]][3];
                        else hydv2= 0; // PHASE 3 flow velocity
                        if ( hydh3 > sico.HFLOWMIN )
                            hydv3 = pow( pow( aw[hydi[hydj]][7], 2 ) + pow( aw[hydi[hydj]][8], 2 ), 0.5 ) / aw[hydi[hydj]][6];
                        else hydv3= 0; // PHASE 3 flow velocity
                        hyde = aw[hydi[hydj]][9]; // entrained PHASE 1 height
                        hyde2 = aw[hydi[hydj]][10]; // entrained PHASE 3 height
                        hyde3 = aw[hydi[hydj]][11]; // entrained PHASE 3 height
                    }

                    hydq = 0; // resetting (PHASE 1) discharge
                    if ( sico.MODEL > 3 ) hydq2 = 0; // resetting PHASE 3 discharge
                    if ( sico.MODEL == 7 ) hydq3 = 0; // resetting PHASE 3 discharge

                    if ( hydalpha[hydj] == 0 || hydalpha[hydj] == sico.PI / 2 || hydalpha[hydj] == sico.PI
                        || hydalpha[hydj] == 3 * sico.PI / 2 ) hydfalpha = 1;
                    else if ( fabs( 1 / sin( hydalpha[hydj] )) < fabs( 1 / cos( hydalpha[hydj] )))
                        hydfalpha = fabs( 1 / sin( hydalpha[hydj] ));
                    else hydfalpha = fabs( 1 / cos( hydalpha[hydj] )); // correction factor for profile direction

                    for ( hydk = 1; hydk <= hydp[0][hydj]; hydk++ ) { // loop over all pixels of profile:

                        alpha = falpha( betax[hydp[hydk][hydj]], betay[hydp[hydk][hydj]], sico ); // aspect
                        hydbeta = atan ( tan( betaxy[hydp[hydk][hydj]] ) * cos ( alpha - hydalpha[hydj] + sico.PI / 2 )); // corrected slope

                        hydfcorr = sico.CSZ * hydfalpha / cos( hydbeta ); // reference length for discharge

                        hydm0 = pow( pow(aw[hydp[hydk][hydj]][1], 2 ) + pow( aw[hydp[hydk][hydj]][2], 2), 0.5 );

                        if ( aw[hydp[hydk][hydj]][0] > sico.HFLOWMIN && hydm0 > 0 ) {

                            hydmx = aw[hydp[hydk][hydj]][1];
                            hydmy = aw[hydp[hydk][hydj]][2];

                            if ( hydmy > 0 ) hydmalpha = acos( hydmx / hydm0 ); else hydmalpha = 2 * sico.PI - acos( hydmx / hydm0 );
                            hydnalpha = hydalpha[hydj] - sico.PI / 2; if ( hydnalpha < 0 ) hydnalpha += 2 * sico.PI;
                            hydm = hydm0 * cos( hydnalpha - hydmalpha ); // momentum in direction perpendicular to profile
                            hydq += hydm * hydfcorr; // updating (PHASE 1) discharge
                        }

                        if ( sico.MODEL > 3 ) {

                            hydm0 = pow( pow(aw[hydp[hydk][hydj]][4], 2 ) + pow( aw[hydp[hydk][hydj]][5], 2), 0.5 );

                            if ( aw[hydp[hydk][hydj]][3] > sico.HFLOWMIN && hydm0 > 0 ) {

                                hydmx = aw[hydp[hydk][hydj]][4];
                                hydmy = aw[hydp[hydk][hydj]][5];

                                if ( hydmy > 0 ) hydmalpha = acos( hydmx / hydm0 ); else hydmalpha = 2 * sico.PI - acos( hydmx / hydm0 );
                                hydnalpha = hydalpha[hydj] - sico.PI / 2; if ( hydnalpha < 0 ) hydnalpha += 2 * sico.PI;
                                hydm = hydm0 * cos ( hydnalpha - hydmalpha ); // momentum in direction perpendicular to profile
                                hydq2 += hydm * hydfcorr; // updating PHASE 2 discharge
                            }
                        }
                        else hydq2 = 0;

                        if ( sico.MODEL == 7 ) {

                            hydm0 = pow( pow(aw[hydp[hydk][hydj]][7], 2 ) + pow( aw[hydp[hydk][hydj]][8], 2), 0.5 );

                            if ( aw[hydp[hydk][hydj]][6] > sico.HFLOWMIN && hydm0 > 0 ) {

                                hydmx = aw[hydp[hydk][hydj]][7];
                                hydmy = aw[hydp[hydk][hydj]][8];

                                if ( hydmy > 0 ) hydmalpha = acos( hydmx / hydm0 ); else hydmalpha = 2 * sico.PI - acos( hydmx / hydm0 );
                                hydnalpha = hydalpha[hydj] - sico.PI / 2; if ( hydnalpha < 0 ) hydnalpha += 2 * sico.PI;
                                hydm = hydm0 * cos ( hydnalpha - hydmalpha ); // momentum in direction perpendicular to profile
                                hydq3 += hydm * hydfcorr; // updating PHASE 3 discharge
                            }
                        }
                        else hydq3 = 0;
                    }

                    fprintf(f_hydinfo[hydj], "%.1f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\n",
                        tsum, hydh, hydv, hyde, hydq, hydh2, hydv2, hyde2, hydq2, hydh3, hydv3, hyde3, hydq3);
                        // writing hydrograph info to file

                }
            }


// ----- STOP --- Writing hydrograph infos to files ----------------------------------------------------------------


// ----- START -- Writing output files and raster maps -------------------------------------------------------------


            #ifdef WITHGRASS


            // Writing output (if GRASS is used)

                for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels:


                // Converting depths into heights

                    if ( cdomain[i] != 0 ) {

                        if ( sico.MODEL <= 3 ) {

                            hflowi = fconvout( i, aw, hflow_max, 1, 1,
                                px, py, pelev, betax, betay, betaxy, sico );
                            hentri = fconvout( i, aw, hflow_max, 1, 2,
                                px, py, pelev, betax, betay, betaxy, sico );
                        }
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {

                            hflowi = fconvout( i, aw, hflow_max+hflow_max2, 1, 1,
                                px, py, pelev, betax, betay, betaxy, sico );
                            hflowi2 = fconvout( i, aw, hflow_max+hflow_max2, 2, 1,
                                px, py, pelev, betax, betay, betaxy, sico );

                            hentri = fconvout( i, aw, hflow_max+hflow_max2, 1, 2,
                                px, py, pelev, betax, betay, betaxy, sico );
                            hentri2 = fconvout( i, aw, hflow_max+hflow_max2, 2, 2,
                                px, py, pelev, betax, betay, betaxy, sico );
                        }
                        else if ( sico.MODEL == 7 ) {

                            hflowi = fconvout( i, aw, hflow_max+hflow_max2+hflow_max3, 1, 1,
                                px, py, pelev, betax, betay, betaxy, sico );
                            hflowi2 = fconvout( i, aw, hflow_max+hflow_max2+hflow_max3, 2, 1,
                                px, py, pelev, betax, betay, betaxy, sico );
                            hflowi3 = fconvout( i, aw, hflow_max+hflow_max2+hflow_max3, 3, 1,
                                px, py, pelev, betax, betay, betaxy, sico );

                            hentri = fconvout( i, aw, hflow_max+hflow_max2+hflow_max3, 1, 2,
                                px, py, pelev, betax, betay, betaxy, sico );
                            hentri2 = fconvout( i, aw, hflow_max+hflow_max2+hflow_max3, 2, 2,
                                px, py, pelev, betax, betay, betaxy, sico );
                            hentri3 = fconvout( i, aw, hflow_max+hflow_max2+hflow_max3, 3, 2,
                                px, py, pelev, betax, betay, betaxy, sico );
                        }
                    }
                    else {
                        hflowi = 0; hentri = 0;
                        if ( sico.MODEL > 3 ) {
                            hflowi2 = 0; hentri2 = 0;
                        }
                        if ( sico.MODEL == 7 ) {
                            hflowi3 = 0; hentri3 = 0;
                        }
                    }


                // Writing output parameters to segment files

                    v[0] = hflowi; // (PHASE 1) flow height
                    if ( aw[i][0] > sico.HFLOWMIN ) v[1] = aw[i][1] / aw[i][0]; else v[1] = 0; // (PHASE 1) velocity in x direction
                    if ( aw[i][0] > sico.HFLOWMIN ) v[2] = aw[i][2] / aw[i][0]; else v[2] = 0; // (PHASE 1) velocity in y direction

                    if ( sico.MODEL <= 3 ) { // for one-phase model:

                        v[3] = hentri; // entrained height
                        v[4] = aw[i][4]; // velocity
                        v[5] = aw[i][5]; // flow kinetic energy
                        v[6] = aw[i][6]; // flow pressure
                    }
                    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:

                        v[3] = hflowi2; // PHASE 2 flow height
                        if ( aw[i][3] > sico.HFLOWMIN ) v[4] = aw[i][4] / aw[i][3]; else v[4] = 0; // PHASE 2 velocity in x direction
                        if ( aw[i][3] > sico.HFLOWMIN ) v[5] = aw[i][5] / aw[i][3]; else v[5] = 0; // PHASE 2 velocity in y direction
                        v[6] = hentri; // entrained or deposited PHASE 1 height
                        v[7] = hentri2; // entrained or deposited PHASE 2 height
                        v[8] = aw[i][8]; // PHASE 1 velocity
                        v[9] = aw[i][9]; // PHASE 2 velocity
                        v[10] = hflowi + hflowi2; // total flow height
                        v[11] = aw[i][11]; // PHASE 1 flow kinetic energy
                        v[12] = aw[i][12]; // PHASE 2 flow kinetic energy
                        v[13] = aw[i][13]; // total flow kinetic energy
                        v[14] = aw[i][14]; // PHASE 1 flow pressure
                        v[15] = aw[i][15]; // PHASE 2 flow pressure
                        v[16] = aw[i][16]; // total flow pressure
                        v[17] = hentri + hentri2; // total entrained or deposited height
                    }
                    else if ( sico.MODEL == 7 ) { // for three-phase model:

                        v[3] = hflowi2; // PHASE 3 flow height
                        if ( aw[i][3] > sico.HFLOWMIN ) v[4] = aw[i][4] / aw[i][3]; else v[4] = 0; // PHASE 3 velocity in x direction
                        if ( aw[i][3] > sico.HFLOWMIN ) v[5] = aw[i][5] / aw[i][3]; else v[5] = 0; // PHASE 3 velocity in y direction
                        v[6] = hflowi3; // PHASE 3 flow height
                        if ( aw[i][6] > sico.HFLOWMIN ) v[7] = aw[i][7] / aw[i][6]; else v[7] = 0; // PHASE 3 velocity in x direction
                        if ( aw[i][6] > sico.HFLOWMIN ) v[8] = aw[i][8] / aw[i][6]; else v[8] = 0; // PHASE 3 velocity in y direction
                        v[9] = hentri; // entrained or deposited PHASE 1 height
                        v[10] = hentri2; // entrained or deposited PHASE 2 height
                        v[11] = hentri3; // entrained or deposited PHASE 3 height
                        v[12] = aw[i][12]; // PHASE 1 velocity
                        v[13] = aw[i][13]; // PHASE 2 velocity
                        v[14] = aw[i][14]; // PHASE 3 velocity
                        v[15] = hflowi + hflowi2 + hflowi3; // total flow height
                        v[16] = aw[i][16]; // PHASE 1 flow kinetic energy
                        v[17] = aw[i][17]; // PHASE 2 flow kinetic energy
                        v[18] = aw[i][18]; // PHASE 3 flow kinetic energy
                        v[19] = aw[i][19]; // total flow kinetic energy
                        v[20] = aw[i][20]; // PHASE 1 flow pressure
                        v[21] = aw[i][21]; // PHASE 2 flow pressure
                        v[22] = aw[i][22]; // PHASE 3 flow pressure
                        v[23] = aw[i][23]; // total flow pressure
                        v[24] = hentri + hentri2 + hentri3; // total entrained or deposited height
                    }

                    //for ( k=0; k<nvect_red; k++ ) Segment_put(&seg_v[k], &v[k], px[i], py[i]); // writing values to segment files
                    for ( k=0; k<nvect_red; k++ ) testv[px[i]][py[i]][k] = v[k];
                }
                //for ( k=0; k<nvect_red; k++ ) Segment_flush(&seg_v[k]); // forcing to immediately write values to segment file


            #endif


        // Summary of state of the flow at output time step

            if ( sico.MODEL <= 3 ) { // for one-phase model:


               #ifdef WITHGRASS


                    printf("   %i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.*f\n",
                        nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max,
                        prec_vol, vol_flow/1000, prec_ekin, ekin_flow/1000000); // displaying parameters


               #else


                    printf("   %i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.*f",
                        nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max,
                        prec_vol, vol_flow/1000, prec_ekin, ekin_flow/1000000); // displaying parameters


                #endif


                fflush(stdout); // forcing immediate display

                fprintf(f_summary, "%i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.*f\n",
                    nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max,
                    prec_vol, vol_flow, prec_ekin, ekin_flow); // writing parameters to summary file
            }
            else if ( sico.MODEL > 3 && sico.MODEL < 7 ) { // for two-phase model:


               #ifdef WITHGRASS


                    printf("   %i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.*f\t%.*f\n",
                        nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max, prec_hflow, hflow_max2, vflow_max2,
                        prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000, prec_ekin, ekin_flow/1000000); // displaying parameters


                #else


                    printf("   %i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.*f\t%.*f",
                        nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max, prec_hflow, hflow_max2, vflow_max2,
                        prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000, prec_ekin, ekin_flow/1000000); // displaying parameters


                #endif


                fflush(stdout); // forcing immediate display
                fprintf(f_summary, "%i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.*f\t%.*f\n",
                    nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max, prec_hflow, hflow_max2, vflow_max2,
                    prec_vol, vol_flow, prec_vol, vol_flow2, prec_ekin, ekin_flow); // writing parameters to summary file
            }
            else if ( sico.MODEL == 7 ) { // for three-phase model:


               #ifdef WITHGRASS


                    printf("   %i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.*f\t%.*f\t%.*f\n",
                        nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max, prec_hflow, hflow_max2, vflow_max2,
                        prec_hflow, hflow_max3, vflow_max3, prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000,
                        prec_vol, vol_flow3/1000, prec_ekin, ekin_flow/1000000); // displaying parameters


                #else


                    printf("   %i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.*f\t%.*f\t%.*f",
                        nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max, prec_hflow, hflow_max2, vflow_max2,
                        prec_hflow, hflow_max3, vflow_max3, prec_vol, vol_flow/1000, prec_vol, vol_flow2/1000,
                        prec_vol, vol_flow3/1000, prec_ekin, ekin_flow/1000000); // displaying parameters


                #endif


                fflush(stdout); // forcing immediate display
                fprintf(f_summary, "%i\t%i\t%.3f\t%.1f\t%.1f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.2f\t%.*f\t%.*f\t%.*f\t%.*f\n",
                    nout, nsum, cflmax, 100*tlength, tsum, prec_hflow, hflow_max, vflow_max, prec_hflow, hflow_max2, vflow_max2,
                    prec_hflow, hflow_max3, vflow_max3, prec_vol, vol_flow, prec_vol, vol_flow2, prec_vol, vol_flow3,
                    prec_ekin, ekin_flow); // writing parameters to summary file
            }

            sprintf(path, "%s%snout%d.txt", outfiles, prefix, xint); // writing number of time steps and control for success to file:
            f_nout=fopen(path, "w");
            fprintf(f_nout, "%i\n%i\n%.0f\n%0f\n%0f\n", nout, csuccess, vol_entr, vol_entr2, vol_entr3 );
            fclose(f_nout);


            if ( sico.MULT == 0 ) { // for single model run:


            // Writing output raster maps

                if ( nout < 10 ) sprintf( madd, "000"); // fill string (for maintaining correct order in list of maps)
                else if ( nout < 100 ) sprintf( madd, "00");
                else if ( nout < 1000 ) sprintf( madd, "0");

                for ( k=0; k<nvect_red; k++ ) { // preparing, writing and closing output raster maps for time steps
                    sprintf( mv, "%s%s%s%i", prefix, mv0[k], madd, nout );


                    #ifdef WITHGRASS


                        foutrast ( mv, testv, sico, k, tsum ); // if GRASS is used, writing GRASS raster maps


                    #endif


                    if ( sico.MODEL <= 3 ) foutasc ( aw, px, py, outmaps, mv, hflow_max, pelev, betax, betay, betaxy,
                        k, sico ); // writing ascii raster maps (one-phase model)
                    else if ( sico.MODEL > 3 && sico.MODEL < 7 ) foutasc ( aw, px, py, outmaps, mv, hflow_max+hflow_max2, pelev, betax, betay, betaxy,
                        k, sico ); // writing ascii raster maps (two-phase model)
                    else if ( sico.MODEL == 7 ) foutasc ( aw, px, py, outmaps, mv, hflow_max+hflow_max2+hflow_max3, pelev, betax, betay, betaxy,
                        k, sico ); // writing ascii raster maps (three-phase model)
                }


            // Writing maximum flow height at time step to ascii raster

                if ( sico.MODEL > 3 && sico.MODEL < 7 ) {
                    sprintf( mv, "%s%s%s%i", prefix, mv0[22], madd, nout );
                    foutasc ( aw, px, py, outmaps, mv, hflow_max+hflow_max2, pelev, betax, betay, betaxy,
                    22, sico );
                }
                else if ( sico.MODEL == 7 ) {
                    sprintf( mv, "%s%s%s%i", prefix, mv0[31], madd, nout );
                    foutasc ( aw, px, py, outmaps, mv, hflow_max+hflow_max2+hflow_max3, pelev, betax, betay, betaxy,
                    31, sico );
                }


            // Writing file for display of flow vectors as arrows

                if ( nout == 1 && sico.MODEL <= 3 ) foutdircoord ( f_directions, 1, px, py, sico );
                else if ( nout == 1 && ( sico.MODEL > 3 && sico.MODEL < 7 )) {
                    foutdircoord ( f_directions, 2, px, py, sico );
                    foutdircoord ( f_directions2, 3, px, py, sico );
                }
                else if ( nout == 1 && sico.MODEL == 7 ) {
                    foutdircoord ( f_directions, 4, px, py, sico );
                    foutdircoord ( f_directions2, 5, px, py, sico );
                    foutdircoord ( f_directions3, 6, px, py, sico );
                } // writing coordinates to file

                if ( sico.MODEL <= 3 ) for ( k=0; k<3; k++ ) foutdir ( f_directions, aw, k, 1, px, py, sico );
                else if ( sico.MODEL > 3 && sico.MODEL < 7 ) {
                    for ( k=0; k<3; k++ ) foutdir ( f_directions, aw, k, 2, px, py, sico );
                    for ( k=3; k<6; k++ ) foutdir ( f_directions2, aw, k, 3, px, py, sico );
                }
                else if ( sico.MODEL == 7 ) {
                    for ( k=0; k<3; k++ ) foutdir ( f_directions, aw, k, 4, px, py, sico );
                    for ( k=3; k<6; k++ ) foutdir ( f_directions2, aw, k, 5, px, py, sico );
                    for ( k=6; k<9; k++ ) foutdir ( f_directions3, aw, k, 6, px, py, sico );
                    // writing parameters to file
                }
            }


        // Writing output raster maps for last time step

            if ( tsum >= tmax || ccontinue == 0 ) { // if time step is last one:

                for ( k=0; k<nvect_red; k++ ) { // preparing, writing and closing output raster maps for last time step:
                    if (( sico.MODEL <= 3 && ( k == 0 || k == 3 ))
                        || (( sico.MODEL > 3 && sico.MODEL < 7 ) && ( k==0 || k==3 || k==6 || k==7 || k==10 || k==17 ))
                        || ( sico.MODEL == 7 && ( k==0 || k==3 || k==6 || k==9 || k==10 || k==11 || k==15 || k==24 ))) {

                        if ( sico.MULT == 0 ) sprintf( mv, "%s%s_fin", prefix, mv0[k] );
                        else sprintf( mv, "%s%s_fin%d", prefix, mv0[k], xint ); // names of output raster maps


                        #ifdef WITHGRASS


                            foutrast ( mv, testv, sico, k, tsum ); // if GRASS is used, writing GRASS raster maps


                        #endif


                        if ( sico.MODEL <= 3 ) foutasc ( aw, px, py, outmaps, mv, hflow_max, pelev, betax, betay, betaxy,
                            k, sico ); // writing Ascii raster maps (one-phase model)
                        else if ( sico.MODEL > 3 && sico.MODEL < 7 ) foutasc ( aw, px, py, outmaps, mv, hflow_max+hflow_max2, pelev, betax, betay, betaxy,
                            k, sico ); // writing Ascii raster maps (two-phase model)
                        else if ( sico.MODEL == 7 ) foutasc ( aw, px, py, outmaps, mv, hflow_max+hflow_max2+hflow_max3, pelev, betax, betay, betaxy,
                            k, sico ); // writing Ascii raster maps (three-phase model)
                    }
                }
            }


// ----- STOP --- Writing output files and raster maps -------------------------------------------------------------


            cflmax = 0; // resetting maximum cfl value
            nout += 1; // updating number of output time steps
        }

        nsum += 1; // updating total number of time steps


// *** End of loop over time steps ---------------------------------------------------------------------------------


    }


// Writing output rasters of cumulative and maximum values


    #ifdef WITHGRASS


    // if GRASS is used, writing data to segment files:

        for ( k=nvect_red; k<nvect_all; k++ ) { // loop over all components of the vector:

            for ( i=0; i<sico.IMAX; i++ ) { // loop over all pixels:

                if ( cdomain[i] != 0 ) { // if pixel is not a 1st row edge pixel:

                    if ( sico.MODEL <= 3 && k == 7 ) pout = fconvout( i, aw, hflow_maxmax, 1, 3,
                        px, py, pelev, betax, betay, betaxy, sico );

                    else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && k == 18 ) pout = fconvout( i, aw, hflow_maxmax, 1, 3,
                        px, py, pelev, betax, betay, betaxy, sico );
                    else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && k == 20 ) pout = fconvout( i, aw, hflow_maxmax, 2, 3,
                        px, py, pelev, betax, betay, betaxy, sico );
                    else if (( sico.MODEL > 3 && sico.MODEL < 7 ) && k == 22 ) pout = fconvout( i, aw, hflow_maxmax, 4, 3,
                        px, py, pelev, betax, betay, betaxy, sico );

                    else if ( sico.MODEL == 7 && k == 25 ) pout = fconvout( i, aw, hflow_maxmax, 1, 3,
                        px, py, pelev, betax, betay, betaxy, sico );
                    else if ( sico.MODEL == 7 && k == 27 ) pout = fconvout( i, aw, hflow_maxmax, 2, 3,
                        px, py, pelev, betax, betay, betaxy, sico );
                    else if ( sico.MODEL == 7 && k == 29 ) pout = fconvout( i, aw, hflow_maxmax, 3, 3,
                        px, py, pelev, betax, betay, betaxy, sico );
                    else if ( sico.MODEL == 7 && k == 31 ) pout = fconvout( i, aw, hflow_maxmax, 4, 3,
                        px, py, pelev, betax, betay, betaxy, sico );
                    else pout = aw[i][k]; // converting depths into heights
                }
                else pout = aw[i][k]; // for 1st row edge pixels, applying depths as heights

                //Segment_put(&seg_v[k], &pout, px[i], py[i]); // writing values to segment file
                testv[px[i]][py[i]][k] = pout;
            }
        }


    #endif


    for ( k=nvect_red; k<nvect_all; k++ ) { // loop over all elements of vector:

        if ( sico.MULT == 0 ) sprintf( mv, "%s%s", prefix, mv0[k] );
        else sprintf( mv, "%s%s%d", prefix, mv0[k], xint ); // name of output raster map


        #ifdef WITHGRASS


            //Segment_flush(&seg_v[k] ); // forcing to write values to segment file immediately
            foutrast ( mv, testv, sico, k, tsum ); // writing GRASS raster maps


        #endif


        foutasc ( aw, px, py, outmaps, mv, hflow_maxmax, pelev, betax, betay, betaxy, k, sico ); // writing Ascii raster maps
    }


    #ifdef WITHGRASS


    // Releasing segment data (if GRASS is used)

        if ( sico.NMESH == 1 ) Segment_release(&seg_nmesh);
        Segment_release(&seg_elev);
        if ( sico.RELM == 1 ) Segment_release(&seg_hrelease);
        if ( sico.RELV == 1 ) { Segment_release(&seg_vinx); Segment_release(&seg_viny); }
        if ( sico.ENTR == 1 ) Segment_release(&seg_hentrmax);
        if ( sico.MODEL > 3 && sico.RELM2 == 1 ) Segment_release(&seg_hrelease2);
        if ( sico.MODEL == 7 && sico.RELM3 == 1 ) Segment_release(&seg_hrelease3);
        if ( sico.MODEL > 3 && sico.RELV2 == 1 ) { Segment_release(&seg_vinx2); Segment_release(&seg_viny2); }
        if ( sico.MODEL == 7 && sico.RELV3 == 1 ) { Segment_release(&seg_vinx3); Segment_release(&seg_viny3); }
        if ( sico.MODEL > 3 && sico.ENTR2 == 1 ) Segment_release(&seg_hentrmax2);
        if ( sico.MODEL == 7 && sico.ENTR3 == 1 ) Segment_release(&seg_hentrmax3);
        if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) Segment_release(&seg_centr);
        if ( sico.STOPPING == 6 && sico.CDEP == 1 ) Segment_release(&seg_cdep);
        if ( sico.PHI == 1 ) Segment_release(&seg_phi);
        if ( sico.PHI2 == 1 ) Segment_release(&seg_phi2);
        if ( sico.PHI3 == 1 ) Segment_release(&seg_phi3);
        if ( sico.DELTA == 1 ) Segment_release(&seg_delta);
        if ( sico.DELTA2 == 1 ) Segment_release(&seg_delta2);
        if ( sico.DELTA3 == 1 ) Segment_release(&seg_delta3);
        if ( sico.NYSS == 1 ) Segment_release(&seg_nyss);
        if ( sico.NYFS == 1 ) Segment_release(&seg_nyfs);
        if ( sico.NYFF == 1 ) Segment_release(&seg_nyff);
        if ( sico.AMBDRAG == 1 ) Segment_release(&seg_ambdrag);
        if ( sico.FLUFRI == 1 ) Segment_release(&seg_flufri);
        if ( sico.TRANSSSFS == 1 ) Segment_release(&seg_transssfs);
        if ( sico.TRANSSSFF == 1 ) Segment_release(&seg_transssff);
        if ( sico.TRANSFSFF == 1 ) Segment_release(&seg_transfsff);
        if ( sico.TRELEASE == 1 ) Segment_release(&seg_trelease);
        if ( sico.TRELSTOP == 1 ) Segment_release(&seg_trelstop);
        //Segment_release(&seg_v);


    #endif


// Closing output files

    fclose ( f_summary );
    if ( hydrograph == 1 ) {

        fclose ( f_hydout );
        for ( i = 0; i < hydnin + hydnout; i++ ) {
            if ( sico.MULT == 0 ) fclose( f_hydinfo[i] );
        }
    }

    if ( sico.MULT == 0 ) fclose ( f_directions );
    if ( sico.MULT == 0 && sico.MODEL > 3 ) fclose ( f_directions2 );
    if ( sico.MULT == 0 && sico.MODEL == 7 ) fclose ( f_directions3 );


// Releasing memory


    #ifdef WITHGRASS


        //free ( seg_v );
        free ( v );


    #endif


    free(flowpar); if ( sico.NMESH == 1 ) free(pnmesh); free(phrelease); free(pvinx); free(pviny); free(phentrmax); free(dentr); free(cflux);
    free(cdomain); free(cdomain2); free(cstopped);
    free(in[0]); free(in); free(ib); free(ibasket[0]); free(ibasket); free(icheck[0]); free(icheck); free(cinner);
    if ( sico.MODEL > 3 ) { free(phrelease2); free(pvinx2); free(pviny2); free(phentrmax2); free(dentr2); free(cinner2); }
    if ( sico.MODEL == 7 ) { free(phrelease3); free(pvinx3); free(pviny3); free(phentrmax3); free(dentr3); free(cinner3); }
    if ( sico.STOPPING > 0 && sico.STOPPING < 4 ) free(fom);
    free(aw[0]); free(awt[0]); free(af[0]); free(ag[0]); free(as[0]); free(ad[0]); free(asigma_x[0]); free(asigma_y[0]);
    free(asigma_f[0]); free(asigma_g[0]); free(wintd[0]); free(wintdtest[0]); free(f[0]); free(g[0]); free(s[0]); free(wintelev[0]); free(wintelev);
    free_dmatrix3(winta, sico.IMAX, 4); free_dmatrix3(wintb, sico.IMAX, 6); free_dmatrix3(wintc, sico.IMAX, 4);
    free_dmatrix3(d, sico.IMAX, 6); 
    free(aw); free(awt); free(af); free(ag); free(as); free(ad); free(asigma_x); free(asigma_y); free(asigma_f); free(asigma_g);
    free(asigma_xelev); free(asigma_yelev); free(wintd); free(wintdtest); free(f); free(g); free(s); free(mv0);

    if ( hydrograph == 1 ) {
        free_dmatrix3(hydhyd, hydtmaxx+1, 5);
        free(hydp[0]), free(hydp);
        free(hydelev); free(hydalpha); free(hydx_metric); free(hydy_metric); free(hydl); free(hydtmax); free(hydi); free(hydx); free(hydy);
        for ( hydj=0; hydj<hydnin; hydj++ ) free(hydreleasename[hydj]);
    }

    if ( sico.NMESH == 1 ) free(nmeshname); free(pelev); free(pelev0); free(px); free(py); free(dx); free(dy); free(betax); free(betay); free(betaxy);
    free(madd); free(prefix); free(outmaps); free(outfiles); free(elevname);
    if ( sico.RELM == 1 ) free(hreleasename); if ( sico.RELV == 1 ) { free(vinxname); free(vinyname); }
    if ( sico.ENTR == 1 ) free(hentrmaxname); free(sico.MAINMAPSET);
    if ( sico.ENTRAINMENT != 0 && sico.CENTR == 1 ) free(centrname);
    if ( sico.STOPPING == 6 && sico.CDEP == 1 ) free(cdepname);
    if ( sico.PHI == 1 ) free(phiname); if ( sico.PHI2 == 1 ) free(phi2name); if ( sico.PHI3 == 1 ) free(phi3name);
    if ( sico.DELTA == 1 ) free(deltaname); if ( sico.DELTA2 == 1 ) free(delta2name); if ( sico.DELTA3 == 1 ) free(delta3name);
    if ( sico.NYSS == 1 ) free(nyssname); if ( sico.NYFS == 1 ) free(nyfsname); if ( sico.NYFF == 1 ) free(nyffname);
    if ( sico.AMBDRAG == 1 ) free(ambdragname); if ( sico.FLUFRI == 1 ) free(flufriname);
    if ( sico.TRANSSSFS == 1 ) free(transssfsname); if ( sico.TRANSSSFF == 1 ) free(transssffname); if ( sico.TRANSFSFF == 1 ) free(transfsffname);
    if ( sico.TRELEASE == 1 ) free(treleasename);
    if ( sico.TRELSTOP == 1 ) free(trelstopname);
    if ( sico.MODEL > 3 ) { if ( sico.RELM2 == 1 ) free(hreleasename2); if ( sico.RELV2 == 1 ) { free(vinxname2); free(vinyname2); }
    if ( sico.ENTR2 == 1 ) free(hentrmaxname2); }
    if ( sico.MODEL == 7 ) { if ( sico.RELM3 == 1 ) free(hreleasename3); if ( sico.RELV3 == 1 ) { free(vinxname3); free(vinyname3); } 
    if ( sico.ENTR3 == 1 ) free(hentrmaxname3); }
    if ( sico.DIFFCTRL != 0 ) { free(cedge[0]); free(cready[0]); free(cedge); free(cready); free(cedge0); free(cneighbours); }
    if ( sico.DIFFCTRL != 0 && sico.MODEL > 3 ) {
        free(cedge2[0]); free(cready2[0]); free(cedge2); free(cready2); free(cedge02); free(cneighbours2);
    }
    if ( sico.DIFFCTRL != 0 && sico.MODEL == 7 ) {
        free(cedge3[0]); free(cready3[0]); free(cedge3); free(cready3); free(cedge03); free(cneighbours3);
    }


// Computing and displaying time needed for model execution

    time_stop = clock(); // time at end of model execution
    time_elapsed = ( ( float ) ( time_stop - time_start ) ) / CLOCKS_PER_SEC; // time needed for model execution
    printf("Model execution completed in %.2f seconds.\n", time_elapsed);
    fflush(stdout); // forcing immediate display

    return 0;
}
